
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

/*
 *  _dyn_.h:  Derived class for nbody systems using a Hermite integrator
 *	      Base class for hdyn, sdyn3, and sdyn.
 *.............................................................................
 *    version 1:  Dec 1992   Piet Hut, Steve McMillan, Jun Makino
 *    version 2:
 *.............................................................................
 *     This file includes:
 *  1) definition of class _dyn_
 *.............................................................................
 */

#ifndef  STARLAB__DYN__H
#  define  STARLAB__DYN__H

#include  "dyn.h"

/*-----------------------------------------------------------------------------
 *  _dyn_  --  a derived class of dynamical particles, with enough information
 *             to integrate the equations of motions with a 4th-order Hermite
 *             polynomial integrator.
 *-----------------------------------------------------------------------------
 */
class  _dyn_ : public dyn
{
    protected:

	real    time;           //
	real    timestep;       //
	  
	real    pot;            //  potential
	vector  jerk;           //  (d/dt) acc
	vector  old_acc;        //
	vector  old_jerk;       //

	vector  pred_pos;       //
	vector  pred_vel;       //
	real    t_pred;		// time corresponding to pred_pos and pred_vel

        real    radius;		// effective (or actual) radius of a node.

    public:

        _dyn_(hbpfp the_hbpfp = new_hydrobase, sbpfp the_sbpfp = new_starbase)
	   : dyn(the_hbpfp, the_sbpfp)
	    {
	    time = timestep = pot = 0;
            jerk = old_acc = old_jerk = pred_pos = pred_vel = 0;
	    t_pred = -VERY_LARGE_NUMBER;
	    radius = 0;
	    }
 
	void  set_timestep(real dt)       {timestep = dt;}
	real  get_timestep()              {return timestep;}

	real  get_time()                  {return time;}
	void  set_time(real t)            {time = t;}

	void  clear_interaction()         {acc=jerk=pot = 0;}

	void  set_acc_and_jerk_and_pot(const vector & a,
				       const vector & j, real p)
	   {acc = a; jerk = j; pot = p;}

	vector  get_pred_pos()            {predict_loworder(system_time);
					   return pred_pos;}
	vector  get_pred_vel()            {predict_loworder(system_time);
					   return pred_vel;}
	void clear_t_pred()		  {t_pred = -VERY_LARGE_NUMBER;}
	real  get_t_pred()                {return t_pred;}
	
	real  get_pot()                   {return pot;}
	void  set_pot(real p)		  {pot = p;}
	void  clear_pot()                 {pot = 0;}
	void  inc_pot(const real& d_pot)  {pot += d_pot;}

        void  set_jerk(const vector& new_jerk)     {jerk = new_jerk;}
        void  set_old_acc(const vector& new_acc)   {old_acc = new_acc;}
        void  set_old_jerk(const vector& new_jerk) {old_jerk = new_jerk;}
	void  clear_jerk()                         {jerk = 0;}
	void  inc_jerk(const vector& d_jerk)       {jerk += d_jerk; }
	void  inc_old_acc(const vector& d_acc)     {old_acc += d_acc; }
	void  inc_old_jerk(const vector& d_jerk)   {old_jerk += d_jerk; }
	vector  get_jerk()                         {return jerk;}
	vector  get_old_acc()                      {return old_acc;}
	vector  get_old_jerk()                     {return old_jerk;}

	real  get_radius()			{return radius;}
	void  set_radius(real r)		{radius = r;}

	_dyn_ * get_parent()
	    {return (_dyn_*) node::get_parent();}
	_dyn_ * get_oldest_daughter()
	    {return (_dyn_*)node::get_oldest_daughter();}
	_dyn_ * get_younger_sister()
	    {return (_dyn_*) node::get_younger_sister();}
	_dyn_ * get_elder_sister()
	    {return (_dyn_*) node::get_elder_sister();}

        _dyn_ * get_root()
            {return (_dyn_*) node::get_root();}
        _dyn_ * get_top_level_node()
            {return (_dyn_*) node::get_top_level_node();}
        _dyn_ * get_binary_sister()
            {return (_dyn_*) node::get_binary_sister();}

	void  init_pred()                 {pred_pos = pos;
					   pred_vel = vel;
				           t_pred = time;}

	void  store_old_force()	          {old_acc = acc; old_jerk = jerk;}

	void _dyn_::predict_loworder(const real t)
	    {
		if (kep != NULL) {
		    init_pred();
		} else {
		    if (t_pred < t) {
			real dt = t - time;
			real dt3 = dt * ONE_THIRD;
			real dt2 = dt * 0.5;
			pred_pos = ((old_jerk * dt3 + old_acc) * dt2 + vel) 
			             * dt + pos;
			pred_vel = (old_jerk * dt2 + old_acc) * dt + vel;
			t_pred = t;
		    }
		}
	    }

        bool  correct();		// Changed to bool by Steve, 9/15/98

	void  update(const real, const real);
	void  flat_update(const real, const real);
	void  correct_and_update(const real, const real, const real);
    };

typedef _dyn_ * _dyn_ptr;  // to enable dynamic array declarations such as
                         //    _dyn_** _dyn__ptr_array = new _dyn_ptr[n];
                         // (note that the following expression is illegal:
                         //    _dyn_** _dyn__ptr_array = new (_dyn_ *)[n];)

inline  node * new__dyn_(hbpfp the_hbpfp, sbpfp the_sbpfp)
    {return (node *) new _dyn_(the_hbpfp, the_sbpfp);}

inline _dyn_ * mk_dyn_(int n, hbpfp the_hbpfp = new_hydrobase,
	                    sbpfp the_sbpfp = new_starbase)
    {return  (_dyn_ *) mk_flat_tree(n, new__dyn_, the_hbpfp, the_sbpfp);}

#define  put__dyn_  put_node

#endif

//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\
//  |  the end of:  |         /|\         |  inc/_dyn_.h
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\

