
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//  hdyn.h: derived class for nbody systems, when using a Hermite integrator
//.............................................................................
//    version 1:  Dec 1992   Piet Hut, Steve McMillan, Jun Makino
//    version 2:
//.............................................................................
//     This file includes:
//  1) definition of class hdyn
//.............................................................................

#ifndef  STARLAB_HDYN_H
#  define  STARLAB_HDYN_H

#define MAX_PERTURBERS				256
#define ALLOW_LOW_LEVEL_PERTURBERS		true
#define RESOLVE_UNPERTURBED_PERTURBERS		false

//#include "sint.h"

#include  "_dyn_.h"
#include  "kira.h"

//-----------------------------------------------------------------------------
//  hdyn  --  a derived class of dynamical particles, with enough information
//             to integrate the equations of motions with a 4th-order Hermite
//             polynomial integrator.
//-----------------------------------------------------------------------------

class  hdyn : public _dyn_ {
    protected:

        // -------------------- Global variables! --------------------

        static kira_counters* kc;

        static real d_min_sq;	// scale term governing tree adjustment
        static real mbar;	// mass scale

        static real gamma2;	// perturbation criterion
        static real gamma23;	// gamma^{-2/3}

        static real omega;	// system angular velocity (for tidal field)
        static real omega_sq;	// omega squared

        static bool use_dstar;  // activate binary evolution

        static real stellar_merger_criterion_sq;
        static real stellar_encounter_criterion_sq;
  
	// -----------------------------------------------------------

        // Variables for unperturbed kepler motion:

        real  perturbation_squared;  // relative perturbation squared
        real  unperturbed_timestep;  // timestep for the unperturbed motion
        bool  fully_unperturbed;     // true if orbit is fully unperturbed

        // Perturber information:

        int  n_perturbers;	// number of perturbers of this node
        hdyn** perturber_list;  // pointer to perturber array
        bool valid_perturbers;  // true if any particle is within the
			        // perturbation radius and the perturber
				// has not overflowed
        real  perturbation_radius_factor;

        // Other (colliding) neighbor information:

        hdyn* nn;               // pointer to the nearest neighbor
        real d_nn_sq;           // distance squared to the nearest neighbor
    
        hdyn* coll;		// pointer to the neighbor whose surface is
                                // closest to 'this' surface (i.e. as
                                // measured by  distance - r1 - r2 )
        real d_coll_sq;		// distance squared to this neighbor
  
        // Variables for HARP-3:

        int harp3_index;	// address of this particle on harp3 memory
        real harp3_rnb_sq;      // the neighbour sphere radius of the particle
			        // used on HARP-3

	// Bookkeeping (keeping track of cost):

	real steps;
	real direct_force;
	real indirect_force;

        // protected functions...

        bool is_weakly_perturbed();		    // unperturbed criteria for
        bool is_stable(bool top_level = true);	    // a multiple system...

    public:

        hdyn(hbpfp the_hbpfp = new_hydrobase, sbpfp the_sbpfp = new_starbase)
	    : _dyn_(the_hbpfp, the_sbpfp) {

	    perturbation_squared = -1;
            unperturbed_timestep  =  -VERY_LARGE_NUMBER;
            fully_unperturbed = false;

            n_perturbers = 0;
	    perturber_list = NULL;
            valid_perturbers = false;
	    perturbation_radius_factor = -1;

            nn = NULL;
            d_nn_sq = 0;

            coll = NULL;
	    d_coll_sq = 0;

	    harp3_index = 0;
	    harp3_rnb_sq = 0;

	    steps = direct_force = indirect_force = 0;

        }

	virtual ~hdyn() {
	    if (perturber_list)
		delete [] perturber_list;	// added 7/29/98 (SLWM & JM)
	}

        kira_counters* get_kira_counters() {return kc;}
        void set_kira_counters(kira_counters *k) {kc = k;}
  
        inline real get_next_time() {
            if (kep == NULL)
                return time + timestep;
            else
                return time + unperturbed_timestep;
        }

        void set_d_min_sq(real d)		{d_min_sq = d;}
        real get_d_min_sq()			{return d_min_sq;}
        void set_mbar(real m)			{mbar = m;}
        real get_mbar()				{return mbar;}

        void set_gamma2(real g)			{gamma2 = g;}
        real get_gamma2()			{return gamma2;}
        void set_gamma23(real g)		{gamma23 = g;}
        real get_gamma23()			{return gamma23;}

        void set_omega(real o)			{omega = o;
						 omega_sq = o*o;}
        real get_omega()			{return omega;}
        real get_omega_sq()			{return omega_sq;}

	bool get_use_dstar()			{return use_dstar;}
	void set_use_dstar(bool u)		{use_dstar = u;}

	hdyn * get_nn()				{return nn;}
	void set_nn(hdyn * new_nn)		{nn = new_nn;}
	real get_d_nn_sq()			{return d_nn_sq;}
	void set_d_nn_sq(real d)		{d_nn_sq = d;}
  
        hdyn * get_coll()			{return coll;}
        void  set_coll(hdyn * new_coll)		{coll = new_coll;}
        real get_d_coll_sq()			{return d_coll_sq;}
        void  set_d_coll_sq(real d)		{d_coll_sq = d;}

        real get_stellar_merger_criterion_sq()
		{return stellar_merger_criterion_sq;} 
        void set_stellar_merger_criterion_sq(real d_sq)
		{stellar_merger_criterion_sq = d_sq;}
        real get_stellar_encounter_criterion_sq()
		{return stellar_encounter_criterion_sq;} 
        void set_stellar_encounter_criterion_sq(real d_sq)
		{stellar_encounter_criterion_sq = d_sq;}

	inline void inc_steps(int i = 1)	{steps += i;}
	inline void inc_steps(real i)		{steps += i;}
	real get_steps()			{return steps;}
	inline void inc_direct_force(int i = 1)	{direct_force += i;}
	inline void inc_direct_force(real i)	{direct_force += i;}
	real get_direct_force()			{return direct_force;}
	inline void inc_indirect_force(int i = 1) {indirect_force += i;}
	inline void inc_indirect_force(real i)	{indirect_force += i;}
	real get_indirect_force()		{return indirect_force;}

        int  get_harp3_index()			{return harp3_index;}
        void set_harp3_index(int hindex)	{harp3_index = hindex;}
        real get_harp3_rnb_sq()			{return harp3_rnb_sq;}
        void set_harp3_rnb_sq(real rnb_sq)	{harp3_rnb_sq = rnb_sq;}

        hdyn * get_parent()
		{return (hdyn*) node::get_parent();}
        hdyn * get_oldest_daughter()
		{return (hdyn*)node::get_oldest_daughter();}
        hdyn * get_younger_sister()
		{return (hdyn*) node::get_younger_sister();}
        hdyn * get_elder_sister()
		{return (hdyn*) node::get_elder_sister();}

        hdyn * get_root()
		{return (hdyn*) node::get_root();}
        hdyn * get_top_level_node()
		{return (hdyn*) node::get_top_level_node();}
        hdyn * get_binary_sister()
		{return (hdyn*) node::get_binary_sister();}

        virtual  istream& scan_dyn_story(istream&);
        virtual  ostream& print_dyn_story(ostream &);
	
        void  init_pred()                 {pred_pos = pos;
					   pred_vel = vel;
					   t_pred = time;}

        void  store_old_force()	          {old_acc = acc; old_jerk = jerk;}

        bool  get_valid_perturbers()              {return valid_perturbers;}
        void  set_valid_perturbers(const bool vp) {valid_perturbers=vp;}

        inline void hdyn::predict_loworder(const real t) {
	    if (kep != NULL) {
		init_pred();
	    } else {
		if (t_pred < t) {
		    real dt = t - time;
		    real dt3 = dt * ONE_THIRD;
		    real dt2 = dt * 0.5;
		    pred_pos = ((old_jerk * dt3 + old_acc) * dt2 + vel) 
		    * dt + pos;
		    pred_vel = (old_jerk * dt2 + old_acc) * dt + vel;
		    t_pred = t;
		}
	    }
	}

        void  set_first_timestep(real, real);
        void  flat_set_first_timestep(real, real);

        bool  correct();		// Changed to bool by Steve, 9/15/98

        void  update(const real, const real);
        void  flat_update(const real, const real);
        void  correct_and_update(const real, const real, const real);

        inline void flat_accumulate_acc_and_jerk(hdyn *, real);

        void  flat_calculate_acc_and_jerk(hdyn *, real);

        // Overload this function for use in kira...

        void flat_calculate_acc_and_jerk(hdyn*, real, bool);

	hdyn* find_perturber_node();
	inline void check_add_perturber(hdyn* p, vector this_pos);

        void perturber_acc_and_jerk_on_leaf(real eps2,
					    vector & a,
					    vector & j,
					    real & p,
					    real & distance_squared,
					    hdyn*& nn,
					    hdyn* pnode,
					    hdyn* step_node);

        void  integrate_node(hdyn* root,
			     real eps, real eta, real max_timestep,
			     real gamma2,
			     bool integrate_unperturbed = true,
			     bool force_unperturbed_time = false);
        void  calculate_acc_and_jerk(real, real, bool);
        void  calculate_acc_and_jerk_on_top_level_node(real, real, bool);
        void  calculate_acc_and_jerk_on_low_level_node(real, real);

        void  calculate_partial_acc_and_jerk(hdyn*, hdyn*, hdyn*,real, vector&,
					     vector&, real&, real&, hdyn*&,
					     bool, hdyn*, hdyn*);
        void  calculate_partial_acc_and_jerk_on_leaf(hdyn*,hdyn*, hdyn*, real,
						     vector&, vector&,
						     real&, real&, hdyn*&,
						     bool, hdyn*, hdyn*);
        void  tree_walk_for_partial_acc_and_jerk_on_leaf(hdyn*, hdyn*, vector,
							 vector, real, vector&,
							 vector&, real&, real&,
							 hdyn*&, bool, hdyn*);

        void  setup_binary_node();

        void  synchronize_node(real eps, real eta , real max_timestep);

        real  distance_to_sister_squared();
        hdyn* new_sister_node(real, bool&);
        int adjust_tree_structure(real, real, real, real);

        real get_unperturbed_timestep() {return unperturbed_timestep;}
        bool get_fully_unperturbed() {return fully_unperturbed;}
        void set_fully_unperturbed(bool f) {fully_unperturbed = f;}
	
        bool is_unperturbed_and_approaching();
        bool is_close_pair();
        void startup_unperturbed_motion();

        // int get_unperturbed_steps(bool to_apo = true,
	//			     bool predict = false);
        real get_unperturbed_steps(bool to_apo = true,
				   bool predict = false);
        // unsigned long get_unperturbed_steps(bool to_apo = true,
	//				       bool predict = false);

        // int set_unperturbed_timestep(bool test_phase);
        real set_unperturbed_timestep(bool test_phase);
        // sint set_unperturbed_timestep(bool test_phase);
        // unsigned long set_unperturbed_timestep(bool test_phase);

        void update_dyn_from_kepler();
        void update_kepler_from_hdyn();
        void reinitialize_kepler_from_hdyn();
        void integrate_unperturbed_motion(real eta, real eps, real dt_max,
					  bool& reinitialize,
					  bool force_time = false);
        void print_unperturbed_binary_params();

        void clear_perturber_list()		{n_perturbers = 0;};
        void set_n_perturbers(int n) {
	    if (is_leaf())
		cerr << "warning: setting n_perturbers for leaf "
		     << format_label() << endl;
	    n_perturbers = n;
	}
        int get_n_perturbers()			{return n_perturbers;};
        hdyn ** get_perturber_list()		{return perturber_list;}

        void remove_perturber_list()		{n_perturbers = 0;
					         valid_perturbers = false;
					         if (perturber_list)
						     delete [] perturber_list;
						 perturber_list = NULL;}

        void print_perturber_list(ostream & s = cerr, char* pre = "");
        void find_print_perturber_list(ostream & s = cerr, char* pre = "");

        void new_perturber_list() {
	    if (perturber_list == NULL) {
		perturber_list = new hdyn *[MAX_PERTURBERS];
	    }
	}

        real get_perturbation_radius_factor()
		{return perturbation_radius_factor;}

        real get_perturbation_squared()		{return perturbation_squared;}
        void set_perturbation_squared(real p)	{perturbation_squared = p;}

	void create_low_level_perturber_list(hdyn* pnode, real pert_factor_sq);

        hdyn* check_merge_node();
        hdyn*  merge_nodes(hdyn * bcoll,real eps, real eta,
			   real max_timestep);

        void top_level_node_prologue_for_force_calculation(real eps2,
							   real pert_factor_sq,
							   bool exact);
        void top_level_node_real_force_calculation(real eps2);
        void top_level_node_epilogue_force_calculation(real eps2);

        bool nn_stats(real energy_cutoff, real kT,
		      vector center, bool verbose, int which = 0);
        real print_pert(int indent = BIN_INDENT);

	hdyn* next_node(hdyn*, bool resolve_unperturbed = true);

};

typedef hdyn * hdynptr; // to enable dynamic array declarations such as
			//    hdyn** hdyn_ptr_array = new hdynptr[n];
			// (note that the following expression is illegal:
			//    hdyn** hdyn_ptr_array = new (hdyn *)[n];)

inline  node * new_hdyn(hbpfp the_hbpfp, sbpfp the_sbpfp)
	{return (node *) new hdyn(the_hbpfp, the_sbpfp);}

inline hdyn * mkhdyn(int n, hbpfp the_hbpfp = new_hydrobase,
		     sbpfp the_sbpfp = new_starbase)
	{return (hdyn *) mk_flat_tree(n, new_hdyn, the_hbpfp, the_sbpfp);}


// Ugly and incomplete!  Clean up once everything is working.
// See hdyn/evolve/kira_init.C...

inline  hdyn * get_hdyn(istream & s, hbpfp the_hbpfp = new_hydrobase,
			sbpfp the_sbpfp = new_starbase)
{
    hdyn * h = (hdyn *) get_node(s, new_hdyn, the_hbpfp, the_sbpfp);

    // Unperturbed motion may be recognized by unperturbed_timestep > 0.
    // Fully inperturbed motion will also have fully_unperturbed = true.

    // Unperturbed binary (kepler) details are not saved, so reconstruct
    // them from the hdyn.

    for_all_nodes(hdyn, h, hh)
        if (hh->is_low_level_node()
	    && hh->get_unperturbed_timestep() > 0
	    && hh->get_elder_sister() == NULL) {

          //	    cerr << "Reestablishing unperturbed motion for "
          //		 << "(" << hh->format_label() << ",";
          //	    cerr << hh->get_binary_sister()->format_label() << ")\n";
          //	    PRL(hh->get_fully_unperturbed());

	  set_kepler_tolerance(1);
	  hh->reinitialize_kepler_from_hdyn();
      }

    return h;
}


void split_top_level_node(hdyn * bi, real eps, real eta,
			  real max_timestep);
void combine_top_level_nodes(hdyn * bj, hdyn * bi,
			     real eps, real eta, real max_timestep);
    
inline hdyn * common_ancestor(hdyn * bi, hdyn * bj)
	{return (hdyn*) common_ancestor((node*)bi, (node*)bj);}

#define  put_hdyn  put_node

#include  "hdyn_kepler.h"

// Declaration of member defined in hdyn_tt.C

// The second definition below, of hdyn_MF_ptr, should probably be local to
// the file hdyn_tt.C  -- we should look at this later.

typedef vector (hdyn::*hdyn_VMF_ptr)(void);   // vector member function pointer
typedef void (hdyn::*hdyn_MF_ptr)(const vector &);   // member function pointer

//  NOTE:
//  all the "hdyn" declarations below should be changed to "node" here
//  (just as was done with get_node() ).

void check_consistency_of_node(hdyn *, hdyn_VMF_ptr, char *);
void check_consistency_of_nodes(hdyn *);
void move_node(hdyn * node_to_move, hdyn * place_to_insert);
void move_node_by_index(int, int, hdyn *);
void create_binary_from_toplevel_nodes(hdyn *, hdyn *);

void pp3(hdyn *, ostream & s = cerr, int i = 0);
void pp3_minimal(hdyn *, ostream & s = cerr, int i = 0);
void pp3_tree(hdyn *, ostream & s = cerr, int i = 0);

vector something_relative_to_ancestor(hdyn*, hdyn*, hdyn_VMF_ptr);

// This is the same function as something_relative_to_root, but
// g++ versions after 2.7.0 complain about the ambiguity for pos,
// vel, and acc if we name it something_relative_to_root!
// This version is actually needed (and used) only for get_jerk.

// DEC C++ also complains about all uses of something_relative_to_root
// when hdyn member functions are involved...

vector hdyn_something_relative_to_root(hdyn*, hdyn_VMF_ptr);

void remove_node_and_correct_upto_ancestor(hdyn * ancestor, hdyn * node);

hdyn* node_to_move(hdyn*, real&);

void get_nodes_to_move(hdyn*, hdyn **, int&, real&);


#define  for_all_leaves_or_unperturbed(dyntype, base, node_name)              \
         for (dyntype* node_name = base;                                      \
	      node_name != NULL;                                              \
	      node_name = (dyntype*) node_name->next_node(base, false))       \
	      if (node_name->get_oldest_daughter() == NULL                    \
		  || (node_name->get_oldest_daughter()->get_kepler()	      \
		      && node_name->get_oldest_daughter()		      \
		      		  ->get_fully_unperturbed()))


void initialize_system_phase1(hdyn*, real);
void set_harp_debug(bool check);

void predict_loworder_all(hdyn*, real);
void update_binary_sister(hdyn* bi);
void correct_acc_and_jerk(hdyn *, real, bool &, bool);

void correct_leaf_for_change_of_mass(hdyn * node, real dm);

void correct_leaf_for_change_of_vector(hdyn * node,
				       vector d_something,
				       hdyn_VMF_ptr get_something,
				       hdyn_MF_ptr inc_something);

void synchronize_tree(hdyn * b, real eps, real eta, real max_timestep);

void correct_perturber_lists(hdyn * b, hdyn ** list, int n, hdyn * cm = NULL);
void expand_perturber_lists(hdyn * b, hdyn * bb, bool verbose = false);

void harp_calculate_acc_and_jerk(hdyn ** next_nodes,
				 int n_next,
				 real eps2,
				 real pert_factor_sq,
				 real time,
				 bool restart);
void harp_calculate_densities(hdyn* root, real h2_crit = 1);

void fast_get_nodes_to_move(hdyn * b,
			    hdyn * list[],
			    int  & nlist,
			    real & tmin,
			    bool & reset);
void dump_node_list(int n = 1000000000);

// From kira.C:

void set_n_top_level(hdyn* b);
int get_n_top_level();

void set_n_check(int n);
bool kira_use_harp();
void kira_set_harp_cpu(real cpu);
void kira_calculate_energies(hdyn* b, real eps2,
			     real& epot, real& ekin, real& etot);
void kira_calculate_top_level_acc_and_jerk(hdyn ** next_nodes,
					   int n_next,
					   real eps2,
					   real gamma23,
					   real time,
					   bool & restart_harp);
void compute_densities(hdyn* b, vector& cod_pos, vector& cod_vel);

// From kira_escape.C:

void check_and_remove_escapers(hdyn* b,
			       real scaled_stripping_radius,
			       real eps, real eta, real dt_max,
			       real initial_step_limit,
			       real& ttmp, hdyn** next_nodes,
			       int& n_next, bool& tree_changed);

// From reset_counters.C:

void reset_counters(hdyn* bi);

// From kira_log.C:

void print_counters(kira_counters* kc, kira_counters* kc_prev = NULL);
void update_cpu_counters(hdyn * b);
void print_statistics(hdyn* b);
void log_output(hdyn * b, real steps, real eps, kira_counters* kc_prev);
void snap_output(hdyn * b, real steps, int& snaps,
		 bool reg_snap, bool last_snap,
		 char * snap_save_file,
		 real t, real ttmp, real t_end,
		 real& t_snap, real dt_snap, bool verbose);

// From kira_ev.C:

void calculate_acc_and_jerk_for_list(hdyn * b,
				     real eps, 
				     hdyn ** next_nodes,
				     int n_next,
				     real time,
				     bool exact,
				     bool tree_changed,
				     bool & reset_force_correction,
				     bool & restart_harp,
				     bool check_all = false);

void calculate_acc_and_jerk_on_all_top_level_nodes(hdyn * b,
						   real eps);

void initialize_system_phase2(hdyn * b,
			      real eps,
			      real eta,
			      real initial_step_limit,
			      bool set_dt = true);

// From kira_energy.C:

void calculate_energies_with_tidal(hdyn* b, real eps2,
				   real& epot, real& ekin, real& etot);

void print_recalculated_energies(hdyn* b, real eps2,
				 bool print_dde = false,
				 bool save_story = false);

void calculate_energies_by_harp3(hdyn * b,
				 real eps2,
				 real &epot,
				 real &ekin,
				 real & etot);

void print_recalculated_energies_by_harp3(hdyn* b,
					  int mode,
					  real eps2,
					  real e_corr);

void add_tidal(hdyn * b, bool pot_only = false);
real de_tidal_pot(hdyn * b);

void print_nn(hdyn* b, int level = 0, ostream& s = cerr);
void print_coll(hdyn* b, int level = 0, ostream& s = cerr);

// From kira_init.C:

real total_mass(hdyn * b);
bool kira_initialize(int argc, char** argv,
		     hdynptr& b,	// hdyn root node
		     real& delta_t,	// time span of the integration
		     real& dt_log,	// output interval--power of 2
		     real& dt_snap,	// snap output interval
		     real& dt_sstar,	// stellar evolution timestep
		     real& dt_esc,	// escaper removal
		     real& dt_reinit,	// reinitialization interval
		     real& eta,		// time step parameter
		     real& eps,		// softening length
		     real& d_min_sq,	// N^2 * node combine threshhold
		     real& lag_factor,	// node hysteresis factor
		     real& gamma2,	// unperturbed threshhold
		     bool& exact,	// no perturber list if true
		     real& scaled_stripping_radius, // stripping radius
		     // for unit mass
		     bool& verbose,
		     bool& save_last_snap,
		     char* snap_save_file,
		     int& n_stop);	// n to terminate simulation

// From kira_runtime.C:

void clean_up_files();
bool check_file(char* name, bool del = true);
void check_runtime(hdyn* b,
		   real eps, real& eta, real& lag_factor, real& t_end,
		   bool& tree_changed);

// From kira_stellar.C:

bool evolve_stars(hdyn* b, real eps, real eta,
		  real initial_step_limit,
		  real dt_log, real dt_sstar);

// From kira_unpert.C:

void print_unperturbed_options();
void toggle_unperturbed(int level);

#endif

//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\
//  |  the end of:  |         /|\         |  inc/hdyn.h
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\


