
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

#include "hdyn.h"

// Remove escaping or destroyed nodes from all perturber lists.
//
// Externally visible functions:
//
//	void correct_perturber_lists


local void check_and_correct_perturbers(hdyn* bi, hdyn ** ilist, int n,
					hdyn * cm)
{
    if (bi->get_perturber_list() == NULL) {
	cerr << endl << "check_and_correct_perturbers: "
	     << bi->format_label()
	     << " has valid_perturbers but no perturber_list" << endl;
	return;
    }

    hdynptr* plist = bi->get_perturber_list();		// convenient shorthand

    bool first = true;
    int null_count = 0;

    for (int i = 0; i < bi->get_n_perturbers(); i++) {
	hdyn* p = plist[i];
	for (int j = 0; j < n; j++) {
	    if (ilist[j] == p) {

		// Replace first plist member found to be on ilist by cm
		// (which may be NULL).  Set all others NULL.

		if (first)
		    plist[i] = cm;
		else
		    plist[i] = NULL;

		first = false;
		if (plist[i] == NULL) null_count++;

		break;		    
	    }
	}
    }

    // Do a second pass through the list to remove NULLs.

    if (null_count > 0) {
	int offset = 0;
	for (int i = 0; i < bi->get_n_perturbers(); i++) {
	    if (offset > 0) plist[i-offset] = plist[i];
	    if (plist[i] == NULL) offset++;
	}
	bi->set_n_perturbers(bi->get_n_perturbers()-null_count);
    }
}

// correct_perturber_lists:  Remove all nodes on the list from any perturber
//			     list that may contain them, replacing the first
//			     by cm, if non-NULL.

void correct_perturber_lists(hdyn * b, hdyn ** list, int n, hdyn * cm)
{
    if (n <= 0) return;

    cerr << endl << "Entering correct_perturber_lists,  ";
    PRL(ALLOW_LOW_LEVEL_PERTURBERS);

    cerr << "removing nodes:  ";
    for (int i = 0; i < n; i++) cerr << list[i]->format_label() << "  ";
    cerr << endl;
    cerr << "cm = ";
    if (cm)
	cerr << cm->format_label() << endl;
    else
	cerr << "NULL" << endl;
    cerr << flush;

    if (ALLOW_LOW_LEVEL_PERTURBERS) {
	for_all_nodes(hdyn, b, bi)
	    if (bi->get_valid_perturbers())
		check_and_correct_perturbers(bi, list, n, cm);
    } else {
	for_all_daughters(hdyn, b, bi)
	    if (bi->get_valid_perturbers())
		check_and_correct_perturbers(bi, list, n, cm);
    }

    cerr << "Leaving correct_perturber_lists." << endl << flush;
}

local int expand_perturber_list(hdyn* bi, hdyn* bb)
{
    if (bi->get_perturber_list() == NULL) {
	cerr << endl << "expand__perturber_list: "
	     << bi->format_label()
	     << " has valid_perturbers but no perturber_list" << endl;
	return 0;
    }

    hdynptr* plist = bi->get_perturber_list();		// convenient shorthand
    int np = bi->get_n_perturbers();

    for (int i = 0; i < np; i++) {
	hdyn* p = plist[i];
	if (plist[i] == bb) {

	    if (np >= MAX_PERTURBERS - 1) {

		// List will overflow if expanded.  Make it invalid.

		bi->set_valid_perturbers(false);

	    } else {

		// Remove bb from the list...

		for (int j = i+1; j < np; j++)
		    plist[j-1] = plist[j];
		np--;

		// ...then add its components to the end.

		for_all_leaves(hdyn, bb, bbb)
		    plist[np++] = bbb;

		bi->set_n_perturbers(np);
	    }

	    return 1;
	}
    }
    return 0;
}

// expand_perturber_lists:  Expand the specified node into components on all
//			    perturber lists that contain it.

void expand_perturber_lists(hdyn* b, hdyn* bb, bool verbose)
{
    int np = 0;
    if (ALLOW_LOW_LEVEL_PERTURBERS) {
	for_all_nodes(hdyn, b, bi)
	    if (bi != bb && bi->get_valid_perturbers())
		np += expand_perturber_list(bi, bb);
    } else {
	for_all_daughters(hdyn, b, bi)
	    if (bi != bb && bi->get_valid_perturbers())
		np += expand_perturber_list(bi, bb);
    }

    if (verbose && np > 0)
	cerr << "Expanded " << np << " perturber lists to resolve "
	     << bb->format_label() << endl;
}
