
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//  hdyn_ev.C: functions related to orbit integration within the hdyn class.
//.............................................................................
//    version 1:  Nov 1994   Piet Hut, Jun Makino, Steve McMillan
//    version 2:
//.............................................................................
//  This file includes:
//    - node synchronization functions
//    - functions related to time steps
//    - function for finding the next node to integrate the orbit of
//    - functions for initialization
//    - predictor and corrector steps for the Hermite integration scheme
//    - functions for calculating acceleration & jerk, for two given nodes
//    - functions for determining which nodes should exert forces on each other
//    - driver function for a one-time-step integration of a node
//.............................................................................

//  More memo...
//
//  The procedure to determine nn/coll is somewhat sorted out. For nn,
//  the following functions clear nn and d_nn_sq as initialization
//  --- flat_calculate_acc_and_jerk
//  --- perturber_acc_and_jerk
//
//  The following functions do not clear
//  --- calculate_partial_acc_and_jerk(_on_leaf)
//  --- tree_walk_for_...
//  These functions still update nn and d_nn_sq (which are supplied as
//  arguments, not the class members)
//  Thus, if these functions are called to determine nn without preceding
//  call to (e.g.)  flat_calculate_acc_and_jerk, d_nn_sq must be initialized
//  to some large number.
//
//  The pointer coll and  d_coll_sq are initialized in the following three
//  functions. 
//   --- flat_calculate_acc_and_jerk
//   --- perturber_acc_and_jerk
//   --- calculate_acc_and_jerk (The top-level routine for force calculation)
//   In practice, it might be okay to initialize only at the top-level routine.
//   The pointer coll is different from nn not just by the definition but
//   because of the fact that it is meaningfully defined only for leafs
//   (physical particles) now. Because of this difference, coll is
//   changed always in place, without propagating the pointer to nodes in
//   higher levels.
//   This means that the pointer coll of a leaf can be overwritten during
//   the force calculation of a CM node. This cannot cause any problem, I
//   believe. 

// MEMO 8 Aug 1996
//
// collision radius is calculated only for leaves (physical particles)
//
// the actual quantity stored in d_coll_sq is
// rij^2 - (d_i + d_2)^2
//

// MEMO from JM (23 Dec 1995):
//
// Changes over the last few days:
//
// calculate_partial_acc_and_jerk(_on_leaf) has changed
// interface (and algorithm!!!).  It used to take two
// node pointers, top and mask.  Top was assumed to be
// the ancestor of the mask, and mask the ancestor of this.
// It calculated the force from all leaves under top except
// for those under mask.
//
// Now it takes three arguments, top, common and mask.  Common
// must be the common ancestor of top and mask (common = top is OK)
// and the routine calculates the force from the subtree under top.
//
// This change was necessary to implement perturbation correction.

//.........................................................................

// NOTE from Steve to Steve, 7/2/97...
//
// Perturber lists are associated with TOP-LEVEL nodes.
// Kepler pointers are associated with the elder binary component...
// Perturbations   are associated with the binary components...
//
// Perturber lists are recomputed at each top-level CM step
//
// Problem: a binary in a deeply nested hierarchy may be handled
//	    very inefficiently, as it uses the top-level perturber
//	    list, which may be quite large.

// ****  Perturber lists added to all CM nodes (Steve, 8/98).  ****
// ****  Updated at CM step, based on parent perturber list.   ****
// ****  Binary component perturber lists are INDEPENDENT.     ****

// #define ALLOW_LOW_LEVEL_PERTURBERS		true	// now defined
							// in hdyn.h

// Also, no longer resolve an unperturbed binary into components for
// purposes of computing its center of mass motion, and optionally allow
// unperturbed centers of mass to remain on a perturber list without
// resolving them into their components.  (See the ~warning below.)

// #define RESOLVE_UNPERTURBED_PERTURBERS	false	// now defined
							// in hdyn.h

//.........................................................................

#include "hdyn.h"
#include "hdyn_inline.C"
#include "kira_debug.h"

#define INITIAL_STEP_LIMIT	0.0625	// Arbitrary limit on the first step
#define USE_POINT_MASS		true

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define DEBUG_EV false			// General debugging flag
//
//#define EV_FUNCTION_ID false		// Force-evaluation functions will
//					// identify themselves if set
//
//#define HARP_DEBUG	true		// Check for GRAPE-related problems

static bool dbg = false;
static bool harp_debug = HARP_DEBUG;

// Accessor function:

void set_harp_debug(bool check)
{
    harp_debug = check;
}


//=============================================================================
//  node synchronization functions
//=============================================================================

//-----------------------------------------------------------------------------
// update_binary_sister: make sure that the companion of binary component
//			 bi is consistent by updating its time and mirroring
//			 bi's pos, vel, acc, and jerk.
//-----------------------------------------------------------------------------

void update_binary_sister(hdyn * bi)
{
    hdyn *sister = bi->get_binary_sister();

    sister->set_time(bi->get_time());
    sister->set_timestep(bi->get_timestep());

    real factor = -bi->get_mass() / sister->get_mass();
    sister->set_pos(factor * bi->get_pos());
    sister->set_vel(factor * bi->get_vel());
    sister->set_acc(factor * bi->get_acc());

    // Note from Steve, 10/12/98: It is apparently possible for an
    // unperturbed binary component to have jerk = Infinity if it
    // has never taken a perturbed step, and this can lead to overflow
    // here.  Check for this explicitly.  (Not entirely clear how the
    // infinity arises...  May be HARP overflow.)

    real j0 = bi->get_jerk()[0];

    if (j0 < VERY_LARGE_NUMBER && j0 > -VERY_LARGE_NUMBER)
	sister->set_jerk(factor * bi->get_jerk());
    else {
	cerr << "update_binary_sister():  setting jerk = 0 for "
	     << bi->format_label() << " at time " << bi->get_system_time()
	     << endl;
	bi->set_jerk(vector(0,0,0));
	sister->set_jerk(vector(0,0,0));
    }

    sister->set_pot(-factor * bi->get_pot());
    sister->store_old_force();

    sister->set_perturbation_squared(bi->get_perturbation_squared());
}

//-----------------------------------------------------------------------------
// synchronize_node -- force the time of a node to a specified value,
//                     by integrating it forward in time to system_time.
//-----------------------------------------------------------------------------

void hdyn::synchronize_node(real eps,		// softening length
			    real eta,		// accuracy parameter
			    real max_timestep)	// maximum time step
{
    if (DEBUG_EV)
	cerr << "Enter synchronize_node for " << format_label() << endl;

    // Note: as currently coded, does NOT take external tidal fields
    // into account.  This is OK for low-level nodes so long as the
    // tidal field is applied only at the top level (see dstar_kira.C).

    if (time == system_time) return;

    if (is_low_level_node()
	&& (get_parent()->get_oldest_daughter() != this)) {
	if (DEBUG_EV)
	    cerr << "Low-level node: "<<format_label()<<endl<<flush;
	get_elder_sister()->synchronize_node(eps, eta, max_timestep);
	update_binary_sister(get_elder_sister());
	return;
    }

    real old_timestep = timestep;
    timestep = system_time - time;

    if (timestep < 0.0) {

	if (DEBUG_EV)
	    cerr << "synchronize_node: negative timestep..." << system_time
		 << " " << time << endl;
	put_node(cerr, *this);
	exit(-1);
    }

    if (DEBUG_EV) {
	cerr << "Synchronize node " << format_label()
	     << " at time " << system_time << endl
	     <<" ...before integrate_node " << endl << flush;
	if (is_low_level_node()) pp3(get_parent());
	cerr << flush;
    }

    //cerr << "Energy before "; print_recalculated_energies(get_root(),
    //							    1, eps*eps);

    // NOTE: the "false" here means that we do NOT attempt to
    //	     synchronize unperturbed binaries...

    integrate_node(get_root(), eps, eta, max_timestep, gamma2, false, true);

    // cerr << "Energy after "; print_recalculated_energies(get_root(),
    // 							    1, eps*eps);

    if (DEBUG_EV)
	cerr <<"After integrate_node "<<endl;

    // NOTE: Timestep adjustment below is not clean...

    while (fmod(time, old_timestep) != 0.0)
	old_timestep *= 0.5;
    timestep = old_timestep;

    if (is_low_level_node())
	update_binary_sister(this);

    if (DEBUG_EV)
	cerr << "Leave synchronize_node for " << format_label() << endl;
}

//=============================================================================
//  functions related to time steps
//=============================================================================

//-----------------------------------------------------------------------------
// set_first_timestep -- calculate the first timestep using acc, jerk,
//                       and pot (acceleration, jerk, potential).
//                       The expression used here is:
//                                                 ________
//               1              /   | acc |       V | pot |`  \
//   timestep = ---- eta * min (   ---------  ,  -----------   )
//               32             \   | jerk |       | acc |    /
//
//  The first expression gives the time scale on which the acceleration is
//  changing significantly.
//  The second expression gives an estimate for the free fall time
//  in the case of near-zero initial velocity.
//-----------------------------------------------------------------------------

void hdyn::set_first_timestep(real eta,		// accuracy parameter
			      real max_dt)	// maximum time step size
{
    real eta_init = eta / 32;	// initial accuracy parameter

    real j2 = jerk * jerk;
    real a2 = acc * acc;

    real dt_adot = eta_init/4;	// time step based on rate of change of acc
    real dt_ff   = eta_init/4;	// time step based on free-fall time
    real dt;			// minimum of the two

    if (j2 > 0)
	dt_adot = eta_init * sqrt(a2 / j2);

    if (pot < 0 && a2 > 0)
	dt_ff = eta_init * sqrt(-pot / a2);

    dt = min(dt_adot, dt_ff);

    // Apply an upper limit to the time step:

    dt = min(dt, eta_init/4);	// eta_init/4 is arbitrary limit...

    timestep = adjust_number_to_power(dt, max_dt);

    if (time != 0)
	while (fmod(time, timestep) != 0)
	    timestep *= 0.5;

    real tnext = time + timestep;

    if (timestep == 0 || time == tnext) {

	cerr << endl << "set_first_timestep: dt = 0 at time "
	     << get_system_time() << endl;
	PRC(dt_adot); PRL(dt_ff);
	PRC(a2); PRC(j2); PRL(pot);
	pp3(this);

	cerr << endl << endl << "System dump:" << endl << endl;

	pp3(get_root());
	exit(0);
    }
}


//-----------------------------------------------------------------------------
// new_timestep:  Calculate the next timestep following the Aarseth
//                formula (Aarseth 1985), including necessary adjustments
//                for the hierarchical timestep scheme.
//-----------------------------------------------------------------------------

real new_timestep(vector & at3,		// 3rd order term
		  vector & bt2,		// 2nd order term
		  vector & jerk,	// 1st order term
		  vector & acc,		// 0th order term
		  real timestep,	// old timestep
		  real &time,		// present time
		  real eta,		// accuracy parameter
		  real dtmax,		// maximum stepsize
		  real correction_factor,
		  hdyn * b,
		  real pert_sq)
{
    if (timestep == 0)
	cerr << "new_timestep: old timestep = 0\n";

    real dt2 = timestep * timestep;
    real dt4 = dt2 * dt2;
    real dt6 = dt4 * dt2;
    real a3scaled2 = 36 * at3 * at3 / dt6;
    real a2scaled2 = 4 * bt2 * bt2 / dt4;
    real a1scaled2 = jerk * jerk;
    real a2 = acc * acc;

    real newstep;

    // Simple criterion:
    // real newstep = eta * sqrt(sqrt(a2 / a2scaled2));

    // Aarseth criterion:

    real aarsethstep = 0;

    if (harp_debug) {

	// Numerator:

	real sqrt1 = 0;
	real tmp1 = a2 * a2scaled2;
	if (tmp1 > 0 && tmp1 < VERY_LARGE_NUMBER)
	    sqrt1 = sqrt(tmp1);
	real sqrt2 = 0;
	real tmp2 = sqrt1 + a1scaled2;

	// Denominator:

	real sqrt3 = 0;
	real tmp3 = a1scaled2 * a3scaled2;
	if (tmp3 > 0 && tmp3 < VERY_LARGE_NUMBER)
	    sqrt3 = sqrt(tmp3);
	real sqrt4 = 0;
	real tmp4 = sqrt3 + a2scaled2;

	if (tmp2 > 0 && tmp2 < VERY_LARGE_NUMBER
	    && tmp4 > 0 && tmp4 < VERY_LARGE_NUMBER)
	    aarsethstep = eta * sqrt(tmp2/tmp4);

	if (tmp1 >= 0 && tmp2 >= 0 && tmp3 >= 0 && tmp4 >= 0
	    && tmp1 < VERY_LARGE_NUMBER
	    && tmp2 < VERY_LARGE_NUMBER
	    && tmp3 < VERY_LARGE_NUMBER
	    && tmp4 < VERY_LARGE_NUMBER) {	// test negation to catch NaN,
						// which always tests false
	} else {

	    cerr << endl << "new_timestep:  found errors at time "
		 << b->get_system_time() << endl;
	    PRC(tmp1); PRC(tmp2); PRC(tmp3); PRL(tmp4);
	    PRC(sqrt2); PRL(sqrt4);
	    pp3(b);

	    cerr << endl << endl << "System dump:" << endl << endl;

	    pp3(b->get_root());
	    exit(0);
	}

#if 0
	real true_step = eta * sqrt((sqrt(a2 * a2scaled2) + a1scaled2)
				  / (sqrt(a1scaled2 * a3scaled2) + a2scaled2));
	if (abs(aarsethstep/true_step - 1) > 1.e-6) {
	    PRC(aarsethstep); PRL(true_step);
	}
#endif

    } else

	aarsethstep = eta * sqrt((sqrt(a2 * a2scaled2) + a1scaled2)
				  / (sqrt(a1scaled2 * a3scaled2) + a2scaled2));

#if 0
    if (b->get_time() > 0) {
	PRI(4); PRC(b->get_index()), PRL(aarsethstep);
	PRI(4); PRC(a1scaled2), PRC(a2scaled2), PRL(a3scaled2);
    }
#endif

    newstep = aarsethstep * correction_factor;

#if 0
    if (b->is_low_level_node() && pert_sq < 0.0001) {

	// Use a different time step criterion if b is a binary and
	// the perturbation is small

	real dist = abs(b->get_pos() - b->get_younger_sister()->get_pos());
	real dtff = sqrt(dist*dist*dist / (2*b->get_parent()->get_mass()));
	real dtv  = sqrt(square(b->get_pos()) / square(b->get_vel()));

	newstep = eta*correction_factor*min(dtff,dtv);

	// PRC(dist); PRC(square(b->get_vel())); PRC(dtff); PRL(dtv);
    }
#endif

    // The natural time step is newstep.  Force it into an appropriate
    // power-of-two block.  A halving of timestep is always OK.  To
    // preserve the synchronization, a doubling is OK only if the
    // current time could be reached by an integral number of doubled
    // time steps (see use of fmod below).

    if (newstep < timestep)
	return 0.5 * timestep;
    else if (newstep < 2 * timestep)
	return timestep;
    else if (fmod(time, 2 * timestep) != 0.0 || 2 * timestep > dtmax)
	return timestep;
    else {

	// Added by Steve 7/28/98 to deal with pathological ICs from SPZ...
	// Do not double if this is a strongly perturbed center of mass node...

	if (b->is_leaf()
	    || b->get_oldest_daughter()->get_perturbation_squared() < 1)
	    return 2 * timestep;
	else
	    return timestep;
    }
}

//-----------------------------------------------------------------------------
// update:  Update the time and the timestep.
//-----------------------------------------------------------------------------

void hdyn::update(const real eta,
		  const real dtmax)
{
    vector at3 = 2 * (old_acc - acc) + timestep * (old_jerk + jerk);
    vector bt2 = -3 * (old_acc - acc) - timestep * (2 * old_jerk + jerk);

    // Reminder:	at3  =  a''' dt^3 / 6		(a' = j)
    //			bt2  =  a''  dt^2 / 2

    time += timestep;

#if 0
    if (time > 0) {
	PRI(4); PRC(index); PRL(timestep);
	PRI(4); PRL(old_acc);
	PRI(4); PRL(acc);
	PRI(4); PRL(old_jerk);
	PRI(4); PRL(jerk);
	PRI(4); PRL((acc-old_acc)/timestep);
	PRI(4); PRL(2*bt2/(timestep*timestep));
	PRI(4); PRL((jerk-old_jerk)/timestep);
	PRI(4); PRL(6*at3/(timestep*timestep*timestep));
    }
#endif

    // Define correction factor for use in new_timestep.

    real correction_factor = 1;

    if (is_low_level_node()) {

	real m = mass/mbar;
	real pot_sq = m*m*d_min_sq/(pos*pos);

	// correction_factor = pow(pot_sq, -0.1);

	if (pot_sq > 1)
	    correction_factor = pow(pot_sq, -0.125);

	// Steve 8/98:  1. 0.125 here is conservative -- expect *cumulative*
	//		   energy error to be ~fourth order.
	//		2. Could rewrite to replace pow() if necessary...
	//		   *** see kira_approx.C ***

	// correction_factor = pow(pos*pos/d_min_sq, 0.025);
	// if (correction_factor > 1.0) correction_factor = 1;
	// if (correction_factor < 0.2) correction_factor = 0.2;

    }

    timestep = new_timestep(at3, bt2, jerk, acc, timestep, time, eta, dtmax,
			    correction_factor, this, perturbation_squared);

    if (timestep <= 0) {
	cerr << "update:  dt = " << timestep
	     << " at time " << get_system_time();
	pp3(this,cerr);
    }
}

//=============================================================================
//  function for finding the next node to integrate
//=============================================================================

//-----------------------------------------------------------------------------
// node_to_move:  Return the node with minimum t+dt.
//-----------------------------------------------------------------------------

hdyn *node_to_move(hdyn * b,
		   real & tmin)
{
    hdyn *bmin = NULL;
    if (b->is_parent())
	for_all_daughters(hdyn, b, d) {
	    hdyn *btmp;
	    real ttmp = VERY_LARGE_NUMBER;
	    btmp = node_to_move(d, ttmp);
	    if (ttmp < tmin) {
		tmin = ttmp;
		bmin = btmp;
	    }
	}

    if (b->get_parent() != NULL)
	if (tmin > b->get_next_time()) {
	    bmin = b;
	    tmin = b->get_next_time();
	}
    return bmin;
}

//-----------------------------------------------------------------------------
// get_nodes_to_move:  Return the nodes with minimum t+dt.  Use a recursive
//		       loop to traverse the system.
//-----------------------------------------------------------------------------

void get_nodes_to_move(hdyn * b,
		       hdyn * list[],
		       int &nlist,
		       real & tmin)
{
    hdyn *bmin = NULL;

    if (b->get_parent() == NULL) {			// b is root: initialize
	nlist = 0;
	tmin = VERY_LARGE_NUMBER;
    }

    if (b->is_parent())
	for_all_daughters(hdyn, b, d) {
	    get_nodes_to_move(d, list, nlist, tmin);
	}

    if (b->get_parent() != NULL) {

	if (tmin > b->get_next_time()) {
	    nlist = 1;
	    tmin = b->get_next_time();
	    list[0] = b;

	} else if (tmin == b->get_next_time()) {	// NOTE: we are assuming
	    list[nlist] = b;				// that equality is OK
	    nlist++;					// here because all
	    if (b->is_low_level_node()) {		// times are powers of 2
		hdyn *bs = b->get_binary_sister();
		for (int j = nlist - 2; j >= 0; j--) {
		    if (list[j] == bs) {
			nlist--;
			j = -1;
		    }
		}
	    }
	}
    }
}

//=============================================================================
//  functions for initialization
//=============================================================================

//-----------------------------------------------------------------------------
// initialize_system_phase1:  Set current time and "predicted" position
//                            and velocity for all nodes.
//-----------------------------------------------------------------------------

void initialize_system_phase1(hdyn * b,  real t)
{
    if (b->get_oldest_daughter() != NULL)
	for_all_daughters(hdyn, b, d)
	    initialize_system_phase1(d, t);

    if ((b->get_kepler() == NULL)
	  &&(b->get_unperturbed_timestep() <= 0)) {
	b->set_time(t);
	b->set_system_time(t);
	b->init_pred();
    }
}

//=============================================================================
//  Predictor and corrector steps for the Hermite integration scheme
//=============================================================================

//-----------------------------------------------------------------------------
// predict_loworder_all:  Predict the positions of all nodes at time t.
//-----------------------------------------------------------------------------

void predict_loworder_all(hdyn * b,
			  real t)
{
    if (!b) return;			// shouldn't happen -- flag as error?

    if (b->is_parent())
	for_all_daughters(hdyn, b, d)
	    predict_loworder_all(d, t);

    b->predict_loworder(t);
}

//-----------------------------------------------------------------------------
// correct:  Apply the corrector for 3rd-order hermite scheme.
//           See Makino & Aarseth (PASJ 1992) for derivation.
//-----------------------------------------------------------------------------

#define ONE12 0.0833333333333333333333
#define ONE3  0.3333333333333333333333

bool hdyn::correct()
{

#if 0
    if (is_low_level_node()) {
	cerr << "old_acc  " << old_acc << endl;
	cerr << "    acc  " << acc << endl;
	cerr << "old_jerk " << old_jerk << endl;
	cerr << "    jerk " << jerk << endl;
	cerr << " dt " << timestep << endl;
	cerr << "pos " << pos << endl;
	cerr << "vel " << vel << endl;
	cerr << "pos " << pred_pos << endl;
	cerr << "vel " << pred_vel << endl;
	cerr << "pos " << get_younger_sister()->pos << endl;
	cerr << "vel " << get_younger_sister()->vel << endl;
	cerr << "pos " << get_younger_sister()->pred_pos << endl;
	cerr << "vel " << get_younger_sister()->pred_vel << endl;
    }
#endif

    vector at3 = 2 * (old_acc - acc) + timestep * (old_jerk + jerk);
    vector bt2 = -3 * (old_acc - acc) - timestep * (2 * old_jerk + jerk);
    real dt2 = timestep * timestep;

    // Reminder:	at3  =  a''' dt^3 / 6		(a' = j)
    //			bt2  =  a''  dt^2 / 2

    real old_v = 0;
    if (harp_debug)
	old_v = abs(vel);

    vector new_pos = pred_pos + (0.05 * at3 + ONE12 * bt2) * dt2;
    vector new_vel = pred_vel + (0.25 * at3 + ONE3 * bt2) * timestep;

    if (harp_debug && old_v > 0) {
	real new_v = abs(new_vel);

	if (new_v/old_v > 1000 || new_v > 1.e6) {   // numbers are arbitrary

	    // Possible runaway -- speed has increased by a large factor.

	    cerr << endl << "correct:  possible runaway at time "
		 << get_system_time() << endl;

	    cerr << endl << "pp3 with old pos and vel:" << endl;
	    pp3(this);

	    PRL(pred_pos);
	    PRL(pred_vel);
	    PRL(new_pos);
	    PRL(new_vel);

	    // cerr << endl << endl << "System dump:" << endl << endl;
	    // pp3(get_root());

	    // Options:		quit
	    //			restart
	    //			flag and continue
	    //			discard new_pos and new_vel, retain
	    //			    old acc and jerk and continue
	    //			flag, recompute acc and jerk, and continue  <--

	    cerr << endl << endl;
	    // exit(0);

	    // Flag the problem.

	    char tmp[128];
	    sprintf(tmp, "runaway in correct at time %f", time);
	    log_comment(tmp);

	    int n_runaway = 0;
	    if (find_qmatch(get_log_story(), "n_runaway"))
		n_runaway = getiq(get_log_story(), "n_runaway");

	    n_runaway++;
	    PRL(n_runaway);

	    if (n_runaway > 10)
		exit(0);

	    putiq(get_log_story(), "n_runaway", n_runaway);

	    return false;	// trigger a retry on return
	}
    }

    pos = new_pos;
    vel = new_vel;

    return true;		// normal return value
}


//
// Some notes on the force-evaluation functions (Steve, 7/98).
// ----------------------------------------------------------
//
// Functions (note: "force" <--> "acc, jerk, and potential"):
// 
// accumulate_acc_and_jerk
// 	compute force due to a specified particle 
// 
// hdyn::flat_calculate_acc_and_jerk
// 	compute force on top-level node 'this' due to all other
// 	top-level nodes in the system (masking allowed)
// 
// hdyn::perturber_acc_and_jerk_on_leaf
// 	compute force on 'this' leaf due to its top-level
// 	perturber list
// 
// hdyn::tree_walk_for_partial_acc_and_jerk_on_leaf
// 	accumulate force on 'this' leaf while traversing the portion
// 	of the tree under a specified node, excluding everything
// 	under a mask node.  Currently, the descent STOPS at an
// 	unperturbed center of mass.
// 
// hdyn::calculate_partial_acc_and_jerk_on_leaf
// 	compute the force on `this' leaf from all particles under
// 	a specified node, excluding everything under a mask node
// 
// hdyn::calculate_partial_acc_and_jerk
// 	compute the force on `this' node or leaf from all particles
// 	under a specified node, excluding everything under a mask node
// 
// hdyn::calculate_acc_and_jerk_on_low_level_node
// 	compute the force on `this' low-level node or leaf
// 
// hdyn::calculate_acc_and_jerk_on_top_level_node
// 	compute the force on `this' top-level node or leaf
// 
// hdyn::calculate_acc_and_jerk
// 	compute the force on `this' node or leaf
// 
// correct_acc_and_jerk
// 	correct computed force in situations where GRAPE-style functions
// 	lead to incorrect results
//
//
// To allow for efficient utilization of the GRAPE hardware from kira,
// function calculate_acc_and_jerk_on_top_level_node is divided into
// calls to the following functions:
//
// hdyn::top_level_node_prologue_for_force_calculation
// 	performs entire calculation in exact case; setup otherwise
// hdyn::top_level_node_real_force_calculation
// 	top-level force calculation only
// 	    calls flat_calculate_acc_and_jerk
// 	REPLACED by harp_calculate_acc_and_jerk in kira if HARP flag set
// hdyn::top_level_node_epilogue_force_calculation
// 	performs remainder (non-GRAPE) of force calculation
// 	    calls calculate_partial_acc_and_jerk
//

// 
// Calling sequences (schematic, "exact" = false)
// 
//                               calculate_acc_and_jerk
// 
//                                          |
//                                          |
//                     ------------------------------------------
//                    |                                          |
//                    v                                          v
// 
// calculate_acc_and_jerk_on_top_level_node   calculate_acc_and_jerk_on_low_level_node
//
//                    |                                          |
//                    v                                          v
// 
//     flat_calculate_acc_and_jerk               calculate_partial_acc_and_jerk
//       (GRAPE-style, or GRAPE itself)            (relative motion, perturbation)
//                  +
//     calculate_partial_acc_and_jerk                            |
//       (perturbations, etc.)                                   |
//                                                               |
//                   |                                           |
//                    -------------------------------------------
//                                          |
//                                          |
//                                          v
// 
//                           (calculate_partial_acc_and_jerk)
// 
//                                          |
//                                          |
//                    -------------------------------------------
//                   |                                           |
//                   v                                           v
// 
// calculate_partial_acc_and_jerk_on_leaf        calculate_partial_acc_and_jerk
//                                                 (RECURSIVE, for acc and jerk
//                   |                                         on a node)
//                   |
//                   |-------------------------------------------
//                   |                                           |
//                   v                                           v
// 
//     perturber_acc_and_jerk_on_leaf       tree_walk_for_partial_acc_and_jerk_on_leaf
//       (if perturber list used)
//                                                               |
//                   |-------------------------------------------
//                   |                                           |
//                   v                                           v
// 
// 	accumulate_acc_and_jerk             tree_walk_for_partial_acc_and_jerk_on_leaf
//                                            (RECURSIVE, if source node is not
//                                                        a leaf)
//


//=============================================================================
//  functions for calculating acceleration & jerk, for given nodes this and b
//=============================================================================

//-----------------------------------------------------------------------------
// accumulate_acc_and_jerk:  Accumulate acc, jerk, and potential
//           from particle b.  Note that the relative position and
//           velocity of b relative to the calling particle are
//           calculated by the caller and passed as arguments.
//           The only reason to pass b here is to access its mass.
//
//	     Note: this function is only referenced from three places:
//
//		flat_calculate_acc_and_jerk()			--> direct
//		perturber_acc_and_jerk_on_leaf()		--> indirect
//		tree_walk_for_partial_acc_and_jerk_on_leaf()	--> indirect
//
//-----------------------------------------------------------------------------

inline void accumulate_acc_and_jerk(hdyn* b,
				    vector d_pos,
				    vector d_vel,
				    real eps2,
				    vector & a,
				    vector & j,
				    real & p,
				    real & distance_squared)
{
    distance_squared = d_pos * d_pos;
    real r2inv = 1.0 / (distance_squared + eps2);
    real a3 = -3.0 * (d_pos * d_vel) * r2inv;
    real mrinv = b->get_mass() * sqrt(r2inv);
    p -= mrinv;
    real mr3inv = mrinv * r2inv;
    a += mr3inv * d_pos;
    j += mr3inv * (d_vel + a3 * d_pos);
}


//-----------------------------------------------------------------------------
// flat_calculate_acc_and_jerk: calculate acc and jerk on top node 'this'
// due to all other top-level nodes in the system except 'mask', always in
// the point-mass approximation, optionally determining a perturber list.
//
// (Overloaded function)
//-----------------------------------------------------------------------------

void hdyn::flat_calculate_acc_and_jerk(hdyn * b,    	// root node
				       real eps2,
				       bool make_perturber_list)
{
    if (EV_FUNCTION_ID) {
	cerr << "    flat_calculate_acc_and_jerk for "
	     << format_label() << endl;
    }

    acc = jerk = 0;
    pot = 0;

    // Initialize both nn and coll:

    nn = NULL;
    d_nn_sq = VERY_LARGE_NUMBER;
    coll = NULL;
    d_coll_sq = VERY_LARGE_NUMBER;

    real distance_squared;

    for_all_daughters(hdyn, b, bi)	
	if (bi != this) {

	    accumulate_acc_and_jerk(bi,				// (inlined)
				    bi->get_pred_pos() - pred_pos,
				    bi->get_pred_vel() - pred_vel,
				    eps2, acc, jerk, pot, distance_squared);

	    // Note that the nn and coll passed to update_nn_coll here
	    // are the actual pointers, so this update changes the
	    // data in b.

	    update_nn_coll(this, 1,				// (inlined)
			   distance_squared, bi, d_nn_sq, nn,
			   radius + bi->radius, d_coll_sq, coll);

	    if (make_perturber_list
		&& is_perturber(mass, bi->mass,			// (inlined)
				distance_squared,
				perturbation_radius_factor)) {
		if (n_perturbers < MAX_PERTURBERS)
		    perturber_list[n_perturbers] = bi;
		n_perturbers++;
	    }
	}
}

void hdyn::perturber_acc_and_jerk_on_leaf(real eps2,
					  vector & a,
					  vector & j,
					  real & p,
					  real & distance_squared,
					  hdyn * &p_nn,
					  hdyn * pnode,
					  hdyn * step_node)
{
    // Calculate acc and jerk on leaf 'this' due to the perturber
    // list associated with node pnode.

    // The input arguments distance_squared and p_nn may be the actual
    // d_nn_sq and nn, or copies.

    if (EV_FUNCTION_ID) {
	cerr << "        perturber_acc_and_jerk_on_leaf for "
	     << format_label() << endl;
    }

    dbg_message("perturber_acc_and_jerk_on_leaf", this);

    a = j = 0.0;
    p = 0;

    distance_squared = VERY_LARGE_NUMBER;
    d_coll_sq = VERY_LARGE_NUMBER;

    if (!pnode->valid_perturbers || pnode->n_perturbers <= 0)
	return;

    // Added &hdyn:: to four "hdyn_something_relative_to_root" (Steve, 6/27/97).

    // Removed all four "hdyn_something_relative_to_root" references, which are
    // quite inefficient (Steve, 8/20/98).

    // Determine absolute position and velocity of 'this', using explicit code.

    vector d_pos = -pred_pos;
    vector d_vel = -pred_vel;

    hdyn* par = get_parent();
    if (par) {
 	hdyn* gpar = par->get_parent();
 	while (gpar) {
 	    d_pos -= par->pred_pos;
 	    d_vel -= par->pred_vel;
 	    par = gpar;
 	    gpar = gpar->get_parent();
 	}
    }

    // vector d_pos = -hdyn_something_relative_to_root(this,
    // 						    &hdyn::get_pred_pos);
    // vector d_vel = -hdyn_something_relative_to_root(this,
    // 						    &hdyn::get_pred_vel);

    // Loop over the perturber list.

    vector d_pos_p;
    vector d_vel_p;

    for (int i = 0; i < pnode->n_perturbers; i++) {

	hdyn *bb = pnode->perturber_list[i];

	// Seems to be necessary the first time through -- perturbers are
	// apparently not (necessarily) predicted elsewhere (Steve 8/13/98):

	if (!elder_sister)
	    predict_loworder_all(bb, time);

	// Determine absolute position and velocity of perturber, using
	// explicit code rather than hdyn_something_relative_to_root.

// 	if (bb->is_top_level_node()) {		// (time saver)
// 	    d_pos_p = d_pos + bb->pred_pos;
// 	    d_vel_p = d_vel + bb->pred_vel;
//  	} else {
//  	    d_pos_p = d_pos +
//  		hdyn_something_relative_to_root(bb, &hdyn::get_pred_pos);
//  	    d_vel_p = d_vel +
//  		hdyn_something_relative_to_root(bb, &hdyn::get_pred_vel);
//  	}

 	d_pos_p = d_pos + bb->pred_pos;
 	d_vel_p = d_vel + bb->pred_vel;

 	hdyn* par = bb->get_parent();
 	if (par) {
 	    hdyn* gpar = par->get_parent();
 	    while (gpar) {
 		d_pos_p += par->pred_pos;
		d_vel_p += par->pred_vel;
 		par = gpar;
 		gpar = gpar->get_parent();
 	    }
 	}

	real d2_bb;
	
	accumulate_acc_and_jerk(bb, d_pos_p, d_vel_p,		// (inlined)
				eps2, a, j, p, d2_bb);
	
	// Note: distance_squared and p_nn come from the argument list,
	// and may be copies of d_nn_sq and nn, or the real thing.
	// However, d_coll_sq and coll are real, so this update
	// actually changes the coll data.

	update_nn_coll(this, 2,					// (inlined)
		       d2_bb, bb, distance_squared, p_nn,
		       radius + bb->radius, d_coll_sq, coll);
    }

    step_node->inc_indirect_force(pnode->n_perturbers);		// bookkeeping
}


//-----------------------------------------------------------------------------
// tree_walk_for_partial_acc_and_jerk_on_leaf:  Accumulate acc and jerk
//      while traversing the portion of the tree under source node b,
//	excluding everything under node mask (inclusive).  Vectors offset_pos
//	and offset_vel MUST be the position and velocity of b relative to
//	the field point (i.e. the point where acc and jerk are wanted).
//
// 	On return, a, j, and p are the acc, jerk, and pot at the field
//	point, updated to include the effect of b (modulo the value of
//	point_mass_flag), distance_squared is the minimum distance squared
//	from any component of b to the field point, and p_nn is an
//	updated pointer to the the nearest neighbor of the field point.
//
//	In the present code, the field point is the location of 'this'.
//	It is ASSUMED that 'mask' is 'this' or an ancestor of 'this'.
//-----------------------------------------------------------------------------

// This routine calculates the force by simple recursion.  If a leaf is
// reached, calculate the force.  Otherwise descend the tree.
//
// NOTE: RECURSIVE LOOPS are quite inefficient on some (most?) systems.
// However, by construction, this function is only used in circumstances
// where a system-wide O(N) traversal of the system is NOT needed, so the
// inefficiency is tolerable.  All functions that require O(N) operations
// are coded as explicit loops (see calculate_acc_and_jerk_on_top_level_node).

void hdyn::tree_walk_for_partial_acc_and_jerk_on_leaf(hdyn * b,
						      hdyn * mask,
						      vector offset_pos,
						      vector offset_vel,
						      real eps2,
						      vector & a,
						      vector & j,
						      real & p,
						      real & distance_squared,
						      hdyn * &p_nn,
						      bool point_mass_flag,
						      hdyn * step_node)
{
    // The input arguments distance_squared and p_nn may be the actual
    // d_nn_sq and nn, or copies.

    if (EV_FUNCTION_ID) {
	cerr << "        tree_walk_for_partial_acc_and_jerk_on_leaf for "
	     << format_label() << endl;
	cerr << "        b = " << b->format_label();
	cerr << ", mask = " << mask->format_label() << endl;
    }

    dbg_message("tree_walk_for_partial_acc_and_jerk_on_leaf", this, b);

    if (b == mask)
	return;

    // Stop descent of the tree at an unperturbed CM, *except* when
    // that CM is the parent of 'this' node (to get the correct 2-body
    // internal force).

    if (b->is_leaf()
	|| (b->is_top_level_node() && point_mass_flag)
	|| (b->kep != NULL && b != parent)
	) {

	if (b != this) {
	    real d2;

	    accumulate_acc_and_jerk(b, offset_pos, offset_vel,	  // (inlined)
				    eps2, a, j, p, d2);

	    step_node->inc_indirect_force();

	    // Note: distance_squared and p_nn come from the argument list,
	    // and may be copies of d_nn_sq and nn, or the real thing.
	    // However, d_coll_sq and coll are real, so this update
	    // actually changes the coll data.

	    update_nn_coll(this, 3,
			   d2, b, distance_squared, p_nn,
			   radius + b->radius, d_coll_sq, coll);

	} else {

	    // Probably can't occur...

	    cerr << "In tree_walk_for_partial_acc_and_jerk_on_leaf:\n";
	    cerr << "b = "; b->pretty_print_node(cerr);
	    cerr << "  mask = "; mask->pretty_print_node(cerr);

	}

    } else {

	for_all_daughters(hdyn, b, d) {		       // recursive loop

	    tree_walk_for_partial_acc_and_jerk_on_leaf(d, mask,
					       offset_pos + d->get_pred_pos(),
					       offset_vel + d->get_pred_vel(),
						       eps2, a, j, p,
						       distance_squared, p_nn,
						       point_mass_flag,
						       step_node);
	}
    }
}

//=============================================================================
//  functions for determining which nodes should exert forces on each other
//=============================================================================

//-----------------------------------------------------------------------------
// calculate_partial_acc_and_jerk_on_leaf:  Calculate the force on `this' from
//      all particles under `top', excluding everything under node mask.
//      This routine first calculates the position of `top' relative to `this',
//      then calls tree_walk_for_partial_acc_and_jerk_on_leaf.
//-----------------------------------------------------------------------------

void hdyn::calculate_partial_acc_and_jerk_on_leaf(hdyn * top,
						  hdyn * common,
						  hdyn * mask,
						  real eps2,
						  vector & a,
						  vector & j,
						  real & p,
						  real & distance_squared,
						  hdyn * &nn,
						  bool point_mass_flag,
						  hdyn * pnode,
						  hdyn * step_node)
{
    // The input arguments distance_squared and nn may be the actual
    // d_nn_sq and nn, or copies.

    if (EV_FUNCTION_ID) {
	cerr << "      calculate_partial_acc_and_jerk_on_leaf for "
	     << format_label() << endl;
	cerr << "      top = " << top->format_label();
	cerr << ", common = " << common->format_label();
	cerr << ", mask = " << mask->format_label() << endl;
    }

    dbg_message("calculate_partial_acc_and_jerk_on_leaf", this);
    dbg_message("                 from particle", top);

    a = j = 0.0;
    p = 0;
    
    // Determine absolute position and velocity of "this":

    vector d_pos = 0;
    vector d_vel = 0;

    // Loop over the appropriate perturber list for external perturbations,
    // and traverse the tree below top, masked by mask, for internal
    // perturbations due to other members of the parent clump.

    if (pnode)
	perturber_acc_and_jerk_on_leaf(eps2, a, j, p,
				       distance_squared, nn,
				       pnode, step_node);

    if (top == mask) return;

    hdyn *b;
    for (b = this; b != common; b = b->get_parent()) {
	d_pos -= b->get_pred_pos();
	d_vel -= b->get_pred_vel();
    }

    for (b = top; b != common; b = b->get_parent()) {
	d_pos += b->get_pred_pos();
	d_vel += b->get_pred_vel();
    }

    tree_walk_for_partial_acc_and_jerk_on_leaf(top, mask, d_pos, d_vel,
					       eps2, a, j, p,
					       distance_squared, nn,
					       point_mass_flag,
					       step_node);
}

//-----------------------------------------------------------------------------
// calculate_partial_acc_and_jerk:  Calculate the force on `this' from
//	all particles under `top', excluding everything under node mask.
//      This routine recursively calls itself and calculates the weighted
//      average of all forces on its daughters.  For a leaf, it calls
//      calculate_partial_acc_and_jerk_on_leaf.  The recursion is OK in
//      this case because it is used only to descend within nodes, not to
//      traverse the entire system.
//-----------------------------------------------------------------------------

void hdyn::calculate_partial_acc_and_jerk(hdyn * top,
					  hdyn * common,
					  hdyn * mask,
					  real eps2,
					  vector & a,
					  vector & j,
					  real & p,
					  real & distance_squared,
					  hdyn * &p_nn,
					  bool point_mass_flag,
					  hdyn* pnode,
					  hdyn* step_node)
{
    // The input arguments distance_squared and p_nn may be the actual
    // d_nn_sq and nn, or copies.  If this is a node, the nn returned
    // is the closer of its component nns.

    if (EV_FUNCTION_ID) {
	cerr << "    calculate_partial_acc_and_jerk for "
	     << format_label() << endl;
	cerr << "    top = " << top->format_label();
	cerr << ", common = " << common->format_label();
	cerr << ", mask = " << mask->format_label() << endl;
    }

    dbg_message("calculate_partial_acc_and_jerk", this);

    // Operations of this function:
    //
    // (1) Calculate acc and jerk on a top-level node due to all
    //     other top-level nodes, in the point-mass approximation
    //     (this is what the GRAPE-4 does...).
    // (2) Calculate masked acc and jerk on a node due to the portion
    //     of the tree below top.
    // (3) Calculate acc and jerk on "this" node due to the particles
    //     in its perturber list, recursively looping over daughters.
    // (4) Calculate acc and jerk on "this" node due to the particles
    //     in its perturber list, in the point-mass approximation.
    // (5) Determine the perturber list of a binary during its
    //     center-of-mass step.

    a = j = 0;
    p = 0;
    real mtot = 0;

    // Don't resolve an unperturbed binary (Steve, 8/20/98).

    if (is_leaf() || point_mass_flag || get_oldest_daughter()->kep) {

	calculate_partial_acc_and_jerk_on_leaf(top, common, mask, eps2,
					       a, j, p,
					       distance_squared, p_nn,
					       point_mass_flag,
					       pnode, step_node);
    } else {

	vector a_daughter, a_save;
	vector j_daughter;
	real p_daughter;
	real m_daughter;

	for_all_daughters(hdyn, this, bd) {

	    // Note from Steve, 7/24/98.  Because of the way nns and
	    // colls are passed around, if we compute the acc and jerk
	    // on a leaf in order to obtain the acc and jerk on its parent
	    // node, the nn information of the leaf will be preserved
	    // because p_nn here is actually the node nn, but the coll
	    // information will be overwritten.  If the coll of bd is the
	    // other binary component (as is often the case), an error
	    // will be made.

	    // The best solution would be to treat colls in exactly the
	    // same way as nn pointers.  For now, however, we just save
	    // and restore the coll data of any node referenced here.

	    hdyn* save_coll = bd->get_coll();		// save the
	    real save_d_coll_sq = bd->get_d_coll_sq();	// coll data

	    m_daughter = bd->get_mass();
	    bd->calculate_partial_acc_and_jerk(top, common, mask, eps2,
					       a_daughter,
					       j_daughter, p_daughter,
					       distance_squared, p_nn,
					       point_mass_flag,
					       pnode, step_node);
	    a += m_daughter * a_daughter;
	    j += m_daughter * j_daughter;
	    p += m_daughter * p_daughter;
	    mtot += m_daughter;

	    if (bd->get_elder_sister() == NULL)		// in case we want
		a_save = a_daughter;			// the perturbation

	    bd->coll = save_coll;			// restore the
	    bd->d_coll_sq = save_d_coll_sq;		// coll data

	}
	real minv = 1 / mtot;
	a *= minv;
	j *= minv;
	p *= minv;

	// Special case:  It is helpful to know the perturbation on a
	// top-level node whose neighbor list has overflowed (or which
	// is being calculated exactly).  In that case, the following
	// should all be true:
	//
	//	top = common = root
	//	mask = this
	//	this is top-level node
	//	point_mass_flag = false
	//	pnode = NULL
	//
	//	valid_perturbers = false
	//	n_perturbers <= 0
	//

	// In this case, we should set perturbation_squared before
	// leaving the function.  Note that perturbation_squared is
	// attached to the daughter nodes, but the perturber list,
	// valid_perturbers, and n_perturbers are properties of the
	// parent.

	if (!point_mass_flag
	    && !pnode
	    && top->is_root()
	    && common == top
	    && mask == this
	    && !valid_perturbers) {

	    // (We probably don't need to check all these!)

	    // "Daughter" quantities refer to the younger daughter on
	    // leaving the loop.  The vector a_2b is the two-body
	    // relative acceleration of the components (computed here
	    // in the point-mass approximation).

	    hdyn* od = get_oldest_daughter();
	    hdyn* yd = get_oldest_daughter()->get_younger_sister();

	    real pscale = m_daughter / (m_daughter + mass);
	    real r2 = square(od->pos - yd->pos);
	    od->perturbation_squared =
			square(pscale * (a_save - a_daughter))
			        * square(r2/mass);
 	}
    }
}


// check_add_perturber:  check if p is a perturber of 'this' and add it
//			 to the perturber list if necessary.

inline void hdyn::check_add_perturber(hdyn* p, vector this_pos)
{
    vector ppos = hdyn_something_relative_to_root(p, &hdyn::get_pred_pos);

//     bool print = get_top_level_node()->n_leaves() >= 4;
//     if (print) {
// 	cerr << "check_add_perturber:  checking " << p->format_label()
// 	     << " at time " << system_time;
// 	cerr << " for node " << format_label() << endl;
// 	PRL(mass);
// 	PRL(square(this_pos - ppos));
// 	PRL(perturbation_radius_factor);
//    }

    if (is_perturber(mass, p->mass,
		     square(this_pos - ppos),
		     perturbation_radius_factor)) {	  // (inlined)

	// If p is a compound system, include all components, without
	// checking if they satisfy the perturbation criteria individually.

	if (RESOLVE_UNPERTURBED_PERTURBERS) {
	    for_all_leaves(hdyn, p, pp) {
		if (n_perturbers < MAX_PERTURBERS)
		    perturber_list[n_perturbers] = pp;
		n_perturbers++;
	    }
	} else {
	    for_all_leaves_or_unperturbed(hdyn, p, pp) {
		if (n_perturbers < MAX_PERTURBERS)
		    perturber_list[n_perturbers] = pp;
		n_perturbers++;
	    }
	}
//	if (print) cerr << "...accepted" << endl;
    } //else
//	if (print) cerr << "...rejected" << endl;
}

void hdyn::create_low_level_perturber_list(hdyn* pnode, real pert_factor_sq)
{
    valid_perturbers = true;
    if (perturber_list == NULL)
	perturber_list = new hdyn *[MAX_PERTURBERS];

    perturbation_radius_factor
	= set_perturbation_radius_factor(pert_factor_sq,
					 d_min_sq,
					 binary_scale(this),
					 mass);

    vector this_pos = hdyn_something_relative_to_root(this,
						      &hdyn::get_pred_pos);
    n_perturbers = 0;

    // First add sisters, aunts, etc. up to pnode...

    hdyn* p = this;
    while (p != pnode) {
	check_add_perturber(p->get_binary_sister(), this_pos);
	p = p->get_parent();
    }

    // ...then accept any node on the pnode list that satisfies the
    // inner node perturbation criterion.

    for (int i = 0; i < pnode->n_perturbers; i++)
	check_add_perturber(pnode->perturber_list[i], this_pos);

    if (n_perturbers > MAX_PERTURBERS) {

	valid_perturbers = false;
	delete [] perturber_list;

    } else {

// 	if (get_top_level_node()->n_leaves() >= 4) {
// 	    cerr << ">>>> this->"; PRL(format_label());
// 	    cerr << ">>>> "; PRL(pnode->format_label());
// 	    cerr << ">>>> "; PRL(pnode->n_perturbers);
// 	    cerr << ">>>> this->"; PRL(n_perturbers);
// 	    print_perturber_list();
// 	}

    }
}

//-----------------------------------------------------------------------------
// calculate_acc_and_jerk_on_low_level_node:  Calculate the acc (etc) on
//         one component of a binary node in the following steps:
//
// First, calculate the perturbation on the relative motion as the
// the difference between the accelerations of the node and that of
// its sister, multiplied by the fractional mass of the sister.
// These accelerations do not include the mutual interaction between
// the node and its sister.
//
// Second, calculate the acceleration due to the sister.
//
// Third, calculate the total acceleration, given by the sum of the
// perturbation acc and the acc from the sister.
//
// Note that all three components are calculated by one call to
// calculate_partial_acc_and_jerk.
//-----------------------------------------------------------------------------

void hdyn::calculate_acc_and_jerk_on_low_level_node(real eps2,
						    real pert_factor_sq)
{
    if (EV_FUNCTION_ID) {
	cerr << "  calculate_acc_and_jerk_on_low_level_node for "
	     << format_label() << endl;
    }

    dbg_message("calculate_acc_and_jerk_on_low_level_node", this);

    hdyn *root = get_root();

    if (parent->get_oldest_daughter()->get_younger_sister()
				     ->get_younger_sister() != NULL)
	err_exit("calculate_acc_and_jerk_on_low_level_node: Not binary node");

    hdyn *sister = get_binary_sister();

    real mtot = 0;

    vector apert1, jpert1;
    vector apert2, jpert2;
    real p_dummy;
    hdyn *top;

    // New formulation of perturber lists introduced by Steve 8/98:

    // Perturber list may not necessarily be associated with the
    // top-level node.  Any CM node above this node may have a
    // valid perturber list.  Use the lowest-level one.

    // Now pass a pointer to the node containing the perturber list
    // (NULL ==> no valid list) instead of bool "use_perturber_list"
    // flag.  Also, since the low-level perturber list includes all
    // members of the parent "clump" except those below the parent
    // node, we need only descend the tree below pnode to pick up the
    // remaining perturbations.

    hdyn* top_level = get_top_level_node();
    hdyn* pnode = find_perturber_node();

    if (pnode) {
	top = pnode;
	kc->pert_step++;
	kc->pert_with_list += pnode->n_perturbers;
    } else {
	top = root;
	kc->pert_without_list++;
    }

    // Acceleration and jerk on this component due to rest of system:

    d_nn_sq = VERY_LARGE_NUMBER;
    calculate_partial_acc_and_jerk(top, top, get_parent(), eps2,
				   apert1, jpert1, p_dummy,
				   d_nn_sq, nn,
				   !USE_POINT_MASS,		// explicit loop
				   pnode,			// with list
				   this);			// node to charge

    // Acceleration and jerk on other component due to rest of system:

    sister->d_nn_sq = VERY_LARGE_NUMBER;
    sister->calculate_partial_acc_and_jerk(top, top, get_parent(), eps2,
					   apert2, jpert2, p_dummy,
					   sister->d_nn_sq, sister->nn,
					   !USE_POINT_MASS,	// explicit loop
					   pnode,		// with list
					   this);		// node to charge
 
    // Note:  The first two calls to calculate_partial_acc_and_jerk pass
    // the d_nn_sq and nn from the hdyn, so the hdyn data are actually
    // updated by update_nn_coll.  The third call (below) passes local
    // variables, so it has no direct effect on d_nn_sq and nn.  (Test
    // afterwards if d_nn_sq and nn must be updated.)  The coll data
    // are always updated by these calls.

    // Relative acceleration and jerk due to other (sister) component:

    vector a_2b, j_2b;
    real p_2b;

    hdyn *p_nn_sister;
    real d_min_sister = VERY_LARGE_NUMBER;
    calculate_partial_acc_and_jerk(get_parent(), get_parent(), this, eps2,
				   a_2b, j_2b, p_2b,
				   d_min_sister, p_nn_sister,
				   !USE_POINT_MASS,
				   NULL,			// no perturbers
				   this);			// node to charge

    real m2 = sister->get_mass();
    real pscale = m2 / (m2 + mass);

    set_acc_and_jerk_and_pot(a_2b + pscale * (apert1 - apert2),
			     j_2b + pscale * (jpert1 - jpert2), p_2b);

    perturbation_squared = square(pscale * (apert1 - apert2))
			        / square(a_2b);

    if (d_nn_sq > d_min_sister) {
	d_nn_sq = d_min_sister;
	nn = p_nn_sister;
    }

    // if (name_is("xxx")) {
    //     print_nn(this, 2);
    //     print_coll(this, 2);
    // }

    // The above procedure will fail to correctly determine the sister's
    // nearest neighbor.  Correct this here.  Note that, in this case,
    // the sister's nn may be a node (this), not a leaf; however, the
    // stored distance will accurately reflect the actual distance to the
    // nn leaf.  Too inconvenient and inefficient to routinely descend the
    // tree below this to (re)locate the nn leaf here -- MUST take this
    // "feature" into account in new_sister_node (hdyn_tree.C) when
    // assessing the need for tree reconfiguration.

    if (d_min_sister < sister->d_nn_sq) {
        sister->d_nn_sq = d_min_sister;
        sister->nn = this;

	// (OK to have sister->nn be a CM node.)

    }

    if ((nn == NULL) || (get_binary_sister()->nn == NULL)) {
	cerr << "calculate_acc_and_jerk_on_low_level_node:  nn = NULL:\n";
	PRL(top_level->n_perturbers);
	pp3(this, cerr);
	pp3(top_level, cerr);
    }

    valid_perturbers = false;

    if (ALLOW_LOW_LEVEL_PERTURBERS && pnode) {

	// Create/revise the low-level perturber list(s).

	// Could revise list as we compute the force, but the distances used
	// should be from the center of mass, not from one of the components.
	// This may be a little less efficient, but it is much easier to code,
	// so keep it for now.  (Steve, 8/98)

	if (is_parent())
	    create_low_level_perturber_list(pnode, pert_factor_sq);

	if (get_binary_sister()->is_parent())
	    get_binary_sister()->
		create_low_level_perturber_list(pnode, pert_factor_sq);
    }
}

local int n_leaves_or_unperturbed(hdyn* b)
{
    // Like n_leaves, but counting fully unperturbed center of mass nodes
    // as single particles.

    hdyn* od = b->get_oldest_daughter();
    if (od == NULL || (od->get_kepler() && od->get_fully_unperturbed())) {
	return 1;
    } else {
	int n = 0;
	for_all_daughters(hdyn, b, bb)
	    n += n_leaves_or_unperturbed(bb);
	return n;
    }
}

// expand_nodes:  expand center-of-mass nodes on a perturber list into
//		  single stars (i.e. leaves).
//
//		  NOTE from Steve, 8/20/98:  we (optionally) no longer
//		  expand fully unperturbed binaries.  This leaves the system
//		  in a slightly dangerous state, as it is possible that
//		  an unperturbed binary center of mass perturbing another
//		  binary may vanish before it can be resolved back into
//		  its components.  This is unlikely to occur dynamically,
//		  as the binary will first become perturbed and the
//		  components restored on the next center of mass time step.
//		  However, it can (and has!) happen.  To avoid this, all
//		  unperturbed binaries are expanded in integrate_unperturbed
//		  if the unperturbed motion is terminated.
//
//		  Nodes may be deleted as a result of stellar evolution (see
//		  merge_nodes), but the components are merged into the center
//		  of mass in that case.  (Nodes are also deleted as they
//		  escape, but this affects all particles, not just binaries.)
//
//		  Binaries undergoing partial unperturbed motion (pericenter
//		  reflection) are always expanded.

local bool expand_nodes(int &n, hdyn ** list, bool debug = false)
{
     if (debug)
 	cerr << endl << "in expand_nodes, n = " << n << endl;

    for (int i = 0; i < n; i++) {

 	if (debug)
 	    cerr << i << ".  " << list[i] << "  "
 		 << list[i]->format_label() << endl;

	if (!list[i]->is_leaf()) {

	    if (debug)
		cerr << "not a leaf" << endl;

	    int nl;

	    if (RESOLVE_UNPERTURBED_PERTURBERS)
		nl = list[i]->n_leaves();
	    else
		nl = n_leaves_or_unperturbed(list[i]);		// <-- new

	    if (debug) {
		pp3(list[i]->get_top_level_node());
		PRL(nl);
	    }

	    if (n + nl - 1 > MAX_PERTURBERS)
		return false;

	    // Make room for the expansion.

	    int j;
	    for (j = n - 1; j > i; j--)
		list[j + nl - 1] = list[j];

	    // Add the new nodes to the list.

	    j = 0;
	    hdyn *tmp = list[i];

	    if (debug)
		PRL(tmp->format_label());

	    if (RESOLVE_UNPERTURBED_PERTURBERS) {
		for_all_leaves(hdyn, tmp, b) {
		    list[i + j] = b;
		    j++;
		}
	    } else {

		for_all_leaves_or_unperturbed(hdyn, tmp, b) {	// <-- new
		    list[i + j] = b;
		    if (debug)
			cerr << "added " << b->format_label() << endl;
		    j++;
		}
	    }

	    n += nl - 1;
	    i += nl - 1;

	    if (debug && j != nl)
		cerr << "expand_nodes -- oops... "
		     << j << " != " << nl << endl;
	}
    }

    return true;
}

//-----------------------------------------------------------------------------
//  NOTE: calculate_acc_and_jerk_on_top_level_node is now split into three
//  pieces to allow use of HARP hardware.
//-----------------------------------------------------------------------------

void hdyn::calculate_acc_and_jerk_on_top_level_node(real eps2,
						    real pert_factor_sq,
						    bool exact)
{
    if (EV_FUNCTION_ID) {
	cerr << "  calculate_acc_and_jerk_on_top_level_node for "
	     << format_label() << endl;
    }

    top_level_node_prologue_for_force_calculation(eps2,pert_factor_sq,
						  exact);
    if (!exact) {
	top_level_node_real_force_calculation(eps2);
	top_level_node_epilogue_force_calculation(eps2);
    }

}


void hdyn::top_level_node_prologue_for_force_calculation(real eps2,
							 real pert_factor_sq,
							 bool exact)
{
    if (EV_FUNCTION_ID) {
	cerr << "  top_level_node_prologue_for_force_calculation for "
	     << format_label() << endl;
    }

    hdyn *root = get_root();
    d_coll_sq = VERY_LARGE_NUMBER;
    coll = NULL;

    if (exact) {

	// Perform the entire force calculation.

	n_perturbers = 0;
	valid_perturbers = false;
	d_nn_sq = VERY_LARGE_NUMBER;

	calculate_partial_acc_and_jerk(root, root, this, eps2,
				       acc, jerk, pot, d_nn_sq, nn,
				       !USE_POINT_MASS,
				       NULL,		// no perturbers
				       this);		// node to charge
    } else if (is_parent()) {

	// Set up computation of perturber list.

	if (perturber_list == NULL)
	    perturber_list = new hdyn *[MAX_PERTURBERS];

	perturbation_radius_factor
		= set_perturbation_radius_factor(pert_factor_sq,
						 d_min_sq,
						 binary_scale(this),
						 mass);
    }
}

void hdyn::top_level_node_real_force_calculation(real eps2)
{
    if (EV_FUNCTION_ID) {
	cerr << "  top_level_node_real_force_calculation for "
	     << format_label() << endl;
    }

    // Special treatment of traversal of top level only (for
    // efficiency, and for GRAPE-4 implementation).

    // Calculate forces first in the POINT-MASS approximation,
    // without using any perturber list, and construct a new list,
    // in the case of CM nodes.

    // Should NOT be called in exact case.

    hdyn * root = get_root();

    if (is_parent()) {
        n_perturbers = 0;
        valid_perturbers = true;
    }

    flat_calculate_acc_and_jerk(root, eps2, is_parent());

    if (nn == NULL) {
        pretty_print_node(cerr); cerr << " nn NULL after flat " << endl;
    }
}

void hdyn::top_level_node_epilogue_force_calculation(real eps2)
{
    if (EV_FUNCTION_ID) {
	cerr << "  top_level_node_epilogue_force_calculation for "
	     << format_label() << endl;
    }

    // Should NOT be called in exact case.

    // Take care of the case where the nearest neighbor is
    // a complex node, since otherwise nn may disappear.
    // (Jan 21, 1998, JM and SPZ)

    // Exclude the case of nn = this (possible when GRAPE is used.)
    // (Sep 9, 1998, SLWM)

    if (nn != this) {
	while (nn->is_parent()) {
	    // cerr << "nn correction for "; pretty_print_node(cerr);
	    // cerr << "and  "; nn->pretty_print_node(cerr);

	    nn = nn->get_oldest_daughter();

	    // cerr << " as   ";   nn->pretty_print_node(cerr);
	    // cerr << endl;
	}
    }
    
    if (!is_parent()) return;

    hdyn * root = get_root();
	
    // if (is_parent()) {
    //     print_label(cerr); cerr << " nn before epilogue ";
    //     n->print_label(cerr); cerr << " ";
    //     RL(d_nn_sq);
    // }

    if (n_perturbers > MAX_PERTURBERS) {
	    
	// Perturber list has overflowed.  Use the entire tree (below).

	valid_perturbers = false;
	kc->perturber_overflow++;
	    
    } else if (n_perturbers > 0) {
	
	// Use the perturber list to correct the center-of-mass force.
	// First, subtract out center-of-mass contribution:
	
	vector a_p, j_p;
	real p_p;
	calculate_partial_acc_and_jerk(this, this, this, eps2,
				       a_p, j_p, p_p, d_nn_sq, nn,
				       USE_POINT_MASS,	    	    // explicit
				       this,	   		    // loop
				       this);

	// (Note: 'this' is OK here because this is a top-level node.)

	acc -= a_p;
	jerk -= j_p;
	pot -= p_p;
	
	// Replace nodes by components on the perturber list and
	// add in contribution due to components.

	bool debug = false;

	if (!expand_nodes(n_perturbers, perturber_list, debug)) {
	    
	    // Perturber list has overflowed.  Use the entire tree (below).
	    
	    valid_perturbers = false;
	    kc->perturber_overflow++;
	    
	} else {

	    // (Explicit loop)
	    
	    calculate_partial_acc_and_jerk(this, this, this, eps2,
					   a_p, j_p, p_p, d_nn_sq, nn,
					   !USE_POINT_MASS,
					   this,
					   this);
	    acc += a_p;
	    jerk += j_p;
	    pot += p_p;
	    
	}

	// if (nn == NULL) {
	//     pretty_print_node(cerr);
	//     cerr << " nn NULL after add back "<<endl;
	// }

    }
    
    // Recompute the force exactly if the perturber list has overflowed.

    if (!valid_perturbers) {
	n_perturbers = -1;
	calculate_partial_acc_and_jerk(root, root, this, eps2,
				       acc, jerk, pot, d_nn_sq, nn,
				       !USE_POINT_MASS,
				       NULL,		// no perturbers
				       this);
    }

    // NOTE:  On exit, if valid_perturbers is true, and
    //        RESOLVE_UNPERTURBED_PERTURBERS is true, then
    //	      perturber_list contains *leaves only.*
    //
    //	      If valid_perturbers is false, then the force on 'this'
    //	      node has been computed exactly; no correction is needed.

    // if (nn == NULL) {
    //     pretty_print_node(cerr);
    //     cerr << " nn NULL after corrections "<<endl;
    // }

    // if (is_parent()) {
    //     print_label(cerr); cerr << " nn after epilogue ";
    //     nn->print_label(cerr); cerr << " ";
    //     PRL(d_nn_sq);
    //     PRC(valid_perturbers); PRL(n_perturbers);
    // }

}


//-----------------------------------------------------------------------------
// calculate_acc_and_jerk:  Generic routine to calculate the acceleration
// of a node.  Call appropriate functions depending on whether or not the
// node is at the top level of the tree.
//-----------------------------------------------------------------------------

void hdyn::calculate_acc_and_jerk(real eps2, real pert_factor_sq, bool exact)
{
    if (EV_FUNCTION_ID) {
	cerr << "calculate_acc_and_jerk for "
	     << format_label() << endl;
    }

    dbg_message("calculate_acc_and_jerk", this);

    d_coll_sq = VERY_LARGE_NUMBER;
    coll = NULL;
    if (is_low_level_node())
	get_binary_sister()->d_coll_sq = VERY_LARGE_NUMBER;

    if (is_top_level_node())
	calculate_acc_and_jerk_on_top_level_node(eps2, pert_factor_sq, exact);
    else
	calculate_acc_and_jerk_on_low_level_node(eps2, pert_factor_sq);

    // NOTE: On return, we must correct the acc and jerk on:
    //
    //       (1) a top-level C.M. node due to other nodes not on its
    //           perturber list, but on whose perturber lists it lies.
    // and
    //       (2) a single top-level node due to any C.M.s on whose perturber
    //           lists it lies.

    if (nn == NULL) {
         pretty_print_node(cerr); cerr << " nn NULL after calc " << endl;
    }
}

// apply_correction: bi is a top-level node on the integration list,
//		     whose interaction with bj has not been correctly
//		     calculated.  Include the missing (tidal) terms here.

void apply_correction(hdyn * bj, hdyn * bi, real eps2)
{
    vector a_c = 0, j_c = 0;
    real p_c = 0;
    vector a_p = 0, j_p = 0;
    real p_p = 0;
    real dum_d_sq = VERY_LARGE_NUMBER;
    hdyn *dum_nn = NULL;

//     if (bi->name_is("1")) {
// 	cerr << "correcting " << bi->format_label();
// 	cerr << " for " << bj->format_label() << " at time "
// 	     << bi->get_system_time() << endl;
// 	PRL(bi->get_acc());
// 	PRL(bj->get_acc());
//     }

    bi->calculate_partial_acc_and_jerk(bj, bi->get_parent(), bi,
				       eps2,
				       a_c, j_c, p_c, dum_d_sq, dum_nn,
				       USE_POINT_MASS,
				       NULL,		// no perturbers
				       bi);
    bi->calculate_partial_acc_and_jerk(bj, bi->get_parent(), bi,
				       eps2,
				       a_p, j_p, p_p, dum_d_sq, dum_nn,
				       !USE_POINT_MASS,
				       NULL,		// no perturbers
				       bi);

    bi->inc_acc(a_p - a_c);
    bi->inc_jerk(j_p - j_c);
    bi->inc_pot(p_p - p_c);

    // Note: changes to indirect_force counter are handled at lower levels.

    // We should update nn etc here, since otherwise
    // perturbers of a perturbed binary see the binary
    // as the nearest neighbor, which is unsafe
    // the following correction added (Jan 21 1998)

    if (bi->get_d_nn_sq() > dum_d_sq) {
	bi->set_d_nn_sq(dum_d_sq);
	bi->set_nn(dum_nn);

        //cerr << "nn correction for ";   bi->pretty_print_node(cerr);
	//cerr << "and  ";   bj->pretty_print_node(cerr);
	//cerr << " as   ";   bi->get_nn()->pretty_print_node(cerr);
	//cerr << " new distance = " << bi->get_d_nn_sq() <<endl;
    }
    
    bi->get_kira_counters()->force_correction++;

//     if (bi->name_is("1")) {
// 	cerr << "after correction" << endl;
// 	PRL(bi->get_acc());
// 	PRL(bj->get_acc());
//     }

}

//-----------------------------------------------------------------------------
// need_correction : return a pointer to a node whose force must be corrected.
//
//     bj: a perturbed top-level CM node
//     bi: one of bj's perturbers
//     t:  present system time
//
// The requirement is as follows:
//
//     a) the actual pointer returned is the top_level_node of bi (btop)
//     b) bi must be the left-most leaf of the tree under btop, to 
//        guarantee uniqueness (i.e. that the correction is applied only once
//     c) btop must be in the present integration block
//     d) bj must not be on the perturber list of btop

local hdyn *need_correction(hdyn * bj, hdyn * bi,
				   bool check_all)
{
    hdyn *btop = bi->get_top_level_node();
    real t = bj->get_system_time();

    if (!check_all && btop->get_next_time() != t)
	return NULL;

    // Now we know that btop is on the integration list.
    // If it is a single star, it needs correction.

    if (btop ->is_leaf())
	return btop;

    hdyn *b_oldest = btop;
    hdyn *bret = NULL;

    // btop is the top-level node of a clump requiring correction.
    // Test if bi is the left-most ("oldest") leaf of the clump, in
    // order to guarantee that btop is corrected only once.

    // Find the left-most leaf.

    while (b_oldest->is_parent())
	b_oldest = b_oldest->get_oldest_daughter();

    // Note that bi can be a top-level node in the case
    // of correction from a node without valid perturber
    // list.

    if ( (b_oldest == bi) || (btop == bi)) {

	bret = btop;

	// No need to apply a correction if some component of bj is a
	// perturber of btop (interaction already properly calculated).

	// Top-level node btop is on the perturber list of node bj.
	// Return NULL iff bj is a perturber of btop.

	// Well, what if perturber list of btop is INVALID?  This in practice
	// *should not* occur unless previous force calculation is exact...
	// ...or if neighbor-list overflow has occurred (Steve 7/98).

	if (btop->get_valid_perturbers()) {
	    for (int j = 0; j < btop->get_n_perturbers(); j++)
		if (btop->get_perturber_list()[j]->get_top_level_node() == bj)
		    bret = NULL;
	}

	// Code modified by Steve 7/98

	else
	    bret = NULL;	// btop force should be correct in this case...

    }

    return bret;
}

// correct_acc_and_jerk: By construction, the acc and jerk on a top-level
//			 node bi_top will be computed incorrectly if it is a
//			 perturber of another top-level node bj that is not
//			 itself a perturber of bi_top.  (Note that if bi_top
//			 is a leaf, then this latter criterion is necessarily
//			 satisfied.)  The reason that the force calculation
//			 requires correction is that it is written to enable
//			 efficient computation of the top-level interactions
//			 with the GRAPE-4.
//
//			 To avoid the use of inverse perturber lists, we
//			 identify nodes bi_top and bj by scanning across the
//			 entire top-level (bj), then checking for perturbers
//			 (bi).  Correction, if necessary, is applied to
//			 bi_top, the top-level node of bi.  Membership on
//			 the integration list is determined by checking a
//			 node's next_time, so there is no need for the list
//			 itself to be provided to this routine.  If flag
//			 check_all is true, a node will be considered for
//			 correction even if its next_time doesn't match
//			 the system time.

// Note from Steve, 9/98.  This is a very inefficient function, as it checks
// all the perturber lists of all nodes, even though in practice very few
// corrections are actually applied.  The most viable way to speed things up
// seems to be the use of inverse perturber lists (*not* implemented yet).

void correct_acc_and_jerk(hdyn * root,
			  real eps2,
			  bool& reset,
			  bool check_all)
{
    static int work_size = 0;
    static hdyn ** nodes;
    static int nnodes = -1;

    if (nnodes == -1) reset = true;

    if (reset) {

	// initialize the array of node pointers

        // cerr << "correct_acc_and_jerk, reset\n";

	int n = 0;
	for_all_daughters(hdyn, root, bb)
	    if (bb->is_parent()) n++;
	if  (work_size < n) { 
	    nodes = new hdynptr[n*2+10];	// DEC C++ doesn't like (hdyn*)
	    work_size = n*2+10;
	}
	n = 0;
	for_all_daughters(hdyn, root, bbb) {
	    if (bbb->is_parent()) {
		nodes[n] = bbb;
		n++ ;
	    }
	}
	nnodes = n;
    }

    reset = false;
    for (int j = 0; j < nnodes; j++) {

	hdyn * bj = nodes[j];

	// for_all_daughters(hdyn, root, bj) {

	if (!bj->is_valid()) {

	    // Invalid node ==> should not occur; better reconstruct
	    // the internal list if found.

	    cerr << "correct_acc_and_jerk: warning: invalid node at time "
		 << root->get_system_time() << endl;
	    reset = true;

	} else if (bj->is_parent() && bj->get_kepler() == NULL) {

	    hdyn *bi_top;

	    // bj is a perturbed top-level CM node.

	    if (bj->get_valid_perturbers()) {

		// Look in bj's perturber list for something to correct.

		for (int i = 0; i < bj->get_n_perturbers(); i++) {

		    hdyn *bi = bj->get_perturber_list()[i];

		    // Flag an invalid perturber...

		    if (!bi->is_valid()) {

			cerr << "correct_acc_and_jerk: warning: "
			     << "invalid perturber #" << i
			     << " for " << bj->format_label()
			     << " at time " << root->get_system_time()
			     << endl;

		    } else {

			if ((bi_top = need_correction(bj, bi, check_all))
			    != NULL) {
			    apply_correction(bj, bi_top, eps2);
			}
		    }
		}

	    } else {

		// Look at all top-level nodes for something to correct.

		// Note (J.M. 96-Aug-5):
		// Actually, one needs to loop over particles in current
		// block only...  Not implemented yet.

		for_all_daughters(hdyn, root, bi) {
		    if (bi != bj) {
			if ((bi_top = need_correction(bj, bi, check_all))
			      != NULL) {
			    apply_correction(bj, bi_top, eps2);
			}
		    }
		}
	    }
	}
    }
}


//-----------------------------------------------------------------------------

// NOTE: The following two functions are the *only* places where
//	 the tidal field is specified.
//
//	 Routines add_tidal and calculate_energies_with_tidal
//	 MUST be kept consistent when changes are made!


void add_tidal(hdyn * b, bool pot_only)
{
    // Add a tidal component to the aceleration of body b.
    // CANNOT neglect the contribution to the jerk.
    // Also include the tidal potential in pot.

    // ASSUME that omega is set and that b is a top-level node.

    if (!pot_only) {

	vector da_tidal_centrifugal = b->get_omega_sq()
      					* vector(3*b->get_pred_pos()[0],
						 0.0,
						 -b->get_pred_pos()[2]);

	vector dj_tidal_centrifugal = b->get_omega_sq()
      					* vector(3*b->get_pred_vel()[0],
						 0.0,
						 -b->get_pred_vel()[2]);

	vector da_coriolis = 2 * b->get_omega()
      				* vector(b->get_pred_vel()[1],
					 -b->get_pred_vel()[0],
					 0.0);

	// Must update acc BEFORE computing dj!

	b->inc_acc(da_tidal_centrifugal + da_coriolis);

	vector dj_coriolis = 2 * b->get_omega()
      				* vector(b->get_acc()[1],
					 -b->get_acc()[0],
					 0.0);

	b->inc_jerk(dj_tidal_centrifugal + dj_coriolis);

	// PRL(da_tidal_centrifugal);
	// PRL(da_coriolis);
	// PRL(abs(da_tidal_centrifugal)/abs(b->get_acc()));

    }

    real x = b->get_pred_pos()[0];
    real z = b->get_pred_pos()[2];
    real dp = 3*x*x - z*z;

    b->inc_pot(-0.5*dp*b->get_omega_sq());
}


real de_tidal_pot(hdyn * b)
{
    // Determine tidal component to the potential of body b.
    // ASSUME that omega is set and that b is a top-level node.

    // Add tidal and centrifugal terms for top-level nodes only.
    // (No potential term for the Coriolis force, note.)

    real dpot = 0;
    for_all_daughters(hdyn, b, bb) {
	real x = bb->get_pos()[0];
	real z = bb->get_pos()[2];
	real dp = 3*x*x - z*z;
	dpot += bb->get_mass() * dp;
    }

    dpot *= -0.5 * b->get_omega_sq();
    return dpot;
}

//-----------------------------------------------------------------------------

//=============================================================================
//  driver function for a one-time-step integration of this node
//=============================================================================

//-----------------------------------------------------------------------------
//  integrate_node -- advance this node by one time step
//-----------------------------------------------------------------------------

void hdyn::integrate_node(hdyn * root,
			  real eps,
			  real eta,
			  real max_timestep,
			  real gamma2,			// not needed...
			  bool integrate_unperturbed,
			  bool force_unperturbed_time)

// I don't understand the use of force_unperturbed_time, so I have
// introduced yet another argument...	(Steve, 7/97)

{
    if (EV_FUNCTION_ID) {
	cerr << "integrate_node for " << format_label()
	     <<" at time " << time  + timestep << endl;
	// pretty_print_node(cerr);
    }

    if (kep == NULL) {

	clear_interaction();
	if (get_gamma2() > 0.0)
	    calculate_acc_and_jerk(eps * eps, get_gamma23(), true);
	else
	    calculate_acc_and_jerk(eps * eps, 0.0, true);
	set_valid_perturbers(false);

	if (omega > 0)
	    if (is_top_level_node())
		add_tidal(this);

	correct();				// note: no retry on error

	update(eta, INITIAL_STEP_LIMIT);
	store_old_force();

	// Note that old_acc = acc at the end of a step.

    } else if (integrate_unperturbed) {

	if (eps == 0) {

	    bool reinitialize;
	    integrate_unperturbed_motion(eta, eps, max_timestep,
					 reinitialize,
					 force_unperturbed_time);

	    if (reinitialize)
		cerr << "\nReceived reinitialization flag "
		     << "during synchronization..." << endl;

	}
	else
	    err_exit(
	        "integrate_node: unperturbed binary with non-zero softening");
    }
}


//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\
//  |  the end of:  |         /|\         |  src/dyn/evolve/hdyn_ev.C
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\

