
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//
// Known Problems
//
// 1) Determination of h2 for unperturbed binary is not quite okay.
//    it should maintain the nearest neighbour? --- Well, seems to
//    do that already for most cases
//
// Question: To what extent should HARP-3 be integrated into KIRA?
//           To have a few garbage variables should be no problem.
//
//           One could define a derived class which might make
//           it easier to move to newer (if any...) hardware or
//           different software interface
//
//
//  hdyn_harp3.C: functions to use HARP-3 (GRAPE-4)
//.............................................................................
//    version 1:  Aug 1996   Jun Makino
//    version 2:  Aug 1998   Steve McMillan
//.............................................................................
//  This file includes:
//
//    -     harp_calculate_acc_and_jerk(next_nodes, n_next, eps * eps)
//          A single function which takes care of everything...
//
//    -     harp_release()
//          Releases the HARP-3 hardware so that someone else can use it
//          (can be called internally from  harp_calculate_acc_and_jerk
//          after a prescribed amount of CPU time has passed)
//
//    -	    Routines to determine neighbour lists (called from within
//	    harp_calculate_acc_and_jerk)
//
//  When called, it performs the following:
//
// -- In the first call, initialize the HARP-3 hardware
//
// -- If the tree structure is changed (or in the first call),
//    send all particles anew and store the number of particles
//
// -- If there are particles in "previous blockstep", which are not in
//    present block step, update these particles as well
//
// -- update the particles in the present block
//
// -- For block of particles,
//    --- Perform force calculation
//    --- store the force
//    --- retrieve the neighbor list
//    --- If no neighbor is found for any of particles, change
//        its neighbor sphere radius and try over again
//    --- In the case of CM particle, store the neighbors to the
//        perturber list
//    --- In the case of single particle, find the nearst neighbor
//    --- and "closest surface" particle. In order to be able to
//        find the closest surface particle, the search radius of
//        must be larger than 2*radius (at least the star with
//        larger radius can find colliding star)
//
// -- store the "previous block" particles

// Class variables related to HARP-3
//
// harp3_rnb_sq : neighbour radius (squared) sent to HARP to
// 		  calculate force
//
// It must be initialized to some value when a particle/node
// is created (or at the first call).  A reasonable value is
// r_min_sq (could be too small, but will be adjusted anyway;
// the details depend on the information required by kira).
//
// - For SINGLE particles, the internal routine should take care
//   of the updating of this variable.
//
// - For CENTER_OF_MASS, there are two possibilities
//   a) if it has a non-null valid perturber list, the perturbation
//	radius can be used.
//   b) if it does not have a non-null valid perturber list, ....
//   c) if it has an overflowed perturber list, I guess you
// 	should still set the perturbation radius as the neighbour
// 	sphere radius... 

//.............................................................................

// Note from Steve to Steve (7/9/97):
//
// force_by_harp3()	     is called only from harp_calculate_acc_and_jerk()
// force_by_harp3_on_leaves()		"	 force_by_harp3_on_all_leaves()
// force_by_harp3_on_all_leaves()	"	 calculate_energies_by_harp3()

// force_by_harp3 computes forces between TOP-LEVEL NODES in the
// POINT-MASS approximation (like flat_calculate_acc_and_jerk).

// force_by_harp3_on_all_leaves loads all LEAVES into the hardware and
// hence effectively resolves all multiples into components.

//.............................................................................

#include "hdyn.h"
#include "harp3.h"
#include "hdyn_inline.C"

static int nboards;

#define MINIMUM_HARP3_RNB_PERT 1e-20 // was 1e-20 changed by SPZ at 11/04/1998


#if 0

#   define HARPNMAX 100000
#   define HARPNBMAX HARPNMAX

    static hdyn* nodes[HARPNMAX];
    static hdyn* next_top[HARPNMAX];
    static hdyn* previous_nodes[HARPNMAX];
    static int nb_check_counter[HARPNMAX];
    static int h3nb[HARPNBMAX];

#else

    static hdyn** nodes = NULL;
    static hdyn** next_top;
    static hdyn** previous_nodes;
    static int* nb_check_counter;
    static vector * pxj = NULL;
    static vector * pvj = NULL;
    static vector * paj = NULL;
    static vector * pjj = NULL;
    static real   * ptj = NULL;
    static real   * pmj = NULL;
    static int    * ppj = NULL;
    static int    * h3nb = NULL;

#endif

static int n_previous_nodes = 0;

static int harp_n_max = 0;
local int put_harp3_index_to_top_level_nodes(hdyn*b)
{
    int index = 0;
    int node_count = 0;
    for_all_daughters(hdyn, b, bb)node_count ++;

    if (harp_n_max == 0) {

	harp_n_max = (int) (node_count * 1.5) + 10;

    } else if (harp_n_max < node_count) {

	harp_n_max = (int) (node_count * 1.5) + 10;

	delete nodes;
	delete next_top;
	delete previous_nodes;
	delete nb_check_counter;
	delete pxj;
	delete pvj;
	delete paj;
	delete pjj;
	delete ptj;
	delete pmj;
	delete ppj;
	delete h3nb;
	nodes = NULL;
    }

    if (nodes == NULL) {
	nodes = new hdynptr[harp_n_max];
	next_top = new hdynptr[harp_n_max];
	previous_nodes = new hdynptr[harp_n_max];

	// Initialize nb_check_counter[] to zero when first created.

	nb_check_counter = new int [harp_n_max];
	for (int i = 0; i < harp_n_max; i++)
	    nb_check_counter[i] = 0;

	pxj = new vector[harp_n_max];
	pvj = new vector[harp_n_max];
	paj = new vector[harp_n_max];
	pjj = new vector[harp_n_max];
	ptj = new real[harp_n_max];
	pmj = new real[harp_n_max];
	ppj = new int[harp_n_max];
	h3nb = new int[harp_n_max];
    }
    
    for_all_daughters(hdyn, b, bbb) {
	nodes[index] = bbb;
	index ++;
	bbb->set_harp3_index(index);
    }
    return index;
}

local int put_harp3_index_to_all_leaves_nodes(hdyn*b)
{
    int index = 0;
    for_all_leaves(hdyn, b, bb) {
	nodes[index] = bb;
	index ++;
	bb->set_harp3_index(index);
    }
    return index;
}

void jpdma_nodes(int nnodes, hdyn * nodelist[], bool predicted,
		 real predicted_time)

// Load the nodes on the list into the HARP.

{
    int jpmax;
    int j, jp;
    hdyn * b;
    jpmax = h3jpmax_();

    jp = 0;
    for (j = 0; j < nnodes; j++) {
	int jj, jm ;
	b = nodelist[j];

	jj =  b->get_harp3_index();
	jm = jj-1;
	if (! predicted) {
	    pxj[jm] = b->get_pos();
	    pvj[jm] = b->get_vel();
	    paj[jm] = b->get_acc()*0.5;
	    pjj[jm] = b->get_jerk()*0.1666666666666666666666666666666;
	    ptj[jm] = b->get_time();
	    pmj[jm] = b->get_mass();
	} else {
	    pxj[jm] = b->get_pred_pos();
	    pvj[jm] = b->get_pred_vel();
	    ptj[jm] = predicted_time;
	    pmj[jm] = b->get_mass();
	}
	ppj[jp] = jj;
	
	jp ++;
	int bufid = 0;
	if ((jp == jpmax) || (j == nnodes - 1)) {
	    int zero = 0;
	    int one = 1;
	    h3mjpdma_indirect_(&jp, ppj, pxj, pvj, paj, pjj, pmj, ptj,&one,
			       &bufid);
	    h3mjpdma_start_(&bufid);
	    bufid ++;
	    if (bufid > 5) bufid = 0; //this 5 should be < JPDMA_BUFFMAX...
	    h3wait_();
	    h3wait_();

	    jp = 0;
	}
    }
}

int jpdma_all_leaves(hdyn *root, bool predicted, real predicted_time)
{
    int jpmax = h3jpmax_();
    int nj = put_harp3_index_to_all_leaves_nodes(root);
    int jp = 0;
    for_all_leaves(hdyn,root,b) {
	int jj, jm ;
	jj =  b->get_harp3_index();
	jm = jj-1;

	// Allow the possibility of using already predicted pos and vel
	// rather than having the HARP do the prediction.

	if (!predicted) {
	    pxj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_pos);
	    pvj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_vel);
	    paj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_acc)*0.5;
	    pjj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_jerk)
       			   * 0.1666666666666666666666666666666;
	    ptj[jm] = b->get_time();
	    pmj[jm] = b->get_mass();
	} else {
	    pxj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_pred_pos);
	    pvj[jm] = hdyn_something_relative_to_root(b, &hdyn::get_pred_vel);
	    ptj[jm] = predicted_time;
	    pmj[jm] = b->get_mass();
	}
	ppj[jp] = jj;
	
	jp ++;
	int bufid = 0;
	if ((jp == jpmax) || (jj == nj)) {
	    int zero = 0;
	    int one = 1;
	    h3mjpdma_indirect_(&jp, ppj, pxj, pvj, paj, pjj, pmj, ptj,&one,
			       &bufid);
	    h3mjpdma_start_(&bufid);
	    bufid ++;
	    if (bufid > 5)	//this 5 should be < JPDMA_BUFFMAX...
	        bufid = 0;
	    h3wait_();
	    h3wait_();
	    
	    jp = 0;
	}
    }
    return nj;
}

void force_by_harp3(real time, int ni, hdyn * nodes[], int nj)

// Calculate the force on top-level nodes.

{
    static vector * px = NULL;
    static vector * pv = NULL;
    static vector * pa = NULL;
    static vector * pj = NULL;
    static real   * pp = NULL;
    static real   * peps2 = NULL;
    static real   * ph2 = NULL;
    int npipe;
    int i, ip,k;
    hdyn * b;
    npipe = h3npipe_();
    
    if (px == NULL) {
	px = new vector[npipe];
	pv = new vector[npipe];
	pa = new vector[npipe];
	pj = new vector[npipe];
	peps2 = new real[npipe];
	ph2 = new real[npipe];
	pp = new real[npipe];
	for (i=0; i<npipe; i++) {
	    peps2[i] = ph2[i] = 0.0;
	}
    }

#ifdef USE_HALF_CHIP    
    if (ni*2 > npipe)
	err_exit("force by harp 3: ni too large");
#else    
    if (ni > npipe)
	err_exit("force by harp 3: ni too large\n");
#endif    

    static real ti;
    ti = time;
    int dbg_mode = 0;
    if (dbg_mode) {cerr << "force by harp 3 "; PRL(time);}
    h3setti_(&ti);
    int j;
    for (i = 0; i < ni; i++) {
	b = nodes[i];

#ifndef USE_HALF_CHIP
	j = i;
#else	
	j= 2*i;
#endif	

	px[j] = b->get_pred_pos();
	pv[j] = b->get_pred_vel();
	ph2[j] = b->get_harp3_rnb_sq();
	if (dbg_mode) {
	    cerr << "force on  particle "; b->print_label(cerr);
	    PRL(j);
	    PRL(px[j]);
	    PRL(pv[j]);
	    PRL(ph2[j]);
	}

#ifdef USE_HALF_CHIP
	j++;
	px[j] = b->get_pred_pos();
	pv[j] = b->get_pred_vel();
	ph2[j] = b->get_harp3_rnb_sq();
#endif	

    }
    for (int jj = j + 1; jj < npipe; jj++) {
	px[jj] = b->get_pred_pos();
	pv[jj] = b->get_pred_vel();
	ph2[jj] = b->get_harp3_rnb_sq();
    }
    //    h3calc_(&nj, &npipe, px, pv,peps2,ph2, pa,pj,pp);
    h3calc_firsthalf_(&nj, &npipe, px, pv,peps2,ph2);
    h3wait_();
    h3calc_lasthalf_(&npipe,  pa,pj,pp);

#if 0
    {
	int i;
	for (i = 0; i < npipe; i++) {

	    fprintf(stderr,
	"##### %d %21.14e %21.14e %21.14e %21.14e %21.14e %21.14e %21.14e\n",
		    i,
		    pa[i*3+0], pa[i*3+1], pa[i*3+2],
		    pj[i*3+0], pj[i*3+1], pj[i*3+2],
		    pp[i]);
	}
    }
#endif



#if 0

    cerr.precision(18);
    cerr << endl << "time = " << time << "  ni = " << ni << endl;

#endif



    for (k = 0;k<ni;k++) {

#ifndef USE_HALF_CHIP
	int j = k;
#else
	int j= 2*k;
#endif

	nodes[k]->set_acc_and_jerk_and_pot(pa[j], pj[j], pp[j]);

	//	pp3(nodes[k], cerr);
	//	PR(pa[j]);PRL(pj[j]); PRL(pp[j]);



#if 0

	fprintf(stderr,	"%6d %25.18e %25.18e %25.18e\n",
		j, px[j][0], px[j][1], px[j][2]);
	fprintf(stderr,	"       %25.18e %25.18e %25.18e\n",
		pv[j][0], pv[j][1], pv[j][2]);
	fprintf(stderr,	"       %25.18e %25.18e %25.18e\n",
		pa[j][0], pa[j][1], pa[j][2]);
	fprintf(stderr,	"       %25.18e %25.18e %25.18e\n",
		pj[j][0], pj[j][1], pj[j][2]);

#endif


    }

    nboards = h3getnboards_();
    h3nbread_(&nboards);
    if (dbg_mode) {
	cerr << "exit force by harp 3 "; PRL(time);
    }
}

void force_by_harp3_on_leaves(real time, int ni, hdyn * nodes[], int nj)

// Note: All LEAVES, and all pipelines are used.

{
    static vector * px = NULL;
    static vector * pv = NULL;
    static vector * pa = NULL;
    static vector * pj = NULL;
    static real   * pp = NULL;
    static real   * peps2 = NULL;
    static real   * ph2 = NULL;
    int npipe;
    int i, ip,k;
    hdyn * b;
    npipe = h3npipe_();
    
    if (px == NULL) {
	px = new vector[npipe];
	pv = new vector[npipe];
	pa = new vector[npipe];
	pj = new vector[npipe];
	peps2 = new real[npipe];
	ph2 = new real[npipe];
	pp = new real[npipe];
	for (i=0;i<npipe;i++) {
	    peps2[i] = ph2[i] = 0.0;
	}
    }

    if (ni > npipe)
	err_exit("force by harp 3: ni too large\n");

    static real ti;
    ti = time;
    int dbg_mode = 0;
    if (dbg_mode) {cerr << "force by harp 3 "; PRL(time);}
    h3setti_(&ti);
    int j;
    for (i = 0; i < ni; i++) {
	b = nodes[i];

// 	if (b->get_harp3_index()%500 == 0) {
// 	    cerr << endl;
// 	    PRC(b->get_harp3_index()); PRL(b->get_pos() - b->get_pred_pos());
// 	    if (b->get_pos()[0] - b->get_pred_pos()[0] != 0) {
// 		PRL(b->get_system_time());
// 		PRL(b->get_time());
// 		PRL(b->get_t_pred());
// 		PRL(b->get_timestep());
// 		PRL(b->get_mass());
// 		PRL(b->get_pos());
// 		PRL(b->get_vel());
// 		PRL(b->get_pred_pos());
// 		PRL(b->get_pred_vel());
// 	    }
// 	}
	    
	j = i;
	px[j] = hdyn_something_relative_to_root(b, &hdyn::get_pred_pos);
	pv[j] = hdyn_something_relative_to_root(b, &hdyn::get_pred_vel);
	ph2[j] = 0;
	if (dbg_mode) {
	    cerr << "force on  particle "; b->print_label(cerr);
	    PRL(j);
	    PRL(px[j]);
	    PRL(pv[j]);
	    PRL(ph2[j]);
	}
    }
    for (int jj = j + 1; jj < npipe; jj++) {
	px[jj] = b->get_pred_pos();
	pv[jj] = b->get_pred_vel();
	ph2[jj] = b->get_harp3_rnb_sq();
    }
    h3calc_firsthalf_(&nj, &npipe, px, pv,peps2,ph2);
    h3wait_();
    h3calc_lasthalf_(&npipe,  pa,pj,pp);
    for (k = 0;k<ni;k++) {
	int j = k;
	nodes[k]->set_acc_and_jerk_and_pot(pa[j],pj[j],pp[j]);
    }
    if (dbg_mode) {cerr << "exit force by harp 3 "; PRL(time);}
}

void force_by_harp3_on_all_leaves(real time, hdyn * b, int nj)
{
    int npipe = h3npipe_();
    static hdyn ** nodes = NULL;

    if (!nodes)
	nodes = new hdynptr[npipe];

    int i = 0;
    int ip = 0;
    for_all_leaves(hdyn,b,bb) {
	nodes[ip] = bb;
	i ++;
	ip++;
	if (ip == npipe) {
	    force_by_harp3_on_leaves(time, ip, nodes, nj);
	    ip = 0;
	}
    }
    if (ip)
	force_by_harp3_on_leaves(time, ip, nodes, nj);
}

static bool harp3_was_used_to_calculate_potential = false;

void calculate_energies_by_harp3(hdyn * b,
				 real eps2,
				 real &epot,
				 real &ekin,
				 real &etot)
{
    real time = b->get_system_time();
    int nj =  jpdma_all_leaves(b, true, time);

    //                            ^^^^  use predicted pos and vel (unpert. OK)

    force_by_harp3_on_all_leaves(time, b,  nj);
    harp3_was_used_to_calculate_potential = true;

    epot =  ekin = etot = 0;

    for_all_leaves(hdyn,b,bb) {
	real mi = bb->get_mass();
	epot += 0.5*mi*bb->get_pot();
	vector vel = hdyn_something_relative_to_root(bb, &hdyn::get_vel);
	ekin += 0.5*mi*vel*vel;
    }
    etot = ekin + epot;


#if 0
    cerr << "grape: "; PRC(etot); PRC(ekin); PRL(epot);
    calculate_energies(b, eps2, epot, ekin, etot);
    cerr << "host:  "; PRC(etot); PRC(ekin); PRL(epot);
#endif

}

bool get_neighbors_and_adjust_h2(int chip, hdyn * b)

    // This function actually sets nn, coll, d_nn_sq and d_coll_sq.

    // As in the case of flat_calculate_..., these values overwrite
    // previous ones.  Previously set values are ignored.
{

    int nnb;
    bool no_nb = false;

    nnb = h3nblist_(&nboards, &chip, h3nb);	// get neighbor list for
						// current neighbor radius

    // cerr << "get_neighbors_and_adjust_h2 ";
    // PRC(chip);
    // PRL(nnb);
    // for (int j=0; j < nnb; j++) {
    //     PRC(j);PRL(h3nb[j]);
    // }
    // PRL(nnb);
    // pp3(b,cerr);

    // We will determine the nn of all stars (and coll, for leaves),
    // and the perturber lists for (the top-level) nodes.

    if (b->is_parent())
	b->new_perturber_list();

    // pp3(b,cerr);
    // cerr << "test_new_perturber end\n";

    if (nnb < 2) {

	no_nb = true;

    } else {

	// Found at least one neighbor -- find the nearest and
	// determine perturber list for a center-of-mass node.

        // cerr << "start search for the nearest\n";

	real dmin_sq = VERY_LARGE_NUMBER;
	hdyn * bmin = NULL;
	real dcmin_sq = VERY_LARGE_NUMBER;
	hdyn * cmin = NULL;
	int npl = 0;

	hdyn ** pl;
	real rpfac;

	if (b->is_parent()) {
	    pl = b->get_perturber_list();
	    rpfac = b->get_perturbation_radius_factor();
	}

	for (int j = 0; j < nnb; j++) {

	    hdyn * bb = nodes[h3nb[j]];

	    // bb is the j-th neighbor of b.

	    if (bb != b) {

		// Note that it is possible that pred_pos of bb
		// is garbage (never updated)...  Use get_pos
		// instead of get_pred_pos here.

		// Now, get_pred_pos performs Just-In-Time prediction
		// so the above case cannot occur (as of Dec 12, 1996)

		vector diff = b->get_pred_pos() - bb->get_pred_pos();
		real d2 = diff * diff;

		update_nn_coll(b, 100,		// (100 = ID)	    // inlined
			       d2, bb, dmin_sq, bmin,
			       b->get_radius() + bb->get_radius(),
			       dcmin_sq, cmin);

		// Recompute the perturber list for parent nodes.

		if (b->is_parent())
		    if (is_perturber(b->get_mass(),bb->get_mass(),
				     d2, rpfac)) {		    // inlined
			if (npl < MAX_PERTURBERS)
			    pl[npl] = bb;
			npl++;
		    }

	    }
	}

	if (b->is_parent())
	    b->set_n_perturbers(npl);

	if (bmin == NULL)
	    no_nb = true;
	else {
	    b->set_nn(bmin);
	    b->set_d_nn_sq(dmin_sq);
	}

	if (cmin != NULL) {
	    b->set_coll(cmin);
	    b->set_d_coll_sq(dcmin_sq);
	}	  

    }

    if (no_nb) {

    	// No nearest neighbor found --  enlarge the neighbour-sphere
	// radius and try again...

	b->set_harp3_rnb_sq(b->get_harp3_rnb_sq()*2);

	// b->print_label(cerr); cerr << " increasing h2 to ";
	// cerr << b->get_harp3_rnb_sq() << endl;

	return false;
    }

    return true;
}

// set_harp3_neighbour_radius: adjust the harp3 neighbour radius
//			       to some reasonable value.

void set_harp3_neighbour_radius(hdyn * b)
{
    int hindex = b->get_harp3_index()-1;
    
    if (b->is_leaf()) {

	// For a single particle, adjust the nnb radius so that it will
	// contain just 1-2 neighbours (set r = sqrt(2)*d_nn, if known).

	if (b->get_nn() != NULL
	    && b->get_d_nn_sq() < 0.1* VERY_LARGE_NUMBER) {

	    // Node seems to have a valid nearest neighbor pointer.
	    // We can perhaps use the distance as well.
	    
	    // Set zero neighbor sphere radius if nb_check_counter is non-zero.

	    if (nb_check_counter[hindex] == 0) {

		// Radius information included here to allow coll
		// criterion to be applied elsewhere.

		b->set_harp3_rnb_sq(max(1.0*b->get_d_nn_sq(),
					4*b->get_radius()*b->get_radius()));

		// In this case, if the actual value set is zero,
		// one might have to do something.

		if (b->get_harp3_rnb_sq() < MINIMUM_HARP3_RNB_PERT) {

		    // (Jun says this should never happen...)

		    cerr << "h2 set to zero for \n";
		    pp3(b,cerr);
		    PRL(b->get_d_nn_sq());
		    b->set_harp3_rnb_sq(b->get_d_min_sq());
		}

	    } else {

		b->set_harp3_rnb_sq(0.0);
	    }

	    // b->pretty_print_node(cerr); 
	    // PRC(b->get_d_nn_sq());
	    // PRC(b->get_radius());
	    // PRL(b->get_harp3_rnb_sq());

	} else {

	    // Node does not know its nearest neighbor.

	    nb_check_counter[hindex] = 0;
	    b->set_harp3_rnb_sq(pow(b->get_d_min_sq(), 1.0/3.0));

	    // Note: d_min_sq^(1/3) ~ square of the average interparticle
	    // spacing for standard N-body units -- OK for leaves and nn.
	    
	}

    } else {

	// For a node, we will want to compute the perturbers, so
	// we need a larger value of harp3_rnb_sq.

	// If the node does not have a valid perturber list,
	// either it has overflowed or nothing has been set before.

	// If it does have non-zero valid perturber list,
	// the perturbation radius should be used.

	if (b->get_nn() != NULL
	    && b->get_d_nn_sq() < 0.1* VERY_LARGE_NUMBER) {

	    b->set_harp3_rnb_sq(max(b->get_perturbation_radius_factor(),
				    b->get_d_nn_sq()));

	} else {

	    b->set_harp3_rnb_sq(b->get_perturbation_radius_factor());

	}	
    
	// Note that this can cause overflow, resulting in
	// not_valid_perturber anyway...

	nb_check_counter[hindex] = 0;

    }

    if (nb_check_counter[hindex] == 0 &&
	b->get_harp3_rnb_sq() < MINIMUM_HARP3_RNB_PERT) {
	cerr << "set_harp2_neighbor ... error \n",
	pp3(b,cerr);
	PRL(nb_check_counter[hindex]);
	PRL(b->get_harp3_rnb_sq());
	PRL(b->get_nn());
	PRL(b->get_d_min_sq());
    }
}

local int initialize_array(hdyn * root)
{
    // Make list of and index top-level nodes.

    int nj = put_harp3_index_to_top_level_nodes(root);

    for (int i = 0; i < (nj+10); i++)
	nb_check_counter[i] = 0;

    // Copy the nodes to the HARP.

    jpdma_nodes(nj, nodes, false, 0.0);

    // cerr << "init_array "; PRL(nj);

    return nj;
}

extern int  harp_check_count;
extern real harp_max_cpu;

void harp_calculate_acc_and_jerk(hdyn ** next_nodes,
				 int n_next,
				 real eps2,
				 real pert_factor_sq,
				 real time,
				 bool restart)
{
    static bool harp3_is_open = false;
    static int nj_on_harp3;
    static int step_counter = 0;
    static real harp_last_cpu  = 0;

    hdyn * root = next_nodes[0]->get_root();

    // if (restart) {
    // 	   cerr << "harp_calculate: restart\n";
    // }

    // Test the state of HARP-3 and open it if necessary.

    step_counter += n_next;

    bool harp3_reattached = false;

#ifdef SHARE_GRAPE4  
  
    if (step_counter > harp_check_count) {
	step_counter = 0;
	if (cpu_time() - harp_last_cpu > harp_max_cpu) {
	    cerr << "\nReleasing harp-3 at time " << time << " after "
		 << cpu_time() - harp_last_cpu <<" CPU sec\n";
	    h3close_();
	    harp3_is_open = false;
	    if (!restart) harp3_reattached = true;
	}
    }

#endif

    if (!harp3_is_open) {
	h3open_();
	set_time_check_mode(0);
	harp_last_cpu = cpu_time();
	int dbgl = 0;
	h3setdebuglevel_(&dbgl);
	harp3_is_open = true;

	// If newly opened, restart irrespective of the actual argument.

	restart = true;
    }

    // It is necessary to know when a restart has been triggered ONLY by
    // the release/reattachment of the HARP hardware.  Indicator is:
    //
    //		harp3_reattached = true

    if (harp3_was_used_to_calculate_potential) {
	restart = true;
	harp3_was_used_to_calculate_potential = false;
    }

    if (restart) {

        // cerr << "harp3 restart\n";

	if (harp3_reattached)		// Save the nb_check_counter[] array
	    for_all_daughters(hdyn, root, bi) {

// 		if (bi->name_is("(1376b,1376a)")) {
// 		    cerr << endl << bi->format_label() << "  dump 1" << endl;
// 		    dump_story(bi->get_dyn_story());
// 		}

		putiq(bi->get_dyn_story(), "nb_check_counter",
		      nb_check_counter[bi->get_harp3_index()-1]);

// 		if (bi->name_is("(1376b,1376a)")) {
// 		    cerr << endl << bi->format_label() << "  dump 2" << endl;
// 		    dump_story(bi->get_dyn_story());
// 		}

	    }

	nj_on_harp3 = initialize_array(root);

	if (harp3_reattached)		// Restore the nb_check_counter[] array
	    for_all_daughters(hdyn, root, bi) {

// 		if (bi->name_is("(1376b,1376a)")) {
// 		    cerr << endl << bi->format_label() << "  dump 3" << endl;
// 		    dump_story(bi->get_dyn_story());
// 		}

		if (find_qmatch(bi->get_dyn_story(), "nb_check_counter")) {
		    nb_check_counter[bi->get_harp3_index()-1] = 
			getiq(bi->get_dyn_story(), "nb_check_counter");
		    rmq(bi->get_dyn_story(), "nb_check_counter");
		}

// 		if (bi->name_is("(1376b,1376a)")) {
// 		    cerr << endl << bi->format_label() << "  dump 4" << endl;
// 		    dump_story(bi->get_dyn_story());
// 		}

	    }

	n_previous_nodes = 0;
    }

    // Create the list of TOP_LEVEL nodes in the present blockstep.

    int i, j;

    for (i = j = 0; i < n_next; i++) {
	if (next_nodes[i]->is_top_level_node()) {

	    next_top[j] = next_nodes[i];

	    // Whether prediction should be performed here
	    // or not is pretty unclear....

	    next_top[j]->predict_loworder(time);
	    j++;
	}
    }

    int n_top = j;	// number of top-level nodes in current list

    // Store the predicted positions of top-level
    // current-block nodes to HARP memory.

    jpdma_nodes(n_top, next_top, true, time);

    // cerr << "harp3 jpdma end\n";

    // Store the particles in the previous block which are not in
    // the present block (update HARP for previous step).

    // cerr << "prev_nodes ";

    for (i = j = 0; i < n_previous_nodes; i++) {
	if (previous_nodes[i]->get_next_time() > time) {
	    previous_nodes[j] = previous_nodes[i];
	    j ++ ;
	}
    }

    // cerr << "\nharp3 previous_node end\n";

    n_previous_nodes = j;
    jpdma_nodes(n_previous_nodes, previous_nodes, false, 0.0);

    // cerr << "\n harp3 jpdma2 end\n";

    for (i = 0; i < n_top; i++) {

	// store current list
        // cerr << "store h "; PRL(i);

	hdyn * b = previous_nodes[i] = next_top[i];

	// Set some reasonable h2 value.

	set_harp3_neighbour_radius(b);

	if (b->is_parent())
	    b->set_valid_perturbers(true);

    }

    n_previous_nodes = n_top;

    // cerr << "\nharp3 set h end\n";

    // Now we actually calculate the force...

#ifndef USE_HALF_CHIP    
    int nimax = h3npipe_();
#else
    int nimax = h3npipe_()/2;
#endif

    real h2_critical = next_top[0]->get_d_min_sq()*8192;

    // We will stop expanding the HARP neighbor sphere once its size
    // exceeds this critical value.  However, it is legal to set
    // harp3_rnb_sq greater than h2_critical -- the neighbor sphere
    // then simply won't be expanded if no neighbors are found.

    // *** Should contain a factor of ~(m_max/<m>)^2, but not crucial...

    // Note:  for equal-mass systems and standard units, this critical
    //	      radius is less than the interparticle spacing for N > ~1000.

    real d2_max = h2_critical * 2;

    i = 0;

    while (i < n_top) {

	int inext = i;
	int ip = min(nimax, n_top - i);

	force_by_harp3(time, ip, next_top+i, nj_on_harp3);

	for (int ichip = 0; ichip < ip ; ichip ++) {
	    int real_ichip;

#ifndef USE_HALF_CHIP
	    real_ichip = ichip;
#else
	    real_ichip = ichip*2;
#endif

	    hdyn * b = next_top[i+ichip];
	    int hindex = b->get_harp3_index()-1;

	    // PRC(ichip); PRL(real_ichip);

	    if (nb_check_counter[hindex] == 0) {

		if (b->get_harp3_rnb_sq() <= 0) {

		    // Should not happen!  About to enter an infinite loop...

		    err_exit("harp_calculate_acc_and_jerk: rnb_sq = 0");

		}

		// Compute neighbors/perturbers iff nb_check_counter = 0.

		if (next_top[i+ichip]->get_harp3_rnb_sq() < 1e-20) {
		    cerr << "h2 = 0 for \n";
		    pp3(b,cerr);
		    PRC(hindex);
		    PRL(nb_check_counter[hindex]);
		}

		int inew = i+ichip;
		int h2_too_large = 0;

		while (!(h2_too_large ||
			 get_neighbors_and_adjust_h2(real_ichip,b))) {

		    // PRC(inew); PRC(i); PRL(ichip);
		    // pp3(next_top[inew], cerr);

		    if (b->get_harp3_rnb_sq() > h2_critical) {

			h2_too_large = 1;	// ==> exit while loop

			b->set_nn(b);
			b->set_d_nn_sq(d2_max);

			// cerr << "no nb found for particle ";
			// b->print_label(cerr);
			// cerr << " at position " << b->get_pos();
			// cerr << " with radius " << b->get_harp3_rnb_sq()
			//      << endl;

		    } else {

			// Expand the neighbor sphere -- must recompute
			// the force.

			// cerr << "RETRY: no nb found for particle ";
			// b->print_label(cerr);
			// cerr << " at (" << b->get_pos() << ") with radius "
			//      << b->get_harp3_rnb_sq()<< endl;

			force_by_harp3(time, 1, next_top+inew, nj_on_harp3);

			real_ichip = 0;
			ichip = ip;
		    }
		}
	    }
	    inext ++;
	}
	i = inext;
    }

    for (i = 0; i < n_top ; i++) {

	hdyn * b = next_top[i];
	int hindex = b->get_harp3_index()-1;

	// Reduce frequency of nn checks (every fourth force calculation).

	if (b->is_leaf())
	    nb_check_counter[hindex] = (nb_check_counter[hindex] + 1)%4;
	else
	    nb_check_counter[hindex] = 0;

    }
}


//------------------------------------------------------------------------
//
// Functions relating to computation of densities.
//
//------------------------------------------------------------------------

#define DEBUG 0

local void set_harp3_density_radius(hdyn * b, real h2_max)
{
    int hindex = b->get_harp3_index() - 1;
    
    // For a single particle, adjust the radius so that it will
    // contain just ~10 neighbours (set r = 4*d_nn, if known).

    if (b->get_nn() != NULL
	&& b->get_d_nn_sq() < 0.1* VERY_LARGE_NUMBER
	&& b->get_d_nn_sq() > 0) {

	// Node seems to have a valid nearest neighbor pointer.

	b->set_harp3_rnb_sq(9 * b->get_d_nn_sq());

    } else

	// Node does not know its nearest neighbor.

	b->set_harp3_rnb_sq(9 * pow(b->get_d_min_sq(), 1.0/3.0));

}

local bool count_neighbors_and_adjust_h2(int chip, hdyn * b)
{
    int nnb = h3nblist_(&nboards, &chip, h3nb);		// get neighbor list

    if (nnb < 15) {

	if (DEBUG)
	    cerr << "increasing harp3_rnb_sq for " << b->format_label()
		 << " (nnb = " << nnb << ", harp_rnb = "
		 << sqrt(b->get_harp3_rnb_sq()) << ")" << endl;

	real fac = 2;
	if (nnb < 4) fac = 4;
	b->set_harp3_rnb_sq(fac * b->get_harp3_rnb_sq());

	return false;
    }

    // Make a list of nodes to send to compute_com().

    dyn** dynlist = new dynptr[nnb];

    real d_max = 0;
    for (int j = 0; j < nnb; j++) {

	hdyn * bb = nodes[h3nb[j]];
	dynlist[j] = (dyn*)bb;

	// bb is the j-th neighbor of b.

	if (DEBUG) {
	    if (bb != b) {
		vector diff = b->get_pred_pos() - bb->get_pred_pos();
		real d2 = diff * diff;
		d_max = max(d_max, d2);
	    }
	}
    }

    if (DEBUG) {
	real harp_rnb = sqrt(b->get_harp3_rnb_sq());
	d_max = sqrt(d_max);
	cerr << b->format_label() << ": ";
	PRC(nnb), PRC(harp_rnb), PRL(d_max);
    }

    compute_density(b, 12, dynlist, nnb);	// writes to dyn story

    // Strangely, compute_density sometimes fails to write the proper
    // info to the dyn story.  Reason and circumstances still unknown.

    if (find_qmatch(b->get_dyn_story(), "density_time")
	&& find_qmatch(b->get_dyn_story(), "density_k_level")
	&& find_qmatch(b->get_dyn_story(), "density")) {

	real density_time = getrq(b->get_dyn_story(), "density_time");
	int  density_k    = getiq(b->get_dyn_story(), "density_k_level");
	real density      = getrq(b->get_dyn_story(), "density");

	if (DEBUG)
	    PRL(density);
    }

    delete [] dynlist;
    return true;
}

void harp_calculate_densities(hdyn* b, real h2_crit)

// Determine particle densities, assigning zero density to particles
// with no neighbor within sqrt(h2_crit).

{
    int nj_on_harp3 = initialize_array(b);

    // Make a list of all top-level nodes.

    hdyn** top_nodes = new hdynptr[b->n_daughters()];

    int n_top = 0;
    for_all_daughters(hdyn, b, bb)
	top_nodes[n_top++] = bb;

    // Store the predicted positions of top-level
    // current-block nodes to HARP memory.

    real time = b->get_system_time();
    jpdma_nodes(n_top, top_nodes, true, time);

    // Set h2 values.

    for (int i = 0; i < n_top; i++)
	set_harp3_density_radius(top_nodes[i], h2_crit);

#ifndef USE_HALF_CHIP
    int nimax = h3npipe_();
#else
    int nimax = h3npipe_()/2;
#endif

    int n_retry = 0;

    int i = 0;
    while (i < n_top) {

	int inext = i;
	int ip = min(nimax, n_top - i);

	force_by_harp3(time, ip, top_nodes+i, nj_on_harp3);

	for (int ichip = 0; ichip < ip ; ichip ++) {

#ifndef USE_HALF_CHIP
	    int real_ichip = ichip;
#else
	    int real_ichip = ichip*2;
#endif

	    hdyn * bb = top_nodes[i+ichip];
	    int hindex = bb->get_harp3_index()-1;

	    // Just count neighbors, for now.

	    int inew = i+ichip;
	    int h2_too_large = 0;

	    while (!(h2_too_large ||
		     count_neighbors_and_adjust_h2(real_ichip, bb))) {

		if (bb->get_harp3_rnb_sq() > h2_crit) {

		    // Write zero density to dyn story.

		    putrq(bb->get_dyn_story(), "density_time",
			  bb->get_system_time());
	            putrq(bb->get_dyn_story(), "density", 0.0);

		    if (DEBUG) {
			PR(bb->get_harp3_rnb_sq());
			cerr << " too large for "
			     << bb->format_label() << endl;
		    }

		    h2_too_large = 1;	// ==> exit while loop

		} else {

		    // Expand the neighbor sphere -- must recompute
		    // the force.

		    force_by_harp3(time, 1, top_nodes+inew, nj_on_harp3);
		    n_retry++;

		    real_ichip = 0;
		    ichip = ip;
		}
	    }

	    if (DEBUG) {
		real dens = getrq(bb->get_dyn_story(), "density");
		cerr << i << "  (" << bb->format_label() << ")  "; PRL(dens);
	    }

	    inext++;
	}
	i = inext;
    }

    // Timestamp the root node.

    putrq(b->get_dyn_story(), "density_time", b->get_system_time());

    if (n_retry > 10) {
	cerr << "harp_calculate_densities:  ";
	PRL(n_retry);
    }

    // Force cleanup later.

    harp3_was_used_to_calculate_potential = true;
    delete [] top_nodes;

}

    
//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\
//  |  the end of:  |         /|\         |  src/dyn/evolve/hdyn_harp3.C
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\
