
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//  hdyn_schedule.C: scheduling routines
//.............................................................................
//    version 1:  Nov 1996   Jun Makino
//    version 2:
//.............................................................................
//  This file includes:
//    -     fast_get_nodes_to_move
//          A single function which takes care of everything ...
//.............................................................................

#include "hdyn.h"

//-----------------------------------------------------------------------------
// get_nodes_to_move:  Return the nodes with minimum t+dt.  Use a recursive
//		       loop to traverse the system.
//-----------------------------------------------------------------------------

inline void swap(hdyn *a[], int i, int j)
{
    hdyn * tmp;
    tmp = a[i];
    a[i] = a[j];
    a[j] = tmp;
}


void quicksort(hdyn *a[], int left, int right)
{
    // Sort nodes in array a, according to get_next_time().

    int i,j;
    real v;

    if (right > left) {
	bool is_sorted = true;
	for (i=left; i< right; i++) {
	    if (a[i]->get_next_time() > a[i+1]->get_next_time()) {
		is_sorted = false;
		i = right;
	    }
	}
	if (is_sorted) return;
	i = (left + right)/2;
	swap(a,i,right);
	v = a[right]->get_next_time(); i = left- 1; j = right;
	for (;;) {
	    while(a[++i]->get_next_time() <  v && i < right) ;
	    while(a[--j]->get_next_time() >  v&& j > left) ;
	    if (i >= j) break;
	    swap(a, i, j);
	}
	swap(a, i, right);
	quicksort(a, left, i - 1);
	quicksort(a, i + 1, right);
    }
}


void sort_node_array(hdyn * nodes[], int nprev, int nbody)
{

#if 0
    cerr << "before sort\n";
    for (int i = 0; i < nprev; i++) {
	PRC(i);nodes[i]->print_label(cerr); PRL(nodes[i]->get_next_time());
    }
#endif

    // Sort the previous list (bodies last stepped forward).

    quicksort(nodes, 0, nprev-1);

#if 0
    for (int i = 0; i < nprev; i++) {
	PRC(i);nodes[i]->print_label(cerr); PRL(nodes[i]->get_next_time());
    }
#endif

    // Perhaps some check is necessary for unpert motion,
    // whose timestep is integer multiple of the basic step.

    if (nprev < nbody) {

	if (nodes[nprev-1]->get_next_time() > nodes[nprev]->get_next_time()) {
	    int i;
	    for (i = nprev; i < nbody; i++) {
		if (nodes[nprev-1]->get_next_time()
		    <= nodes[i]->get_next_time()) {

#if 0
		    cerr << "sort_node_array: correct sequence for ";
		    cerr << nprev-1  << " "; nodes[nprev-1]->print_label(cerr);
		    cerr << " and ";
		    cerr << i  << " "; nodes[i]->print_label(cerr);
		    cerr << " "<< nodes[nprev-1]->get_next_time() << endl;
#endif		    

		    quicksort(nodes, 0, i);
		    return;
		}
	    }
	    quicksort(nodes, 0, nbody-1);
	}
    }
}


static int work_size = 0;
static hdyn ** nodes;
static int nbody = 0;
static int nprev;

// ***** NOTE that the scheduling may become badly corrupted if bodies not
// ***** on the list have been integrated (e.g. by synchronize_tree()).
// ***** If this is done, then a reset should be forced.

void fast_get_nodes_to_move(hdyn * b,
			    hdyn * list[],
			    int &nlist,
			    real & tmin,
			    bool & reset)
{
    if (nbody == 0) reset = true;

    if (reset) {

	// Initialize the array of node pointers.

	int n = 0;
	for_all_nodes(hdyn,b,bb) n ++;
	if (work_size < n) {
	    nodes = new hdynptr[n*2];
	    work_size = n*2;
	}
	n = 0;
	for_all_nodes(hdyn,b,bbb) {
	    if (!bbb->is_root()
		&& (bbb->is_top_level_node()
		    || bbb->get_elder_sister() == NULL)) {
		nodes[n] = bbb;
		n++ ;
	    }
	}
	nbody = nprev = n;
    }

    reset = false;

    bool debug = false;
    // debug = (b->get_system_time() > 501.97944641);

    sort_node_array(nodes, nprev, nbody);

    nlist = 0;
    tmin = nodes[0]->get_next_time();
    for (int i = 0; i < nbody; i++) {
	if (nodes[i]->get_next_time() < tmin) {
	    cerr << endl << "Error in fast_get_nodes_to_move():" << endl;
	    PRL(tmin);
	    PRC(i); cerr << "node " << nodes[i]->format_label() << endl;
	    PRL(nodes[i]->get_next_time());
	    for (int j = 0; j <= i; j++) {
		PRC(j);
		cerr << nodes[j]->format_label() << "  "
		     << nodes[j]->get_time() << "  "
		     << nodes[j]->get_next_time() << endl;
	    }
	    err_exit("fast_get_nodes_to_move: internal error: sort failed\n");
	} else if (nodes[i]->get_next_time() > tmin) {
	    i = nbody;	// aka break!
	} else {
	    list[i] = nodes[i];
	    nlist = i + 1;
	}
    }
    nprev = nlist;

    if (debug) {
      cerr << endl;
      cerr.precision(12);
      PRC(nbody); PRC(nlist); PRC(tmin);
      cerr.precision(8);
      cerr << list[0]->format_label() << endl;
    }
}


void dump_node_list(int n)
{
    cerr << "\nCurrent node list (nbody = " << nbody << "):\n";
    for (int i = 0; i < nbody && i < n; i++) {
        cerr << i << "  " << nodes[i]->format_label()
	     << "  " << nodes[i]->get_next_time() << endl;
    }
    cerr << endl;
}
