
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//
//  hdyn_tree.C: functions related to tree evolution in kira.
//.............................................................................
//    version 1:  Nov 1995   Jun Makino, Steve McMillan
//.............................................................................
//
//  Externally visible functions:
//
//	real   hdyn::distance_to_sister_squared
//	hdyn*  hdyn::new_sister_node
//	int    hdyn::adjust_tree_structure
//	hdyn*  hdyn::merge_nodes
//	bool   hdyn::check_merge_nodes
//
//.............................................................................
//
//  NOTE (SLWM, 7/97): when a binary node is created, update_binary_sister
//  is not called, so the components' timesteps, etc. may not be up to date
//  until the next time step is taken.  This is not generally a problem,
//  but may cause difficulties in set_unperturbed_timestep if a lower-level
//  node takes a step before the higher-level binary is advanced.
//
//.............................................................................

#include "hdyn.h"
#include <star/dstar_to_kira.h>
#include <star/single_star.h>
#include "kira_local.h"
#include "kira_debug.h"

// Note:  The *only* places in kira where "close" is defined, for
//	  purposes of making binary trees, are too_close() and
//	  new_sister_node().

#define CLOSE_CRITERION	2	// 0: "close" criterion based on distance
				// 1:				 potential
				// 2:				 force

// Both criteria 1 and 2 expand the binary distance criterion as masses
// increase.  Both appear to work (Steve, 7/29/98).

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define DEBUG_TREE		false
//#define TREE_DEBUG_LEVEL	0

local void init_predictor_tree(hdyn * b)
{
    if (! b->is_root()) {
	b->init_pred();
	b->store_old_force();
    }
    for_all_daughters(hdyn,b,bb) init_predictor_tree(bb);
}

void synchronize_tree(hdyn * b, real eps, real eta, real max_timestep)
{
    if (!b->is_root())
	b->synchronize_node(eps, eta, max_timestep);

    for_all_daughters(hdyn, b, bb)
	synchronize_tree(bb, eps, eta, max_timestep);
}

real hdyn::distance_to_sister_squared()
{
    vector d_sep = pos - get_binary_sister()->get_pos();
    return d_sep * d_sep;
}

local inline real max_mass(hdyn* b)
{
    real m_max = 0;
    if (b->is_parent()) {
	for_all_daughters(hdyn, b, bb)
	    if (bb->get_mass() > m_max)
		m_max = max(m_max, max_mass(bb));
    } else
	m_max = b->get_mass();

    return m_max;
}

local inline real n_mass(hdyn* b)
{
    // Choose which mass we use in the binary criterion.

    // Must be careful to avoid "polymerization" of clumps containing
    // very massive stars.  Try limiting the mass of a node to the
    // maximum mass of any of its leaves...

    // return b->get_mass();

    return max_mass(b);
}

local bool too_close(hdyn * bi, hdyn * bj, real limit_sq,
		     bool print = true)
{
    if (bi == bj)
	return false;

    real gap_sq = square(bi->get_pos() - bj->get_pos());

    // Binary criterion modified 7/98 by SLWM and SPZ to depend on
    // acceleration; modified again by SLWM and JM to include potential.
    // The criteria below are chosen to reduce to the simple distance
    // expressions when all masses are equal.

    bool close;
    real mass_factor = 1;		// default (DISTANCE criterion:
					//		rij  <  d_min)

    if (CLOSE_CRITERION > 0) {

	mass_factor = 0.5 * (n_mass(bi) + n_mass(bj)) / bi->get_mbar();

	if (CLOSE_CRITERION == 1) {

	    // POTENTIAL criterion:  (mi + mj) / rij  >  2 <m> / d_min
	    //
	    //		==>  rij  <  0.5 * (mi+mj) * d_min / <m>

	    mass_factor *= mass_factor;

	} else if (CLOSE_CRITERION == 2) {

	    // FORCE criterion:  (mi + mj) / rij^2  >  2 <m> / d_min^2
	    //
	    //		==>  rij^2  <  0.5 * (mi + mj) * d_min^2 / <m>

	}

    }

    close = (gap_sq < mass_factor * limit_sq);

    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0 && print && close) {
	cerr << endl << "*** "; bi->print_label(cerr);
	cerr << " too close to "; bj->print_label(cerr);
	cerr << " (criterion " << CLOSE_CRITERION << ")" << endl;
    }

    return close;
}

local bool too_big(hdyn * bi, real limit_sq)
{
    hdyn *od = bi->get_oldest_daughter();
    if (od == NULL)
	return FALSE;

    // The next if statement prevents an unperturbed binary from being
    // split even if it is large.  This must not cause any real error
    // at present (as of Nov 26, 1994), the synchronization of 
    // the node is not quite correctly implemented for low-level
    // node, particularily for unperturbed ones.  We must fix this
    // anyway. 

    if (od->get_kepler() != NULL)
	return FALSE;

    bool big = !too_close(od, od->get_younger_sister(), limit_sq, false);

    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0 && big)
	cerr << endl << "*** " << bi->format_label() << " too big\n";

    return big;
}

// new_sister_node:  Check to see if a given node is too near any other.
//                   Return a pointer to the node that should become the
//                   new sister, or NULL if the tree is OK as is.

hdyn* hdyn::new_sister_node(real lag_factor, bool & top_level_combine)
{
    top_level_combine = false;	// if true, combine top level nodes on return;
    				// if false, combine lower level nodes if a
				//    non-NULL value is returned

    // Note on multiples from Steve, 10/21/98.  Some care is needed
    // in dealing with multiples because of a "feature" of function
    // calculate_acc_and_jerk_on_low_level_node (hdyn_ev.C):
    //
    // If (A, B) is the top-level binary of a multiple, and if A is
    // a leaf with nn under B, then both A and B's nn pointers will
    // be correctly set (A to the leaf under B, B to A).  However, if
    // A is a composite node containing B's nn, then A's nn will be
    // B, but B's nn will be A (efficiency issue in hdyn_ev).  Thus,
    // a close encounter between triple members will never be picked
    // up on integration of the outer binary if A is composite (but
    // all is well if A is a leaf).  An exchange in this case must
    // be detected when A's internal motion is integrated (leaves
    // under A have the correct nn pointers), but proper account must
    // then be taken of the fact that B's nn is not a leaf, but A.
    // Failure to do this can lead to runaway timesteps (--> 0) when
    // a close encounter goes undetected.

    if (nn->get_parent() == parent)	// already sisters,
					// so no change needed
	return NULL;

    real mass_factor = 1;		// default (DISTANCE criterion)

    if (CLOSE_CRITERION > 0) {

	real m = n_mass(this);
	mass_factor = (m + n_mass(get_binary_sister())) / (m + n_mass(nn));

	if (CLOSE_CRITERION == 1) {

	    // POTENTIAL criterion:

	    mass_factor *= mass_factor;

	} else if (CLOSE_CRITERION == 2) {

	    // FORCE criterion:

	}

    }

    if (distance_to_sister_squared() > d_nn_sq * lag_factor * mass_factor) {
	
	// Nearest neighbor is closer than binary sister.
	//
	// Should check whether or not this and nn are mutual nearest
	// neighbors.  If they are, then we are done.  If not, then we
	// must be a little more careful.  We should further check if
	// one of nn's ancestors can become the new sister.
	//
	//    Criterion:  If the nearest neighbor of any ancestor
	//    of nn is a component of this node, connect these two.

	hdyn * new_sister = nn;
	hdyn * root = get_root();

	// The nn of new_sister may be a node (see note above).  Take care
	// of this possibility before proceeding.  Expect that the nn is
	// the binary sister (should be the only way in which this can
	// occur -- see hdyn_ev), but do the search generally, anyway.

	if (new_sister->nn != NULL
	    && new_sister->nn->is_valid()
	    && !new_sister->nn->is_leaf()) {

	    // The true nn lies below new_sister->nn.  Locate it, and
	    // modify new_sister->nn accordingly.  Its distance should
	    // be new_sister->d_nn_sq, but let's not tempt fate...

	    real d_sq_min = VERY_LARGE_NUMBER;
	    hdyn * local_nn = NULL;
	    vector s_pos = hdyn_something_relative_to_root(new_sister,
							   &hdyn::get_pos);

	    for_all_leaves(hdyn, new_sister->get_nn(), bb) {
		vector b_pos = hdyn_something_relative_to_root(bb,
							       &hdyn::get_pos);
		real d_sq = square(b_pos - s_pos);
		if (d_sq < d_sq_min) {
		    local_nn = bb;
		    d_sq_min = d_sq;
		}
	    }

	    if (local_nn) {

		new_sister->set_nn(local_nn);

		if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0) {
		    cerr << "new_sister_node:  computed local nn for "
			 << new_sister->format_label();
		    PR(local_nn);
		    cerr << "  (" << local_nn->format_label() << ")" << endl;
		    PRC(new_sister->d_nn_sq); PRL(d_sq_min);
		}
	    }
	}

#if 1
	if (new_sister->nn != NULL
	    && new_sister->nn->is_valid()
	    && common_ancestor(new_sister->nn, this) != this) {

	    hdyn* ns = new_sister;

	    do {
		new_sister = new_sister->get_parent();
	    } while(new_sister != root
		    && new_sister->nn != NULL
		    && new_sister->nn->is_valid()
		    && new_sister->parent != parent
		    && common_ancestor(new_sister->nn, this) != this);

	    if (new_sister == root
		|| new_sister->nn == NULL
		|| !new_sister->nn->is_valid()
		|| new_sister->parent == parent
		|| common_ancestor(new_sister, this)== this
		|| common_ancestor(new_sister, this)== new_sister
		) return NULL;

	    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0) {
		cerr << "new_sister_node:  ascended tree from "
		     << ns->format_label();
		cerr << " to " << new_sister->format_label() << endl;
	    }
	}
#endif

	hdyn * ancestor = common_ancestor(this, new_sister);
	if (ancestor->is_root()) {
	    top_level_combine = TRUE;
	    return new_sister;
	}

	// The new sister node is in the same subtree.
	// Attach this node to the location as high up as possible.

	if (parent != ancestor) {
	    while (new_sister->get_parent() != ancestor) {
		new_sister = new_sister->get_parent();
	    }
	}

	if (new_sister->kep == NULL) {
	    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0) {
	        cerr << "new sister node for "; pretty_print_node(cerr);
		cerr << " (nn = "; nn->pretty_print_node(cerr);
		cerr << ") is " ; new_sister->pretty_print_node(cerr);
		PRI(2); PRL(top_level_combine);
	    }
	    return new_sister;
	}
    }
    return NULL;
}

local void halve_timestep(hdyn * b)
{
    b->set_timestep(0.5 * b->get_timestep());
}

local void print_relative_energy_and_period(hdyn* bi, hdyn* bj)
{
    real M = bi->get_mass() + bj->get_mass();
    real E = -M / abs(bi->get_pos() - bj->get_pos())
			+ 0.5 * square(bi->get_vel() - bj->get_vel());

    cerr << "relative E/mu = " << E << flush;

    if (E < 0) {
	real a = -0.5 * M / E;
	real P = TWO_PI * sqrt(a*a*a/M);
    	cerr << "  P = " << P;				// no endl, note
    }
}

void combine_top_level_nodes(hdyn * bj, hdyn * bi,
			     real eps, real eta, real max_timestep)
{
    if (DEBUG_TREE) {
        cerr << endl << "combine_top_level_nodes: combining ",
	bi->pretty_print_node(cerr);
	cerr << " and ", bj->pretty_print_node(cerr);
	cerr << " at time " << bj->get_system_time();
	if (TREE_DEBUG_LEVEL > 0) {
	    cerr << endl;
	    pp2(bi, cerr);
	    print_binary_from_dyn_pair(bi, bj);
	    if (bi->is_parent())
		print_binary_from_dyn_pair(bi->get_oldest_daughter(),
					   bi->get_oldest_daughter()
					     ->get_younger_sister());
	    if (bj->is_parent())
		print_binary_from_dyn_pair(bj->get_oldest_daughter(),
					   bj->get_oldest_daughter()
					     ->get_younger_sister());
	    if (TREE_DEBUG_LEVEL > 1) {

//		Not really necessary if pp3 (new node) occurs below:
//
// 		cerr << endl;
// 		pp3(bi, cerr);
// 		pp3(bj, cerr);

		if (TREE_DEBUG_LEVEL > 2) {
		    cerr << endl;
		    put_node(cerr, *bi);
		    put_node(cerr, *bj);
		}
		cerr << endl;
	    }
	} else {
	    cerr << endl << "                         ";
	    print_relative_energy_and_period(bi, bj);
	    cerr << endl;
	}
    }

    // Make sure bi and bj are up to date.

    predict_loworder_all(bi->get_root(), bi->get_system_time());

    // print_recalculated_energies(bi->get_root(), 1, eps * eps);
    // pp3((bi->get_root()), cerr);
    // synchronize_tree(bi->get_root(),time, eps, eta, max_timestep);
    // print_recalculated_energies(bi->get_root(), 1, eps * eps);
    // pp3((bi->get_root()), cerr);

    bi->synchronize_node(eps, eta, max_timestep);
    bj->synchronize_node(eps, eta, max_timestep);

    create_binary_from_toplevel_nodes(bi, bj);

    // Reset all perturber lists below the newly-formed top-level
    // node -- probably not necessary for bi and bj if low-level
    // perturber lists are allowed, but cleaner to clear and
    // recompute the lists.

    for_all_nodes(hdyn, bj->get_parent(), bb) {
	bb->set_valid_perturbers(false);
	bb->set_perturbation_squared(-1);
    }

    // Shrink timesteps for safety...

    // halve_timestep(bi);
    // halve_timestep(bj);
    halve_timestep(bj->get_parent());	// parent step should be less than the
					// daughter steps so that the parent
					// perturber list is computed before
					// the internal motion is advanced

    bi->init_pred();
    bj->init_pred();
    bj->get_parent()->init_pred();

    // for DEBUG
    init_predictor_tree(bj->get_parent());

    bi->get_kira_counters()->top_level_combine++;

    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 1)
	pp3(bi->get_parent(), cerr);
}

// synchronize_branch:  Synchronize parent and sister nodes up to, but
//                      not including, the specified ancestor.

local void synchronize_branch(hdyn * bi, hdyn * ancestor,
			      real eps, real eta, real max_timestep)
{
    if (!is_descendent_of(bi, ancestor, 0))
	err_exit("synchronize_branch: bi not descended from ancestor");

    for (hdyn * bb = bi; bb != ancestor; bb = bb->get_parent()) {

	// Synchronize sister node:

	bb->get_binary_sister()->synchronize_node(eps, eta, max_timestep);

	// Synchronize parent:

	bb->get_parent()->synchronize_node(eps, eta, max_timestep);
    }
}

void split_top_level_node(hdyn * bi, real eps, real eta, real max_timestep)
{
    if (!bi->is_top_level_node())
	err_exit("split_top_level_node: not at top level");

    if (DEBUG_TREE) {
        cerr << endl << "split_top_level_node: splitting ";
	bi->pretty_print_node(cerr);
	cerr << " at time " << bi->get_system_time();
	if (TREE_DEBUG_LEVEL >= 1) {
	    cerr << endl;
	    pp2(bi, cerr);
	    print_binary_from_dyn_pair(bi->get_oldest_daughter(),
				       bi->get_oldest_daughter()
				         ->get_younger_sister());
	    if (TREE_DEBUG_LEVEL > 1) {
		cerr << endl;
		pp3(bi, cerr);
		put_node(cerr, *bi);
	    }
	} else {
	    cerr << endl << "                      ";
	    print_relative_energy_and_period(bi->get_oldest_daughter(),
					     bi->get_oldest_daughter()
					       ->get_younger_sister());
	    cerr << endl;
	}
    }

    hdyn *od = bi->get_oldest_daughter();
    hdyn *yd = od->get_younger_sister();

    // Synchronize the nodes involved.

    predict_loworder_all(bi->get_root(), bi->get_system_time());

    // for DEBUG ...
    // synchronize_tree(bi->get_root(),eps, eta, max_timestep);
    // synchronize_tree(bi, eps, eta, max_timestep);

    bi->synchronize_node(eps, eta, max_timestep);
    od->synchronize_node(eps, eta, max_timestep);
    update_binary_sister(od);

    // Express od and yd quantities relative to root.

    od->set_pos(hdyn_something_relative_to_root(od, &hdyn::get_pos));
    od->set_vel(hdyn_something_relative_to_root(od, &hdyn::get_vel));
    od->set_acc(hdyn_something_relative_to_root(od, &hdyn::get_acc));
    od->set_jerk(hdyn_something_relative_to_root(od, &hdyn::get_jerk));
    od->store_old_force();
    yd->set_pos(hdyn_something_relative_to_root(yd, &hdyn::get_pos));
    yd->set_vel(hdyn_something_relative_to_root(yd, &hdyn::get_vel));
    yd->set_acc(hdyn_something_relative_to_root(yd, &hdyn::get_acc));
    yd->set_jerk(hdyn_something_relative_to_root(yd, &hdyn::get_jerk));
    yd->store_old_force();

    od->init_pred();
    yd->init_pred();

    // May not be necessary to do this if low-level perturbers are
    // allowed, but cleaner to recompute the lists.  Clear all
    // perturber lists below the new top-level nodes.

    {for_all_nodes(hdyn, od, bb) {
	bb->set_valid_perturbers(false);
	bb->set_perturbation_squared(-1);
    }}
    {for_all_nodes(hdyn, yd, bb) {
	bb->set_valid_perturbers(false);
	bb->set_perturbation_squared(-1);
    }}

    // Dissolve the parent node and attach od and yd at top level.

    hdyn *root = bi->get_root();

    detach_node_from_general_tree(*bi);

    // Imperfect correction of neighbor pointers:

    if (bi->get_nn()->get_nn() == bi)
	bi->get_nn()->set_nn(NULL);

    // Should also check for the *very* unlikely possibility that:
    //
    //	(1) binary bi is unperturbed, and
    //	(2) binary bi is on the perturber list of some other binary.
    //
    // If so, bi must be removed from the perturber list before being
    // deleted.  NOT done yet... (Steve, 8/20/98)

    delete bi;

    add_node(*od, *root);
    add_node(*yd, *root);

    // halve_timestep(od);
    // halve_timestep(yd);

    bi->get_kira_counters()->top_level_split++;
}

local void combine_low_level_nodes(hdyn * bi, hdyn * bj,
				   real eps, real eta, real max_timestep)
{
    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 0) {
        cerr << "\ncombine_low_level_nodes:  combining ";
	bi->pretty_print_node(cerr);
	cerr << " and ";
	bj->pretty_print_node(cerr);
	cerr << " at time " << bi->get_system_time() << "\n";
    }

    hdyn *ancestor;
    ancestor = common_ancestor(bi, bj);

#if 0
    cerr << "combine_low_level_nodes:" << endl;
    cerr << "bi = " << bi->format_label() << endl;
    cerr << "bj = " << bj->format_label() << endl;
    cerr << "ancestor = " << ancestor->format_label() << endl;
    pp3_minimal(ancestor);
#endif

    // Make sure bi and bj are up to date.

    predict_loworder_all(bi->get_root(), bi->get_system_time());

    bi->synchronize_node(eps, eta, max_timestep);
    bj->synchronize_node(eps, eta, max_timestep);

    synchronize_branch(bi, ancestor, eps, eta, max_timestep);
    synchronize_branch(bj, ancestor, eps, eta, max_timestep);

    move_node(bi, bj);

    // halve_timestep(bi);
    // halve_timestep(bj->get_parent());

    predict_loworder_all(bi->get_root(), bi->get_system_time());

    // Force all perturber lists in the present subtree to be rebuilt.
    // Recompute the ancestor (should simply be the new parent node of
    // bi and bj), since move_nodes may delete the original.

    ancestor = common_ancestor(bi, bj);

    for_all_nodes(hdyn, ancestor, bb) {
	bb->set_valid_perturbers(false);
	bb->set_perturbation_squared(-1);

	// cerr << "cleared perturber lists for " << bb->format_label() << endl;
    }

    bi->get_kira_counters()->low_level_combine++;
}

local bool number_of_daughter_nodes_exceeds(node * b, int n)
{
    int nn = 0;

    for_all_daughters(node, b, bb)
	if (++nn > n)
	    return true;

    return false;
}

local int adjust_low_level_node(hdyn * bi, real lag_factor,
				real eps, real eta, real max_timestep)
{
    int adjusted = 1;

    bool top_level_combine;	// TRUE iff top-level nodes are to be combined.
				// Takes precedence over the returned value of
				// new_sister_node.  If new_sister_node is
				// non-NULL, it is the node which should become
				// the new sister of node bi within the same
				// binary subtree.

    hdyn *sister = bi->new_sister_node(lag_factor, top_level_combine);

#if 0
    if (sister != NULL){
	cout << "adjust low_level, "; bi->pretty_print_node(cout);
	cout << "-- "; sister->pretty_print_node(cout);
	cout << "-- "; sister->get_nn()->pretty_print_node(cout); cout<<endl;
    }
#endif

    if (sister != NULL)
	predict_loworder_all(bi->get_root(), bi->get_system_time());

    if (top_level_combine) {

	if (number_of_daughter_nodes_exceeds(bi->get_root(), 2)) {

	    if (DEBUG_TREE && TREE_DEBUG_LEVEL > 1) {
	        cerr << "\nadjust_low_level_node: "
		     << "normal top level combine at time "
		     << bi->get_time() << endl;
		// print_recalculated_energies(bi->get_root(), 1, eps * eps);
	    }

	    // cerr<<"Call combine_top_level_nodes from adjust_low_level_node"
	    //     << endl;

	    combine_top_level_nodes(sister->get_top_level_node(),
				    bi->get_top_level_node(),
				    eps, eta, max_timestep);

	} else {

	  if (DEBUG_TREE && TREE_DEBUG_LEVEL > 1) {
	        cerr << "adjust_low_level_node:"
		     << " top level combine with two top level nodes at time "
		     << bi->get_time() << endl;
		// print_recalculated_energies(bi->get_root(), 1, eps * eps);
	    }

	    // cerr << "Call split_top_level_node from adjust_low_level_node"
	    //      << endl;

	    split_top_level_node(bi->get_top_level_node(),
				 eps, eta, max_timestep);

	}

    } else if (sister != NULL) {

	combine_low_level_nodes(bi, sister, eps, eta, max_timestep);

    } else

	adjusted = 0;

    return adjusted;
}

int hdyn::adjust_tree_structure(real lag_factor,
				real eps, real eta, real max_timestep)
{
    int status = 1;

    // cerr << "\n*** in adjust_tree_structure for " << format_label() << endl;

    hdyn *br = get_root();
    if (is_top_level_node()) {		// this is a top-level node
      
	bool cm = false;
	if (oldest_daughter) cm = true;

	if (nn == NULL) {
            pretty_print_node(cerr); cerr << " nn is NULL " << endl;
	    return 0;
        }

	hdyn *nn_top = nn->get_top_level_node();

	if (too_close(this, nn_top, d_min_sq)) {

	    if (number_of_daughter_nodes_exceeds(parent, 2)) {

		predict_loworder_all(get_root(), system_time);

		// cerr << "\ntime = " << system_time << endl;
		// print_recalculated_energies(get_root(), 1, eps*eps);
		// pp3(get_root(), cerr);

		// cerr << "Call combine_top_level_nodes from "
		//      << "adjust_tree_structure"<<endl;

		combine_top_level_nodes(nn_top, this,
					eps, eta, max_timestep);

		// cerr << "Time = " << system_time<< endl;
		// print_recalculated_energies(get_root(), 1, eps*eps);
		// pp3(get_root(), cerr);

	    } else {
		status = 0;
	    }

	} else if (too_big(this, d_min_sq * lag_factor)) {

	    // print_recalculated_energies(get_root(), 1, eps*eps);

	    predict_loworder_all(get_root(), system_time);

	    // cerr << "Call split_top_level_node from adjust_tree_structure"
	    //      << endl;

	    split_top_level_node(this, eps, eta, max_timestep);

	    // Better exit next -- 'this' has already been deleted...

	} else 
	    status = 0;

    } else {				// this is a low-level node
					// check both this and our sister
	if (kep == NULL) {
	    hdyn *old_top_level_node = get_top_level_node();
	    if (!adjust_low_level_node(this, lag_factor,
				       eps, eta, max_timestep))
	    status = adjust_low_level_node(get_binary_sister(),
					   lag_factor,
					   eps, eta, max_timestep);
	    if (status) {
		if (old_top_level_node != get_top_level_node())
		    get_top_level_node()->clear_perturber_list();
	    }
	}else{
	    status = 0;
	}
    }

    return status;
}

local real print_encounter_elements(hdyn* bi, hdyn* bj,
				    char* s = "Collision")
{
    kepler k;
    initialize_kepler_from_dyn_pair(k, bi, bj, true);	// minimal kepler
							// -- no phase info

    cerr.precision(6);
    cerr << endl;
    cerr << s << " at time = " << bi->get_time();

    if (bi->get_use_sstar()) {

	cerr << " ("
	     << bi->get_starbase()->conv_t_dyn_to_star(bi->get_time())
	     << " [Myr])";
	cerr << " between \n"
	     << "     " << bi->format_label() << " (";
	put_state(make_star_state(bi), cerr);
	cerr << "; M = " << bi->get_starbase()->get_total_mass()
	     << " [Msun], "
	     << " R = " << bi->get_starbase()->get_effective_radius()
	     << " [Rsun]) and\n"
	     << "     " << bj->format_label()
	     << " (";
	put_state(make_star_state(bj), cerr);
	cerr << "; M = " << bj->get_starbase()->get_total_mass()
	     << " [Msun], "
	     << " R = " << bj->get_starbase()->get_effective_radius()
	     << " [Rsun]) "
	     <<"\n     at distance "
	     << k.get_periastron()
	     << " ("
	     << bi->get_starbase()->conv_r_dyn_to_star(k.get_periastron())
	     << " [Rsun]).";

    } else {

	cerr << " between " << bi->format_label();
	cerr << " and " << bj->format_label();

    }

    cerr << endl;

    if (k.get_energy() < 0) {
        cerr << "     Orbital parameters:" << endl;

        print_binary_params(&k, bi->get_mass(), 0.0,
			    abs(bi->get_pos()), true, 10, 10);
    }
    else
	cerr << "     E = " << k.get_energy() << endl;
      

    return k.get_energy();
}

// Flag close encounter distances between stars.  Output occurs on any
// unbound encounter, and on the first encounter in a bound system.

local void print_close_encounter(hdyn* bi,
				 hdyn* bj) {

    if (bi->get_mass() < bj->get_mass())  // Note: this "<" means that equal-
	return;				  // mass stars get printed twice!

    real d2cc_2 = -1;
    real d2cc_1 = -1;
    real d2cc   = -1;

    // Variables:	bi is primary, bj is current coll
    //			cc_name is label of previous coll
    //			cc_time is time of previous coll
    //			d2cc is squared distance to coll
    //			d2cc_1 is previous d2cc
    //			d2cc_2 is previous d2cc_1
    //			pcc_name is label of coll at last pericenter
    //			pcc_time is time of last pericenter
    //			pcc_cntr counts bound pericenters
    //
    // All but bi and bj are stored in the bi log story.

    char *cc_name = getsq(bi->get_log_story(), "cc_name");

#if 0
    cerr << "\nIn print_close_encounter with bi =  " << bi->format_label()
	 << "  at time " << bi->get_system_time() << endl;
    cerr << "     "; print_coll(bi,2);
    cerr << "     "; print_nn(bi,2); 
    cerr << "     (bj = " << bj->format_label();
    cerr << ":  coll = "; print_coll(bj);
    cerr << ",  nn = "; print_nn(bj); cerr << ")" << endl;

    cerr << "     cc_name = " << getsq(bi->get_log_story(), "cc_name")
	 << endl;
#endif
  
    if (cc_name && streq(cc_name, bj->format_label())) {

	// Retrieve coll information from the log story.

	d2cc_2 = getrq(bi->get_log_story(), "d2cc_1");
	d2cc_1 = getrq(bi->get_log_story(), "d2cc");
	d2cc   = square(bi->get_pos() - bj->get_pos());

	if (d2cc_2 > d2cc_1 && d2cc >= d2cc_1) {    // just passed pericenter

	    int pcc_cntr = 0;
	    real E = get_total_energy(bi, bj);

	    if (E > 0) {

	        // Always print unbound encounter elements.

		print_encounter_elements(bi, bj, "Close encounter");

	    }
	    else {

	        // Only print first bound encounter.

		char *pcc_name = getsq(bi->get_log_story(), "pcc_name");
	
		if (pcc_name && streq(pcc_name, bj->format_label())) {
		    pcc_cntr = getiq(bi->get_log_story(), "pcc_cntr");
		    if (pcc_cntr == -1) 
			pcc_cntr=0;
		}

		pcc_cntr++;

		if (pcc_cntr == 1) 
		    print_encounter_elements(bi, bj, "First bound encounter");
	    }

	    // Save data on last pericenter (bound or unbound).
            // Note that an unbound pericenter resets pcc_cntr to zero.

	    putiq(bi->get_log_story(), "pcc_cntr", pcc_cntr);
	    putsq(bi->get_log_story(), "pcc_name", bj->format_label());
	    putrq(bi->get_log_story(), "pcc_time", bi->get_time());  

	}

    }
    else if (getsq(bi->get_log_story(), "pcc_name"))

	if (cc_name == NULL
	    || strcmp(cc_name, getsq(bi->get_log_story(), "pcc_name"))) {

	    // Unlikely multiple encounter(?).  Has been known to occur
	    // when one incoming star overtakes another.

	    cerr << endl << "print_close_encounter:  "
	         << "bi = " << bi->format_label()
		 << " at time " << bi->get_system_time() << endl;
	    cerr << "     current coll = " << bj->format_label();
	    cerr << "  previous coll = ";
	    if (cc_name)
	        cerr << cc_name;
	    else
	        cerr << "(null)";
	    cerr << endl;
	    cerr << "     coll at last pericenter = "
	         << getsq(bi->get_log_story(), "pcc_name");
	    cerr << "  (periastron counter = "
	         << getiq(bi->get_log_story(), "pcc_cntr") << ")\n";
	}

    putsq(bi->get_log_story(), "cc_name", bj->format_label());
    putrq(bi->get_log_story(), "d2cc_1", d2cc_1);  
    putrq(bi->get_log_story(), "d2cc", d2cc);  
    putrq(bi->get_log_story(), "cc_time", bi->get_time());  

#if 0
    cerr << "\nAt end of print_close_encounter at time "
	 << bi->get_system_time() << endl;
    cerr << "bi = " << bi->format_label();
    cerr << ", cc_name = " << getsq(bi->get_log_story(), "cc_name") << endl;
#endif

}

hdyn* hdyn::check_merge_node() {
  
    if (use_dstar) {
	if (is_low_level_node()) {
	    if (binary_is_merged(this)) {
		cerr << "Merging a binary with logs:" << endl;
		print_log_story(cerr);
		cerr << "and\n";
		get_binary_sister()->print_log_story(cerr);
		return get_binary_sister();
	    }
	}
    }

    if (is_leaf() && coll != NULL && coll->is_leaf() && (kep == NULL)) {
      
	real sum_of_radii_sq = pow(radius+coll->radius, 2);
	if (d_coll_sq <= (stellar_encounter_criterion_sq-1)
						* sum_of_radii_sq) {

	    // cerr << endl << "Encounter between " << format_label();
	    // cerr << " and " << coll->format_label()
	    //      << " at time = " << get_time() << endl;

	    // pp3_minimal(get_top_level_node(), cerr);
	    // if (get_top_level_node()->get_nn())
	    //     pp3_minimal(get_top_level_node()->get_nn(), cerr);
	
	    // Merge (true) or fly-by (false).
	    // Merge  if: d <= alpha (r1 + r2)
	    // Report if: d > alpha (r1 + r2)

	    // Use kepler pericenter rather than sampled orbit points
	    // to determine whether or not a collision will occur.
	    // Probably desirable to have stellar_encounter_criterion
	    // a few times greater than stellar_merger_criterion to
	    // ensure that we get close enough to the collision for a
	    // more precise determination of its elements.

	    // Make a "minimal" kepler -- no phase information.

	    kepler k;
	    initialize_kepler_from_dyn_pair(k, this, coll, true);

	    // if (d_coll_sq > (stellar_merger_criterion_sq-1)
	    //					* sum_of_radii_sq) {

	    if (k.get_periastron() > (stellar_merger_criterion_sq-1)
						* sum_of_radii_sq) {
		print_close_encounter(this, coll);
		return NULL;
	    }
	
	    return coll;
	}
    }

    return NULL;
}

#define MASS_LOSS	0.0
#define KICK_VELOCITY	0.0

hdyn* hdyn::merge_nodes(hdyn * bcoll, real eps, real eta, real max_timestep)
{
    if (DEBUG_TREE) {
	if (TREE_DEBUG_LEVEL > 0) {
	    cerr << endl << "Merging leaves ";
	    pretty_print_node(cerr);
	    cerr << " and ";
	    bcoll->pretty_print_node(cerr);
	    cerr.precision(15);
	    cerr << " at time " << system_time << endl;
	    cerr.precision(PRECISION);
	    if (TREE_DEBUG_LEVEL > 1) {
		cerr << endl;
		pp3(this, cerr);
		pp3(bcoll, cerr);
		if (TREE_DEBUG_LEVEL > 2) {
		    cerr << endl;
		    put_node(cerr, *this);
		    put_node(cerr, *bcoll);
		}
		cerr << endl;
	    }
	}
    }

    real radius1 = bcoll->get_starbase()->conv_r_dyn_to_star(radius);
    real radius2 = bcoll->get_starbase()->conv_r_dyn_to_star(bcoll->radius);
    real sum_of_radii_sq = pow(radius+bcoll->radius, 2);
    real mass1 =        get_starbase()->get_total_mass();
    real mass2 = bcoll->get_starbase()->get_total_mass();

    cerr << "calling synchronize_tree from merge_nodes" << endl << flush;
    synchronize_tree(get_root(), eps, eta, max_timestep);
    cerr << "back" << endl << flush;

    if (is_top_level_node() || get_binary_sister() != bcoll) {

        // Nodes to be merged ('this' and bcoll) are not binary
	// sisters.  Force them to become sisters prior to actually
	// merging them, temporarily placing them in the same
	// subtree if necessary.
	
	hdyn* ancestor = common_ancestor(this, bcoll);
	
	if (ancestor->is_root()) {

	    // Nodes are not in the same subtree.  Combine their
	    // subtrees into a single clump to allow use of function
	    // move_node below.  Undo the combination before 
	    // continuing, if necessary.
	    
	    bool decombine = !is_top_level_node()
				&& !bcoll->is_top_level_node();

	    // cerr << "Call combine_top_level_nodes from merge_nodes" << endl;

	    combine_top_level_nodes(get_top_level_node(),
				    bcoll->get_top_level_node(),
				    eps, eta, max_timestep);

	    // pp2(get_top_level_node(), cerr);

	    if (get_binary_sister() != bcoll)
	        move_node(bcoll, this);		// Move bcoll to become
		  				// sister of this.
	    
	    if (decombine) {
		//cerr << "Call split_top_level_node from merge_nodes" << endl;
		split_top_level_node(get_top_level_node(),
				     eps, eta, max_timestep);
	    }
	    
	} else {

	    // pp2(get_top_level_node(), cerr);

	    move_node(bcoll, this);		// Move bcoll to become
						// sister of this.
	}		
    }
    
    // Nodes 'this' and bcoll are binary sisters (in all cases).
    // Merge them into their center-of-mass node and remove them.
    
    // pp2(get_top_level_node(), cerr);
    
    // Compute energies for bookkeeping:
    
    real epot0, ekin0, etot0;
    calculate_energies(get_root(), eps*eps, epot0, ekin0, etot0);
    
    hdyn* cm = get_parent();
    
    real epot_int = -mass * bcoll->mass / abs(pos - bcoll->pos);
    real ekin_int = 0.5 * (mass * square(vel)
			   + bcoll->mass * square(bcoll->vel));
    real etot_int = epot_int + ekin_int;
    
    PRC(epot0); PRC(ekin0); PRL(etot0);
    PRC(epot_int); PRC(ekin_int); PRL(etot_int);
    //pp3(cm, cerr);
    //pp3(cm->get_root(), cerr);
    
    cm->set_radius(radius + bcoll->radius);
    label_merger_node(cm);

    // pp3(cm->get_root(), cerr);
    
    real dm;
    vector dv;
    if (!use_sstar){

	// Test code for experiment with no stellar evolution
	// set some mass loss and kick
	
	// Mass loss and velocity kick:
	
	dm = -MASS_LOSS*cm->mass;
	
	real costh = randinter(-1, 1);
	real sinth = sqrt(1 - costh*costh);
	if (randinter(0, 1) < 0.5) sinth = -sinth;
	real phi = TWO_PI*randinter(0, 1);
	dv = KICK_VELOCITY * sqrt(cm->mass/cm->radius)
	* vector(sinth*cos(phi),
		 sinth*sin(phi), costh);
	cm->set_radius((1-MASS_LOSS)*cm->get_radius());

	print_encounter_elements(this, bcoll, "Collision");

    } else {

	// real code asking for collision product to stellar
	// evolution code
	
	hdyn * primary = this;
	hdyn * secondary = bcoll;
	
	cerr << "Merging two STARS"<<endl;
	// pp2(cm->get_root(), cerr);

	((star*)(primary->sbase))->dump(cerr, false);
	((star*)(secondary->sbase))->dump(cerr, false);
	
	if (!merge_with_primary(dynamic_cast(star*, primary->sbase),
			        dynamic_cast(star*, secondary->sbase))) {
	    primary = bcoll;
	    secondary = this;
	}

	print_encounter_elements(primary, secondary, "Collision");
	
	((star*)(primary->sbase))
	                ->merge_elements(((star*)(secondary->sbase)));
	
	cerr << "Merger product: "<< endl;
	cerr << format_label() << " (";
	//	        put_state(make_star_state(sbase), cerr);
	        put_state(make_star_state(primary), cerr);
	cerr << "; M=" << get_total_mass(primary) << " [Msun], "
	     << " R=" << get_effective_radius(primary) << " [Rsun]). " << endl;
	((star*)(primary->sbase))->dump(cerr, false);

	
	real old_mass = cm->mass;

	// For now let's keep the disintegrated star and make a log of it.
	// Special case if the merger product ceases to exist.
	//
	// if (((star*)(primary->sbase))->get_element_type()==Disintegrated) {
	//   cerr << "Disintegrated star, set CM mass to zero" << endl;
	//   cm->mass = 0;
	// }
	
	// Must (1) merge the logs of primary and secondary, and      <---
	//      (2) add a line to the log describing the merger.      <---
	
	// Attach the new merger star to cm.
	// Old information of star in CM (if any) is lost here......
	
	cm->sbase = primary->sbase;
	cm->sbase->set_node(cm);
	primary->sbase = NULL;
	
	real new_mass = get_total_mass(cm);
	dm = new_mass - old_mass;
	dv = anomalous_velocity(cm);
    }

    correct_leaf_for_change_of_mass(cm, dm);
    if (cm->mass < 0) {
	cerr << "check and merge, negative mass ";
	PRL(cm->mass);
    }

    cm->coll = NULL;			// Set its coll to null.
    
    // Add kick velocity, if any...
    
    if (square(dv) > 0)
	correct_leaf_for_change_of_vector(cm, dv, &hdyn::get_vel,
					  &hdyn::inc_vel);
    
    cm->set_oldest_daughter(NULL);

    
    // To try to free memory by simply 
    //		delete bcoll;
    //		delete this;
    // caused very strange error.  What is wrong?
    //
    // *** Delete this and bcoll in the calling function. ***

    
    real epot, ekin, etot;
    calculate_energies(get_root(), eps*eps, epot, ekin, etot);
    
    PRC(epot); PRC(ekin); PRL(etot);
    // pp3(cm, cerr);
    
    real de_total = etot - etot0;
    vector vcm = hdyn_something_relative_to_root(cm, &hdyn::get_vel);
    real de_kick = 0.5 * cm->mass * (vcm*vcm - square(vcm-dv));
    real de_int = -etot_int;
    
    PRC(de_total); PRC(de_int); PRL(de_kick);
    
    // Update statistics on mass and energy change components:
    
    kc->dm_massloss -= dm;	// dm < 0, note

    kc->de_total += de_total;
    kc->de_merge += de_int;
    kc->de_massloss += de_total - de_int - de_kick;
    kc->de_kick += de_kick;
    
    hdyn* root = cm->get_root();
    real time = cm->time;
    
    // Special treatment of case cm->mass = 0.
    
    if (cm->mass <= 0) {
	
	// NOTE: Must copy cm's log entry before deleting the node.	<---
	//       Place it in the root log.				<---
	
	remove_node_and_correct_upto_ancestor(cm->get_parent(), cm);
    }

    predict_loworder_all(root, time);
}
