
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

// Initiate and maintain unperturbed motion.
//
// Externally visible functions:
//
//	void hdyn::update_dyn_from_kepler
//	bool hdyn::is_close_pair
//	bool hdyn::is_unperturbed_and_approaching
//	void hdyn::startup_unperturbed_motion
//	void hdyn::print_unperturbed_binary_params
//	void hdyn::integrate_unperturbed_motion
//	real hdyn::set_unperturbed_timestep
//	bool hdyn::is_weakly_perturbed
//	bool hdyn::is_stable


// To do:  Keep track of tidal errors associated with termination
//	   of unperturbed motion.


#include "hdyn.h"
#include <star/dstar_to_kira.h>
#include "kira_debug.h"

//-------------------------------------------------------------------------

// Define what level of unperturbed motion is permitted by default:

#define ALLOW_UNPERTURBED		true

#define ALLOW_MULTIPLES			false	// "false" here turns off all
						// unperturbed multiple motion

static bool allow_unperturbed = ALLOW_UNPERTURBED;
static bool allow_multiples   = ALLOW_MULTIPLES;

// (Only one reference to each of these variables in the code below.)

// Accessors:

void print_unperturbed_options()
{
    cerr << "Unperturbed options:  ";
    PRC(allow_unperturbed); PRL(allow_multiples);
}

void toggle_unperturbed(int level)		// toggle unperturbed options
{
    if (level == 0)
	allow_unperturbed = !allow_unperturbed;
    else
	allow_multiples = !allow_multiples;

    // Consistency:

    if (!allow_unperturbed) allow_multiples = false;
}

//-------------------------------------------------------------------------

// Control output on unperturbed motion.

// Warning: these can produce a *lot* of output!
// Probably could (should) make some or all of these into variables...
//
// Possible variables:	report unperturbed start/finish (level 0)
//			FULL_MERGE_TOLERANCE
//			PARTIAL_MERGE_TOLERANCE
//			MULTIPLE_MERGE_TOLERANCE

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define REPORT_START_UNPERTURBED        	true
//#define REPORT_CONTINUE_UNPERTURBED		false
//#define REPORT_END_UNPERTURBED		true
//#define REPORT_PERICENTER_REFLECTION		false
//#define REPORT_IMPENDING_MULTIPLE_STATUS	false
//#define REPORT_ZERO_UNPERT_STEPS		false
//#define REPORT_MULTIPLE			true
//
//#define UNPERT_REPORT_LEVEL			0
//#define END_UNPERT_REPORT_LEVEL		0  // allow more flexibility
//
//#define UNPERT_FUNCTION_ID		false  	// Selected functions will
//						// identify themselves if set
//
// Note:  constructs such as the following are legal, as all of the
//	  above macros are only used in member functions:
//
// #define UNPERT_FUNCTION_ID	(system_time > 100 && system_time < 101)

//-------------------------------------------------------------------------

// The following parameters define "unperturbed" motion.

// Note that gamma2 is used in determining membership in the perturber
// list.  Note also that d_min_sq does not appear in this file, so the
// criteria for unperturbed motion are independent of the criteria used
// in constructing and maintaining the tree (except that unperturbed
// motion can only occur in low-level nodes, so the binary must first
// be picked up by hdyn_tree).
//
// Where relevant, numbers are tailored to the default gamma = 1.e-7.


// MIN_STEPS:  Minimum number of steps permitted in unperturbed motion.

#define MIN_STEPS			5


// OUTSIDE_SEMI:  Comparison of orbital separation with semi-major axis for
// full merging.  Criterion is
//
//	SEP_FACTOR * separation > semi-major axis
//
// Macro is used three times, in is_unperturbed_and_approaching(),
// is_stable(), and set_unperturbed_timestep().
//
// SEP_FACTOR is slightly greater than 1 in order to accommodate nearly
// circular binaries

#define SEP_EPSILON			0.01	// arbitrary, but small
#define SEP_FACTOR			(1.0 + SEP_EPSILON)

#define OUTSIDE_SEMI(sep, sma)		((SEP_FACTOR)*(sep) > (sma))

// The comparison is often performed on elements of a kepler structure (k):

#define KEP_OUTSIDE_SEMI(k) \
    ((SEP_FACTOR)*((k).get_separation()) > ((k).get_semi_major_axis()))


// FULL_MERGE_TOLERANCE:  Factor above gamma2 at which we permit
// merging of an entire binary orbit.  Used once, in function
// is_unperturbed_and_approaching().

#define FULL_MERGE_TOLERANCE		1e4
#define RELAX_FACTOR			10	// relaxed factor to continue

// Allowing unperturbed motion at a greater value of gamma than is
// used in determining neighbors is reasonable since dominant
// perturbative terms are expected to be periodic, and so are
// ineffective over a full orbit period.

// Relaxed criterion to continue is probably reasonable (for true
// binaries only), as the first unperturbed step takes them to
// apocenter, where the external perturbation is greater.


// PARTIAL_MERGE_TOLERANCE:  Factor below gamma2 at which we permit
// merging of at least part of an orbit.  Used once, in function
// is_unperturbed_and_approaching().

#define PARTIAL_MERGE_TOLERANCE		1e-2


// FULL_MERGE_TOL_FOR_CLOSE_BINARY:  Absolute limit on perturbation_squared
// (INDEPENDENT of gamma2, note) at which an unperturbed binary is regarded
// as "close."  Used once, in function is_close_pair().

#define FULL_MERGE_TOL_FOR_CLOSE_BINARY	1e-4


// The following parameters are used in is_stable() for assessing
// multiple stability.

// MULTIPLE_MERGE_TOLERANCE:  Factor above gamma2 at which we permit
// merging of an inner binary orbit.  Defines "weakly perturbed
// outer orbit."

#define MULTIPLE_MERGE_TOLERANCE		1e6	// was 1.e4

// UNCONDITIONAL_STABLE_FAC:  Regard a multiple as stable if
//
//	inner semi-major axis < outer periastron / UNCONDITIONAL_STABLE_FAC
//
// for (both) inner binaries.

#define UNCONDITIONAL_STABLE_FAC	5	// was 9

// PARTIAL_STABLE_FAC:  Allow inner binary to be unperturbed for part of
// the outer orbit if
//
//	inner semi-major axis < outer periastron / PARTIAL_STABLE_FAC

#define PARTIAL_STABLE_FAC		30	// large to reduce tidal effect

//-------------------------------------------------------------------------

// STATIC flags to keep track of unperturbed motion:

static int binary_type;

#define UNKNOWN_STATUS			0
#define STABLE_OUTER 			1
#define STABLE_INNER			2
#define NOT_APPROACHING			3
#define PERICENTER_REFLECTION		4
#define FULL_MERGER			5
#define CONTINUE_MERGER			6
#define PERTURBED			7
#define MULTIPLE			8
#define UNKNOWN_PERTURBERS		9

static char* bt[10] = {"unknown status",
		      "stable multiple outer binary",
		      "stable multiple inner binary",
		      "components not approaching",
		      "pericenter reflection",
		      "full binary merger",
		      "continue existing merger",
		      "perturbed binary",
		      "multiple system",
		      "no valid perturber list"};

static int multiple_type;

#define NOT_MULTIPLE			0
#define FULL_MULTIPLE			1
#define APOCENTER_REFLECTION		2

static char* tt[3] = {"not multiple",
		      "full_multiple merger",
		      "apocenter reflection"};

//------------------------------------------------------------------------

// Local functions:

local inline void print_found_multiple(hdyn* b,
				     bool real_multiple,
				     kepler& outer,
				     real inner_semi1,
				     real inner_semi2)
{
    cerr.precision(12);
    cerr << endl;

    if (real_multiple)
	cerr << "Found hierarchical stable multiple CM ";
    else
	cerr << endl << "Impending multiple: ";

    cerr << b->get_parent()->format_label();

    if (!real_multiple)
	cerr << " is a multiple CM";

    cerr << " at time " << b->get_system_time() << endl;
    cerr << "components " << b->format_label();
    cerr << " and " << b->get_binary_sister()->format_label();

    if (!real_multiple)
	cerr << "; outer orbit is weakly perturbed:\n";

    cerr.precision(8);

    cerr << "    perturbation_squared = " << b->get_perturbation_squared()
	 << endl
	 << "    outer peri = " << outer.get_periastron()
	 << "  inner semi = ";

    if (inner_semi1 > 0)
	cerr << inner_semi1 << "  ";

    if (inner_semi2 > 0)
	cerr << inner_semi2;

    cerr << endl
	 << "    outer period = " << outer.get_period()
	 << "  parent dt = " << b->get_parent()->get_next_time()
	     				- b->get_system_time()
	 << endl;
}

local inline void print_startup_message(hdyn * b, int type)
{
    cerr.precision(12);
    cerr << endl << "starting unperturbed motion for "
	 << b->format_label();
    cerr << " and " << b->get_binary_sister()->format_label()
	 << " at time " << b->get_time() << endl
	 << bt[type] << " (binary_type = " << type << ")";
    cerr.precision(8);
    if (b->get_kepler())
	cerr << "  period = " << b->get_kepler()->get_period();
    cerr << endl;
}

local inline void print_binary_data(hdyn* bi)
{
    PRL(bi->get_perturbation_squared());
    PRL(bi->get_unperturbed_timestep());
    PRL(bi->get_fully_unperturbed());
    PRL(bi->get_d_nn_sq());
    PRL(bi->get_perturbation_radius_factor());
    PRL(bi->find_perturber_node()->get_valid_perturbers());
    PRL(bi->get_d_coll_sq());
    PRL(bi->get_coll()->format_label());
}

local inline bool is_multiple(hdyn* b)   // ("multiple" means 3 or more stars)
{
    // Return true if low-level node b is one of the outer components of
    // a multiple system.

    if (b->is_top_level_node()) return false;
    return !( b->is_leaf() && b->get_binary_sister()->is_leaf() );
}

local inline real get_energy(hdyn* b, real& separation)
{
    // Compute the energy of b relative to its binary sister.

    if (b->is_top_level_node()) return 0;

    hdyn* s = b->get_binary_sister();
    separation = abs(b->get_pos() - s->get_pos());

    return 0.5*square(b->get_vel() - s->get_vel())
	      - b->get_parent()->get_mass() / separation;
}

local inline real get_semi(hdyn* bcm)
{
    // Compute the semi-major axis of the binary daughters of bcm.

    real semi = 0;
    if (bcm->is_parent()) {

	// Get semi-major axis of the orbit.
	// Use existing kep if orbit is already unperturbed.

	if (bcm->get_oldest_daughter()->get_kepler()) {
	    semi = bcm->get_oldest_daughter()->get_kepler()
					     ->get_semi_major_axis();
	} else {
	    real sep;
	    real energy = get_energy(bcm->get_oldest_daughter(), sep);
	    semi = -0.5 * bcm->get_mass() / energy;
	}
    }

    if (semi < 0) semi = -VERY_LARGE_NUMBER;
    return semi;
}

local inline real get_period(real mass, real sma)
{
    // Compute the period of a binary with specified mass and
    // semi-major axis.

    return TWO_PI * sqrt( sma*sma*sma / mass );
}



void hdyn::print_unperturbed_binary_params()
{
    cerr << "    perturbed timestep for " << format_label()
	 << " is " << timestep << endl;

    cerr << "    pos*vel = " << pos*vel
	 << endl
	 << "    a = " << kep->get_semi_major_axis()
	 << "  r = " << kep->get_separation()
	 << endl
	 << "    peri = " << kep->get_periastron()
	 << "  apo = " << kep->get_semi_major_axis()
				* (1 + kep->get_eccentricity())
	 << endl
	 << "    period = " << kep->get_period()
	 << "  perturbation_squared = " << perturbation_squared
	 << endl;

    PRI(4); PRL(unperturbed_timestep);

    PRI(4); PRL(get_parent()->get_next_time());
    PRI(4); PRL(get_parent()->timestep);
    PRI(4); print_nn(get_parent(), 2);
}

void hdyn::update_dyn_from_kepler()
{
    if (UNPERT_FUNCTION_ID) {
	cerr << ">> update_dyn_from_kepler for "
	     << format_label() << endl;
    }

    hdyn *sister = get_binary_sister();
    time = system_time;
    kep->transform_to_time(system_time);

    real factor = sister->mass / (mass + sister->mass);
    real sfactor = factor - 1.0;
    pos = kep->get_rel_pos() * factor;
    vel = kep->get_rel_vel() * factor;
    pred_pos = pos;
    pred_vel = vel;

    sister->time = time;
    sister->pred_pos = kep->get_rel_pos() * sfactor;
    sister->pred_vel = kep->get_rel_vel() * sfactor;
    sister->pos = kep->get_rel_pos() * sfactor;
    sister->vel = kep->get_rel_vel() * sfactor;

    clear_interaction();
    calculate_acc_and_jerk(0.0, 0.0, true);  // computes perturbation_squared
    store_old_force();

    update_binary_sister(this);
}

bool hdyn::is_close_pair()
{
    if (kep == NULL) return false;
    if (kep->get_energy() >= 0) return false;

    real rp = kep->get_periastron();
    real sum_radius = radius + get_binary_sister()->radius;

    if (rp < 5 * sum_radius) {

	// This criterion is *very bad*
	// it should reflect the Roche radius...

	if (perturbation_squared < FULL_MERGE_TOL_FOR_CLOSE_BINARY) {

	    // cerr << "time = " << system_time << " ";

	    pretty_print_node(cerr);

	    // cerr << " close pair criterion ";
	    // cerr << perturbation_squared << " " ;
	    // PRC(rp); PRL(sum_radius);

	    return true;
	}
    }
    return false;
}



// is_weakly_perturbed:  Return true iff 'this' system is lightly perturbed
//			 and satisfies some other basic acceptance criteria
//			 for the outer binary of an unperturbed multiple
//			 system.
//
//			 New function written by Steve, 8/98.

bool hdyn::is_weakly_perturbed()
{
    if (UNPERT_FUNCTION_ID) {
	cerr << ">> check is_weakly_perturbed for "
	     << format_label() << " at time " << system_time << endl;
    }

    if (!is_low_level_node())
	return false;

    // 'This' is a component of a binary, possibly a multiple.
    // Return true if the binary defined by this and its binary
    // sister is weakly perturbed (i.e. not strongly perturbed).

    real separation, energy;

    if ((energy = get_energy(this, separation)) >= 0)
	return false;

    hdyn* pnode = find_perturber_node();
    if (!pnode) return false;

    if (!pnode->valid_perturbers)
	return false;

    // Definition of "not weakly perturbed":

    if (perturbation_squared > gamma2 * MULTIPLE_MERGE_TOLERANCE)
	return false;

    // Avoid merging if the outer orbit is too extended.

    real semi_major_axis = -0.5 * parent->get_mass() / energy;

    if (semi_major_axis > 5 * separation) return false;	  // 5 is ~arbitrary...

    // Require the outer orbit to be outside its semi-major axis,
    // or completely unperturbed.

    bool is_weakly_perturbed = false;

    if (OUTSIDE_SEMI(separation, semi_major_axis))
	is_weakly_perturbed = true;
    else if (pnode->n_perturbers == 0)
	is_weakly_perturbed = true;

    if (is_weakly_perturbed) {

	// The following isn't directly related to the strength of the
	// perturbation, but will be checked if the binary turns out to
	// be stable.  May as well ckeck these here and avoid the stability
	// check if it is not needed.

	// The minimum unperturbed step for a multiple is currently thew
	// outer orbit period.  Return here if that will not be possible,
	// given the parent time step.

	real period = get_period(parent->get_mass(), semi_major_axis);
	real pdt2 = get_parent()->get_next_time()
				+ 0.5 * get_parent()->timestep - time;

	if (period <= pdt2) {

	    return true;			// ONLY way for this function
						// to return true...

	} else if (REPORT_MULTIPLE && UNPERT_REPORT_LEVEL > 0) {

	    cerr << "Multiple " << get_parent()->format_label()
		 << " is weakly perturbed, but parent time step is too short"
		 << endl
		 << "period = " << period << "  parent_step = " << pdt2
		 << endl;
#if 0
	    pp3(get_parent());

	    print_binary_from_dyn_pair(this, get_younger_sister());
	    if (is_parent())
		print_binary_from_dyn_pair(get_oldest_daughter(),
					   get_oldest_daughter()
					     ->get_younger_sister());
	    if (get_younger_sister()->is_parent())
		print_binary_from_dyn_pair(get_younger_sister()
					     ->get_oldest_daughter(),
					   get_younger_sister()
					     ->get_oldest_daughter()
					     ->get_younger_sister());
#endif
	}
    }

    return false;
}



#define NEAR_MULTIPLE_FAC	1.2

// is_stable:  Return true iff this is a stable system.
//
//	       New (recursive) function written by Steve, 8/98.

bool hdyn::is_stable(bool top_level)
{
    if (UNPERT_FUNCTION_ID) {
	cerr << ">> check is_stable for "
	     << format_label() << endl;
    }

    if (!is_low_level_node())
	return false;

    // 'This' is a component of a binary, possibly a multiple.
    // Return true iff the binary defined by this and its binary
    // sister is stable.
    //
    // Unperturbed criteria for this to be a stable system:
    //
    // (a) the interior motion is stable, meaning that the ratio of
    //     the semimajor axis of each inner binary to the pericenter
    //	   of the outer binary is less than some critical value.
    //
    // (b) each component is stable: a single star, a binary, or a
    //	   stable multiple.

    hdyn* sister = get_binary_sister();
    if (!sister) return false;

    // Binaries are stable:

    if (is_leaf() && sister->is_leaf()) return true;

    // Set up a kepler structure describing the outer orbit.
    // Use existing kep if the outer orbit is already unperturbed.

    kepler outerkep;

    if (kep == NULL) {
	outerkep.set_time(time);
	outerkep.set_total_mass(parent->get_mass());
	outerkep.set_rel_pos(pos - sister->pos);
	outerkep.set_rel_vel(vel - sister->vel);
	outerkep.initialize_from_pos_and_vel();
    } else
	outerkep = *kep;

    real outer_peri = outerkep.get_periastron();

    real inner_semi1 = get_semi(this);
    real inner_semi2 = get_semi(sister);
    real inner_semi_sum = inner_semi1 + inner_semi2;

    if (inner_semi_sum < 0)
	return false;			// unbound inner orbit
    else if (inner_semi_sum == 0)
	return true;			// binary is stable -- shouldn't happen
					// (already tested above)

    // Check criterion (a):

    bool stable_a = false;

    if (UNCONDITIONAL_STABLE_FAC*inner_semi_sum < outer_peri) {

	// Unconditionally stable multiple.  Note that the perturbations
	// on the inner binaries are never checked explicitly.

	if (top_level)
	    multiple_type = FULL_MULTIPLE;

	stable_a = true;

    } else if (top_level && PARTIAL_STABLE_FAC*inner_semi_sum
				       < outerkep.get_separation()) {

	// Conditionally stable multiple (part of outer orbit only).

	real peri_time = outerkep.pred_advance_to_periastron() - time;

	if (peri_time > 0.8 * outerkep.get_period()) {   // 0.8 ~arbitrary

	    // (Note that the above requirement actually means
	    //  that the outer orbit is separating...)

	    if (top_level)
		multiple_type = APOCENTER_REFLECTION;

	    stable_a = true;

	} else {

	    if (top_level && REPORT_IMPENDING_MULTIPLE_STATUS)
		cerr << "Outer orbit " << parent->format_label()
		     << " is partially unperturbed but not compact...\n";
	}
    }

    // Check criterion (b):

    bool stable_b = stable_a;

    if (stable_a) {

	if (is_parent())
	    stable_b &= get_oldest_daughter()->is_stable(false);

	if (sister->is_parent())
	    stable_b &= sister->get_oldest_daughter()->is_stable(false);
    }


    // Diagnostic output:

    if (top_level) {

	if (REPORT_IMPENDING_MULTIPLE_STATUS
	    && !stable_a
	    && (UNCONDITIONAL_STABLE_FAC*inner_semi_sum
			< NEAR_MULTIPLE_FAC*outer_peri
		|| PARTIAL_STABLE_FAC*inner_semi_sum
			< NEAR_MULTIPLE_FAC*outerkep.get_separation())) {

	    // Getting close to the stable multiple criterion...

	    print_found_multiple(this, false, outerkep,
				 inner_semi1, inner_semi2);
	}

	if (stable_b && REPORT_MULTIPLE) {

	    if (multiple_type == FULL_MULTIPLE)
		print_found_multiple(this, true, outerkep,
				     inner_semi1, inner_semi2);
	    else if (multiple_type == APOCENTER_REFLECTION) {
		cerr.precision(12);
		cerr << "\nfound partially stable multiple "
		     << parent->format_label()
		     << " at time " << time << endl;
		cerr.precision(8);

		hdyn* pnode = find_perturber_node();
		if (!pnode)
		    cerr << "pnode is NULL" << endl;
		else {
		    PRC(pnode->format_label());
		    PRL(pnode->n_perturbers);
		}

		cerr << "outer period = " << outerkep.get_period() << endl;
		cerr << "inner period = ";
		if (inner_semi1 > 0)
		    cerr << get_period(mass, inner_semi1) << "  ";
		if (inner_semi2 > 0)
		    cerr << get_period(sister->mass, inner_semi2);
		cerr << endl;
	    }
	}

	if (stable_a && !stable_b && REPORT_MULTIPLE) {
	    cerr << "\nmultiple " << parent->format_label()
		 << " is stable, but components are not, at time "
		 << time << endl;
	}
    }

    return stable_b;
}



bool hdyn::is_unperturbed_and_approaching()

// Test unperturbed criterion for startup *or* continuation of unperturbed
// motion.  Note that the notion of "unperturbed" is based on the magnitude
// of the perturbation, not on the number of perturbers on the perturber
// list.  Note also that the validity of the perturber list is checked in
// the binary case, but not (yet) for multiples.

// By construction when this function is called, 'this' is already a
// low-level node, so its binary properties are well defined.

{
    if (!allow_unperturbed) return false;

    if (UNPERT_FUNCTION_ID) {
	cerr << ">> check is_unperturbed_and_approaching for "
	     << format_label() << " at time " << system_time << endl;
    }

    binary_type = UNKNOWN_STATUS;		// Should not occur

    // If this function returns true, then binary_type must be one of
    // the following:
    //
    //			PERICENTER_REFLECTION	// for binaries
    //			FULL_MERGER
    //			CONTINUE_MERGER
    //
    //			STABLE_OUTER		// for multiples

    if (is_multiple(this)) {

	binary_type = MULTIPLE;

	if (!allow_multiples) return false;	// Fixes all multiple problems!!

	// Unperturbed criteria for this to be a stable system:
	//
	// (1) the outer binary is bound and stable, meaning that the
	//     perturbation is small; in addition, require that the
	//     current orbital phase be outside the semimajor axis.
	//
	// (2) the interior motion is stable, meaning that the ratio of
	//     the semimajor axis of the (or each) inner binary to the
	//     pericenter of the outer binary is less than some critical
	//     value.
	//
	// (3) each component is stable: a single star, a binary, or a
	//     stable multiple.

	multiple_type = NOT_MULTIPLE;

	if (is_weakly_perturbed() && is_stable()) {

	    binary_type = STABLE_OUTER;
	    return true;

	    // Note that multiples are detected only when the outer orbit
	    // is advanced.  Up to the point of multiple recognition, inner
	    // binaries are subject to the usual (more stringent) unperturbed
	    // criterion.  With this convention, 'this' is always one of the
	    // outer components of the multiple system.

	}

    } else {

	// Rest of function applies only to binaries.

	// Accept nearly circular binaries even if they are
	// separating slowly.

	if (pos * vel >= 1.e-3 * abs(pos)*abs(vel)) {

	    binary_type = NOT_APPROACHING;
	    return false;
	}

	if (!find_perturber_node()) {

	    // Possible that a newly formed center of mass node hasn't
	    // yet taken a step.  Don't want to break up an unperturbed
	    // binary because of this.  Set binary_type appropriately
	    // and return true if this is already unperturbed.

	    binary_type = UNKNOWN_PERTURBERS;
	    if (kep)
		return true;
	    else
		return false;
	}

 	// System is not a multiple and the components are "approaching."

	if ((perturbation_squared < gamma2 * PARTIAL_MERGE_TOLERANCE )
	    && is_low_level_leaf()
	    && younger_sister->is_low_level_leaf()) {

	    // Sufficient to pick up the binary here as partly unperturbed;
	    // partial merger may be promoted to full merger later.

	    binary_type = PERICENTER_REFLECTION;
	    return true;

	} else {

	    if (kep == NULL) {

	        if ((perturbation_squared < gamma2 * FULL_MERGE_TOLERANCE)
		    && is_low_level_leaf()
		    && younger_sister->is_low_level_leaf()) {
		
		    if (get_parent()->timestep
			  > 10 * timestep) {		// 10.0 is ~arbitrary

		        kepler kepl;
			hdyn *sister = get_binary_sister();
		    
			kepl.set_time(time);
			kepl.set_total_mass(parent->get_mass());
			kepl.set_rel_pos(pos - sister->pos);
			kepl.set_rel_vel(vel - sister->vel);
			kepl.initialize_from_pos_and_vel();

			if (kepl.get_energy() < 0.0) {
			    if (KEP_OUTSIDE_SEMI(kepl)) {
			        if ((get_parent()->get_next_time() - time) 
				      >  kepl.get_period()) {

				    binary_type = FULL_MERGER;
				    return true;

				}
			    }
			}
		    }
		}

	    } else {

	        // Note relaxed criterion for continuing unperturbed
	        // binary (but *not* multiple) motion.

	        real crit_pert2 = gamma2 * FULL_MERGE_TOLERANCE
		    			 * RELAX_FACTOR;

	        if (perturbation_squared < crit_pert2 || is_close_pair()) {

		    binary_type = CONTINUE_MERGER;
		    return true;

		}
	    }
	}
    }

    binary_type = PERTURBED;
    return false;
}



// startup_unperturbed_motion: only invoked if is_unperturbed_and_approaching
//			       returns true...

void hdyn::startup_unperturbed_motion()
{
    // Begin treatment of unperturbed motion.  The main actions of this
    // function are (1) to create kepler structure(s) if necessary, and
    // (2) to determine the unperturbed time step(s).  May also modify
    // binary_type.

    // Note: In the case of unperturbed multiple motion, 'this' is one
    // of the outer components (see is_unperturbed_and_approaching).

    if (UNPERT_FUNCTION_ID) {
	cerr << endl << ">> startup_unperturbed_motion for "
	     << format_label() << " at time " << system_time << endl;
    }

    update_kepler_from_hdyn();	// Creates a kepler structure if none exists...

    fully_unperturbed = false;	// May be set true in set_unperturbed_timestep()

    if (REPORT_START_UNPERTURBED) {
	if (binary_type != PERICENTER_REFLECTION
	    || REPORT_PERICENTER_REFLECTION)
	    print_startup_message(this, binary_type);
    }

    int save_binary_type = binary_type;	 // from is_unperturbed_and_approaching

    // Used to be an int, changed to real because integers are more limited.
    // (SPZ:02/1998):

    real steps = set_unperturbed_timestep(true);

    if (REPORT_START_UNPERTURBED && binary_type != save_binary_type) {

	if (save_binary_type == PERICENTER_REFLECTION
	    && !REPORT_PERICENTER_REFLECTION)
	    print_startup_message(this, save_binary_type);

        cerr << "binary_type changed to " << binary_type
	     << ":  " << bt[binary_type] << endl;

    }

    // If steps = 0, there is no reason to start unperturbed motion.
    // Moreover, if steps is 0, program goes into an infinite loop, so
    // discard the kepler structure and return.

    // Avoid kepler if steps is less than 5 or so, in order to avoid
    // unnecessary roundoff due to kepler conversion.

    if (steps < MIN_STEPS) {

	if (REPORT_START_UNPERTURBED || (steps <= 0 && REPORT_ZERO_UNPERT_STEPS)) {
	    cerr << "\nstartup_unperturbed_motion:  calculated step size = "
		 << steps << endl
		 << "do not apply unperturbed motion to " << format_label()
		 << " at time " << system_time << "\n";
	}

	delete kep;
	kep = NULL;
	unperturbed_timestep = -VERY_LARGE_NUMBER;
	get_binary_sister()->unperturbed_timestep = -VERY_LARGE_NUMBER;
	get_binary_sister()->kep = NULL;

	return;
    }

    unperturbed_timestep = timestep * steps;

    if (UNPERT_REPORT_LEVEL > 0
	&& (binary_type != PERICENTER_REFLECTION
	    || REPORT_PERICENTER_REFLECTION))
	print_unperturbed_binary_params();

    get_binary_sister()->unperturbed_timestep = unperturbed_timestep;

    // If this is the outer binary of a multiple, the inner component has
    // already been synchronized (in kira); make it unperturbed too.

    if (is_multiple(this)) {

        binary_type = STABLE_INNER;

	if (is_parent())
	    get_oldest_daughter()->
	        startup_unperturbed_motion();

	if (get_binary_sister()->is_parent())
	    get_binary_sister()->get_oldest_daughter()->
	        startup_unperturbed_motion();
    }

    // cerr << "Leaving startup_unperturbed_motion: "; PRL(fully_unperturbed);

    // On normal exit, a kepler structure exists, timestep is the last
    // perturbed timestep, suitable for restart if no orbital evolution
    // occurs, and unperturbed_timestep is an integral number of timesteps.
    // The hdyn is frozen at the instant at which unperturbed motion was
    // started or continued.  It is updated at each unperturbed step.
    // The kepler structure is never updated, just predicted.

    // *** The only difference between pericenter reflection and full ***
    // *** unperturbed motion is the value of fully_unperturbed.      ***


#if 0
    putrq(get_parent()->get_dyn_story(), "unpert_startup_time",
	  get_system_time());
    int ifull = fully_unperturbed;
    putiq(get_parent()->get_dyn_story(), "fully_unperturbed", ifull);
#endif

}



real hdyn::set_unperturbed_timestep(bool test_phase)
{
    // Determine and set the unperturbed timestep.  Zero return value
    // means no unperturbed motion.
    //
    // *** May change the value of binary_type. ***
    //
    // Flag test_phase indicates that we should check that the system
    // is outside its semi-major axis (using KEP_OUTSIDE_SEMI) before
    // applying full merging.

    // On arrival in this function, binary_type must be one of the
    // "true" values set by is_unperturbed_and_approaching.  The
    // MULTIPLE values get special treatment.  The "binary" values, in
    // order of *increasing* external perturbation, are:
    //
    //		PERICENTER_REFLECTION
    //		FULL_MERGER
    //		CONTINUE_MERGER
    //
    // Since the binary has already been deemed "unperturbed", an
    // unperturbed step should be taken if possible, even if it
    // extends beyond the parent step.

    if (UNPERT_FUNCTION_ID) {
	cerr << ">> set_unperturbed_timestep for "
	     << format_label() << endl;
    }

    if (!test_phase) {
        // cerr << "set_unpert_timestep called with no test phase\n";
        // PRL(kep->get_energy());
    }

    hdyn * sister = get_binary_sister();
    real steps = 0;

    // Establish "default" outcome.

    if (is_multiple(this)) {

	// No default -- unperturbed motion does not occur if the
	// criteria in get_unperturbed_steps() are not met.

    } else {

	// Default is periastron reflection.  Determine the number of
	// perturbed steps to the reflection point.

	binary_type = PERICENTER_REFLECTION;

	real peri_time = kep->pred_advance_to_periastron() - time;
	steps = ceil(2 * peri_time / timestep);

	// (Recall that ceil rounds to the next integer up, so we
	// overshoot slightly.)

    }

    if (kep->get_energy() < 0.0) {

	// Consider the possibility of complete merging in a bound orbit.

	if (is_multiple(this) && multiple_type == APOCENTER_REFLECTION) {

	    // Determine if we are close to apocenter in a multiple.

	    real apo_time = kep->pred_advance_to_periastron()
	                          - time - kep->get_period()*0.5;
	    real steps = floor(((apo_time * 2.0) / timestep));

	    // Note: "floor" here means the separation at the end of
	    //	      the step is slightly greater than at the start.

	    if (REPORT_MULTIPLE) {
		cerr << "multiple (" << tt[multiple_type] << "): "
		     << format_label() << endl;
		PRC(apo_time), PRC(kep->get_period()); PRL(steps);
	    }

	    binary_type = STABLE_OUTER;

	} else if (!test_phase
		   || KEP_OUTSIDE_SEMI(*kep)
		   || (get_parent()->kep != NULL)) {

	    // Full merger criteria:
	    //     (1) not checking phase, or
	    //     (2) system outside its semi-major axis, or
	    //     (3) parent is already unperturbed.

	    // The criteria in get_unperturbed_steps() will return zero
	    // if it is not possible to fit an integral number of orbits
	    // or an advance to apastron into the time allowed by the
	    // parent step.  If usteps <= 0, this must mean that the timing
	    // of the parent step is the problem.  In that case, we fall
	    // back to pericenter reflection.  However, if that occurs
	    // near apocenter (as it will if the orbit had previously been
	    // advanced to apastron), we convert it to a complete orbit.

	    // If usteps > 0, for a true binary, we "overshoot" slightly,
	    // ending at the next perturbed step past apocenter.  For a
	    // multiple, we don't proceed to apocenter.  Instead, we
	    // require an integral number of orbits ending just before
	    // (i.e. at separation outside) a complete period multiple.

	    real usteps = get_unperturbed_steps(!is_multiple(this));

	    if (usteps <= 0) {

		if (REPORT_ZERO_UNPERT_STEPS) {

		    cerr << endl << "    set_unperturbed_timestep: "
			 << "zero step for unperturbed binary\n";

		    cerr.precision(12);
		    PRI(4); PRL(time);
		    cerr << "    parent: " << get_parent()->format_label()
			 << endl;
		    PRI(4); PRL((get_parent()->get_next_time()));
		    PRI(4); PRL(kep->get_period());
		    PRI(4); PRL(usteps);
		    cerr.precision(8);
		}

		// Note that steps and binary_type are left unchanged.

	    } else {

		steps = usteps;
		binary_type = FULL_MERGER;	// may be changed from
						// CONTINUE_UNPERTURBED
		fully_unperturbed = true;
	    }
	}
    }

    // Now steps is the number of perturbed time steps to the reflection
    // point (rounded to the nearest integer), or the number of perturbed
    // steps corresponding to an integral number of orbits (plus mapping
    // to apocenter for true binaries) if full merging is applied.  Note
    // that, in either case, the resulting unperturbed step is *not*
    // necessarily a power of 2, but it does retain the binary's place
    // in the block scheduling scheme.

    // Special case occurs if the reflection is occurring near apocenter.
    // In this case, promote the step to a full merger for a single orbit.

    if (binary_type == PERICENTER_REFLECTION && KEP_OUTSIDE_SEMI(*kep)) {
	steps = ceil(kep->get_period()/timestep);
	binary_type = FULL_MERGER;
	fully_unperturbed = true;
    }

    unperturbed_timestep = timestep * steps;

    sister->kep = kep;
    sister->unperturbed_timestep = unperturbed_timestep;
    sister->fully_unperturbed = fully_unperturbed;

    if (steps < 0) {
	cerr << endl
	     << "steps negative!, the code will break soon...;-("
	     << endl;
	PRC(timestep); PRC(steps); PRL(unperturbed_timestep);
    }
    
    return steps;
}



// get_unperturbed_steps moved here from hdyn_kepler.C by Steve, 8/7/98 

// Note from SPZ.  Previous implementation returned an int.
// Stellar evolution allows binaries with orbital periods as
// small as a few milliseconds (see Portegies Zwart 1998), so
// return type was changed to real.

// Additional (first) argument to_apo added by Steve 8/5/98.

// int hdyn::get_unperturbed_steps(bool predict, bool to_apo)
// real hdyn::get_unperturbed_steps(bool predict, bool to_apo)
// unsigned long hdyn::get_unperturbed_steps(bool predict, bool to_apo)

real hdyn::get_unperturbed_steps(bool to_apo, bool predict)
{
    // Return the number of current time steps to advance
    // unperturbed binary motion up to (but not past) the next
    // CM time step.
    // If to_apo is true (default), the step will continue on
    // to the next apocenter (before the CM step).

    // First check if parent is younger binary sister, in which
    // case its time step may not be "definitive."

    // Note from Steve, 7/17/97:  For unknown reasons, this
    // function fails to compile properly under g++ version
    // cygnus-2.7-96q4 on Dec UNIX V4.0B, at optimization
    // levels higher than 0...

    hdyn * p = get_parent();
    if (p->is_low_level_node() && p->get_elder_sister() != NULL)
	p = p->get_elder_sister();

    real pdt = p->get_next_time() - time;	// parent time step
    real pdt2 = pdt + 0.5*p->timestep;

    if (pdt < 0) {
	cerr << endl << "get_unperturbed_steps: ";
	PRC(p->get_next_time()); PRL(time);
	cerr << "*no* corrective action taken\n";
	return 0;
    }

    // Goal: to advance the binary by as great a time as possible,
    //	     subject to the constraints that (a) we do not wish to
    //	     go too far past the parent node's step, and (b) we do
    //	     not wish to end up with a separation smaller than the
    //	     current separation.  In the event of a conflict, it is
    //	     better to break constraint (a) than constraint (b), as
    //	     we will continue with the current perturbed time step
    //	     on restart.  The actual unperturbed timestep will be
    //	     an integral multiple of the perturbed step, to retain
    //	     proper block scheduling.
    //
    //  (1)  If to_apo is false (e.g. for unperturbed multiples, to
    //       minimize the tidal error), then we simply want to
    //       advance the system by an integral number of orbits.  If
    //       this is not possible within the parent step, or within
    //       half a parent time step after the present parent step,
    //       then we return zero steps (thus preventing unperturbed
    //       motion from starting or continuing).  The unperturbed
    //       step will *never* exceed the parent step by more than 1
    //       orbital period.
    // 
    //       Since the unperturbed time step is constrained by the
    //       value of the perturbed step, it is not in general
    //       possible to advance a binary by an exact number of orbit
    //       periods.  Unperturbed motion is initiated only on the
    //       inward part of the binary orbit, so we take the time
    //       step to be slightly *less* than a whole number of
    //       periods, ensuring that the final separation is slightly
    //       greater than the initial one.  We also ensure that the
    //       step will always pass the next apocenter (assuming that
    //       a step is to be taken), even if that will violate either
    //       or both of constraints (a) and (b) above.
    //
    //  (2)  It to_apo is true (e.g. for true binaries, where we will
    //       neglect the tidal error associated with changing the
    //       binary phase), we will also try to map the orbit to (just
    //       after) the apocenter preceding the parent's next step.
    //       If necessary, we may advance to the apocenter beyond the
    //       parent time step, subject to the same restrictions as in
    //       (1) above.

    // Determine integral number of orbits to take.

    real orb = floor(pdt / kep->get_period());

    // (Recall that floor rounds down to the next integer, so orb
    //  orbital periods end *before* pdt.)

    if (orb <= 0) {

	// See if we can take an orbital step in less than half
	// the next parent step (pdt2).

	if (kep->get_period() <= pdt2)
	    orb = 1;
    }

    // Determine time to next apocenter.

    real peri_time;
    if (predict)
	peri_time = kep->pred_advance_to_periastron() - time;
    else
	peri_time = kep->get_pred_time() - time;

    real apo_time = peri_time - 0.5*kep->get_period();
    if (apo_time < 0)
	apo_time += kep->get_period();		// apo_time is always >= 0

    real steps = 0;

    if (to_apo) {

	steps = ceil( (max(0.0, orb - 1) * kep->get_period() + apo_time)
		     						/ timestep);

	// Use of ceil here ensures that the unperturbed step ends just after
	// the apocenter immediately preceding the last allowed full orbit.
	// If orb = 0, this step takes us to the next apocenter, regardless
	// of the parent step.

    } else {

	steps = floor(orb * kep->get_period() / timestep);

	// Use of floor here ensures that the unperturbed step ends just
	// before an integral number of orbits.

	// However, this means that the binary separation at the end point
	// increases slowly.  Make sure that the step ends after apocenter
	// (and not just before).

	if (orb*kep->get_period() - steps*timestep
	     > kep->get_period() - apo_time)
	    steps += 1;				// should only need 1 correction
						// while loop is dangerous...

    }

    return steps;

}



void hdyn::integrate_unperturbed_motion(real eta,
					real eps,	// eps and dt_max are
					real dt_max,	// simply passed on to
							// create_or_delete_binary
					bool& reinitialize,
					bool force_time)
{
    // Update unperturbed binary motion and check for continuation.
    //
    // Flag reinitialize is set true if a reinitialization of the system
    // is needed on return (e.g. after binary mass loss).
    //
    // Flag force_time will force the binary time to the current system
    // time if set.

    // NOTE: Integrating a binary to a time other than the scheduled
    //	     end of its timestep may result in its being resolved at
    //	     an undesirable part of its orbit, causing large errors.

    if (UNPERT_FUNCTION_ID) {
	cerr << endl << ">> integrate_unperturbed_motion for "
	     << format_label() << " at time " << system_time << endl;
    }

    reinitialize = false;

    hdyn *sister = get_binary_sister();
    time += unperturbed_timestep;

    if (time != system_time) {

	// Note: this can happen in case of changing binary orbit
        //	 because of stellar evolution.  Probably undesirable
	//	 to terminate the motion in this case, and also need
	//	 special care in determining new step.  TO BE FIXED...

	if (!force_time && REPORT_CONTINUE_UNPERTURBED) {
	    cerr << "\nintegrate_unperturbed_motion for "
		 << format_label() << endl;
	    PRC(time); PRC(system_time); cerr << " are different.\n";
	    cerr << "Forcing time to system time...\n";
	}
	time = system_time;
    }

    // Store initial separation, for use below.

    real initial_sep_sq = kep->get_separation() * kep->get_separation();

    // An unperturbed step consists of transforming the kepler
    // structure to the new time, then updating the hdyn time, pos,
    // vel, acc, jerk, and perturbation, then determining if the
    // unperturbed motion is to continue.

    // Note:  Because the period is not exactly an integral number
    // of time steps, the mean and true anomalies, pos, vel, etc.,
    // will  change slightly from one step to another.  This drift
    // may lead to restart problems, as the kepler structure
    // obtained using the new pos and vel may not be precisely the
    // same as the original kepler.  (SLWM, 3/98)

    // (Basically, the operation kepler --> (pos, vel) --> kepler is
    // not quite cyclic, leading to small phase errors that subsequently
    // grow in time.)

    update_dyn_from_kepler();		// update pos, vel, and perturbation

    // This update will introduce a tidal error (especially in unperturbed
    // multiples) if the unperturbed time step is not an integral multiple
    // of orbit periods.  The error may be acceptably small for binaries,
    // which are picked up with a more stringent unperturbed criterion (they
    // are continued with a relaxed criterion, but this is probably OK if
    // they are advanced to apocenter on the first step).  However, both
    // the inner and the outer binaries of an unperturbed multiple may be
    // quite strongly perturbed, and the tidal errors significant.

    // Fix:  always advance the inner and outer binaries by an *integral*
    //	     number of periods in the unperturbed multiple case (and be
    //	     very wary of partially unperturbed multiple motion...).
    //
    //							(Steve, 5/8/98)

    // Possible alternatives are to allow the tidal error and either
    // simply keep a running total of such errors, or to absorb the
    // error into the motion (e.g. by correcting the outer orbit in
    // some way) -- neither is currently implemented.

    bool save_unpert = fully_unperturbed;
    fully_unperturbed = false;

    if (unperturbed_timestep < kep->get_period()) {
        kc->partial_unpert_orbit++;
	//  PRL(kc->partial_unpert_orbit);
    }
    else {
        kc->full_unpert_step++;
        kc->full_unpert_orbit += (int) (unperturbed_timestep
					   	/ kep->get_period());
	// PRC(kc->full_unpert_step);
	// PRL(kc->full_unpert_orbit);
    }

    // Note that we use the same unperturbed criterion as in the main code,
    // and apply the same unperturbed_timestep function.

    bool is_u = is_unperturbed_and_approaching();

    // int set_u = 0;
    // unsigned long set_u = 0;
    real set_u = 0;

    // cerr << "check set_unperturbed_timestep time: ";
    // PRC(cpu_time());

    if (is_u)
	set_u = set_unperturbed_timestep(!force_time);

    // PRL(cpu_time());

    // End of step.  Check to see if unperturbed motion can be extended.
    // Note that the kepler structure is never advanced, just predicted...

    if (is_u && set_u) {

	// Is still unperturbed. Not much to do here now...

        if (REPORT_CONTINUE_UNPERTURBED
	    || (is_multiple(this) && REPORT_MULTIPLE)) {

	    cerr.precision(12);
	    cerr << "\ncontinuing unperturbed motion for "
		 << format_label();
	    cerr << " and " << get_binary_sister()->format_label()
		 << " at time " << time << " (" << set_u << " steps)"
		 << endl;
	    cerr.precision(8);

	    if (UNPERT_REPORT_LEVEL > 0)
		print_unperturbed_binary_params();
	}

    } else {

	// Time to end the unperturbed motion.

	// Start by correcting any perturber lists containing the CM.
	// (This would be repaired anyway at the next perturbee CM step,
	// but that may be too late...)

	bool verbose = REPORT_END_UNPERTURBED
			    || (is_multiple(this) && REPORT_MULTIPLE);

// 	PRL(verbose);
// 	PRL(save_unpert);
// 	PRL(binary_type);

	if (save_unpert && !RESOLVE_UNPERTURBED_PERTURBERS)
	    expand_perturber_lists(get_root(), this, verbose);

	// NOTE: Ending the inner component of an unperturbed multiple
	//       should really end the outer component too.

	if (verbose) {

	    // Pericenter reflection will be separating on termination.

	    if (binary_type != NOT_APPROACHING	// (= end of peri reflection)
		|| REPORT_PERICENTER_REFLECTION) {

		cerr.precision(12);
		cerr << "\nending unperturbed motion for "
		     << format_label();
		cerr << " and " << get_binary_sister()->format_label()
		     << " at time " << time << endl;
		cerr << bt[binary_type] << " (binary_type = " << binary_type
		     << ")";
		cerr.precision(8);
		if (get_kepler())
		    cerr << "  period = " << get_kepler()->get_period();
		cerr << endl;

		if (UNPERT_REPORT_LEVEL > 0 || END_UNPERT_REPORT_LEVEL > 0) {

		    print_unperturbed_binary_params();

		    if (binary_type != NOT_APPROACHING) {

			hdyn* p = get_parent();
			hdyn* pnn = p->get_nn();

			cerr.precision(6);
			print_binary_from_dyn_pair(this, get_binary_sister(),
						   0, 0, true);
			if (pnn)
			    print_binary_from_dyn_pair(p, pnn, 0, 0, true);
			else
			    cerr << "nn is NULL" << endl;
			cerr.precision(8);

			if (pnn
			    && (UNPERT_REPORT_LEVEL > 1
				|| END_UNPERT_REPORT_LEVEL > 1)) {

			    pp3_minimal(get_top_level_node(), cerr);

			    // Attempt to estimate the work required to get
			    // past this encounter.

			    kepler outerkep;

			    outerkep.set_time(time);
			    outerkep.set_total_mass(p->get_mass());
			    outerkep.set_rel_pos(p->pos - pnn->pos);
			    outerkep.set_rel_vel(p->vel - pnn->vel);
			    outerkep.initialize_from_pos_and_vel();

			    real r_end = outerkep.get_separation();
			    if (perturbation_squared
				  > gamma2 * FULL_MERGE_TOLERANCE)
				r_end *= sqrt(perturbation_squared
					      / (gamma2
						   * FULL_MERGE_TOLERANCE));
			    if (r_end > outerkep.get_apastron())
				r_end = 0.9999*outerkep.get_apastron();
			    real transit_time
				= 2 * (outerkep.pred_advance_to_radius(r_end)
				       - time);

			    real ave_step
				= timestep * sqrt(kep->get_periastron()
						  /kep->get_separation());

			    PRI(4); PRL(transit_time);
			    PRI(4); PRL(ave_step);

			    if (find_perturber_node()) {
				cerr << "    estimate "
				     << 2 * find_perturber_node()->n_perturbers
					* (transit_time/ave_step) / 1e6
				     << " million force calculations to"
				     << " end of encounter"
				     << endl;
			    }

			    cerr << "    perturber node: "
				 << find_perturber_node()->format_label()
				 << endl;
			    cerr << "    n_pert = "
				 << find_perturber_node()->n_perturbers
				 << endl;
			    PRI(4); print_nn(find_perturber_node(), 2);

			    PRI(4); PRC(is_u); PRL(set_u);
			}
		    }
		}
	    }
	}

	bool update_dynamics[2] = {false, false};

	// Check for mass loss by binary evolution during the unperturbed
	// step.

	if (get_use_dstar() && has_dstar(this)) {

	    create_or_delete_binary(get_parent(),
				    update_dynamics,
				    eps, eta, dt_max);

	    			    // extra args needed here because
				    // of the possibility of merger

	    //cerr << "integrate_unperturbed_motion: "
	    //     << "back from create_or_delete_binary" << endl << flush;

	    // If a merger occurred in create_or_delete_binary, both
	    // this and its coll particle have already been deleted!
	    // Detect this by checking if this is still a valid node.

	    if (!is_valid()) {
		reinitialize = true;
		return;
	    }

	    // If the final orbit is radically different from the
	    // initial one, we have to recompute the time step.
	    // However, no need to do this if mass loss is forcing
	    // a complete reinitialization of the N-body system.

	    // update_dynamics[0] = true ==> full reinitialization needed
	    // update_dynamics[1] = true ==> new binary timestep needed

	    if (!update_dynamics[0] && update_dynamics[1]) {

		if (REPORT_START_UNPERTURBED    ||
		    REPORT_CONTINUE_UNPERTURBED ||
		    REPORT_END_UNPERTURBED)
		    cerr << "\nCorrected timestep for change "
			 << "in binary elements.\n";

		update_dyn_from_kepler();   // may introduce a tidal error...
					    // -- should include in de_total

		set_first_timestep(eta, 0.5*kep->get_period());
		get_binary_sister()->timestep = timestep;

		get_parent()->mass = mass + get_binary_sister()->get_mass();
	    }

	    reinitialize = update_dynamics[0];
	}

	delete kep;
	kep = NULL;
	get_binary_sister()->kep = NULL;
	unperturbed_timestep = -VERY_LARGE_NUMBER;
	get_binary_sister()->unperturbed_timestep = -VERY_LARGE_NUMBER;

	// Check that the timestep is reasonable.

	if (!(update_dynamics[0] || update_dynamics[1])) {

	    // Separation may have changed if phase was adjusted during step...

	    real sep_sq = square(pos - get_binary_sister()->pos);

	    if (sep_sq < 0.75*initial_sep_sq) {

		// Time step is probably too long.  Reduce it appropriately.

		real target_timestep = timestep
		    			* pow(sep_sq/initial_sep_sq, 0.75);
		while (timestep > target_timestep) timestep /= 2;

		if (REPORT_CONTINUE_UNPERTURBED)
		    cerr << "\nintegrate_unperturbed_motion: timestep for "
		         << format_label() << " reduced to " << timestep
			 << " on restart at time " << time << endl;

		get_binary_sister()->timestep = timestep;
	    }
	}
    }
}
