
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

#include "hdyn.h"

// Functions associated with calculating the energy of the system.
//
// Externally visible functions:
//
//	void calculate_energies_with_tidal
//	void print_recalculated_energies


void calculate_energies_with_tidal(hdyn* b, real eps2,
				   real& epot, real& ekin, real& etot)
{
    kira_calculate_energies(b, eps2, epot, ekin, etot);

    // (Uses calculate_energies_by_harp3 or calculate_energies.)

    // Note: calculate_energies is a dyn function, and so does not
    // set hdyn::pot.  Must assume in that case that the existing pots
    // are usable.  However, the existing pots will already include
    // the tidal contribution (see add_tidal).  The pots returned by
    // calulate_energies_by_harp must be corrected for the tidal terms.

    if (b->get_omega() > 0) {

	// Add tidal terms.

	real dpot = de_tidal_pot(b);
	epot += dpot;
	etot += dpot;

	for_all_daughters(hdyn, b, bb)
	    add_tidal(bb, true);	// "true" ==> pot only.
    }
}

#define MAX_EKIN	1.0e15		// Assume a hardware error if ekin
					// is greater than this value.

void print_recalculated_energies(hdyn* b, real eps2,
				 bool print_dde,
				 bool save_story)
{
    static real de_prev = VERY_LARGE_NUMBER;

    real epot = 0;
    real ekin = 0;
    real etot = 0;

    calculate_energies_with_tidal(b, eps2, epot, ekin, etot);
    cerr << "Energies: " << epot << " " << ekin << " " << etot;

    // Possible hardware problem?

    if (abs(ekin) > MAX_EKIN) {
	cerr << endl << "...retrying..." << endl;

	epot = ekin = etot = 0;
	calculate_energies_with_tidal(b, eps2, epot, ekin, etot);
	cerr << "Energies: " << epot << " " << ekin << " " << etot;
    }

    if (abs(ekin) > MAX_EKIN) print_dde = false;

    real de = 0;

    if (b->get_kira_counters()->initial_etot == 0) {

	b->get_kira_counters()->initial_etot = etot;
	b->get_kira_counters()->de_total = 0;

    } else {

	// real de = (etot - e_corr - initial_etot) / epot;
	// The above normalization fails if many stars have escaped
	// the system with HUGE kicks.

	de = (etot - b->get_kira_counters()->de_total
		- b->get_kira_counters()->initial_etot);
					//   / kira_stats.initial_etot;
					//     ^^^^^^^^ note! ^^^^^^^^
    }

    if (print_dde && de_prev != VERY_LARGE_NUMBER) {
	cerr << endl << "          de = " << de
	     << "  d(de) = " << de - de_prev;
	if (ekin > 0) cerr << "  (" << (de - de_prev) / ekin << ")";
	cerr << endl;
    } else
	cerr << "  de = " << de << endl;

    if (print_dde)
	de_prev = de;

    if (save_story) {
	putrq(b->get_dyn_story(), "energy_time", b->get_system_time());
	putrq(b->get_dyn_story(), "kinetic_energy", ekin);
	putrq(b->get_dyn_story(), "potential_energy", epot);
    }
}
