
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

// Functions associated with integration within kira.
//
// Externally visible functions:
//
//	void calculate_acc_and_jerk_for_list
//	void calculate_acc_and_jerk_on_all_top_level_nodes
//	void initialize_system_phase2


#include "hdyn.h"
#include "kira_debug.h"

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define DEBUG_KIRA_EV		0

#define TIME_INTERNAL	0    // Perform timing of force-evaluation functions
			     // in calculate_acc_and_jerk_for_list().

void calculate_acc_and_jerk_for_list(hdyn * b,
				     real eps, 
				     hdyn ** next_nodes,
				     int n_next,
				     real time,
				     bool exact,
				     bool tree_changed,
				     bool & reset_force_correction,
				     bool & restart_harp,
				     bool check_all)
{
    int i;
    real gamma23 = b->get_gamma23();
    real sys_t = next_nodes[0]->get_system_time();

    // Note that this function uses the same function calls as
    // calculate_acc_and_jerk_on_top_level_node, but in a different
    // order: all prologue calls are performed first, then all
    // top-level (i.e. GRAPE) force calculations, then all epilogue
    // calls.

    // Code to time specific force-calculation operations:

    real cpu0, cpu1, cpu2, cpu3, cpu4;
    int kmax = 1;

    if (TIME_INTERNAL) {
	for (int k = 0; k < n_next; k++) {
	    hdyn* bi = next_nodes[k];

	    if (bi->name_is("13a")
		&& bi->get_kepler() == NULL
		&& bi->get_time() > 6.6
		&& bi->find_perturber_node()
		&& bi->find_perturber_node()->get_n_perturbers() > 0
		) {

		// Found the particle.  Clean up the rest of the list.

		for (int kk = 0; kk < n_next; kk++)
		    if (kk != k)
			next_nodes[kk]->set_timestep(VERY_LARGE_NUMBER);
		n_next = 1;
		next_nodes[0] = bi;

		cerr.precision(12);
		cerr << endl << "timing " << bi->format_label()
		     << " at time " << bi->get_system_time() << endl;
		if (bi->is_low_level_node()) {
		    PRL(bi->get_top_level_node()->format_label());
		    if (bi->find_perturber_node()) {
			PRL(bi->find_perturber_node()->format_label());
			PRL(bi->find_perturber_node()->get_n_perturbers());
		    }
		}
		cerr.precision(8);

		kmax = 25000;
		cpu0 = cpu_time();
	    }
	}
    }

    for (int k = 0; k < kmax; k++) {

    for (i = 0; i < n_next; i++) {

	hdyn *bi = next_nodes[i];

	// bi->predict_loworder(sys_t);	  // redundant if predict_loworder_all
					  // was used in evolve_system()?

	// Probably need to predict the entire clump containing
	// node bi (Steve, 8/98):

	predict_loworder_all(bi->get_top_level_node(), sys_t);

	if (bi->is_top_level_node()) {
	    bi->clear_interaction();
	    bi->top_level_node_prologue_for_force_calculation(eps * eps,
							      gamma23,
							      exact);
	}
    }

    }
    if (kmax > 1) cpu1 = cpu_time();

    if (tree_changed)
	restart_harp = true;

    for (int k = 0; k < kmax; k++) {

    if (!exact) {

	kira_calculate_top_level_acc_and_jerk(next_nodes, n_next, eps * eps,
					      gamma23, time, restart_harp);

	// (Uses harp_calculate_acc_and_jerk or
	//  top_level_node_real_force_calculation, as appropriate.)

	int n = get_n_top_level();

	for (i = 0; i < n_next; i++) {
	    hdyn *bi = next_nodes[i];
	    if (bi->is_top_level_node()) {

		bi->inc_direct_force(n-1);	// direct force counter

		bi->top_level_node_epilogue_force_calculation(eps * eps);
	    }
	}
    }

    }
    if (kmax > 1) cpu2 = cpu_time();

    for (int k = 0; k < kmax; k++) {

    for (i = 0; i < n_next; i++) {
	hdyn *bi = next_nodes[i];

	// Compute forces on low-level nodes.

	if (bi->is_low_level_node() && bi->get_kepler() == NULL) {

	    if (DEBUG_KIRA_EV) {
	        cerr << "\nComputing force on low-level node "
		     << bi->format_label() << endl;
	    }

	    bi->clear_interaction();
	    bi->calculate_acc_and_jerk(eps * eps, gamma23, exact);

	    if (DEBUG_KIRA_EV) {
		PRL(bi->get_acc());
		PRL(bi->get_binary_sister()->get_acc());
		PRL(bi->get_pos());
		PRL(bi->get_binary_sister()->get_pos());
	    }
	}

	// Update all step counters (direct force counters are
	// incremented above; indirect force counters are handled
	// in hdyn_ev.C)

	if (k == 0) bi->inc_steps();
    }

    }
    if (kmax > 1) cpu3 = cpu_time();

    // Complete calculation of accs and jerks by correcting for C.M.
    // interactions.

    for (int k = 0; k < kmax; k++) {

    if (!exact) {

	// Note: correct_acc_and_jerk() does not check list membership to
	// determine if correction is needed, but instead checks that
	// get_next_time == system_time.  Normally, these are equivalent,
	// as the next_nodes list is determined based on that time
	// criterion.  However, when we reinitialize the system, the time
	// condition is not met.  Instead, all nodes except unperturbed
	// binary components are on the list.  Force correction to occur
	// by using the extra flag check_all.

	correct_acc_and_jerk(b, eps * eps,
			     reset_force_correction, check_all);
    }

    }
    if (kmax > 1) cpu4 = cpu_time();

    for (int k = 0; k < kmax; k++) {

    if (b->get_omega() > 0) {

        // Add coriolis, centrifugal, and tidal forces.

        for (i = 0; i < n_next; i++)
	    if (next_nodes[i]->is_top_level_node())
	        add_tidal(next_nodes[i]);

    }

    }
    if (kmax > 1) {
	cerr << "CPU times:  "
	     << (cpu1-cpu0)/kmax << "  "
	     << (cpu2-cpu1)/kmax << "  "
	     << (cpu3-cpu2)/kmax << "  "
	     << (cpu4-cpu3)/kmax << "  "
	     << (cpu_time()-cpu4)/kmax << endl;
	exit(0);
    }
}

void calculate_acc_and_jerk_on_all_top_level_nodes(hdyn * b,
						   real eps)
{
    int n_top = b->n_daughters();
    hdynptr * list = new hdynptr[n_top];

    int i_top = 0;
    for_all_daughters(hdyn, b, bb)
	list[i_top++] = bb;

    bool restart_harp = true;
    bool reset_force_correction = true;

    // Currently *very* unclear about the last four parameters,
    // especially the last two...

    calculate_acc_and_jerk_for_list(b, eps, list, n_top,
				    b->get_system_time(),
				    false,	// usually what we want...
				    false,	// called before tree changes
				    restart_harp,
				    reset_force_correction);
}



// initialize_system_phase2:  Calculate acc, jerk, timestep, etc for all nodes.

// System should be synchronized prior to calling this function.

void initialize_system_phase2(hdyn * b,
			      real eps,
			      real eta,
			      real initial_step_limit,
			      bool set_dt)
{
    dbg_message("initialize_system_phase2", b);

    if (!b->is_root()) {
	err_exit("initialize_system_phase2 called with non-root");
    }

    static int work_size = 0;
    static hdyn ** nodes = NULL;
    static int nnodes = -1;

    int n = 0;
    for_all_nodes(hdyn, b, b1) n++ ;

    if  (work_size < n) {
	if (!nodes) delete nodes;
	nodes = new hdynptr [n+10];
	work_size = n+10;
    }

    // (New variable names necessary here because of DEC C++...)

    for_all_nodes(hdyn, b, b2) {
	if (!b2->get_kepler() && (b2->get_unperturbed_timestep() > 0)) {

	    // When is this necessary? (SLWM 3/98)

	    cerr << "Created kepler for unperturbed binary "
		 << b2->get_parent()->format_label() << "..."
		 << endl;
	    b2->update_kepler_from_hdyn();
	}
    }

    // Make a list of all nodes except unperturbed binary components.

    n = 0;
    for_all_nodes(hdyn, b, b3) {
	if ((b3 != b) && (!b3->get_kepler())) {
	    nodes[n] = b3;
	    n++ ;
	}
    }

    real time = nodes[0]->get_system_time();

    bool tree_changed = true;
    bool reset_force_correction = true;
    bool restart_harp = true;
    bool exact = false;
    bool check_all = true;

    calculate_acc_and_jerk_for_list(b, eps, nodes, n, time,
				    exact,
				    tree_changed,
				    reset_force_correction,
				    restart_harp,
				    check_all);

    // (All perturber lists have been redetermined...)

    real min_dt = VERY_LARGE_NUMBER;

    for_all_nodes(hdyn, b, b4) {
	if ((b4 != b) && (!b4->get_kepler())) {
	    b4->store_old_force();
	    if (set_dt || (b4->get_time() <= 0 || b4->get_timestep() <= 0) )
		b4->set_first_timestep(eta, initial_step_limit);
	    min_dt = min(min_dt, b4->get_timestep());
	}
    }

    // Check for perturbed binaries in the input data...

    if (set_dt && b->get_system_time() <= 0) {
	for_all_nodes(hdyn, b, b5) {
	    if ((b5 != b) && (!b5->get_kepler())) {
		if (b5->is_parent()
		    && b5->get_oldest_daughter()
			 ->get_perturbation_squared() > 1)
		    b5->set_timestep(min_dt);
	    }
	}
    }

    // cerr << "leaving initialize_system_phase2()\n\n";
}
