#define STARLAB_HAS_HARP

       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//// kira:  Hermite N-body integrator with evolving hierarchical tree
////        structure, stellar and binary evolution, and an external
////        tidal field.
////
//// Options:     -a    specify accuracy parameter [0.1]
////              -B    turn on binary evolution [off]
////              -c    include comment [none]
////              -C    specify GRAPE release interval, in seconds [15]
////              -d    specify log output interval [0.25]
////              -D    specify snapshot interval [end of run]
////              -e    specify softening length [0]
////              -E    use exact calculation [false]
////              -f    specify close-encounter distance [1 --> 1/N]
////              -g    specify hysteresis factor [2.5]
////              -G    specify initial stripping radius [none]
////              -h    specify stellar-evolution time step [0.015625 = 1/64]
////              -I    specify (re)initialization timescale [1]
////              -J    specify Jacobi radius [none]
////              -k    specify perturbation factor [1.e-7]
////              -m    specify true mean stellar mass, in solar masses [none]
////              -M    specify true total system mass, in solar masses [none]
////              -n    stop at specified number of particles [10]
////              -N    specify frequency of CPU check output [25000]
////              -O    save (and overwrite) extra snapshot at each output [no]
////              -q    specify virial ratio [0.5]
////              -Q    use tidal field [none]
////              -r    use relaxation time scaling [do not use]
////              -R    specify virial or tidal radius, in pc [none]
////              -s    specify random seed [take from system clock]
////              -S    turn on stellar evolution [off]
////              -t    specify time span of calculation [10]
////              -T    specify half-mass crossing time, in Myr [none]
////              -u    toggle unperturbed multiple motion [disabled]
////              -U    toggle all unperturbed motion [enabled]
////              -v    toggle "verbose" mode [on]
////              -V    specify initial virial radius (code units)
////                                    [1; take from input snap if possible]
////              -X    specify escaper removal timescale [reinit]
////              -z    specify stellar merger criterion [0 stellar radii]
////              -Z    specify stellar encounter criterion [0 stellar radii]
////
//// As a convenient shorthand, any "dt" interval specified less than zero
//// is interpreted as a power of 2, i.e. "-d -3" sets dt_log = 0.125.

// Level-2 help:

//++ Some run-time parameters:
//++
//++ The initial virial radius is read from the input snapshot.  If no
//++ initial virial radius is found there, the value specified by the
//++ "-V" command-line option is used.  If no "-V' option is set, a
//++ default value of 1 is assumed.
//++
//++ The initial mass is read from the input snapshot.  If no initial
//++ mass is found there, the mass is computed if t = 0; otherwise, a
//++ default value of 1 is assumed.  The initial mass is saved in the
//++ output snapshot for future use.
//++
//++ If a tidal field is specified ("-Q"), the initial Jacobi radius
//++ is set from the "initial_rtidal_over_rvirial" entry in the input
//++ file, if such an entry exists.  This value may be overridden by
//++ setting the Jacobi radius on the command line ("-J").  If specified,
//++ the command-line value scales the initial tidal radius (if found),
//++ or the virial radius (otherwise).  The initial Jacobi radius is
//++ written to the output snapshot for future use.  If a
//++ "kira-written" initial Jacobi radius is found in the input file,
//++ it is used regardless of the other settings.  Thus, in a typical
//++ situation, the Jacobi radius computed at t = 0 will be used in
//++ all subsequent restarts.
//++
//++ A stripping radius may be specified with the "-G" option.  If a
//++ command-line value specified, it scales the Jacobi radius if one
//++ is known, or the virial radius otherwise.  The scaled stripping
//++ radius (normalized to unit mass) is written to the output
//++ snapshot for future use.  If a scaled stripping radius is found
//++ in the input file, it is used regardless of the other settings.
//++ As with the initial Jacobi radius, in a typical situation, the
//++ scaled stripping radius computed at t = 0 will be used in all
//++ subsequent restarts.
//++
//++ If the input snapshot does not contain physical scales (and if
//++ stellar evolution/interactions are enabled), they may be set on
//++ the command line.  The "-M" option specifies the total system
//++ mass, in solar units.  The "-R" option specifies the value of the
//++ Jacobi radius (if tidal field) or the virial radius (otherwise),
//++ in parsecs.  The "-T" option specifies the (Heggie & Mathieu)
//++ system time unit, in Myr.  If the "-r" option is specified, the
//++ time unit is scaled to the system relaxation time, rather than
//++ the dynamical time.

//      J. Makino, S. McMillan          12/92
//      S. McMillan, P. Hut              9/94
//	J. Makino, S. McMillan          12/95
//	J. Makino, S. McMillan           7/96
//	J. Makino, S. McMillan           1/97
//	S. McMillan			 6/97
//	S. McMillan, S. Portegies Zwart	 3/98
//	S. McMillan, S. Portegies Zwart	 7/98
//	S. McMillan			 8/98
//
// NOTE:  ALL direct references to STARLAB_HAS_HARP are confined to
//	  this file (referenced externally via functions below).
//

// Known problems:  First timestep is a bit shaky, in particular 
//                  if the system is cold.
//
// July 1997 (SM)
//
//   Need to have the ability to ensure rigorous reproducability on restart.
//   Should restructure evolve_system to allow log output, snap output,
//   escaper removal, stellar evolution, debugging to be performed
//   independently.



#ifdef TOOLBOX

#include "hdyn.h"
#include "star/dstar_to_kira.h"
#include "kira_local.h"
#include "kira_debug.h"

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define DEBUG_MAIN		false

#define CHECK_PROGRESS		true
#define N_CHECK			25000


#define INITIAL_STEP_LIMIT  0.0625	// Arbitrary limit on the first step

real initial_step_limit = INITIAL_STEP_LIMIT;		// Global data!
real step_limit = INITIAL_STEP_LIMIT;
static kira_counters kc_prev; 
static bool dbg = false;

static int n_check = N_CHECK;
void set_n_check(int n)
{
    n_check = n;
}

static int n_top_level = 0;				// should be made part
void set_n_top_level(hdyn* b)				// of the node class
{							// eventually...
    n_top_level = b->n_daughters();
}

int get_n_top_level()
{
    return n_top_level;
}


//========================================================================

// HARP interface, for kira and library functions.
// All HARP-related calls come through here.

#if defined(STARLAB_HAS_HARP)

#define HARP_CHECK_COUNT 15000
#define HARP_MAX_CPU 15

int  harp_check_count = HARP_CHECK_COUNT;		// Global data!
real harp_max_cpu = HARP_MAX_CPU;			// Global data!

#endif

bool kira_use_harp()
{
#if defined(STARLAB_HAS_HARP)
    return true;
#else
    return false;
#endif
}

void kira_set_harp_cpu(real cpu)
{
#if defined(STARLAB_HAS_HARP)
    harp_max_cpu = cpu;
    harp_check_count = (int)harp_max_cpu * 1000;
#endif
}

void kira_calculate_energies(hdyn* b, real eps2,
			     real& epot, real& ekin, real& etot)
{
#if defined(STARLAB_HAS_HARP)
    calculate_energies_by_harp3(b, eps2, epot, ekin, etot);
#else
    calculate_energies(b, eps2, epot, ekin, etot);
#endif
}

void kira_calculate_top_level_acc_and_jerk(hdyn ** next_nodes,
					   int n_next,
					   real eps2,
					   real gamma23,
					   real time,
					   bool & restart_harp)
{
#if defined(STARLAB_HAS_HARP)
    harp_calculate_acc_and_jerk(next_nodes, n_next, eps2,
				gamma23, time, restart_harp);
    restart_harp = false;
#else
    for (int i = 0; i < n_next; i++) {
	hdyn *bi = next_nodes[i];
	if (bi->is_top_level_node())
	    bi->top_level_node_real_force_calculation(eps2);
    }
#endif
}

void compute_densities(hdyn* b, vector& cod_pos, vector& cod_vel)
{
#if defined(STARLAB_HAS_HARP)
    cerr << "Computing densities using HARP..." << endl;
    real cpu0 = cpu_time();

    harp_calculate_densities(b, 1.0);	// place densities in dyn stories
    real cpu1 = cpu_time();

    compute_mean_cod(b, cod_pos, cod_vel);
    real cpu2 = cpu_time();

    cerr << "CPU times:  density " << cpu1 - cpu0
	 << "  cod " << cpu2 - cpu1
	 << endl;
#endif
}

//========================================================================



local void getq(story* s, char* l, int &i)		{i = getiq (s, l);}
local void getq(story* s, char* l, unsigned long &i)	{i = getulq(s, l);}
local void getq(story* s, char* l, real &r)		{r = getrq (s, l);}
local void getq(story* s, char* l, vector &v)		{v = getvq (s, l);}

#  define GETLOG(x) if (find_qmatch(b->get_log_story(), #x)) \
		        getq(b->get_log_story(), #x, \
			     b->get_kira_counters()->x); \
		    else \
		        b->get_kira_counters()->x = 0;

local void initialize_counters_from_log(hdyn* b) {
  
    // Counters may be saved in the root log story.

    b->get_kira_counters()->cpu_time = 0;
    GETLOG(total_cpu_time);

    GETLOG(step_top_single);
    GETLOG(step_top_cm);
    GETLOG(step_low_level);
    GETLOG(force_correction);

    GETLOG(pert_step);
    GETLOG(pert_with_list);
    GETLOG(pert_without_list);
    GETLOG(perturber_overflow);

    GETLOG(full_unpert_step);
    GETLOG(full_unpert_orbit);
    GETLOG(partial_unpert_orbit);

    GETLOG(tree_change);
    GETLOG(top_level_combine);
    GETLOG(low_level_combine);
    GETLOG(top_level_split);

    GETLOG(leaf_merge);

    GETLOG(dm_escapers);
    GETLOG(dm_massloss);

    GETLOG(initial_etot);
    GETLOG(de_total);
    GETLOG(de_merge);
    GETLOG(de_massloss);
    GETLOG(de_kick);

    kc_prev = *b->get_kira_counters();
}


// Handy local functions...

local inline bool node_contains(hdyn * b, int i)	// overloaded
{
    // Hmmm... for_all_leaves() fails on Steve's
    //	       merlot with DEC C++ if b is a leaf...

    if (b->is_parent()) {
        for_all_leaves(hdyn, b, bb)
	    if (bb->get_index() == i)
		return true;
    } else
        if (b->get_index() == i)
	    return true;

    return false;
}

local inline bool node_contains(hdyn * b, char* s)	// overloaded
{
    if (b->is_parent()) {
        for_all_leaves(hdyn, b, bb)
	    if (bb->name_is(s))
		return true;
    } else
        if (b->name_is(s))
	    return true;

    return false;
}

// match_label: return true iff b's id (name, index, or #index)
//              matches the specified string.

local bool match_label(hdyn* b, char* label)
{
    if (b->name_is(label))
        return true;
#if 0
    // Function format_label() prepends a # to numbers.
    // Check for the number without the #.

    if (b->get_index() >= 0) {
	char id[64];
	sprintf(id, "%d", b->get_index());
	if (streq(id, label))
	    return true;
    }
#endif
    return false;
}

// match_label_tree: return true iff the label of any
//                   node below b (or b) matches the
//                   specified string.

local bool match_label_tree(hdyn* b, char* label)
{
    for_all_nodes(hdyn, b, bb)
	if (match_label(bb, label)) return true;
    return false;
}

// cond_print_nn: for each b in the list, print out the IDs of the entire
//                subtree containing b if any of those IDs match the
//		  specified string.

local void cond_print_nn(hdyn** list, int n, char* label, char* header = NULL)
{
    for (int i = 0; i < n; i++) {
	hdyn* b = list[i];
	if (b) {
	    if (match_label_tree(b->get_top_level_node(), label)) {
		if (header) cerr << header << endl;
		cerr << "print_nn triggered by node "
		     << b->format_label() << endl;
		for_all_leaves(hdyn, b, bb)
		    print_nn(bb, 2);
	    }
	}
    }
}

local void test_kepler(hdyn *b)
{
    for_all_nodes(hdyn,b,bb) {
	if (bb->get_kepler()) {
	    if (bb->is_top_level_node()) {
		cerr << " test_kepler: top level ";
		bb->pretty_print_node(cerr); cerr << "has kepler \n";
		pp3(bb, cerr);
	    }
	}
    }
}



// Triple debugging...

local void print_triple_stats(hdyn* root, hdyn* b)
{
    if (!b->is_top_level_node()) return;
    if (b->n_leaves() != 3) return;

    hdyn* ss = b->get_oldest_daughter();
    hdyn* bs = ss->get_younger_sister();
    if (ss->is_parent()) {
	hdyn* tmp = ss;
	ss = bs;
	bs = tmp;
    }

    // Triple is (ss, bs), with ss single, bs double.

    real mss = ss->get_mass();
    vector pos_ss = b->get_pos() + ss->get_pos();
    vector vel_ss = b->get_vel() + ss->get_vel();
    hdyn* bs1 = bs->get_oldest_daughter();
    hdyn* bs2 = bs1->get_younger_sister();
    real mb1 = bs1->get_mass();
    real mb2 = bs2->get_mass();
    vector pos_bs1 = b->get_pos() + bs->get_pos() + bs1->get_pos();
    vector vel_bs1 = b->get_vel() + bs->get_vel() + bs1->get_vel();
    vector pos_bs2 = b->get_pos() + bs->get_pos() + bs2->get_pos();
    vector vel_bs2 = b->get_vel() + bs->get_vel() + bs2->get_vel();

    real pot_1 = -mss * (mb1+mb2) / abs(ss->get_pos() - bs->get_pos());
    real pot_2 = -mss * mb1 / abs(pos_ss - pos_bs1)
		 -mss * mb2 / abs(pos_ss - pos_bs2);
    real pot_b = -mb1 * mb2 / abs(pos_bs1 - pos_bs2);

    real pot_ext = 0;
    for_all_daughters(hdyn, root, x) {
	if (x != b)
	    pot_ext -= x->get_mass() * (mss/abs(x->get_pos() - pos_ss)
					+ mb1/abs(x->get_pos() - pos_bs1)
					+ mb2/abs(x->get_pos() - pos_bs2));
    }

    real ke = 0.5 * (mss*square(vel_ss) + mb1*square(vel_bs1) 
		     			+ mb2*square(vel_bs2));
    // PRC(ke), PRL(pot_ext);
    // PRC(pot_2 + pot_b + ke), PRL(pot_ext + pot_2 + pot_b + ke);

    // print_binary_from_dyn_pair(ss, bs, 0, 0, true);
    // print_binary_from_dyn_pair(bs1, bs2, 0, 0, true);

    cerr << endl << "Triple " << b->format_label() << " at time "
	 << b->get_system_time() << ":" << endl
	 << "    Eint = " << pot_2 + pot_b + ke
	 << "  Eint + phi_ext = " << pot_2 + pot_b + ke + pot_ext << endl;
}



local void print_binary_diagnostics(hdyn* bi)
{
    bool diag;

    kepler *kep;
    hdyn *s = bi->get_binary_sister();
    real M = bi->get_parent()->get_mass();

    if (bi->get_kepler() == NULL) {
	kep = new kepler;
	kep->set_time(bi->get_time());
	kep->set_total_mass(M);
	kep->set_rel_pos(bi->get_pos() - s->get_pos());
	kep->set_rel_vel(bi->get_vel() - s->get_vel());
	kep->initialize_from_pos_and_vel();
    } else
	kep = bi->get_kepler();

    // if (kep->get_separation() < kep->get_semi_major_axis())
    //    diag = false;

    if (diag) {

#if 0		    
	cerr << "\nLow-level node bi = ", bi->print_label(cerr);
	cerr << endl;
	PRI(4); PRL(bi->get_time());
	cerr.precision(10);
	PRI(4); cerr << "top-level: ",
	PRL(bi->get_top_level_node()->get_time());
	cerr.precision(PRECISION);
	pp2(bi->get_top_level_node(), cerr, 2);
	PRI(4);
	PRL(bi->get_top_level_node()->get_valid_perturbers());
	PRI(4); PRL(bi->get_top_level_node()->get_n_perturbers());
#endif

	if (bi->get_kepler())
	    cerr << "    bi unperturbed";
	else
	    cerr << "    bi perturbed";

	real r = kep->get_separation();
	real E = kep->get_energy();
	real a = r;
	if (E < 0) a = min(r, kep->get_semi_major_axis());

	cerr << ", -E/mu = " << -E
	    << "  P = " << 2*M_PI * sqrt(a*a*a/M)
		<< "  e = " << kep->get_eccentricity()
		    << endl
			<< "    sma = " << kep->get_semi_major_axis()
			    << "  r = " << r
				<< endl;

	PRI(4);
	if (bi->get_kepler()) PRC(bi->get_unperturbed_timestep());
	PRL(bi->get_timestep());
    }

    if (bi->get_kepler() == NULL) delete kep;
}



local void check_unperturbed(hdyn* bi, real eps, real eta, real dt_log,
			     bool& tree_changed)
{
    // Check for unperturbed motion.

    // This is the ONLY place in kira where unperturbed motion
    // (of any sort) is initiated.

    // The unperturbed criterion is defined in function
    // is_unperturbed_and_approaching()) (see hdyn_unpert.C).

    if (bi->get_kepler() == NULL && eps == 0
	&& (bi->is_unperturbed_and_approaching())) {

	// Must bring triple components up to date before
	// continuing (about to freeze the entire triple system).

	// *** In general, should bring ANY substructure
	// *** components up to date before proceeding
	// *** with unperturbed startup.

	// cerr << endl << "Unperturbed motion for "
	//      << bi->format_label()
	//      << " at time " << bi->get_time() << endl;

	if (bi->is_parent() || bi->get_binary_sister()->is_parent()) {

	    bool synch = false;

	    if (bi->is_parent()) {
		bool sync = false;
		for_all_nodes(hdyn, bi, bb) {
		    if (bb->get_time() < bi->get_time())
			sync = true;
		}
		if (sync)
		    synchronize_tree(bi, eps, eta, dt_log);
		synch |= sync;
	    }

	    if (bi->get_binary_sister()->is_parent()) {
		bool sync = false;
		for_all_nodes(hdyn, bi->get_binary_sister(), bb) {
		    if (bb->get_time()
			< bi->get_binary_sister()->get_time())
			sync = true;
		}
		if (sync)
		    synchronize_tree(bi->get_binary_sister(),
				     eps, eta, dt_log);
		synch |= sync;
	    }

	    // Synchronize_tree() is OK here only because we
	    // don't (yet) apply tidal fields to low-level nodes.

	    // If a newly-synchronized node is not on the
	    // integration list (which must be the case, as we
	    // only synchronize if a node is not up to date), then
	    // there is a good chance that the scheduling list will
	    // be corrupted, so force the list to be recomputed.

	    if (synch) {
		tree_changed = true;
		cerr << "Synchronized binary components...\n";
	    }
	}

	// Actual initialization of unperturbed motion:

	bi->startup_unperturbed_motion();
    }
}



local void merge_and_correct(hdyn* b, hdyn* bi, hdyn* bcoll,
			     real eps, real eta, real dt_log)
{
    cerr << "Merging node " << bi->format_label();
    cerr << " with node " << bcoll->format_label() 
	 << " at time " << bi->get_system_time() << endl;

    bool is_pert = (bi->get_kepler() == NULL);
    hdyn* cm = bi->merge_nodes(bcoll, eps, eta, dt_log);

    // Check and correct for the presence of either merged node
    // on any perturber list.

    if (RESOLVE_UNPERTURBED_PERTURBERS || is_pert) {
	hdynptr del[2];
	del[0] = bi;
	del[1] = bcoll;
	correct_perturber_lists(b, del, 2, cm);
    }

    delete bi;
    delete bcoll;

    b->get_kira_counters()->leaf_merge++;
}

local hdyn* check_merge(hdyn* bi, real eps, real eta, real dt_log)
{
    hdyn * bcoll;
    if ((bcoll = bi->check_merge_node()) != NULL) {

	hdyn* b = bi->get_root();

	merge_and_correct(b, bi, bcoll, eps, eta, dt_log);

	// Check for multiple mergers.

	bool merge_flag = true;
	while (merge_flag) {

	    merge_flag = false;

	    for (hdyn* bb = b;
		 (bb != NULL) && !merge_flag;   
		 bb = (hdyn*) bb->next_node(b)) {

		if (bb->is_leaf()) {
		    if ((bcoll = bb->check_merge_node())!= NULL) {
			merge_and_correct(b, bb, bcoll, eps, eta, dt_log);
			merge_flag = true;
		    }
		}
	    }
	}
    }
    return bcoll;
}



#define TIME_LIST 0

local int integrate_list(hdyn * b, real eps, real eta, 
			 real lag_factor, real dt_log,
			 hdyn ** next_nodes, int n_next,
			 bool exact, bool & tree_changed)
{
    int i, steps = 0;
    real sys_t = next_nodes[0]->get_system_time();
    static bool restart_harp = true;
    static bool reset_force_correction = true;

    // Code to time specific force-calculation operations:

    real cpu0, cpu1;
    int kmax = 1;

    if (TIME_LIST) {
	for (int k = 0; k < n_next; k++) {
	    hdyn* bi = next_nodes[k];

	    if (bi->name_is("(13a,13b)")
		&& bi->get_kepler() == NULL
		&& bi->find_perturber_node()
		&& bi->find_perturber_node()->get_n_perturbers() > 0
		) {

		// Found the particle.  Clean up the rest of the list.

		for (int kk = 0; kk < n_next; kk++)
		    if (kk != k)
			next_nodes[kk]->set_timestep(VERY_LARGE_NUMBER);
		n_next = 1;
		next_nodes[0] = bi;

		cerr.precision(12);
		cerr << endl << "timing " << bi->format_label()
		     << " at time " << bi->get_system_time() << endl;
		if (bi->is_low_level_node()) {
		    PRL(bi->get_top_level_node()->format_label());
		    if (bi->find_perturber_node()) {
			PRL(bi->find_perturber_node()->format_label());
			PRL(bi->find_perturber_node()->get_n_perturbers());
		    }
		}
		cerr.precision(8);

		kmax = 25000;
		cpu0 = cpu_time();
	    }
	}
    }

    // Separate the force calculation from the rest for
    // implementation on the GRAPE-4.

    // test_kepler(next_nodes[0]->get_root());

    // cond_print_nn(next_nodes, n_next, "1",
    //		    "\nintegrate_list: before calculate_acc_and_jerk");

    real t_next = next_nodes[0]->get_next_time();


    for (int k = 0; k < kmax; k++) {

    calculate_acc_and_jerk_for_list(b, eps, next_nodes, n_next, t_next,
				    exact, tree_changed,
				    reset_force_correction,
				    restart_harp);
    }
    if (kmax > 1) cpu1 = cpu_time();

    // cond_print_nn(next_nodes, n_next, "1",
    //		     "integrate_list: after calculate_acc_and_jerk");

    // Apply corrector and redetermine timesteps.

    bool reinitialize = false;

    for (int k = 0; k < kmax; k++) {

    for (i = 0; i < n_next; i++) {

	hdyn *bi = next_nodes[i];
	bool diag = false;

	if (bi->get_kepler() == NULL) {

	    if (diag) cerr << " perturbed correction for "
			   << bi->format_label() << endl;

	    if (!bi->correct()) {

		// A problem has occurred during the step.

		// Recompute acc and jerk on the front end (no bookkeeping
		// yet) and retry once.  (Steve 9/15/98)

		cerr << "retrying force calculation for "
		     << bi->format_label() << endl;

		// Better do an exact calculation, as we can't (yet) call
		// correct_acc_and_jerk() to correct a single particle...

		bi->clear_interaction();
		bi->calculate_acc_and_jerk(eps*eps, b->get_gamma23(), true);
		bi->set_valid_perturbers(false);
		if (bi->is_top_level_node() && b->get_omega() > 0)
		    add_tidal(bi);

		if (!b->correct()) {
		    cerr << endl << "failed to correct error for particle "
			 << bi->format_label() << " at time "
			 << bi->get_system_time() << endl;
		    err_exit("Run terminated in integrate_list");
		}
	    }

	    bi->update(eta, step_limit);  // Note step_limit...
	    bi->init_pred();

	    bi->store_old_force();

	    // Note that old_acc = acc at the end of a step.

	    // PRL(bi->get_acc());

	} else {

	    if (diag) {
		cerr << "unperturbed motion for "
		     << bi->format_label() << endl;
		cerr << "nn = " << bi->get_nn()->format_label() << endl;
	    }

//  	    if (bi->name_is("(9,5)")) {
//  		pp3(bi->get_top_level_node(), cerr);
//  		print_recalculated_energies(b, 0);
//  		print_triple_stats(b, bi->get_top_level_node());
//  	    }

	    if (eps == 0)
		bi->integrate_unperturbed_motion(eta, eps, dt_log,
						 reinitialize);
	    else
		err_exit("integrate_list() invoked with non-zero softening");

//  	    if (bi->name_is("(9,5)")) {
// 		pp3(bi->get_top_level_node(), cerr);
//  		print_recalculated_energies(b, 0);
//  		print_triple_stats(b, bi->get_top_level_node());
//  	    }

	}
    }

    }
    if (kmax > 1) {
	cerr << "CPU times: "
	     << (cpu1-cpu0)/kmax << "  "
	     << (cpu_time()-cpu1)/kmax << endl;
	exit(0);
    }

    dbg = false;

    if (dbg) {
        cerr << "\nTime = " << next_nodes[0]->get_time()
	     << ",  steps = " << steps << "\n";
	print_recalculated_energies(b, eps * eps);
    }

    // Complete all steps before modifying binary structure...

    for (i = 0; i < n_next; i++) {

	hdyn *bi = next_nodes[i];
	bool diag = false;

	if (!bi->is_low_level_node()) {

	    if (bi->is_leaf())
		b->get_kira_counters()->step_top_single++;
	    else
		b->get_kira_counters()->step_top_cm++;

	    // Diagnostics:

	    if (diag) {
	        cerr << "\nTop-level node bi = " << bi->format_label()
		     << " at time " << bi->get_time() << endl;
		cerr << "timestep = " << bi->get_timestep() << endl;
	    }

	} else {

	    update_binary_sister(bi);
	    b->get_kira_counters()->step_low_level++;

	    // print_binary_diagnostics(bi);

	    // Check for unperturbed motion:

	    check_unperturbed(bi, eps, eps, dt_log, tree_changed);
	}

	steps++;
    }

    // ONLY ONE tree reconstruction (following a collision or
    // otherwise) is currently permitted per block step.

    //++ Note from Steve to Steve, 7/98.  Could we relax the
    //++ requirement of only one tree reconstruction per step?

    for (i = 0; i < n_next; i++) {

	hdyn* bcoll = check_merge(next_nodes[i], eps, eta, dt_log);

	if (bcoll) {

	    // Merger occurred and tree has to be rebuilt.

	    tree_changed = true;
	    restart_harp = true;
	    reset_force_correction = true;

	    // **** Unsynchronized call to initialize_system_phase2()...

	    cerr << "call initialize_system_phase2() from integrate_list\n";

	    initialize_system_phase2(b, eps, eta, initial_step_limit);
	    PRL(tree_changed);

	    // Remove merged star and its merger companion from the list.
	    // Note that the list may be still be incomplete on return, as
	    // multiple merger companions may remain on it.

	    next_nodes[i] = NULL;
	    for (int j = 0; j < n_next; j++) {
		if (next_nodes[j] == bcoll)
		    next_nodes[j] = NULL;
	    }

	    return steps;
	}
    }

    if (reinitialize) {

	// **** Unsynchronized call to initialize_system_phase2()...

	initialize_system_phase2(b, eps, eta, initial_step_limit);

    } else {

	for (i = 0; i < n_next; i++) {

	    hdyn *bi = next_nodes[i];

	    hdynptr* cm_list = NULL;
 	    int n_list = 0;

 	    if (bi->is_low_level_node() || bi->is_parent()) {

 		// Make a list of CM nodes in the current clump.

 		for_all_nodes(hdyn, bi, bb)
 		    if (bb->is_parent()) n_list++;

 		if (n_list > 0) {
 		    cm_list = new hdynptr[n_list];
 		    n_list = 0;
 		    for_all_nodes(hdyn, bi, bb)
 			if (bb->is_parent()) cm_list[n_list++] = bb;
 		}
 	    }

	    // Check (and modify, if necessary) the tree structure:

	    if (bi->adjust_tree_structure(lag_factor, eps, eta, dt_log)) {

		b->get_kira_counters()->tree_change++;
		reset_force_correction = true;
		restart_harp = true;
		tree_changed = true;

#ifndef STARLAB_HAS_HARP

		if (DEBUG_MAIN) {
		    cerr << "\nAfter adjusting tree structure... \n";
		    cerr << "Time = " << next_nodes[0]->get_system_time()
			 << " single_steps = "
			 << b->get_kira_counters()->step_top_single << endl;
		    print_recalculated_energies(b, eps * eps);
		    // pp3(bi->get_top_level_node(), cerr);
		    flush(cerr);
		}

#endif

		// Set next_nodes[i] = NULL for any center of mass in the
		// clump that has been adjusted, so we can use the rest of
		// the array in evolve_system on return from integrate_list.

 		if (cm_list && n_list > 0) {
 		    for (int j = 0; j < n_next; j++) {
 			if (!next_nodes[j]->is_valid())
 			    next_nodes[j] = NULL;
 			else
 			    for (int k = 0; k < n_list; k++)
 				if (next_nodes[j] == cm_list[k]) {
 				    next_nodes[j] = NULL;
 				    break;
 				}
 		    }
 		}
		if (cm_list) delete [] cm_list;

		return steps;		// NOTE: we currently return after
					//	 the FIRST tree rearrangement,
					// 	 so we can only have one
					//	 restructuring per block time
					//	 step.
	    }
	    if (cm_list) delete [] cm_list;
	}
    }

    return steps;
}



local void full_reinitialize(hdyn* b, real t, real eps, real eta, bool verbose)
{
    if (verbose)
	cerr << "\nReinitializing system at time " << t << endl;

    real cpu_0 = cpu_time();

    b->set_system_time(t);
    // b->to_com();
    initialize_system_phase1(b, t);
    initialize_system_phase2(b, eps, eta,
			     initial_step_limit,
			     false);			// "false" here means
							// timesteps are only
							// set if zero

    // Quietly reinitialize all kepler structures, using the same
    // code as is used in get_hdyn() (hdyn.h).  (This repeats actions
    // taken by get_hdyn() on startup.)

    for_all_nodes(hdyn, b, bb)
	if (bb->get_unperturbed_timestep() > 0
	    && bb->get_elder_sister() == NULL)
	    bb->reinitialize_kepler_from_hdyn();

    if (verbose) {
	cerr << "CPU time for reinitialization = "
	     << cpu_time() - cpu_0 << endl;
    }
}

local bool check_sync(hdyn* b)
{
    bool need_new_list = false;

    for_all_nodes(hdyn, b, bb)
	if (bb->get_time() != b->get_system_time()
	     && bb->get_kepler() == NULL) {
	  cerr << "check_sync warning:  node "
	       << bb->format_label() << " not synchronized at time "
	       << b->get_system_time() << "; node time = " << bb->get_time()
	       << endl;
	  need_new_list = true;
	}

    return need_new_list;
}

local void backward_step_exit(hdyn* b, real ttmp, real t,
			      hdyn** next_nodes, int n_next)
{

    cerr << "evolve_system: time went backwards!\n";

    cerr.precision(12);
    PRC(ttmp); PRC(b->get_system_time()); PRL(t);

    for (int i = 0; i < n_next; i++) {
	cerr << endl;
	PRC(i); next_nodes[i]->pretty_print_node(cerr);
	cerr << " "<< next_nodes[i]->get_time()<< " ";
	cerr << next_nodes[i]->get_next_time()<<endl;
	pp3(next_nodes[i]->get_top_level_node(), cerr);
    }
    put_node(cerr, *b);
    exit(0);
}

local void check_binary_scheduling(hdyn* b,
				   hdyn** next_nodes,
				   int n_next)
{
    // Force the elder of two binary sisters to be integrated.

    for (int i = 0; i < n_next; i++) {
	hdyn *bi = next_nodes[i];

	if (bi->is_low_level_node()) {

	    hdyn *od = bi->get_parent()
	                 ->get_oldest_daughter();

	    if (od->get_timestep() != bi->get_timestep()) {
		// cerr << "timesteps of sisters are different !! \n";
		od->set_timestep(bi->get_timestep());
	    }

	    if (od->get_time() != bi->get_time()) {
		od->set_time(bi->get_time());
	    }

	    next_nodes[i] = bi->get_parent()->get_oldest_daughter();
	}
    }
}

local inline void update_step(real t, real& ts, real dts)
{
    while (ts < t)
	ts += dts;
}



// evolve_system:  Main integration loop.

local void evolve_system(hdyn * b,	       // hdyn array                   
			 real delta_t,	       // time span of the integration 
			 real dt_log,	       // time step of the integration 
			 real dt_snap,	       // snapshot output interval
			 real dt_sstar,	       // timestep for single
                                               //     stellar evolution
			 real dt_esc,	       // escaper removal
			 real dt_reinit,       // reinitialization interval
			 real eta,	       // accuracy parameter 
			 real eps,	       // softening length  
			 real d_min_sq,	       // trigger factor squared
			 real lag_factor,      // hysteresis factor squared
			 real gamma2,	       // perturbation factor squared
			 bool exact,	       // exact force calculation
			 real scaled_stripping_radius,
					       // stripping radius with
					       // initial mass factor included
			 bool verbose,
			 bool save_last_snap,  // save snap at log output
			 char* snap_save_file, // filename to save in
			 int n_stop )	       // when to stop
{
    // Initialization:

    clean_up_files();
    cpu_init();
    initialize_counters_from_log(b);

    real steps = 0;
    real count = 0;
    int snaps = 0;

    hdyn **next_nodes = new hdyn *[2 * b->n_daughters()];	// Overkill?
    set_n_top_level(b);

    // Global data!

    b->set_d_min_sq(d_min_sq);
    
    b->set_gamma2(gamma2);
    b->set_gamma23(pow(gamma2, -1.0/3.0));
    b->set_mbar(b->get_mass()/b->n_leaves());	// mass scale
    
    // Establish limits on time steps.

    real dt_max = min(dt_log, dt_sstar);
    dt_max = min(dt_max, dt_reinit);

    initial_step_limit = min(initial_step_limit, dt_max);
    initial_step_limit = min(initial_step_limit, dt_snap);

    step_limit = min(dt_max, dt_snap);  // no initial_step_limit, note...

    // The scheduling is such that the system will be synchronized
    // at *every* integer multiple of step_limit.

    // The behavior of the system is governed by the following times:

    real dt_sync = step_limit;

    // Check and flag some basic relationships between time scales.

    if (dt_reinit < dt_sync)
	err_exit("dt_reinit < dt_sync.");	// Require dt_reinit >= dt_sync
    if (dt_esc < dt_sync)
	err_exit("dt_esc < dt_sync.");		// Require dt_esc >= dt_sync
    if (dt_sstar < dt_sync)
	err_exit("dt_sstar < dt_sync.");	// Require dt_sstar >= dt_sync

    if (verbose) {
	if (dt_log < dt_sync)
	    cerr << "Warning: dt_log < dt_sync, quoted errors unpredictable.\n";
	if (dt_snap < dt_sync)
	    cerr << "Warning: dt_snap < dt_sync, no restart possible.\n";
	else if (dt_snap < dt_reinit)
	    cerr << "Warning: dt_snap < dt_reinit, restart unpredictable.\n";
    }

    real t = b->get_time();		// current time
    real t_end = t + delta_t;		// final time, at end of integration
    real t_log = t;			// time of next log output
//    if (t > 0) t_log += dt_log;
    real t_snap = t + dt_snap;		// time of next snapshot output
    real t_sync = t + dt_sync;		// time system will next be synchronized
    real t_esc = t + dt_esc;		// time of next escaper check
					// Note: do NOT apply immediate check
    real t_reinit = t + dt_reinit;	// time of next reinitialization

    if (verbose) {
	cerr << endl;
	PRC(t), PRL(t_end);
	PRL(dt_sync);
	PRL(dt_log);
	PRL(dt_snap);
	PRL(dt_esc);
	PRL(dt_reinit);
    }

    // Initialize the system.

    full_reinitialize(b, t, eps, eta, verbose);

    bool tree_changed = true;	// Used by fast_get_nodes_to_move.
    				// Set by integration/evolution routines.

    // Note: the system can be reinitialized in several places, resulting
    // in considerable duplication of effort and inefficiency.  It remains
    // to be seen whether this produces any noticeable effect on run time.

    while (t <= t_end) {

	int n_next;
	real ttmp;

	// Create the new time step list.

	// get_nodes_to_move(b, next_nodes, n_next, ttmp);
	fast_get_nodes_to_move(b, next_nodes, n_next, ttmp, tree_changed);

	if (ttmp > t_reinit) {

	    // Time to reinitialize the system.

	    // *** REQUIRE dt_reinit >= dt_sync. ***

	    full_reinitialize(b, t, eps, eta, verbose);
	    tree_changed = true;

	    fast_get_nodes_to_move(b, next_nodes, n_next, ttmp, tree_changed);
	    update_step(ttmp, t_reinit, dt_reinit);
	}

	bool last_snap = false;

	if (ttmp > t_log) {

	    // Standard log output.

	    // System should be synchronized here, but it is not essential.
	    // However, energy errors may not be reliable, and the system
	    // may not be restartable (if save_last_snap is set).

	    real cpu0 = cpu_time();
	    log_output(b, steps, eps, &kc_prev);

	    last_snap = save_last_snap;

	    cerr << endl << "Total CPU time for log output = "
		 << cpu_time() - cpu0 << endl;

	    cerr << endl; flush(cerr);
	    update_step(ttmp, t_log, dt_log);

	}

	if (ttmp > t_esc) {

	    // Check for and remove escapers.

	    // System is reinitialized and time step list is
	    // recomputed if any stars are removed.

	    // *** REQUIRE dt_esc >= dt_sync. ***

	    int n_prev = b->n_leaves();

	    if (scaled_stripping_radius > 0)
		check_and_remove_escapers(b, scaled_stripping_radius,
					  eps, eta, dt_max,
					  initial_step_limit,
					  ttmp, next_nodes,
					  n_next, tree_changed);

	    int n_new = b->n_leaves();

	    if (n_new <= n_stop) {
		cerr << "N_stop reached\n";
		exit(0);
	    }

	    if (n_new != n_prev) {

		// Should be somewhat redundant, but...

		full_reinitialize(b, t, eps, eta, verbose);
		tree_changed = true;
		fast_get_nodes_to_move(b, next_nodes, n_next, ttmp,
				       tree_changed);

		// log_output(b, steps, eps);

		cerr << endl; flush(cerr);
	    }

	    update_step(ttmp, t_esc, dt_esc);
	}

	if (ttmp > t_sync) {

	    // System is synchronized.
	    // Check to see if we should stop the run.

	    if (check_file("STOP")) {

		t_end = ttmp - 1;	// Forces optional snap output 
					// and end of run in snap_output.

		cerr << "\n\n***** Calculation STOPped by user *****\n\n";
	    }

	    tree_changed |= check_sync(b);
	    update_step(ttmp, t_sync, dt_sync);
	}

	// Output a snapshot to cout at the scheduled time, or at end of run.
	// NOTE: This is the *only* place where output to cout can occur.

	// Not essential to have dt_snap >= dt_reinit, but should have this
	// if we want full restart capability.

	bool reg_snap = (ttmp > t_snap
			 || (dt_snap < VERY_LARGE_NUMBER && ttmp > t_end)
			 || (fmod(steps, 1000) == 0 && check_file("DUMP")));

	if (reg_snap || last_snap) {

	    // log_output(b, steps, eps);

	    snap_output(b, steps, snaps,
			reg_snap, last_snap, snap_save_file,
			t, ttmp, t_end, t_snap, dt_snap, verbose);
	    if (reg_snap)
		update_step(ttmp, t_snap, dt_snap);
	}

	// Proceed to the next step.

	if (ttmp < t)
	    backward_step_exit(b, ttmp, t, next_nodes, n_next);
	    
	// Force the elder of two binary sisters to be integrated.

	check_binary_scheduling(b,next_nodes, n_next);

	b->set_system_time(t = ttmp);
	b->set_time(t);

	// Take a new step to time t:

#ifndef STARLAB_HAS_HARP
	
	// We need full prediction if any top-level nodes appear on
	// the integration list.  Otherwise, we really only need to 
	// predict the perturbers (done in perturber_acc_and_jerk...).
	// However, if a perturber list is invalid or overflows, then
	// we need full prediction again...

	bool predict_all = false;
	for (int i = 0; i < n_next; i++) {
	    hdyn* n = next_nodes[i];
	    if (n->is_top_level_node()
		|| !n->get_top_level_node()->get_valid_perturbers()) {
		predict_all = true;
		break;
	    }
	}

	if (predict_all)
	    predict_loworder_all(b, t);	

	// HARP does (top-level) prediction, if present.

#endif

	steps += integrate_list(b, eps, eta,
				lag_factor, dt_max,
				next_nodes, n_next,
				exact, tree_changed);
	count += 1;

#if 0

	if (t > 2.67) {
	    bool print_energy = false;
	    for (int j = 0; j < n_next; j++) {
		hdyn* n = next_nodes[j];
		if (n && n->is_valid()
		    && n->is_top_level_node())
		    print_energy = true;
	    }
	    if (print_energy)
		print_recalculated_energies(b, eps * eps);
	}


	if (t > 1.3 && t < 1.375) {
	    for (int j = 0; j < n_next; j++) {
		hdyn* n = next_nodes[j];
		if (n && n->is_valid()
		    && n->name_is("(1376b,1376a)")) {
		    cerr << "node " << n->format_label() << "  ";
		    PRL(t);
		    dump_story(n->get_dyn_story());
		}
	    }
	}



	if (t > 566.246119)
	    pp3((hdyn*)node_with_name("(1840,2812)", b->get_root()));

	if (fmod(count, 100) == 0) {

	    // Primitive checksum...

	    real checkx = 0, checkv = 0, checka = 0, checkj = 0;
	    for_all_nodes(hdyn, b, bb) {
		checkx += abs(bb->get_pos());
		checkv += abs(bb->get_vel());
		checka += abs(bb->get_acc());
		checkj += abs(bb->get_jerk());
	    }
	    cerr.precision(17);
	    cerr << endl; PRL(t);
	    PRC(checkx); PRL(checkv);
	    PRC(checka); PRL(checkj);
	    cerr.precision(PRECISION);
	}

	if (t > 566.246127) exit(0);
		

//	if (b->get_system_time() > 566.8)
	    for (int j = 0; j < n_next; j++) {
		hdyn* n = next_nodes[j];
		if (n && n->is_valid()
		    && n->is_low_level_node()
		    && fmod(n->get_steps(), 100.0) == 0
		    && n->name_is("1139")
		    && n->get_perturbation_squared() < 0.0001) {

		    pp3(n);

		    real dist = abs(n->get_pos()
				    - n->get_younger_sister()->get_pos());
		    real dtff = sqrt(dist*dist*dist
				     / (2*n->get_parent()->get_mass()));
		    real dtv  = sqrt(square(n->get_pos())
				     / square(n->get_vel()));

		    real kepstep = eta*min(dtff, dtv);
		    real ratio = n->get_timestep()/kepstep;
		    PRC(kepstep); PRL(ratio);
		}
	    }
#endif

	// Optionally check the heartbeat...

	if (CHECK_PROGRESS && fmod(count, n_check) == 0) {
	    PRC(count), PRC(t), PRL(cpu_time());
	    if (fmod(count, 4*n_check) == 0)
		print_recalculated_energies(b, eps*eps);
	}

	// (Removed numerous debugging examples to kira_debug.C -- Steve 8/98)

	// Note: if a node is destroyed as part of the step,
	// the corresponding entry in next_nodes returns NULL.

	// Integrate_list handles tree changes -- allows one per step.

	// Evolve the stellar system (stars and binaries).
	// Entire system will be reinitialized if necessary.

	// This may cause undesirable unperturbed binary updates...
	// *** To be fixed...
	
	tree_changed |= evolve_stars(b, eps, eta,
				     initial_step_limit,
				     dt_log, dt_sstar);

	if (tree_changed) set_n_top_level(b);

	// Miscellaneous checks (see kira_runtime.C):

	if (fmod(count, 2500) == 0)
	    check_runtime(b, eps, eta, lag_factor, t_end, tree_changed);
    }
}



main(int argc, char **argv)
{
    check_help();

    // Parameters to be passed directly to kira:

    hdyn *b;			// hdyn root node

    // Time scales -- all should be powers of 2.

    real delta_t;		// time span of the integration
    real dt_log;		// output interval
    real dt_snap;		// snap output interval
    real dt_sstar;		// standard single-star evolution timestep
    real dt_esc;		// timestep to remove escapers
    real dt_reinit;		// timestep to reinitialize entire system

    real eta;			// time step of the integration
    real eps;			// softening length               

    real d_min_sq;		// threshhold for node combine (scale: 1/N^2)
    real lag_factor;		// hysteresis factor for node removal
    real gamma2;		// max perturbation squared for unperturbed
    bool exact;			// force calculation using perturber list

    bool verbose;		// Toggle "verbose" mode

    real scaled_stripping_radius; // Stripping radius for unit mass

    bool save_last_snap = false;
    char snap_save_file[256];
    int  n_stop;		// n to terminate simulation

    if (!kira_initialize(argc, argv,
			 b, delta_t, dt_log, dt_snap, dt_sstar,
			 dt_esc, dt_reinit,
			 eta, eps, d_min_sq, lag_factor, gamma2, exact,
			 scaled_stripping_radius, verbose,
			 save_last_snap, snap_save_file, n_stop))
	get_help();

    // b->to_com();

    evolve_system(b, delta_t, dt_log, dt_snap, dt_sstar, dt_esc, dt_reinit,
		  eta, eps, d_min_sq, lag_factor, gamma2, exact,
		  scaled_stripping_radius, verbose,
		  save_last_snap, snap_save_file, n_stop);
}

#endif
