
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

// Read in command-line parameters and initialize the system.

#include <star/dstar_to_kira.h>

#define		DEFAULT_ETA		0.1
#define		DEFAULT_GAMMA2		1.e-14

#define		DEFAULT_DT		10
#define		DEFAULT_DT_REINIT	1
#define		DEFAULT_DT_LOG		1
#define		DEFAULT_DT_SSTAR	0.015625

#define		SEED_STRING_LENGTH	60

real total_mass(hdyn * b)
{
    real total_mass_nodes = 0;
    for_all_daughters(hdyn, b, bi)
	total_mass_nodes += bi->get_mass();
    return total_mass_nodes;
}

local void initialize_index(node * b, bool verbose)
{
    // Add indices to otherwise unidentified nodes.

    int index = 0;

    for_all_nodes(node, b, bb)
	if (bb != b)
	    if (bb->get_index() > 0)
		index = max(index, bb->get_index());

    // Start numbering at some well-separated value.

    index = 1000 * (index / 1000 + 1);

    for_all_nodes(node, b, bb)
	if (bb != b && bb->get_index() < 0 && bb->get_name() == NULL) {
	    if (verbose) {
		cerr << "Attached index " << index << " to unnamed ";
		if (bb->is_parent())
		    cerr << "node\n";
		else
		    cerr << "leaf\n";
	    }
	    bb->set_index(index++);
	}
}

local void correct_multiples(hdyn* b,
			     bool verbose)
{
    // Check and fix input problems associated with hierarchical systems.
    // Probably not the best place to do this, but keep it here for now...

    for_all_nodes(hdyn, b, bb) {
	if (bb->get_elder_sister() == NULL
	    && bb->get_kepler() != NULL) {
	    if (bb->is_parent()) {
		if (verbose) {
		    cerr << "unperturbed " << bb->format_label() << endl;
		    cerr << "    multiple system, ";
		}
		if (bb->get_oldest_daughter()->get_kepler() == NULL) {
		    if (verbose)
			cerr << "daughter node "
			     << bb->get_oldest_daughter()->format_label()
			     << " not unperturbed.  Correcting...\n";
		    bb->get_oldest_daughter()->
			reinitialize_kepler_from_hdyn();
		} else
		    if (verbose)
			cerr << "daughter node uperturbed.\n";
	    }
	}
    }
}

local real get_virial_radius(hdyn* b,
			     bool verbose,
			     bool r_virial_set, real input_r_virial)
{
    real r_virial = getrq(b->get_log_story(), "initial_rvirial");

    if (r_virial > 0) {

	// The input snapshot has the information needed.  Snapshot
	// data OVERRIDES any command-line entry.

	if (verbose)
	    cerr << "Read initial r_virial = " << r_virial
		 << " from input snapshot." << endl;

    } else {

	// See if r_virial was set from the command-line, or use default.

	if (r_virial_set) {
	    if (verbose)
		cerr << "Setting initial r_virial = " << r_virial
		     << " from command-line input." << endl;
	    r_virial = input_r_virial;
	} else {
	    r_virial = 1;
	    if (verbose)
		cerr << "Adopting default initial r_virial = "
		     << r_virial << endl;
	}

	// Write the virial radius to the log story if t = 0.

	if (b->get_system_time() <= 0)
	    putrq(b->get_log_story(), "initial_rvirial", r_virial);
    }

    return r_virial;
}

local real get_initial_mass(hdyn* b, bool verbose)
{
    real initial_mass = getrq(b->get_log_story(), "initial_mass");

    if (initial_mass <= 0) {

	// Note: presently no option to set initial_mass from the command line.

	if (b->get_system_time() <= 0)
	    initial_mass = total_mass(b);
	else {
	    initial_mass = 1;
	    if (verbose)
		cerr << "Adopting default initial_mass = "
		     << initial_mass << endl;
	}

	putrq(b->get_log_story(), "initial_mass", initial_mass);
    }

    return initial_mass;
}

local real get_jacobi_radius(hdyn* b,
			     bool verbose,
			     real r_virial,
			     bool r_jacobi_set, real input_r_jacobi)
{
    real r_jacobi = -1;

    // Determine the initial Jacobi radius of the system.  This radius
    // may be specified on the command line, or it may be derived from
    // the information in the input snapshot.  Unlike r_virial, the
    // command-line takes precedence over initial data found in the
    // input snapshot, although it is superceded by any "kira" data
    // found in that file.  Scaling is such that, if a non-kira
    // version of r_jacobi is found in the input snapshot, the
    // command-line value will scale that number.  A kira version is
    // used as is.  Otherwise, the command-line value scales the virial
    // radius (hence the two tests of r_jacobi_set below).

    // See if we can determine the Jacobi radius from the input file.

    if (find_qmatch(b->get_log_story(), "initial_rtidal_over_rvirial")) {

	// The input snapshot contains an initial Jacobi radius.  Most
	// likely the initial model is a King profile and this is the
	// King radius (misnamed for historical reasons -- it may or
	// may not have anything to do with a tidal cutoff).

	real r_jacobi_over_r_virial = getrq(b->get_log_story(),
					    "initial_rtidal_over_rvirial");

	if (r_jacobi_over_r_virial > 0)

	    r_jacobi = r_jacobi_over_r_virial * r_virial;

	else {

	    if (verbose)
		cerr << "Warning: error reading "
		     << "initial_rtidal_over_rvirial from input snapshot."
		     << endl;
	    else
		err_exit(
	      "Error reading initial_rtidal_over_rvirial from input snapshot.");
	}
    }

    if (r_jacobi_set) {

	if (r_jacobi > 0) {

	    // The Jacobi radius has been specified in the input file and
	    // on the command line.  Use the command line as a scaling for
	    // the "known" value.

	    if (verbose)
		cerr << endl << "Command-line (\"-J\") Jacobi radius "
		     << input_r_jacobi << " scales initial value "
		     << r_jacobi << " from input snapshot."
		     << endl;

	    r_jacobi *= input_r_jacobi;

	} else {

	    // Use the command-line data to scale the virial radius.

	    if (verbose)
		cerr << endl << "Command-line (\"-J\") Jacobi radius "
		     << input_r_jacobi << " scales initial virial radius "
		     << r_virial << endl;

	    r_jacobi = input_r_jacobi * r_virial;

	}
    }

    // See if a kira_initial_jacobi_radius exists in the input file.
    // If it does, it TAKES PRECEDENCE over any other setting.  (Do
    // things in this order to maximize information output to user.)

    if (find_qmatch(b->get_log_story(), "kira_initial_jacobi_radius")) {

	real kira_initial_jacobi_radius
	    = getrq(b->get_log_story(), "kira_initial_jacobi_radius");

	if (kira_initial_jacobi_radius > 0) {

	    r_jacobi = kira_initial_jacobi_radius;

	    if (verbose) {
		cerr << "Using initial Jacobi radius ";
		PRL(kira_initial_jacobi_radius);
		cerr << " taken from input snapshot." << endl;
	    }

	} else {

	    if (r_jacobi > 0)
		if (verbose)
		    cerr << "Warning: error reading "
			 << "kira_initial_jacobi_radius"
			 << " from input snapshot."
			 << endl;
		else
		    err_exit(
	      "Error reading kira_initial_jacobi_radius from input snapshot.");
	}
    }

    return r_jacobi;
}

local char* stredit(char* s, char c1, char c2)
{
    if (!s) return NULL;

    char* s1 = new char[strlen(s)+1];
    if (!s1) return NULL;

    strcpy(s1, s);

    char* ss = s1;
    while (*ss != '\0') {
	if (*ss == c1) *ss = c2;
	ss++;
    }

    return s1;
}

local void kira_system_id(char* name)
{
    cerr << "Starlab version " << STARLAB_VERSION;

    char* s = stredit(_COMPILE_DATE_, '_', ' ');
    if (s) {
	cerr << "; kira created on " << s;
	delete s;
    }
    cerr << endl;

    cerr << "Command line reference: " << name << endl;

    if (   (name[0] == '.' && name[1] == '/')	// easier to list
	|| (name[0] == '~' && name[1] == '/')	// excluded strings...
	||  name[0] == '/') {
    } else {

	cerr << "System `which " << name << "` = ";

	char tmp[256];
	strcpy(tmp, "which ");
	strcat(tmp, name);
	system(tmp);		// Note: already adds newline...

    }

    cerr << "Current directory = " << getenv("PWD") << endl;

    cerr << endl;
}

bool kira_initialize(int argc, char** argv,
		     hdynptr& b,	// hdyn root node
		     real& delta_t,	// time span of the integration
		     real& dt_log,	// output interval--power of 2
		     real& dt_snap,	// snap output interval
		     real& dt_sstar,	// stellar evolution timestep
		     real& dt_esc,	// escaper removal
		     real& dt_reinit,	// reinitialization interval
		     real& eta,		// time step parameter
		     real& eps,		// softening length
		     real& d_min_sq,	// N^2 * node combine threshhold
		     real& lag_factor,	// node hysteresis factor
		     real& gamma2,	// unperturbed threshhold
		     bool& exact,	// no perturber list if true
		     real& scaled_stripping_radius, // stripping radius
						    // for unit mass
		     bool& verbose,
		     bool& save_last_snap,
		     char* snap_save_file,
		     int& n_stop)	// n to terminate simulation
{

    // Establish defaults for parameters carried back explicitly:

    delta_t = DEFAULT_DT;
    dt_log = DEFAULT_DT_LOG;

    dt_snap = VERY_LARGE_NUMBER;
    dt_sstar = VERY_LARGE_NUMBER;
    dt_esc = VERY_LARGE_NUMBER;
    dt_reinit = DEFAULT_DT_REINIT;

    eta = DEFAULT_ETA;
    eps = 0.0;
    d_min_sq = 1.0;
    lag_factor = 2.5;
    gamma2 = DEFAULT_GAMMA2;
    exact = false;
    verbose = true;

    save_last_snap = false;
    n_stop = 10;

    // Defaults for ancillary and indirect parameters:

    real initial_stripping_radius = 0;
    char *comment;		// comment string

    int  input_seed, actual_seed;

    // For use by stellar evolution (establish units and conversions):

    real m_bar_true = 1;	// actual mean stellar mass, in solar masses
    real m_tot_true = 1;	// actual total mass ,in solar masses
    real r_vir  = 1;            // radius scaling (virial radius in parsecs)
    real t_vir = 1;             // time scaling (system time unit, in Myr)
    real q_vir = 0.5;           // virial ratio (instead of time scaling)
    real T_start = 0;

    real smc, sec;
    bool smc_flag = false;
    bool sec_flag = false;
    
    bool B_flag = false;
    bool D_flag = false;
    bool G_flag = false;
    bool m_flag = false;
    bool M_flag = false;
    bool Q_flag = false;
    bool R_flag = false;
    bool S_flag = false;
    bool T_flag = false;
    bool c_flag = false;
    bool s_flag = false;

    bool r_virial_set = false;
    real input_r_virial;

    bool r_jacobi_set = false;
    real input_r_jacobi;

    bool use_relax = false;

    char seedlog[SEED_STRING_LENGTH];

    extern char *poptarg;
    int c;
    char* param_string =
	"a:Bc:C:d:D:e:E:f:g:G:h:I:J:k:m:M:n:N:O:q:QrR:s:St:T:uUvV:X:z:Z:";

    kira_system_id(argv[0]);

    while ((c = pgetopt(argc, argv, param_string)) != -1) {
	switch (c) {
	    case 'a':	eta = atof(poptarg);
			break;
	    case 'B':   S_flag = true;	    // Does B_flag = true make sense
	    		B_flag = true;	    // if S_flag = false?
		        break;
	    case 'c':	c_flag = TRUE;
			comment = poptarg;
			break;
	    case 'C':	kira_set_harp_cpu(atof(poptarg));
			break;
	    case 'd':	dt_log = atof(poptarg);
	    		if (dt_log < 0) dt_log = pow(2.0, dt_log);
			break;
	    case 'D':	D_flag = TRUE;
			dt_snap = atof(poptarg);
	    		if (dt_snap < 0) dt_snap = pow(2.0, dt_snap);
			break;
	    case 'e':	eps = atof(poptarg);
			break;
	    case 'E':	if (atoi(poptarg)) exact = true;
			break;
	    case 'f':	d_min_sq = atof(poptarg);
			d_min_sq *= d_min_sq;
			break;
	    case 'g':	lag_factor = atof(poptarg);
			break;
	    case 'G':	G_flag = true;
			initial_stripping_radius = atof(poptarg);
			break;
	    case 'h':	dt_sstar = atof(poptarg);
	    		if (dt_sstar < 0) dt_sstar = pow(2.0, dt_sstar);
			break;
	    case 'I':	dt_reinit = atof(poptarg);
	    		if (dt_reinit < 0) dt_reinit = pow(2.0, dt_reinit);
			break;
	    case 'J':	r_jacobi_set = true;
			input_r_jacobi = atof(poptarg);
			break;
	    case 'k':	gamma2 = atof(poptarg);
			gamma2 *= gamma2;
			break;
	    case 'm':	m_flag = true;
			m_bar_true = atof(poptarg);
			break;
	    case 'M':	M_flag = true;
			m_tot_true = atof(poptarg);
			break;
	    case 'n':	n_stop = atoi(poptarg);
			break;
	    case 'N':	set_n_check(atoi(poptarg));
			break;
	    case 'O':	save_last_snap = true;
	    		strcpy(snap_save_file, poptarg);
	    		break;
            case 'q':	q_vir = atof(poptarg);
			break;
            case 'Q':	Q_flag = true;
	       		break;
            case 'r':	use_relax = true;
	       		break;
	    case 'R':	R_flag = true;
			r_vir = atof(poptarg);
			break;
            case 's':	s_flag = true;
			input_seed = atoi(poptarg);
			break;
            case 'S':	S_flag = true;
	       		break;
	    case 't':	delta_t = atof(poptarg);
			break;
            case 'T':	T_flag = true;
			t_vir = atof(poptarg);
			break;
	    case 'u':	toggle_unperturbed(1);
			break;
	    case 'U':	toggle_unperturbed(0);
			break;
	    case 'v':	verbose = !verbose;
			break;
	    case 'V':	r_virial_set = true;
			input_r_virial = atof(poptarg);
			break;
	    case 'X':	dt_esc = atof(poptarg);
	    		if (dt_esc < 0) dt_esc = pow(2.0, dt_esc);
			break;
	    case 'z':   smc_flag = true;
	                smc = atof(poptarg);
			break;
	    case 'Z':   sec_flag = true;
	                sec = atof(poptarg);
			break;
	    default:
	    case '?':	params_to_usage(cerr, argv[0], param_string);
			return false;
	}
    }

    if ((S_flag || B_flag) && dt_sstar == VERY_LARGE_NUMBER)
	dt_sstar = DEFAULT_DT_SSTAR;

    b = get_hdyn(cin);
    if (b == NULL) err_exit("Can't read input snapshot");

    b->set_kira_counters(new kira_counters);
    
    correct_multiples(b, verbose);

    b->set_use_sstar(S_flag);
    b->set_use_dstar(B_flag && S_flag);
    b->set_stellar_merger_criterion_sq(pow(smc, 2));
    b->set_stellar_encounter_criterion_sq(pow(sec, 2));
    
    if (c_flag)
	b->log_comment(comment);

    b->log_history(argc, argv);

    if (S_flag || B_flag) {
        if (s_flag == FALSE)
	    input_seed = 0;                         // default
        actual_seed = srandinter(input_seed);
        sprintf(seedlog,
		"         random number generator seed = %d",actual_seed);
        b->log_comment(seedlog);
	cerr << "Initial random seed = " << actual_seed << endl;
    }

    initialize_index(b, verbose);

    // Establish defaults in time scales:

    if (!D_flag) dt_snap = delta_t;
    if (dt_snap == 0) dt_snap = VERY_LARGE_NUMBER;

    if (dt_reinit <= 0 || dt_reinit == VERY_LARGE_NUMBER) dt_reinit = dt_log;
    if (dt_esc == VERY_LARGE_NUMBER) dt_esc = dt_reinit;

    int nbody = b->n_leaves();
    if (nbody < 0) err_exit("kira: N <= 0!");

    // Divide by nbody twice separately since int*int may overflow
    d_min_sq /= nbody;
    d_min_sq /= nbody;
    
    lag_factor *= lag_factor;

    if (eps <= 0) {
	if (verbose)
	    cerr << "Using unsoftened gravitational potential." << endl;
	eps = 0;
    } else {
	if (verbose)
	    cerr << "Using softened gravitational potential, eps = "
		 << eps << endl;
    }

    if (verbose) print_unperturbed_options();

    //----------------------------------------------------------------------

    // Determine the initial virial radius and initial mass (both in code
    // units), for use in scaling.

    // Note: the parameters set here must give correct results both on
    //       initialization AND on restart...

    if (verbose) cerr << endl;

    real r_virial = get_virial_radius(b, verbose, r_virial_set, input_r_virial);
    real initial_mass = get_initial_mass(b, verbose);

    //----------------------------------------------------------------------

    // Tidal field (turned on with "-Q" on the command line).

    b->set_omega(0.0);			// parameter determining tidal field
    					// (point-mass Galactic field for now)
    real r_jacobi = -1;

    if (Q_flag) {

	// Using a tidal field probably should imply stripping, but it
	// doesn't have to.  Output a warning in that case.

	if (verbose) {
	    cerr << endl;
	    if (!G_flag)
		cerr << "Warning: tidal field used without stripping."
		     << endl;
	}

	r_jacobi = get_jacobi_radius(b, verbose, r_virial,
				     r_jacobi_set, input_r_jacobi);

	if (r_jacobi <= 0)

	    err_exit("Tidal field enabled but Jacobi radius unknown.");

	else {

	    if (verbose)
		cerr << "Self-consistent tidal field; "
		     << "initial Jacobi radius = " << r_jacobi
		     << endl;

	    // Save the Jacobi radius for future use.

	    putrq(b->get_log_story(), "kira_initial_jacobi_radius", r_jacobi);

	    // Set up the dynamical tidal parameters.

	    // Determine the (constant) parameter omega corresponding to
	    // the angular velocity about the Galactic center (see Binney
	    // & Tremaine, p. 452).

	    b->set_omega(sqrt((initial_mass/3.0) / pow(r_jacobi, 3.0)));

	    if (verbose)
		cerr << "Point mass field, omega = " << b->get_omega()
		     << endl;
	}
    }

    //----------------------------------------------------------------------

    // Kira has the options of stripping stars beyond some specified
    // radius (referred to as the "stripping" radius in the code).  The
    // stripping radius r is set with "-G r".  Implementation of the
    // stripping is INDEPENDENT of the treatment of a tidal field.
    // However, the interpretation of r depends on what tidal information
    // is known.
    //
    // 1. If "-Q" has not been specified, r specifies the stripping radius
    // in units if the virial radius.
    //
    // 2. If "-Q" has been specified, r specifies the stripping radius
    // in units if the jacobi radius.
    //
    // Command-line settings ARE OVERRIDDEN if kira_stripping_radius is
    // specified in the input file.

    scaled_stripping_radius = 0;	// stripping radius for unit mass
    					// (0 ==> no stripping)

    if (G_flag) {

	if (verbose) cerr << endl;

	if (initial_stripping_radius > 0) {

	    if (r_jacobi > 0) {

		if (verbose)
		    cerr << "Command-line -G " << initial_stripping_radius
			 << " used as scaling factor for jacobi radius.\n";

		initial_stripping_radius *= r_jacobi;

	    } else {

		if (verbose)
		    cerr << "Command-line -G " << initial_stripping_radius
			 << " used as scaling factor for virial radius.\n";

		initial_stripping_radius *= r_virial;

	    }

	    PRL(initial_stripping_radius);
	    scaled_stripping_radius = initial_stripping_radius
					/ pow(initial_mass, 1.0/3.0);
	}

	if (find_qmatch(b->get_log_story(), "kira_scaled_stripping_radius")) {

	    real kira_scaled_stripping_radius
		= getrq(b->get_log_story(), "kira_scaled_stripping_radius");

	    if (kira_scaled_stripping_radius > 0) {

		scaled_stripping_radius = kira_scaled_stripping_radius;

		if (initial_stripping_radius > 0)
		    cerr << "Command-line stripping radius superceded by "
			 << "value found in input file." << endl;
	    }
	}

	if (scaled_stripping_radius <= 0) 
	    err_exit("Unable to determine stripping radius.");

	PRL(scaled_stripping_radius);
	cerr << "current stripping radius = "
	     << scaled_stripping_radius * pow(total_mass(b), 1.0/3.0);

	putrq(b->get_log_story(), "kira_scaled_stripping_radius",
	      scaled_stripping_radius);

    }

    //----------------------------------------------------------------------

    // Scaling factors for stellar evolution.

    // On restart, pick up scalings from the root star story (if any),
    // impose new scalings (if any), and initialize star parts.

    if (S_flag) {	// Note that B_flag ==> S_flag (Steve 9/19/97)

	if (verbose)
	    cerr << endl;
	if (b->get_starbase() == NULL) err_exit("No starbase!");

	// See if any scaling information already exists.

	bool scales_from_snapshot
	    = b->get_starbase()->get_stellar_evolution_scaling();

	if (scales_from_snapshot) {
	    if (R_flag || M_flag || T_flag) {
		if (verbose)
		    cerr << "Ignoring scale factors in input snapshot.\n";
		scales_from_snapshot = false;
	    } else {
		if (verbose) {
		    cerr << "Scale factors taken from input snapshot.\n";
		    b->get_starbase()->print_stellar_evolution_scaling(cerr);
		}
	    }
	}

	if (!scales_from_snapshot) {

	    // We will have to set up the scaling.  Check to make
	    // sure that the necessary information exists.

	    if (!M_flag && !m_flag) err_exit("Masses not specified.");
	    if (!R_flag) err_exit("Physical virial radius not specified.");
	    if (r_virial < 0) err_exit("System virial radius not known.");

	    if (!M_flag) m_tot_true = m_bar_true * nbody;
	    if (!m_flag) m_bar_true = m_tot_true / nbody;

	    // The quantity r_vir is the virial radius of the system, in pc.
	    // The quantity r_virial is the virial radius in N-body units.
	    // Use "-R r" to specify the scale on the command line.
	    //
	    // Convention:  if there is no tidal field, r is the virial radius.
	    //	  	    if there is  a tidal field, r is the Jacobi radius.

	    if (r_jacobi > 0) {

		if (verbose) cerr << "Input -R " << r_vir
				  << " interpreted as Jacobi radius.\n";
		r_vir /= (r_jacobi/r_virial);

	    } else

		if (verbose) cerr << "Input -R " << r_vir
				  << " interpreted as virial radius.\n";

	    if (!T_flag) {

		// If no time unit is explicitly specified, derive it from
		// the other units.  However, allow the possibility that the
		// mass of the system being modeled and the total mass of the
		// N-body system are not the same (e.g. when modeling a large
		// cluster by a smaller N-body system).

		// Standard (N-body / Heggie & Mathieu) time unit, in Myr:

		t_vir = 21.0 * sqrt(q_vir/m_tot_true) * pow(r_vir, 1.5);
	    }

	    // Now t_vir is the number of Myr corresponding to a standard
	    // time unit (an "N-body unit") IN THE SYSTEM BEING MODELED
	    // (having virial radius r_vir pc and total mass m_tot_true solar
	    // masses).  Passing this to set_stellar_evolution_scaling
	    // preserves (say) the number of orbits before the first supernova,
	    // i.e. performs "dynamical" scaling of the evolution.  In order to
	    // perform "relaxation" scaling, i.e. to have the right number of
	    // relaxation times before the first supernova, we must modify t_vir
	    // to take the N-dependence of the relaxation time into account.
	
	    if (use_relax) {
		real Gamma = 0.1;			// (or whatever...)
		real n_true = m_tot_true/m_bar_true;
		real tr_true = n_true / log10(1 + Gamma*n_true);
		real tr = nbody / log10(1 + Gamma*nbody);
		
		// (Probably don't really want to use N, eventually.)

		// More evolution per orbit in a smaller system,
		// so each orbit must correspond to more years...

		t_vir *= tr_true / tr;
	    }

	    real t_virial = sqrt(pow(r_virial, 3)/b->get_mass());
	    if (verbose) {
		cerr << "Units (N-body / physical):" << endl;
		cerr << "    <m>: " << b->get_mass()/nbody
		     << " / " << m_bar_true << " M_sun" << endl;
		cerr << "    [M]: " << b->get_mass()
		     << " / " << m_tot_true << " M_sun" << endl;
		cerr << "    [T]: " << t_virial << " / " << t_vir
		     << " Myr" << endl;
		cerr << "    [L]: " << r_virial << " / " << r_vir
		     << " pc"  << endl;
	    }

	    // **** NOTE: Length/radius scaling is almost certainly wrong
	    // ****       for use with binary evolution...

	    b->get_starbase()->set_stellar_evolution_scaling(nbody*m_bar_true,
							     r_vir/r_virial,
							     t_vir/t_virial);
	}

	// Add star parts to node part.
	// Required dynamic_cast to hdyn in order to notify hdyn of
	// the stellar size.

	addstar(b,                             // Note that T_start and
		T_start,                       // Main_Sequence are
		Main_Sequence,                 // defaults. They are
		true);                         // ignored if a star
					       // story exists.

	if (verbose) {
	    cerr << "stellar_merger_criterion = "
		 <<  sqrt(b->get_stellar_merger_criterion_sq()) << endl;

	    cerr << "stellar_encounter_criterion = "
		 <<  sqrt(b->get_stellar_encounter_criterion_sq()) << endl;
	}
    }

    return true;
}
