
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

// Functions associated with stellar evolution.
//
// Externally visible functions:
//
//	bool evolve_stars

#include "hdyn.h"
#include "star/dstar_to_kira.h"
#include "kira_local.h"
#include "kira_debug.h"

// *** Debugging/diagnostic flags are now defined in kira_debug.h. ***

//#define REPORT_STELLAR_EVOLUTION false
//#define REPORT_STELLAR_MASS_LOSS false
//#define REPORT_BINARY_MASS_LOSS  false

#define MINIMUM_MASS_LOSS 1.e-16  // Do not allow for mass loss below
                                  // the numerical precision

#define MINIMUM_REPORT_MASS_LOSS  1.e-6

local void dissociate_binary(hdyn* bi)
{
    if (bi->get_kepler() == NULL) {
	cerr << "dissociate_binary: no kepler!\n";
	return;
    }

    // Discard kepler if sudden mass loss took place.

    // The orbital phase of the binary is randomized before the
    // sudden mass loss is applied, since an unperturbed binary
    // tends to be at its apocenter when integrated.

    if (REPORT_BINARY_MASS_LOSS) {
	cerr.precision(8);
	cerr << "\nsudden mass loss from binary, ";
	cerr << bi->format_label() << endl;

	PRC(bi->get_time()); PRL(bi->get_system_time());
	PRL(bi->get_kepler()->get_time());
	pp3(bi->get_top_level_node(), cerr);

	if (has_dstar(bi))
	    ((double_star*)(bi->get_parent()
			      ->get_starbase()))
			      ->dump(cerr);
	// bi->get_kepler()->print_all(cerr);
	cerr.precision(PRECISION);
    }

    // The kepler structure has not yet been modified due
    // to rapid binary evolution (e.g. supernova), so the
    // component positions and velocities retain their values
    // prior to the start of the evolution step in which the
    // supernova occurred.  Any mass transfer occurring during
    // the step, but prior to the supernova, has already been
    // accounted for.  Component masses will be adjusted
    // individually later (as two single stars).

    // A combination of fast and several slow mass-loss episodes
    // is currently handled as a single slow episode followed by
    // a fast episode.

    bi->update_dyn_from_kepler();

    // Note that this update will in general produce a tidal error since it
    // changes the phase of an unperturbed binary.  However, this error will
    // be absorbed in "kira_counters->de_total" after reinitialization.

    // bi->set_first_timestep(eta, initial_step_limit);
    // Not necessary because entire system is scheduled
    // to be reinitialized.

    delete bi->get_kepler();
    bi->set_kepler(NULL);
    bi->get_binary_sister()->set_kepler(NULL);

    if (REPORT_BINARY_MASS_LOSS) {
	cerr << "Deleted kepler...";
	PRC(bi->get_time()); PRC(bi->get_system_time());
    }

    if (has_dstar(bi)) {
	bool update_dynamics[2] = {false, false};
	create_or_delete_binary(bi->get_parent(),
				update_dynamics);	// Defaults probably
							// safe here because
							// merger won't occur.
	
    }
		    
    // NOTE: In case of sudden mass loss, probably have
    // to recompute accs and jerks, at least of neighbors.
}



bool evolve_stars(hdyn* b, real eps, real eta,
		  real initial_step_limit,
		  real dt_log, real dt_sstar)
{
    bool correct_dynamics = false;
    real dt_max = min(dt_log, dt_sstar);
    
    real cpu;

    if (fmod(b->get_system_time(), dt_sstar) == 0.0 && b->get_use_sstar()) {

	if (REPORT_STELLAR_EVOLUTION) {
	    cerr << "\n===============\nStellar evolution start at time "
		 << b->get_system_time() << endl;
	    cpu = cpu_time();
	}

	correct_dynamics |= stellar_evolution(b);

	if (REPORT_STELLAR_EVOLUTION) {
	    real delta_cpu = cpu_time() - cpu;
	    PRC(delta_cpu); PRL(correct_dynamics);
	    cerr << "\n===============\nStellar evolution end\n";
	}
     } 

    if (b->get_use_dstar()) {

	bool save_correct_dynamics = correct_dynamics;
	cpu = cpu_time();

	if (REPORT_STELLAR_EVOLUTION) {
	    cerr << "\n===============\nBinary evolution start at time "
		 << b->get_system_time() << endl;
	}

	correct_dynamics |= binary_evolution(b, eps, eta, dt_log);

	if (correct_dynamics != save_correct_dynamics)
	    cerr << "\nAfter binary evolution, d(CPU) = "
		 << cpu_time() - cpu << endl;

	if (REPORT_STELLAR_EVOLUTION) {
	    if (correct_dynamics == save_correct_dynamics) {
		real delta_cpu = cpu_time() - cpu;
		PRC(delta_cpu);
	    }
	    PRL(correct_dynamics);
	    cerr << "\nBinary evolution end\n===============\n";
	}
    }

    // Evolution is over.  See if we need to correct the system for
    // mass loss, mergers, etc.

    if (correct_dynamics) {

	real time = b->get_system_time();

	if (REPORT_STELLAR_EVOLUTION) {
	    cerr << " synchronizing tree: "; PRL(time);
	}

	// Note that synchronize_tree() does NOT touch unperturbed
	// binaries.  (Still true?)

	synchronize_tree(b, eps, eta, dt_max);

	if (REPORT_STELLAR_EVOLUTION) 
	    cerr << "After synchronize_tree" << endl;

	predict_loworder_all(b, b->get_system_time());	// Necessary??
	if (REPORT_STELLAR_EVOLUTION) 
	    cerr << "After predict_loworder_all" << endl;

	real mass0 = total_mass(b);

	real epot0, ekin0, etot0;
	calculate_energies_with_tidal(b, eps*eps, epot0, ekin0, etot0);

	if (REPORT_STELLAR_EVOLUTION) {
	    cerr << "After calculate_energies_with_tidal"<<endl;
	    PRC(epot0); PRC(ekin0); PRL(etot0);
	}

	// Change stellar masses after evolution.

	real de_kick = 0;

	if (REPORT_STELLAR_EVOLUTION) 
	    cerr << "\n----------\nCorrecting dynamical masses..." << endl;

	for_all_leaves(hdyn, b, bi) {
	    real dm = 0;

	    // Order of business:	slow mass loss
	    //				fast mass loss and dissociation
	    //				single stars

	    if (bi->get_kepler()) {
		if (has_dstar(bi)) {	// elder binary component only...

		    hdyn * parent = bi->get_parent();

		    // Update all binaries for slow mass loss (even if
		    // dm_slow = 0, in which case we simply update the
		    // unperturbed motion).

		    dm = get_total_mass(parent) - parent->get_mass();

		    // Note: dm is the CHANGE in mass of the binary system.
		    // It may be positive or negative.
		    // Sudden_mass_loss is the mass LOST by the system
		    // in an impulsive fashion.  It is always positive,
		    // by convention.

		    // Determine mass changes due to fast and slow processes
		    // (positive numbers mean mass increase!).

		    real dm1_fast = -sudden_mass_loss(bi);
		    real dm2_fast = -sudden_mass_loss(bi->get_binary_sister());
		    real dm_fast  = dm1_fast + dm2_fast;
		    real dm_slow  = dm - dm_fast;

		    if (dm != 0 || dm_slow != 0 || dm_slow != 0) {

			if (abs(dm)      >= MINIMUM_REPORT_MASS_LOSS ||
			    abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS ||
			    abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS) {

			    PRC(dm); PRC(dm_slow); PRL(dm_fast);
			    PRC(dm1_fast); PRL(dm2_fast);
			}
		    }
			
		    // Note: sudden_mass_loss() is reset to 0 after use.

		    if (abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS
			&& REPORT_BINARY_MASS_LOSS) {
			cerr.precision(8);
			cerr << "slow mass loss from binary, "; PRL(dm_slow);
			cerr << bi->format_label() << endl;

			PRC(bi->get_time()); PRL(bi->get_system_time());
			PRL(bi->get_kepler()->get_time());
			pp3(bi->get_top_level_node(), cerr);

			((double_star*)(parent->get_starbase()))->dump(cerr);
			// bi->get_kepler()->print_all(cerr);
			cerr.precision(PRECISION);
		    }

		    // At present, the kepler and dyn structures reflect the
		    // binary configuration at the start of the current
		    // evolution step.  Update them now to reflect changes
		    // due to binary evolution.

		    // Only correct for slow mass loss here.  Fast mass loss
		    // is accounted for after the binary is dissociated.

		    update_kepler_from_binary_evolution(parent, dm_fast);

		    if (abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS &&
			REPORT_BINARY_MASS_LOSS)
			cerr << "After kepler update" << endl;

		    update_dyn_from_binary_evolution(parent, bi, dm1_fast, 
						                 dm2_fast);

		    if (abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS &&
			REPORT_BINARY_MASS_LOSS) 
			cerr << "dyn updated" << endl;

		    // Neither update_kepler_from_binary_evolution nor
		    // update_dyn_from_binary_evolution modify the parent
		    // mass.  Correct it here, changing component offsets
		    // accordingly.

		    correct_leaf_for_change_of_mass(parent, dm_slow);

		    // Correcting the parent but not the components in this
		    // way effectively isotropizes the mass loss.

		    if (abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS &&
			REPORT_BINARY_MASS_LOSS) {
			cerr << "leaf corrected" << endl;
			pp3(bi->get_top_level_node(), cerr);
		    }

		    // Set new unperturbed timestep.
		    
		    // bi->set_unperturbed_timestep(false);  // Changed 3/98...

		    bool reinitialize = false;
		    bi->integrate_unperturbed_motion(eta, eps, dt_log,
						     reinitialize,
						     false);	// ...to this

		    // NOTE: it is possible that the binary will have
		    //	     been deleted in integrate_unperturbed_motion,
		    //	     so we shouldn't assume that kepler or dstar
		    //	     exist on return.  It is also possible that
		    //	     the binary has merged, in which case node
		    //	     bi is no longer valid.

		    // Note from Steve 8/98.  This integration may well
		    // cause the unperturbed motion to be terminated by
		    // leaving the components separating on the next step.
		    // To be repaired...

		    if (!bi->is_valid()) {
			if (abs(dm_slow) >= MINIMUM_REPORT_MASS_LOSS
			    && REPORT_BINARY_MASS_LOSS
			    && REPORT_STELLAR_EVOLUTION) {
			    cerr << "After unperturbed timestep: "<<endl;

			    PRC(bi->get_time());
			    PRL(bi->get_unperturbed_timestep());
			    PRL(bi->get_time()+bi->get_unperturbed_timestep());
			    PRL(bi->get_system_time());
			}

			// Start to handle any fast evolution by dissociating
			// the binary.  (The rest is done by treating the
			// components as single stars.)

			if (dm_fast != 0) {
			    cerr << "SN in binary: ";
			    PRL(dm_fast);
			    dissociate_binary(bi);
			}
		    }

		} else if (bi->is_leaf() && !bi->get_use_dstar()) {

		    // Pretty suspect to have stellar evolution with
		    // no binary evolution...

		    cerr << "SN in non-dstar binary\n";
		    dissociate_binary(bi);
		}
	    }

	    // What to do about possible mass loss if merger occurred...?

	    if (bi->is_valid()) {

		dm = get_total_mass(bi) - bi->get_mass();

		// For single stars, dm is the total mass loss.  We draw no
		// distinction between fast and slow evolution.

		// For ex-binary components, dm is the fast mass loss only.
		// (Slow mass loss was completely accounted for in the
		// preceding loop.)  Check for a kepler to avoid applying
		// rounding error corrections to binary component masses.

		if (abs(dm) > MINIMUM_MASS_LOSS && bi->get_kepler() == NULL) {
		
		    correct_leaf_for_change_of_mass(bi, dm);

		    if (REPORT_STELLAR_MASS_LOSS &&
			abs(dm) >= MINIMUM_REPORT_MASS_LOSS) {

			cerr << "Single star " << bi->format_label()
			     << " lost mass: " << dm << endl;
			if (bi->is_low_level_node()) {
			    cerr << "star is leaf" << endl;
			    pp3(bi->get_top_level_node(), cerr);
			}
		    }

		    if (bi->get_mass() < 0) {
			cerr << "\nEvolved star, negative mass of star "
			     << bi->format_label() 
			     << " with mass = " << dm << endl;;
			pp3(bi, cerr);
			hdyn * btop = bi->get_top_level_node();
			pp3(btop, cerr);
			((star*)btop->get_starbase())->dump(cerr);
		    }
		}
	    
		// Add kick velocity, if any...
	    
		vector dv = anomalous_velocity(bi);
		if (square(dv) > 0)
		    correct_leaf_for_change_of_vector(bi, dv, &hdyn::get_vel,
							      &hdyn::inc_vel);

		vector vnew = hdyn_something_relative_to_root(bi, &hdyn::get_vel);
		de_kick += 0.5 * bi->get_mass() * (vnew*vnew - square(vnew-dv));

		// For unperturbed binary, this treatment        <---------
		// might be quite corrrect.......		 <---------

	    }
	}

	// Reset center of mass and recompute accs and jerks
	// on top-level nodes (latter not needed?).

	b->to_com();

	// calculate_acc_and_jerk_on_all_top_level_nodes(b, eps);
	// predict_loworder_all(b, b->get_system_time());

	real epot, ekin, etot;
	calculate_energies_with_tidal(b, eps*eps, epot, ekin, etot);

	real de_total = etot - etot0;

	if (REPORT_STELLAR_EVOLUTION) {
	    cerr << "End of correction: "; PRL(de_total);
	    PRC(epot); PRC(ekin); PRL(etot);
	}

	// Update statistics on mass and energy change components:

	b->get_kira_counters()->dm_massloss += mass0 - total_mass(b);
	
	b->get_kira_counters()->de_total += de_total;
	b->get_kira_counters()->de_massloss += de_total - de_kick;
	b->get_kira_counters()->de_kick += de_kick;

	predict_loworder_all(b, b->get_system_time());

	if (REPORT_STELLAR_EVOLUTION) 
	    cerr << "initialize_system_phase2 called from evolve_stars\n";
       	initialize_system_phase2(b, eps, eta, initial_step_limit);

       	b->set_mass(total_mass(b));
    }

    // test_kepler(b);

    if (REPORT_STELLAR_EVOLUTION) {
	cerr << "\nEnd of evolve_stars: "; PRL(correct_dynamics);
	cerr << "---------\n\n";
    }

    return correct_dynamics;
}

