
//// mk_aniso_king: construct an anisotropic King model.
////
//// Options:     -a    specify alpha1 [-0.0009]
////              -b    specify alpha3 [0.0003]
////              -c    add a comment to the output snapshot [false]
////              -i    number the particles sequentially [don't number]
////              -n    specify number of particles [no default]
////              -o    echo value of random seed [don't echo]
////              -s    specify random seed [random from system clock]
////              -T    test options [0]
////                        1: print King model to cerr
////                        2: save model and surface brightness profiles
////              -u    leave final N-body system unscaled
////                        [scale to E=-1/4, M = 1, R = 1]
////              -w    specify King dimensionless depth [no default]
////
//// The resulting model fills its Roche lobe in the specified tidal
//// field.  Its length scale is set by G = M = 1 and the value of alpha1.

// version 1:  July 1998    Steve McMillan
//			    steve@zonker.drexel.edu
//                          Physics Dept., Drexel University, Phila PA USA
//                          (Wrapper for D.C. Heggie's Fortran 77 program)

#include "dyn.h"

#ifdef TOOLBOX

#define  SEED_STRING_LENGTH  256
char  tmp_string[SEED_STRING_LENGTH];

#ifdef FORTRAN_TRAILING_UNDERSCORE
#   define AKING aking_
#else
#   define AKING aking
#endif

extern "C" void AKING(real*, real*, real*, real*, real*, real*, real*,
		      real*, real*, int*, int*, real*, int*);

local void mk_aniso_king(dyn * b, int n, real w0, real alpha1, real alpha3,
			 bool u_flag, int test, int seed)

// Create an anisotropic King model using Heggie's Fortran code, and
// optionally initialize an N-body system with total mass = 1, core
// radius = 1 (?).

{
    // These tests are redundant: they duplicate the tests at the
    // start of aking.

    if (alpha3 != 0 && alpha1 >= 0)
	err_exit("mkking: must specify alpha1 < 0");
    if (w0 < 1)
	err_exit("mkking: must specify w0 > 1");
    if (w0 > 12)
	err_exit("mkking: must specify w0 < 12");

    // Compute the cluster density/velocity/potential profile

    real* m = new real[n];
    real* x = new real[n];
    real* y = new real[n];
    real* z = new real[n];
    real* vx = new real[n];
    real* vy = new real[n];
    real* vz = new real[n];

    // Enforce limit on seed required by internal random-number generator
    // in aking (ran2, from Fortran Numerical Recipes, 1ed).

    seed %= 150999;

    AKING(m, x, y, z, vx, vy, vz,
	  &alpha1, &alpha3, &seed, &n, &w0, &test);

    if (b == NULL || n < 1) return;

    // Initialize the N-body system.

    sprintf(tmp_string,
    "         Anisotropic King model, w0 = %.2f, alpha1 = %f, alpha3 = %f",
	      w0, alpha1, alpha3);
    b->log_comment(tmp_string);

    // Assign positions and velocities.

    int i = 0;
    for_all_daughters(dyn, b, bi) {

	bi->set_mass(m[i]);
	bi->set_pos(vector(x[i], y[i], z[i]));
	bi->set_vel(vector(vx[i], vy[i], vz[i]));

	i++;
    }

    delete m, x, y, z, vx, vy, vz;

    // System is in virial equilibrium in a consistent set of units
    // with G and total mass = 1 and length scale determined by the
    // fact that the cluster exactly fills its Roche lobe in the
    // external field.

    // Transform to center-of-mass coordinates and optionally
    // scale to standard parameters.

    b->to_com();

    if (!u_flag && n > 1) {

	// Scaling may be OK because cluster is Roche-lobe filling.

	real kinetic, potential;

	get_energies(b, 0.0, kinetic, potential);
	scale_virial(b, -0.5, kinetic, potential);
	real energy = kinetic + potential;
	scale_energy(b, -0.25, energy);

	putrq(b->get_dyn_story(), "total_energy", -0.25);
	putrq(b->get_log_story(), "initial_rvirial", 1.0);

    }
}

#define OMEGA_2	3.e-4

main(int argc, char ** argv)
{
    real w0;
    int  n;

    // Default tidal parameters are appropriate to a cluster in a
    // a circular orbit of radius D = 1000 length units around a
    // point-mass "galaxy" of mass Mg = 3.0e5
    //
    //	    alpha1  = -3 G Mg / D^3  = -3 Omega^2  (Binney & Tremaine, p. 450)
    //	    alpha3  =    G Mg / D^3  =    Omega^2
	
    real alpha1 = -3 * OMEGA_2;
    real alpha3 = OMEGA_2;
    bool a_flag = false;

    int  input_seed, actual_seed;

    bool c_flag = false;
    bool i_flag = false;
    bool n_flag = false;
    bool o_flag = false;
    bool s_flag = false; 
    bool w_flag = false;
    bool u_flag = false;

    int test = 0;
    char  *comment;

    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "a:b:c:in:os:T:uw:";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
        switch(c) {
            case 'a': alpha1 = atof(poptarg);
	    	      a_flag = true;
                      break;
            case 'b': alpha3 = atof(poptarg);
	    	      a_flag = true;
                      break;
	    case 'i': i_flag = true;
                      break;
            case 'n': n_flag = true;
                      n = atoi(poptarg);
                      break;
	    case 'o': o_flag = true;
                      break;
            case 's': s_flag = true;
                      input_seed = atoi(poptarg);
                      break;
            case 'T': test = atoi(poptarg);
                      break;
	    case 'u': u_flag = true;
                      break;
            case 'w': w_flag = true;
                      w0 = atof(poptarg);
                      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	              get_help();
        }

    if (!w_flag) {
        cerr << "mkking: please specify the dimensionless depth";
        cerr << " with -w #\n";
        exit(1);
    }

    if (!n_flag && test == 0) {
        cerr << "mkking: please specify the number # of";
        cerr << " particles with -n #\n";
        exit(1);
    }

    if (n_flag && n < 1)
	err_exit("mkking: n > 0 required");

    if (!a_flag) {
	cerr << "Using default "; PRC(alpha1); PRL(alpha3);
    }

    dyn *b = NULL;

    if (n_flag) {

	b = new dyn();
	dyn *by, *bo;

	bo = new dyn();
	if (i_flag)
	    bo->set_label(1);
	b->set_oldest_daughter(bo); 
	bo->set_parent(b);

	for (int i = 1; i < n; i++) {
	    by = new dyn();
	    if (i_flag)
		by->set_label(i+1);
	    by->set_parent(b);
	    bo->set_younger_sister(by);
	    by->set_elder_sister(bo);
	    bo = by;
	}

	if (c_flag)
	    b->log_comment(comment);		// add comment to story
	b->log_history(argc, argv);

	if (s_flag == false) input_seed = 0;	// default
	actual_seed = srandinter(input_seed);

	if (o_flag)
	    cerr << "mkking: random seed = " << actual_seed << endl;

	sprintf(tmp_string,
		"         random number generator seed = %d",
		actual_seed);
	b->log_comment(tmp_string);
    }

    mk_aniso_king(b, n, w0, alpha1, alpha3, u_flag, test, actual_seed);

    if (n_flag)
	put_node(cout, *b);
}

#endif

/* end of: mk_aniso_king.C */  
