//
// dyn_di.C: dyn-specific diagnostic functions.
//

#include "dyn.h"

void dbg_message(char* s, dyn* b) {
#ifdef DEBUG    
    cerr << s << " ";
    b->pretty_print_node(cerr); cerr << "\n";
#else
    char* dummy_char = s;   	// to keep the compiler happy
    dyn* dummy = b;       	// to keep the compiler happy
#endif    
}

void dbg_message(char* s, dyn* bj, dyn *bi) {
#ifdef DEBUG    
    cerr << s << " ";
    bj->pretty_print_node(cerr); cerr << "->";
    bi->pretty_print_node(cerr); cerr << "\n";
#else
    char* dummy_char = s;       // to keep the compiler happy
    dyn* dummy_i = bi;       	// to keep the compiler happy
    dyn* dummy_j = bj;       	// to keep the compiler happy
#endif    
}

local inline void accumulate_pot_on_work(dyn * bj,
					 vector & d_pos,
					 vector & d_vel,
					 real eps2,
					 real & p) {
    real r2inv = 1.0 / (d_pos*d_pos + eps2);
    real mrinv = bj->get_mass() * sqrt(r2inv);
    p -= mrinv;

//  silly trick to make the compiler shut up about "warning:  d_vel not used":
    vector dummy = d_vel;

//    cout << "work p = " << p << " \n";    
}

local void tree_calculate_pot(dyn * bj,
			dyn *mask,
			dyn *bi,
			vector offset_pos,
			vector offset_vel,
			real eps2,
			real & p)
{
    if(bj == mask) return;
    if(bj->get_oldest_daughter() != NULL){
	for(dyn * bb = bj->get_oldest_daughter();bb != NULL;
	    bb = bb->get_younger_sister()){
	    tree_calculate_pot(bb,mask,bi,
			       offset_pos + bb->get_pos(),
			       offset_vel + bb->get_vel(),
			       eps2,p);
	}
    }else{
	if(bj != bi){
	    accumulate_pot_on_work(bj, offset_pos, offset_vel, eps2, p);
	}
    }
}


local void calculate_pot_on_leaf(dyn * bj,
			   dyn *mask,
			   dyn *bi,
			   real eps2,
			   real & p)
{
    //    dbg_message("calculate_pot_on_leaf", bi);
    vector d_pos;
    vector d_vel;
    for(dyn * b = bi;b != bj; b = b->get_parent()){
	d_pos -= b->get_pos();
	d_vel -= b->get_vel();
    }
    tree_calculate_pot(bj,mask,bi,d_pos,d_vel,eps2,p);
}
real pot_on_node(dyn * bj,
			   dyn *mask,
			   dyn *bi,
			   real eps2)
{
    //    dbg_message("calculate_pot_on_node", bi);
    real p = 0;
    real mtot = 0;
    if(bi->get_oldest_daughter() == NULL){
	calculate_pot_on_leaf(bj,mask,bi,eps2,p);
    }else{
	real p_daughter;
	for(dyn * bd = bi->get_oldest_daughter(); bd != NULL ;
	    bd = bd->get_younger_sister()){
	    real m_daughter = bd->get_mass();
	    p_daughter = pot_on_node(bj,mask,bd,eps2);
	    p += m_daughter*p_daughter;
	    mtot += m_daughter;
	}
	real minv = 1/mtot;
	p *= minv;
    }
    return p;
}

real pot_on_binary_node(dyn * bi,
			real eps2)
{
    //    dbg_message("calculate_pot_on_binary_node", bi);
    real mtot = 0;
    dyn * parent = bi->get_parent();
    if(parent->get_oldest_daughter()->get_younger_sister()
       ->get_younger_sister()!= NULL){
	cout << "Not a binary node \n";
	exit(1);
    }
    dyn * sister = bi->get_elder_sister();
    if(sister == NULL) sister = bi->get_younger_sister();
    
    return pot_on_node(parent,bi,bi,eps2);
}

real pot_on_general_node(dyn * bj, dyn * bi, real eps2)
{
    //    dbg_message("calculate_pot", bi);
    if(bi->get_parent()->get_parent() == NULL){
	return pot_on_node(bj,bi,bi,eps2);
    }else{
	return pot_on_binary_node(bi,eps2);
    }
}

local void accumulate_energies(dyn * root, dyn*b, real eps2,
			       real & epot, real& ekin, real & etot)
{
    if(b->get_oldest_daughter() != NULL){
	for(dyn * bb = b->get_oldest_daughter();bb != NULL;
	    bb = bb->get_younger_sister()){
	    accumulate_energies(root, bb, eps2, epot,ekin,etot);
	}
    }
    if(b->get_parent()!=NULL){
	real mi = b->get_mass();
	epot += 0.5*mi*pot_on_general_node(root, b, eps2);
	vector vel = b->get_vel();
	ekin += 0.5*mi*vel*vel;
    }
    etot = ekin + epot;
}

void calculate_energies(dyn * root, real eps2,
			real & epot, real & ekin, real & etot)
{
    epot = ekin = etot = 0;
    accumulate_energies(root, root, eps2, epot, ekin, etot);
}

void print_recalculated_energies(dyn* b, int mode, real eps2, real e_corr)
{
    real epot = 0;
    real ekin = 0;
    real etot = 0;
    static    real initial_etot;
    accumulate_energies(b, b, eps2, epot, ekin, etot);
    cerr << "Energies: " << epot << " " << ekin << " " << etot
	 <<" "<<initial_etot;
    if(mode){
	// real de = (etot - e_corr - initial_etot) / epot;
	// The above normalization fails if many stars have escaped
	// the system with HUGE kicks
	real de = (etot - e_corr - initial_etot) / initial_etot;
	cerr << " de = " << de << "\n";
    }else{
	initial_etot = etot;
	cerr << "\n";
    }	
}
