
//  sys_stats.C: Print out various diagnostic statistics on a system.

#include "dyn.h"

#ifndef TOOLBOX

local void print_numbers_and_masses(dyn* b)
{
    // Numbers of stars and nodes:

    int N = b->n_leaves();
    int N_top_level = b->n_daughters();
    PRI(4); PRC(N); PRL(N_top_level);

    // Masses and averages:

    real total_mass_nodes = 0;
    for_all_daughters(dyn, b, bi) total_mass_nodes += bi->get_mass();

    real total_mass_leaves = 0;
    real m_min = VERY_LARGE_NUMBER, m_max = -m_min;
    for_all_leaves(dyn, b, bi) {
	total_mass_leaves += bi->get_mass();
	m_min = min(m_min, bi->get_mass());
	m_max = max(m_max, bi->get_mass());
    }
    real m_av = total_mass_leaves / N;

    PRI(4); PRL(b->get_mass());
    PRI(4); PRC(total_mass_nodes); PRL(total_mass_leaves);
    PRI(4); PRC(m_min); PRC(m_max); PRL(m_av);
}

local int which_zone(dyn* bi, vector& dc_pos, int n_lagr, real* r_lagr)
{
    vector dr = bi->get_pos() - dc_pos;
    real dr2 = dr*dr;

    for (int j = 0; j < n_lagr; j++)
      if (r_lagr[j] > dr2) return j;

    return n_lagr;
}

local void print_numbers_and_masses_by_radial_zone(dyn* b, int which)
{
    if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

        // Make a list of previously computed lagrangian radii.

        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
	real *r_lagr = new real[n_lagr];
	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	// Numbers of top-level nodes:

	int *N = new int[n_lagr+1];

	// Note that r_lagr omits the 0% and 100% radii.

	// Masses and averages:

	real *total_mass_nodes = new real[n_lagr+1];
	real *m_min = new real[n_lagr+1];
	real *m_max = new real[n_lagr+1];

	// Initialization:

	for (int i = 0; i <= n_lagr; i++) {
	    N[i] = 0;
	    total_mass_nodes[i] = 0;
	    m_min[i] = VERY_LARGE_NUMBER;
	    m_max[i] = -VERY_LARGE_NUMBER;
	    r_lagr[i] *= r_lagr[i];
	}

	// As in lagrad, use the geometric center if 
	// density center is unknown.

	vector dc_pos = 0;
	if (find_qmatch(b->get_dyn_story(), "density_center_pos"))
	    dc_pos = getvq(b->get_dyn_story(), "density_center_pos");

	int N_total = 0;
	for_all_daughters(dyn, b, bi)

	    // Count single stars and binaries separately.

	    if ( (which == 0 && bi->get_oldest_daughter() == NULL)
		|| (which == 1 && bi->get_oldest_daughter() != NULL) ) {

	        // Find which zone we are in.

	        int i = which_zone(bi, dc_pos, n_lagr, r_lagr);

		// Update statistics.

		N[i]++;
		N_total++;
		total_mass_nodes[i] += bi->get_mass();
		m_min[i] = min(m_min[i], bi->get_mass());
		m_max[i] = max(m_max[i], bi->get_mass());
	    }

	if (N_total > 0) {

	    for (int i = 0; i <= n_lagr; i++)
	        cerr << "    zone " << i << "  N = " << N[i]
		     << "  m_av = " << total_mass_nodes[i] / N[i]
		     << "  m_min = " << m_min[i]
		     << "  m_max = " << m_max[i]
		     << endl;

	    cerr << "\n  Cumulative:\n";

	    int  Nc = 0;
	    real m_totc = 0;
	    real m_minc = VERY_LARGE_NUMBER;
	    real m_maxc = -VERY_LARGE_NUMBER;

	    for (int i = 0; i <= n_lagr; i++) {
	        Nc += N[i];
		m_totc += total_mass_nodes[i];
		m_minc = min(m_minc, m_min[i]);
		m_maxc = max(m_maxc, m_max[i]);
	        cerr << "    zone " << i << "  N = " << Nc
		     << "  m_av = " << m_totc / Nc
		     << "  m_min = " << m_minc
		     << "  m_max = " << m_maxc
		     << endl;
	    }

	} else

	    cerr << "    (none)\n";
    }
}

local void print_anisotropy_by_radial_zone(dyn* b, int which)
{
    if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

        // Make a list of previously computed lagrangian radii.

        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
	real *r_lagr = new real[n_lagr];
	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	real *total_weight = new real[n_lagr+1];
	real *vr2_sum = new real[n_lagr+1];
	real *vt2_sum = new real[n_lagr+1];

	// Note that r_lagr omits the 0% and 100% radii.

	// Initialization:

	for (int i = 0; i <= n_lagr; i++) {
	    total_weight[i] = 0;
	    vr2_sum[i] = 0;
	    vt2_sum[i] = 0;
	    r_lagr[i] *= r_lagr[i];
	}

	// As in lagrad, use the geometric center if 
	// density center is unknown.

	vector dc_pos = 0;
	if (find_qmatch(b->get_dyn_story(), "density_center_pos"))
	    dc_pos = getvq(b->get_dyn_story(), "density_center_pos");

	int N_total = 0;
	for_all_daughters(dyn, b, bi)

	    // Count single stars and binaries separately.

	    if ( (which == 0 && bi->get_oldest_daughter() == NULL)
		|| (which == 1 && bi->get_oldest_daughter() != NULL) ) {

	        // Find which zone we are in.

	        int i = which_zone(bi, dc_pos, n_lagr, r_lagr);

		// Update statistics.

		N_total++;

		real weight = bi->get_mass();
		weight = 1;			// Heggie/Binney & Tremaine

		real v2 = bi->get_vel()*bi->get_vel();
		real vr = bi->get_pos()*bi->get_vel();
		real vr2 = vr * vr / square(bi->get_pos());

		total_weight[i] += weight;
		vr2_sum[i] += weight * vr2;
		vt2_sum[i] += weight * (v2 - vr2);		
	    }

	if (N_total > 0) {

	    for (int i = 0; i <= n_lagr; i++) {
	        real vv = vr2_sum[i];
		if (vv <= 0) vv = 1;
	        cerr << "    zone " << i
		     << "  beta = " << 1 - 0.5 * vt2_sum[i] / vv
		     << endl;
	    }

	    cerr << "\n  Cumulative:\n";

	    real vr2c = 0;
	    real vt2c = 0;

	    for (int i = 0; i <= n_lagr; i++) {
	        vr2c += vr2_sum[i];
	        vt2c += vt2_sum[i];
	        real vv = vr2c;
		if (vv <= 0) vv = 1;
	        cerr << "    zone " << i
		     << "  beta = " << 1 - 0.5 * vt2c / vv
		     << endl;
	    }

	} else

	    cerr << "    (none)\n";
    }
}

local real system_energy(dyn* b)
{
    real kin = 0;
    for_all_leaves(dyn, b, bi)
	kin += bi->get_mass()
	        * square(something_relative_to_root(bi, &dyn::get_vel));
    kin *= 0.5;

    real pot = 0.0;
    for_all_leaves(dyn, b, bi) {
	real dpot = 0.0;
	for_all_leaves(dyn, b, bj) {
	    if (bj == bi) break;
	    vector dx = something_relative_to_root(bi, &dyn::get_pos)
			  - something_relative_to_root(bj, &dyn::get_pos);
	    dpot += bj->get_mass() / abs(dx);
	}
	pot -= bi->get_mass() * dpot;
    }

    return kin + pot;
}

local void print_energies(dyn* b, real& kT)
{
    // Energies (top-level nodes):

    real kinetic_energy, potential_energy;
    get_energies(b, 0.0, kinetic_energy, potential_energy);

    real total_energy = kinetic_energy + potential_energy;
    real virial_ratio = -kinetic_energy / potential_energy;
    kT = -total_energy / (1.5*b->n_daughters());

    PRI(4); cerr << "top-level nodes:\n";
    PRI(8); PRC(kinetic_energy); PRL(potential_energy);
    PRI(8); PRC(total_energy); PRC(kT); PRL(virial_ratio);

    PRI(4); cerr << "total energy (full system) = "
 		 << system_energy(b) << endl;
}

local real print_binary_params(kepler* k, real m1, real kT,
			       real dist_from_cod, bool verbose)
{
    real m = k->get_total_mass();
    real m2 = m - m1;
    real mu = m1 * m2 / m;
    real mu_kT;

    if (kT > 0) mu_kT = mu / kT;

    if (verbose) {

	cerr << "a = " << k->get_semi_major_axis()
	     << "  e = " << k->get_eccentricity()
	     << "  D = " << dist_from_cod
	     << endl;
	PRI(21); cerr << "E = " << mu*k->get_energy()
	     << "  -E/mu = " << -k->get_energy();
	     if (kT > 0) cerr << "  -E/kT = " << -mu_kT*k->get_energy();
	     cerr << endl;
	PRI(21); cerr << "masses  " << m << "  " << m1 << "  " << m2 << endl;

    } else

	cerr << "  "  << m << "  " << m1 << "  " << m2
	     << "  "  << k->get_semi_major_axis()
	     << "  " << k->get_eccentricity()
	     << "  " << mu*k->get_energy()
	     << "  " << -k->get_energy();
	if (kT > 0) cerr << "  " << -mu_kT*k->get_energy();
	cerr << "  " << dist_from_cod
	     << endl;

    return mu*k->get_energy();
}

local void search_for_binaries(dyn* b, real energy_cutoff, real kT,
			       vector center, bool verbose)
{
    // Search for binary pairs among top-level nodes.

    bool found = false;

    for_all_daughters(dyn, b, bi)
	for_all_daughters(dyn, b, bj) {
	    if (bj == bi) break;

	    real M = bi->get_mass() + bj->get_mass();
	    vector dx = bj->get_pos() - bi->get_pos();
	    vector dv = bj->get_vel() - bi->get_vel();
	    real mu = (bi->get_mass() * bj->get_mass()) / M;
	    real E = mu*(0.5*dv*dv - M / abs(dx));

	    if (E < -energy_cutoff*kT) {

		kepler k;
		k.set_time(0);
		k.set_total_mass(M);
		k.set_rel_pos(dx);
		k.set_rel_vel(dv);
		k.initialize_from_pos_and_vel();

		dyn* primary = bi;
		if (bj->get_mass() > bi->get_mass()) primary = bj;

		if (verbose) PRI(4);
		cerr << "(";
		bi->pretty_print_node(cerr);
		cerr << ",";
		bj->pretty_print_node(cerr);
		cerr << ")  ";

		real dist_from_cod = abs(primary->get_pos()-center);

		print_binary_params(&k, primary->get_mass(), kT,
				    dist_from_cod, verbose);

		found = true;
	    }
	}    

    if (!found) cerr << "    (none)\n";
}

local void print_binaries(dyn* b, real kT, vector center, real rcore,
			  bool verbose, bool allow_n_sq_ops)
{
    // Print out the properties of "known" binaries (i.e. binary subtrees),
    // and get some statistics on binaries in the core.

    // If allow_n_sq_ops is false, then we don't know kT and core parameters.

    if (!allow_n_sq_ops) kT = 0;

    real eb = 0;
    real mbcore = 0;
    int nbcore = 0;

    for_all_daughters(dyn, b, bi)
	if (dyn* od = bi->get_oldest_daughter()) {

	    if (verbose) PRI(4); bi->pretty_print_node(cerr);

	    if (bi->n_leaves() > 2) {

		if (verbose) cerr << " is a multiple system\n";

	    } else {

		if (verbose) cerr << " is a binary:  ";

		bool reset = false;
		if (bi->get_kepler() == NULL) {
		    new_kepler(bi, 0);
		    reset = true;
		}

		real dist_from_cod = abs(bi->get_pos()-center);

		eb += print_binary_params(bi->get_kepler(), od->get_mass(),
					  kT, dist_from_cod, verbose);

		if (allow_n_sq_ops) {
		  if (dist_from_cod <= rcore) {
		    nbcore++;
		    mbcore += bi->get_mass();
		  }
		}

		if (reset) {
		    delete bi->get_kepler();
		    bi->set_kepler(NULL);
		}
	    }
	}

    if (verbose) cerr << endl;
    PRI(4); cerr << "total binary energy = " << eb << endl;

    if (allow_n_sq_ops) {
      PRI(4); PRC(nbcore); PRL(mbcore);
    }
}

void compute_core_parameters(dyn* b, int k, vector& center,
			     real& rcore, int& ncore, real& mcore)
{
    vector vel;

    compute_density(b, k);
    compute_mean_cod(b, center, vel);

    real total_weight = 0;
    real sum = 0;
    for_all_daughters(dyn, b, bi) {
	real density = getrq(bi->get_dyn_story(), "density");
	real dens2 = density * density;                        // weight factor

        total_weight += dens2;
	sum += dens2 * square(bi->get_pos() - center);
    }

    real rcore2 = sum/total_weight;
    rcore = sqrt(rcore2);
    ncore = 0;
    mcore = 0;

    for_all_daughters(dyn, b, bi)
	if (square(bi->get_pos() - center) <= rcore2) {
	    ncore++;
	    mcore += bi->get_mass();
	}
}

local void print_core_parameters(dyn* b, vector& density_center, real& rcore)
{
    real mcore;
    int ncore;
    
    compute_core_parameters(b, 6, density_center, rcore, ncore, mcore);
    PRI(4); PRL(density_center);
    PRI(4); PRC(rcore); PRC(ncore); PRL(mcore);
}

local void print_lagrangian_radii(dyn* b, bool percents)
{
    char tmp[3];

    int nl, indent;
    if (! percents){
	nl = 4;
	indent = 20;
	PRI(4); cerr << "quartile radii: ";
    }else{
	nl = 10;
	indent = 25;
	PRI(4); cerr << "10 percentile radii: ";
    }

    compute_general_mass_radii(b, nl);

    if (find_qmatch(b->get_dyn_story(), "n_lagr")) {
        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
	real *r_lagr = new real[n_lagr];
	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	for (int k = 0; k < n_lagr; k += 5) {
	  if (k > 0) {
	    cerr << endl;
	    for (int kk = 0; kk < indent; kk++) cerr << " ";
	  }
	  for (int i = k; i < k+5 && i < n_lagr; i++)
	    cerr << " " << r_lagr[i];
	}
	cerr << endl;
    }
}

void sys_stats(dyn* b, real energy_cutoff,
	       bool verbose, bool percents, bool binaries,
	       bool print_time = false,
	       bool allow_n_sq_ops = false)
{
    if (print_time) {
      real time = getrq(b->get_dyn_story(), "t");
      if (time > -VERY_LARGE_NUMBER)
	cerr << "Time = " << time << endl;
      else
	cerr << "Time = (undefined)" << endl;
    }

    if (verbose) cerr << "\n  Overall parameters:\n";
    print_numbers_and_masses(b);

    real kT = 0;
    if (allow_n_sq_ops)
      print_energies(b, kT);

    if (verbose) cerr << "\n  Lagrangian radii:\n";
    print_lagrangian_radii(b, percents);

    if (verbose)
        cerr << "\n  Mass distribution by Lagrangian zone (singles):\n";
    print_numbers_and_masses_by_radial_zone(b, 0);

    if (verbose)
        cerr << "\n  Mass distribution by Lagrangian zone (multiples):\n";
    print_numbers_and_masses_by_radial_zone(b, 1);

    if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (singles):\n";
    print_anisotropy_by_radial_zone(b, 0);

    if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (multiples):\n";
    print_anisotropy_by_radial_zone(b, 1);

    vector center;
    real rcore;
    if (allow_n_sq_ops) {
      if (verbose) cerr << "\n  Core parameters:\n";
      print_core_parameters(b, center, rcore);
    }

    if (binaries) {

      if (verbose) cerr << "\n  Binaries:\n";
      print_binaries(b, kT, center, rcore, verbose, allow_n_sq_ops);

      if (allow_n_sq_ops) {
	if (verbose) cerr << "\n  Other bound pairs with |E| > 0.5 kT:\n";
	search_for_binaries(b, energy_cutoff, kT, center, verbose);
      }

    }
}

#else

main(int argc, char **argv)
{
    bool binaries = true, verbose = false, out = false,
         percents = false, n_sq = true;

    extern char *poptarg;
    int c;
    char* param_string = "bnotv";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'b': binaries = !binaries;
		      break;
	    case 'n': n_sq = !n_sq;
		      break;
	    case 'o': out = !out;
		      break;
	    case 't': percents = !percents;
		      break;
	    case 'v': verbose = !verbose;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      exit(1);
	}

//  Loop over input until no more data remain.

    dyn *b;
    int i = 0;
    while (b = get_dyn(cin)) {

//         for_all_daughters(dyn, b, b) {
// 	    PRL(b->get_pos());
// 	    PRL(b->get_vel());
// 	}

        // NOTE:  get_xxx() reads NaN in as legal garbage...

	if (i++ > 0) cerr << endl;
	sys_stats(b, 0.5, verbose, percents, binaries, true, n_sq);

	if (out) put_node(cout, *b);
	rmtree(b);
    }
}

#endif
