
//// scale:  (re)scale an N-body system to specified M, Q (=T/U), and E.
////         Note that if the required Q implies E > 0, no additional
////         energy scaling is applied.
////
//// Options:    -e    specify softening parameter [0]
////             -E    specify total energy [don't scale]
////             -m/M  specify total mass [don't scale]
////             -q/Q  specify virial ratio [don't scale]
////		 -s/S  scale to "standard units [not set]
////
////         NOTE that only top-level nodes are considered in get_energies,
////         scale_virial, and scale_energy.  If standard units are chosen
////         with "-S", then the center of mass of the system is also set
////         to zero.

//  Steve McMillan, April 1993

#include "dyn.h"

#ifndef TOOLBOX

#define ALL_i ni = n->get_oldest_daughter(); ni != NULL; \
					     ni = ni->get_younger_sister()
#define j_ABOVE_i nj = ni->get_younger_sister(); nj != NULL; \
					     nj = nj->get_younger_sister()

void scale_mass(dyn* n, real m)
{
    real mass = 0;
    for_all_daughters(dyn, n, ni)
	mass += ni->get_mass();

    real mscale = m/mass;
    for_all_nodes(dyn, n, nj)
	nj->set_mass(mscale*nj->get_mass());  // No "scale_mass"!

    n->set_mass(m);
}

void get_energies(dyn * n, real eps2, 
		  real& kinetic_energy, real& potential_energy)
{
    dyn  * ni, *nj;

    kinetic_energy = 0;
    for (ALL_i) {
	vector v = ni->get_vel();
        kinetic_energy += ni->get_mass() * v * v;
    }
    kinetic_energy *= 0.5;

    potential_energy = 0;
    for (ALL_i) {
        real dphi = 0;
	vector xi = ni->get_pos();
	for (j_ABOVE_i) {
	    vector xij = nj->get_pos() - xi;
	    dphi += nj->get_mass()/sqrt(xij*xij + eps2);
	}
	potential_energy -= ni->get_mass() * dphi;
    }
}

void scale_virial(dyn * n, real q, real& kinetic_energy, real potential_energy)
{
    // Set the virial ratio by scaling velocities.

    dyn  * ni;

    if (q > 0) q = -q;  // Only need specify |Q|.

    real vscale = sqrt(q*potential_energy/kinetic_energy);

    for (ALL_i) ni->scale_vel(vscale);

    kinetic_energy = q*potential_energy;
}

void scale_energy(dyn * n, real e, real& energy)
{
    // Set the energy by scaling positions and velocities.

    dyn  * ni;

    if (energy >= 0) return;
    if (e > 0) e = -e;  // Only need specify |E| and only E < 0 makes sense!

    real xscale = energy/e;
    real vscale = sqrt(1./xscale);

    // PRL(xscale);

    for (ALL_i) {
	ni->scale_pos(xscale);
	ni->scale_vel(vscale);
    }

    energy = e;
}

#else

#define  FALSE  0
#define  TRUE   1

main(int argc, char ** argv) {
    real m, q, e;
    real eps = 0;
    int m_flag = FALSE;
    int q_flag = FALSE;
    int e_flag = FALSE;
    int eps_flag = FALSE;
    int s_flag = false;
    dyn *root;

    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "m:M:q:Q:e:E:sS";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'M':
	    case 'm': m_flag = TRUE;
		      m = atof(poptarg);
		      break;
	    case 'Q':
	    case 'q': q_flag = TRUE;
		      q = atof(poptarg);
		      break;
	    case 'E': e_flag = TRUE;
		      e = atof(poptarg);
		      break;
	    case 'e': eps_flag = TRUE;
		      eps = atof(poptarg);
		      break;
	    case 'S':
	    case 's': m_flag = TRUE;
		      q_flag = TRUE;
		      e_flag = TRUE;
	    	      s_flag = true;
		      m = 1;
	    	      q = 0.5;
		      e = -0.25;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	              get_help();
                      exit(1);
	}

    if (m_flag && m <= 0) warning("Specified mass <= 0");
    if (e_flag && e >= 0) warning("Specified energy >= 0");
    if (q_flag && q <  0) warning("Specified virial ratio < 0");

    root = get_dyn(cin);

    if (m_flag)
        scale_mass(root, m);

    // NOTE: Scaling the total mass changes both the total energy and
    //	     the virial ratio.  No attempt is made to preserve either
    //	     in cases where they are not specified on the command line.

    real kinetic_energy = 0, potential_energy = 0;

    if (q_flag || e_flag) {

	get_energies(root, eps*eps, kinetic_energy, potential_energy);

	if (q_flag) scale_virial(root, q, kinetic_energy, potential_energy);

	// NOTE: Changing the virial ratio changes the total energy of
	//	 the system, but preserves the total mass.

	real energy = kinetic_energy + potential_energy;
	if (e_flag) {
	    real fac = e/energy;
	    scale_energy(root, e, energy);
	    kinetic_energy *= fac;
	    potential_energy *= fac;
	}

	// NOTE: The energy scaling is done on the assumption that eps = 0,
	//	 and preserves the value of the virial ratio in that case.
	//	 The total mass is always left unchanged.

	putrq(root->get_log_story(), "total_energy", energy);
    }

    // Update any virial radius information found in the root dyn story.
    // Assuming here that we only use the scale tool on initialization.

    real r_vir = getrq(root->get_log_story(), "initial_rvirial");

    if (r_vir > -VERY_LARGE_NUMBER) {
        real mass = 0;
	if (m_flag)
	    mass = m;
	else
	    for_all_daughters(dyn, root, b) mass += b->get_mass();

	if (kinetic_energy <= 0)
	    get_energies(root, eps*eps, kinetic_energy, potential_energy);

        putrq(root->get_log_story(), "initial_rvirial",
	      -0.5 * mass / potential_energy);
    }

    if (m_flag) {

        real initial_mass = getrq(root->get_log_story(), "initial_mass");

	if (initial_mass > -VERY_LARGE_NUMBER)
            putrq(root->get_log_story(), "initial_mass", m);
    }

    if (s_flag)
	root->to_com();

    root->log_history(argc, argv);
    put_node(cout, *root);
}

#endif

// end of: scale.C
