
//// sys_stats:  Print out various diagnostic statistics on the input
////             system.  These include:
////
////                 system time, number, mass, mass distribution
////                 relaxation time
////                 system energy
////                 core parameters
////                 lagrangian radii
////                     for quartiles [default]
////                     for ten-percentiles
////                     for "special" choice of Lagrangian masses,
////                         currently 0.005, 0.01, 0.02, 0.05,
////                                   0.1, 0.25, 0.5, 0.75, 0.9
////                 mass distribution by lagrangian zone
////                 anisotropy by lagrangian zone
////                 binary parameters
////
////             In addition, Lagrangian radii are written to the root
////             dyn story.  If N^2_ops	is selected, then core parameters
////             and particle densities are also written to the root
////             and particle dyn stories.
////
////             If N^2_ops is selected, Lagrangian radii are computed
////             relative to the density center, as are binary radial
////             coordinates.  Otherwise, the geometric center is used.
////                 
//// Options:    -b    include binary statistics [yes]
////             -e    recalculate the total energy (requires -n) [no]
////             -n    perform/don't perform actions requiring O(N^2)
////                   operations (e.g. computation of energy and core
////                   radius) [yes]
////             -o    pipe system to cout [no]
////             -s    use "special" Lagrangian masses [no]
////             -t    use "ten-percentile" Lagrangian masses [no]
//
//   Note:  Because of the reference to libstar.a in sys_stats, the
//          only global function in this file is sys_stats itself.
//          Other "stats-related" functions are found in dyn_stats.C.
//	    Thus, any "dyn-only" tools referencing just the helper
//	    functions will not have to load the "star" libraries.

#include "dyn.h"
#include "star/sstar_to_dyn.h"

#ifndef TOOLBOX

#define PRECISION 6

local bool check_for_doubling(dyn* b)
{
    if (getrq(b->get_log_story(), "fraction_of_masses_doubled") > 0)
	return true;
    else
	return false;
}

local real print_relaxation_time(dyn* b,
				 real& kinetic_energy,
				 real& potential_energy)
{
    // Print the relaxation time

    real r_virial, t_relax;
    real total_mass = 0;
    int N = b->n_leaves();
    
    for_all_leaves(dyn, b, bj)
	total_mass += bj->get_mass();

    get_energies(b, 0.0, kinetic_energy, potential_energy);

    r_virial = -0.5 * total_mass * total_mass / potential_energy;

    t_relax = 9.62e-2 * sqrt(r_virial * r_virial * r_virial / total_mass)
              * N / log10(0.4 * N);

    PRI(4); cerr << "r_virial = " << r_virial << endl;
    PRI(4); cerr << "t_relax = "  << t_relax  << endl;

    return r_virial;
}

local bool print_numbers_and_masses(dyn* b)
{
    // Numbers of stars and nodes:      

    real total_cpu_time = getrq(b->get_log_story(), "total_cpu_time");
    if (total_cpu_time > -VERY_LARGE_NUMBER) {
	PRI(4); PRL(total_cpu_time);
    } else {
	PRI(4); cerr << "total_cpu_time = (undefined)" << endl;
    }

    int N = b->n_leaves();
    int N_top_level = b->n_daughters();
    cerr << "    N = " << N << "  N_top_level = " << N_top_level << endl;

    // Masses and averages:

    real total_mass_nodes = 0;
    for_all_daughters(dyn, b, bi)
	total_mass_nodes += bi->get_mass();

    real total_mass_leaves = 0;
    real m_min = VERY_LARGE_NUMBER, m_max = -m_min;
    for_all_leaves(dyn, b, bj) {
	total_mass_leaves += bj->get_mass();
	m_min = min(m_min, bj->get_mass());
	m_max = max(m_max, bj->get_mass());
    }
    real m_av = total_mass_leaves / max(1, N);

    cerr << "    total_mass = " << b->get_mass();
    cerr << "  nodes: " << total_mass_nodes;
    cerr << "  leaves: " << total_mass_leaves << endl;

    cerr << "    m_min = " << m_min << "  m_max = " << m_max
	 << "  m_av = " << m_av << endl;

    return (m_min < m_max);
}

local int which_zone(dyn* bi, vector& dc_pos, int n_lagr, real* r_lagr)
{
    vector dr = bi->get_pos() - dc_pos;
    real dr2 = dr*dr;

    for (int j = 0; j < n_lagr; j++)
      if (r_lagr[j] > dr2) return j;

    return n_lagr;
}

local void print_numbers_and_masses_by_radial_zone(dyn* b, int which)
{
    if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

        // Make a list of previously computed lagrangian radii.

        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
	real *r_lagr = new real[n_lagr];

	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	// Numbers of top-level nodes:

	int *N = new int[n_lagr+1];

	// Note that r_lagr omits the 0% and 100% radii.

	// Masses and averages:

	real *total_mass_nodes = new real[n_lagr+1];
	real *m_min = new real[n_lagr+1];
	real *m_max = new real[n_lagr+1];

	// Initialization:

	for (int i = 0; i <= n_lagr; i++) {
	    N[i] = 0;
	    total_mass_nodes[i] = 0;
	    m_min[i] = VERY_LARGE_NUMBER;
	    m_max[i] = -VERY_LARGE_NUMBER;
	    if (i < n_lagr) r_lagr[i] *= r_lagr[i];
	}

	// Use the same center as lagrad, or the geometric center
	// in case of error.

	vector dc_pos = 0;
	if (find_qmatch(b->get_dyn_story(), "lagr_pos"))
	    dc_pos = getvq(b->get_dyn_story(), "lagr_pos");
	else
	    warning(
	    "print_numbers_and_masses_by_radial_zone: lagr_pos not found");

	int N_total = 0;

	// PRL(which);

	for_all_daughters(dyn, b, bi)

	    // Count single stars and binaries separately.
	    // Count single and doubled-mass stars separately also.

	    if ( (which == 0 && bi->get_oldest_daughter() == NULL)
		|| (which == 1 && bi->get_oldest_daughter() != NULL)
                || (which == 2 && getiq(bi->get_log_story(),
					"mass_doubled") == 0 ) 
		|| (which == 3 && getiq(bi->get_log_story(),
					"mass_doubled") == 1) ) {
		 //		 || (which == 2)
		 //		 || (which == 3) ) {

	        // Find which zone we are in.

	        int i = which_zone(bi, dc_pos, n_lagr, r_lagr);

		// Update statistics.

		N[i]++;
		N_total++;
		total_mass_nodes[i] += bi->get_mass();
		m_min[i] = min(m_min[i], bi->get_mass());
		m_max[i] = max(m_max[i], bi->get_mass());
	    }

	if (N_total > 0) {

	    int i;
	    for (i = 0; i <= n_lagr; i++) {
	        cerr << "    zone " << i << "  N = " << N[i];
	        if (N[i] > 0)
		    cerr << "  m_av = " << total_mass_nodes[i] / max(N[i], 1)
		         << "  m_min = " << m_min[i]
		         << "  m_max = " << m_max[i];
		cerr << endl;
	    }

	    cerr << "\n  Cumulative:\n";

	    int  Nc = 0;
	    real m_totc = 0;
	    real m_minc = VERY_LARGE_NUMBER;
	    real m_maxc = -VERY_LARGE_NUMBER;

	    for (i = 0; i <= n_lagr; i++) {
	        Nc += N[i];
		m_totc += total_mass_nodes[i];
		m_minc = min(m_minc, m_min[i]);
		m_maxc = max(m_maxc, m_max[i]);
		cerr << "    zone " << i << "  N = " << Nc;
		if (Nc > 0) 
		    cerr << "  m_av = " << m_totc / max(Nc, 1)
		         << "  m_min = " << m_minc
		         << "  m_max = " << m_maxc;
		cerr << endl;
	    }

	} else

	    cerr << "    (none)\n";

	delete [] r_lagr;
	delete [] N;
	delete [] total_mass_nodes;
	delete [] m_min;
	delete [] m_max;
    }
}

local void print_anisotropy_by_radial_zone(dyn* b, int which)
{
    if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

        // Make a list of previously computed lagrangian radii.

        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
	real *r_lagr = new real[n_lagr];
	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	real *total_weight = new real[n_lagr+1];
	real *vr2_sum = new real[n_lagr+1];
	real *vt2_sum = new real[n_lagr+1];

	// Note that r_lagr omits the 0% and 100% radii.

	// Initialization:

	for (int i = 0; i <= n_lagr; i++) {
	    total_weight[i] = 0;
	    vr2_sum[i] = 0;
	    vt2_sum[i] = 0;
	    if (i < n_lagr) r_lagr[i] *= r_lagr[i];
	}

	// Use the same center as lagrad, or the geometric center
	// in case of error.

	vector dc_pos = 0;
	if (find_qmatch(b->get_dyn_story(), "lagr_pos"))
	    dc_pos = getvq(b->get_dyn_story(), "lagr_pos");
	else
	    warning("print_anisotropy_by_radial_zone: lagr_pos not found");

	int N_total = 0;

	// PRL(which);

	for_all_daughters(dyn, b, bi)

	    // Count single stars and binaries separately.

	    if ( (which == 0 && bi->get_oldest_daughter() == NULL)
		|| (which == 1 && bi->get_oldest_daughter() != NULL) 
		|| (which == 2 && getiq(bi->get_log_story(),
					"mass_doubled") == 0) 
		|| (which == 3 && getiq(bi->get_log_story(),
					"mass_doubled") == 1) ) {

	        // Find which zone we are in.

	        int i = which_zone(bi, dc_pos, n_lagr, r_lagr);

		// Update statistics.

		N_total++;

		real weight = bi->get_mass();
		weight = 1;			// Heggie/Binney & Tremaine

		real v2 = bi->get_vel()*bi->get_vel();
		real vr = bi->get_pos()*bi->get_vel();
		real vr2 = vr * vr / square(bi->get_pos());

		total_weight[i] += weight;
		vr2_sum[i] += weight * vr2;
		vt2_sum[i] += weight * (v2 - vr2);		
	    }


	if (N_total > 0) {

	    int k;
	    for (k = 0; k <= n_lagr; k += 5) {
	        if (k > 0) cerr << endl;
		cerr << "           ";
		for (int i = k; i < k+5 && i <= n_lagr; i++) {
		    if (vr2_sum[i] > 0)
		        cerr << " " << 1 - 0.5 * vt2_sum[i] / vr2_sum[i];
		    else
		        cerr << "   ---   ";
		}
	    }
	    cerr << endl;

	    cerr << "\n  Cumulative anisotropy:\n";

	    real vr2c = 0;
	    real vt2c = 0;

	    for (k = 0; k <= n_lagr; k += 5) {
	        if (k > 0) cerr << endl;
		cerr << "           ";
		for (int i = k; i < k+5 && i <= n_lagr; i++) {
		    vr2c += vr2_sum[i];
		    vt2c += vt2_sum[i];
		    if (vr2c > 0)
		        cerr << " " << 1 - 0.5 * vt2c / vr2c;
		    else
		        cerr << "   ---   ";
		}
	    }
	    cerr << endl;

	} else

	    cerr << "    (none)\n";

	delete [] r_lagr;
        delete [] total_weight;
	delete [] vr2_sum;
	delete [] vt2_sum;
    }
}

local real system_energy(dyn* b)
{
    real kin = 0;
    for_all_leaves(dyn, b, bj)
	kin += bj->get_mass()
	        * square(something_relative_to_root(bj, &dyn::get_vel));
    kin *= 0.5;

    real pot = 0.0;
    for_all_leaves(dyn, b, bi) {
	real dpot = 0.0;
	for_all_leaves(dyn, b, bj) {
	    if (bj == bi) break;
	    vector dx = something_relative_to_root(bi, &dyn::get_pos)
			  - something_relative_to_root(bj, &dyn::get_pos);
	    dpot += bj->get_mass() / abs(dx);
	}
	pot -= bi->get_mass() * dpot;
    }

    return kin + pot;
}

local real top_level_kinetic_energy(dyn* b)
{
    // (Probably should compute energy relative to density center...)

    real kin = 0;
    for_all_daughters(dyn, b, bb)
        kin += bb->get_mass()*square(bb->get_vel());
    return 0.5*kin;
}

local void print_energies(dyn* b,
			  real& kinetic_energy,
			  real& potential_energy,
			  real& kT)
{
    // Energies (top-level nodes):

    if (kinetic_energy == 0)
      get_energies(b, 0.0, kinetic_energy, potential_energy);

    real total_energy = kinetic_energy + potential_energy;
    real virial_ratio = -kinetic_energy / potential_energy;
    kT = -total_energy / (1.5*b->n_daughters());

    cerr << "    top-level nodes:\n";
    cerr << "        kinetic_energy = " << kinetic_energy
	 << "  potential_energy = " << potential_energy << endl;
    cerr << "        total_energy = " << total_energy
	 << "  kT = " << kT << "  virial_ratio = " << virial_ratio
	 << endl;

    cerr << "    total energy (full system) = "
 	 << system_energy(b) << endl;
}

local void search_for_binaries(dyn* b, real energy_cutoff, real kT,
			       vector center, bool verbose)
{
    // Search for bound binary pairs among top-level nodes.

    bool found = false;

    for_all_daughters(dyn, b, bi)
	for_all_daughters(dyn, b, bj) {
	    if (bj == bi) break;

	    real E = get_total_energy(bi, bj);

	    // Convention: energy_cutoff is in terms of kT if set,
	    //		   in terms of E/mu if kT = 0.

	    if ((kT > 0 && E < -energy_cutoff*kT)
		|| (kT <= 0 && E * (bi->get_mass() + bj->get_mass())
		    	< -energy_cutoff * bi->get_mass() * bj->get_mass())) {

		print_binary_from_dyn_pair(bi, bj, kT, center, verbose);
		
		found = true;
	    }
	}    

    if (!found) cerr << "    (none)\n";
}

// Convenient to keep this global:

#define P_MAX	0.1
#define NP	12
#define NR	10

static int histogram[NP][NR];	// first index:  log period
				// second index: radius

local void bin_recursive(dyn* bi, real rcore, real rhalf)
{
    // (Note: inner and outer multiple components are counted
    //	      as separate binaries.)

    for_all_nodes(dyn, bi, bb) {
	dyn* od = bb->get_oldest_daughter();
	if (od) {

	    // Get period of binary with CM bb.

	    bool del_kep = false;

	    if (!od->get_kepler()) {
		new_child_kepler(bb);
		del_kep = true;
	    }

	    real P = od->get_kepler()->get_period();

	    if (del_kep) {
		delete od->get_kepler();
		od->set_kepler(NULL);
	    }

	    real R = abs(something_relative_to_root(bb, &dyn::get_pos));

	    // Bin by P and R.

	    if (P > 0) {

		// P bins are half dex, decreasing.
		// P > P_MAX in bin 0, P_MAX >= P > 0.316 P_MAX in bin 1, etc.

		int ip = (int) (1 - 2*log10(P/P_MAX));
		if (ip < 0) ip = 0;
		if (ip >= NP) ip = NP - 1;

		int ir = 0;
		if (R > rcore) {

		    // Bin 0 is the core.
		    // Linear binning in rcore < r < rhalf, logarithmic
		    // outside rhalf (x2); bin NR/2 has rhalf <= R < 2*rhalf.

		    if (R < rhalf) {
			real x = R/rhalf;
			ir = (int) (NR*x/2);
			if (ir < 1) ir = 1;
		    } else {
			real rlim = 2*rhalf;
			ir = NR/2;
			while (R > rlim) {
			    if (ir >= NR-1) break;
			    rlim *= 2;
			    ir++;
			}
		    }
		}

		// Simply count binaries, for now.

		histogram[ip][ir]++;
	    }
	}
    }	
}

local void print_histogram()
{
    cerr << endl
	 << "  Binary distribution by period and radius:" 
	 << endl << endl;

    for (int jr = NR-1; jr >= 0; jr--) {

	if (jr == NR/2)
	    cerr << "      Rhalf-> ";
	else if (jr == 0)
	    cerr << "       core-> ";
	else
	    PRI(14);

	for (int ip = NP-1; ip >= 0; ip--)
	    fprintf(stderr, "%5d", histogram[ip][jr]);
	cerr << endl;
    }

    cerr << endl;
    PRI(18); cerr << "<";
    PRI((NP/2-1)*5-6); cerr << "period (0.5 dex)";
    PRI((NP/2)*5-11); cerr << "> " << P_MAX << endl;

}

local void print_binaries(dyn* b, real kT,
			  vector center, real rcore, real rhalf,
			  bool verbose, void (*dstar_params)(dyn*) = NULL)
{
    // Print out the properties of "known" binaries (i.e. binary subtrees),
    // and get some statistics on binaries in the core.  Note that *all*
    // binary subtrees are printed, regardless of energy.
    // Node b is the root node.

    // Printout occurs in four passes:	(0) perturbed multiples
    //					(1) unperturbed multiples
    //					(2) perturbed binaries
    //					(3) unperturbed binaries

    real eb = 0;
    real nb = 0;
    int n_unp = 0;
    real e_unp = 0;

    real mbcore = 0;
    int nbcore = 0;

    for (int ip = 0; ip < NP; ip++)
	for (int jr = 0; jr < NR; jr++)
	    histogram[ip][jr] = 0;

    bool found = false;

    for (int pass = 0; pass < 4; pass++)
    for_all_daughters(dyn, b, bi) {

	dyn* od = bi->get_oldest_daughter();

	if (od &&
	    (pass == 0 && bi->n_leaves()  > 2 && od->get_kepler() == NULL) ||
	    (pass == 1 && bi->n_leaves()  > 2 && od->get_kepler() != NULL) ||
	    (pass == 2 && bi->n_leaves() == 2 && od->get_kepler() == NULL) ||
	    (pass == 3 && bi->n_leaves() == 2 && od->get_kepler() != NULL)) {

	    if (verbose) {
		if (od->get_kepler() == NULL)
		    cerr << "    ";
		else
		    cerr << "  U ";
		bi->pretty_print_node(cerr);
	    }

	    if (bi->n_leaves() > 2) {

		if (verbose) cerr << "   is a multiple system:\n";

		// Recursively print out the structure :

		eb += print_structure_recursive(bi,
						dstar_params,
						n_unp, e_unp,
						kT, verbose);

		bin_recursive(bi, rcore, rhalf);

	    } else {

		int init_indent = BIN_INDENT - 4 - strlen(bi->format_label());

		if (verbose) {
		    if (od->get_kepler()) {
			PRI(init_indent);
			cerr << "is unperturbed (has a kepler pointer)" << endl;
			init_indent = BIN_INDENT;
		    }
		}

		bool reset = false;
		if (od->get_kepler() == NULL) {
		    new_child_kepler(bi);	// attach temporary kepler
		    reset = true;		// to oldest daughter
		}

		real dist_from_cod = abs(bi->get_pos()-center);

		eb += print_binary_params(od->get_kepler(), od->get_mass(),
					  kT, dist_from_cod, verbose,
					  init_indent);
		nb++;

		// Indirect reference to dstar output:

		if (dstar_params != NULL)
		    dstar_params(bi);

		if (rcore > 0) {
		    if (dist_from_cod <= rcore) {
			nbcore++;
			mbcore += bi->get_mass();
		    }
		}

		bin_recursive(bi, rcore, rhalf);

		if (reset) {
		    delete od->get_kepler();
		    od->set_kepler(NULL);
		}

		// Virtual reference to additional binary output.
		// Must do this *after* temporary keplers have been removed.

		real e = od->print_pert();
		if (e != 0) {
		    n_unp++;
		    e_unp += e;
		}
	    }

	    found = true;
	}
    }

    if (!found)

	cerr << "    (none)\n";

    else {

	if (verbose) cerr << endl;
	cerr << "  Total binary energy ("
	     << nb << " binaries) = " << eb << endl;

	if (n_unp != 0) {
	    cerr << "  Total unperturbed binary energy ("
		 << n_unp << " binaries) = " << e_unp << endl;
	}

	if (rcore > 0)
	    cerr << "  nbcore = " << nbcore
		 << "  mbcore = " << mbcore << endl;

	print_histogram();
    }
}

local void print_core_parameters(dyn* b, bool allow_n_sq_ops,
				 vector& density_center, real& rcore)
{
    real mcore;
    int ncore;
    
    compute_core_parameters(b, 12, allow_n_sq_ops,
			    density_center, rcore, ncore, mcore);

    cerr << "    density_center = " << density_center << endl;
    cerr << "    rcore = " << rcore << "  ncore = " << ncore
	 << "  mcore = " << mcore << endl;

    // Place the data in the root dyn story.

    putrq(b->get_dyn_story(), "core_radius_time", b->get_system_time());
    putrq(b->get_dyn_story(), "core_radius", rcore);
    putiq(b->get_dyn_story(), "n_core", ncore);
    putrq(b->get_dyn_story(), "m_core", mcore);
}

// Linear interpolation routine.

local real linear_interpolation(const real x,
				const real x1, const real x2,
				const real y1, const real y2) {

        real a = (y2-y1)/(x2-x1);
	real b = y1 - a*x1;

	real y = a*x + b;
	return y;
}

#if 0
// Currently in /star/util/kingfit.C   

// Fitting function for N-body Rc/Rvir to King's concentration parameter c. 
// Argument is core radius / virial radius.

local real kingCfit(const real rc_rvir) {

  int nfit = 12;
  real WoKing[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  real Cking[] = {0., 0.295643, 0.505349, 0.673898, 0.840487,
		  1.03059, 1.25598, 1.5284, 1.83414, 2.12028,
		  2.35155, 2.5495, 2.7396};
  real Rc_Rvir[] = {1., 0.497, 0.492, 0.428, 0.376, 0.314,
		    0.241, 0.179, 0.110, 0.0568, 0.0369,
		    0.0185, 0.0129};
  
  if (rc_rvir>=Rc_Rvir[1])
    return 0;
  else if(rc_rvir<=Rc_Rvir[nfit])
    return 3;
  else
    for (int i=1; i<=nfit; i++) 
      if (rc_rvir>=Rc_Rvir[i])
	return linear_interpolation(rc_rvir,
				    Rc_Rvir[i-1], Rc_Rvir[i],
				    Cking[i-1], Cking[i]);
  return 0;
}

// Fitting function for N-body Rc/Rvir to King's dimentionless Wo.
// Argument is core radius / virial radius.

local real kingWfit(const real rc_rvir) {

  int nfit = 12;
  real WoKing[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  real Rc_Rvir[] = {1., 0.497, 0.492, 0.428, 0.376, 0.314,
		    0.241, 0.179, 0.110, 0.0568, 0.0369,
		    0.0185, 0.0129};

  if (rc_rvir>=Rc_Rvir[1])
    return 1;
  else if(rc_rvir<=Rc_Rvir[nfit])
    return 12;
  else
    for (int i=1; i<=nfit; i++) 
      if (rc_rvir>=Rc_Rvir[i])
	return  linear_interpolation(rc_rvir,
				     Rc_Rvir[i-1], Rc_Rvir[i],
				     WoKing[i-1], WoKing[i]);
  return 1;
}

local void print_king_parameters(const real rc_rvir) {

  real Wo = kingWfit(rc_rvir);
  real C  = kingCfit(rc_rvir);
  
    PRI(4); cerr << "Wo = " << Wo 
		 << "  c = " << C
		 << endl;
  
}

#endif

local bool testnode(dyn * b)
{
    if (b->is_top_level_node())
	return true;
    else
	return false;
}

local bool single_fn(dyn * b)
{

  //  int test = getiq(b->get_log_story(), "mass_doubled");
  //  PRL(test);
  if (getiq(b->get_log_story(), "mass_doubled") == 0 )
    return true;
  else
    return false;
}

local bool double_fn(dyn * b)
{
  if (getiq(b->get_log_story(), "mass_doubled") == 1 )
    return true;
  else
    return false;
}  

local real print_lagrangian_radii(dyn* b, int which_lagr,
				  bool verbose = true, int which = 0)
{
    char tmp[3];
    bool nonlin = false;

    real rhalf = 1;
    int ihalf;

    int nl, indent;
    if (which_lagr == 0) {
	nl = 4;
	indent = 15;
	ihalf = 1;
    } else if (which_lagr == 1) {
	nl = 10;
	indent = 20;
	ihalf = 4;
    } else {
	nl = 10;
	indent = 26;
	nonlin = true;
	ihalf = 4;
    }

    if (which == 0)
	compute_general_mass_radii(b, nl, nonlin);

    if (which == 2)
	compute_general_mass_radii(b, nl, nonlin, single_fn);

    if (which == 3)
	compute_general_mass_radii(b, nl, nonlin, double_fn);

    if (find_qmatch(b->get_dyn_story(), "n_lagr")
	&& getrq(b->get_dyn_story(), "lagr_time") == b->get_system_time()) {

	vector lagr_pos = getvq(b->get_dyn_story(), "lagr_pos");

	if (verbose) {
	    cerr << endl << "  Lagrangian radii relative to ("
		 << lagr_pos << "):" << endl;
	    if (find_qmatch(b->get_dyn_story(), "pos_type"))
		cerr << "                               ( = "
		     << getsq(b->get_dyn_story(), "pos_type")
		     << ")" << endl;
	    if (which_lagr == 0)
		cerr << "    quartiles: ";
	    else if (which_lagr == 1)
		cerr << "    10-percentiles: ";
	    else
		cerr << "    selected percentiles: ";
	}

        int n_lagr = getiq(b->get_dyn_story(), "n_lagr");  // should be nl - 1
	real *r_lagr = new real[n_lagr];

	getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

	for (int k = 0; k < n_lagr; k += 5) {
	    if (k > 0) {
		cerr << endl;
		for (int kk = 0; kk < indent; kk++) cerr << " ";
	    }
	    for (int i = k; i < k+5 && i < n_lagr; i++)
		cerr << " " << r_lagr[i];
	}
	cerr << endl << flush;

	rhalf = r_lagr[ihalf];
	delete [] r_lagr;
    }
    return rhalf;
}

void sys_stats(dyn* b,
	       real energy_cutoff,
	       bool verbose,
	       bool binaries,
	       int which_lagr,
	       bool print_time,
	       bool compute_energy,
	       bool allow_n_sq_ops,
	       void (*dstar_params)(dyn*),
	       bool (*print_dstar_stats)(dyn*, bool, vector, bool))

// The last two arguments are a nasty C way of allowing this dyn function
// to output additional hdyn/dstar information when called from kira...

{
    cerr.precision(PRECISION);

    if (print_time) {

	real time = getrq(b->get_dyn_story(), "t");
	if (time <= -VERY_LARGE_NUMBER)
	    time = b->get_system_time();

	if (time > -VERY_LARGE_NUMBER)
	    cerr << "Time = " << time << endl;
	else
	    cerr << "Time = (undefined)" << endl;

	if (b->get_use_sstar())
	    print_sstar_time_scales(b);
    }

    if (verbose) cerr << "\n  Overall parameters:\n";
    bool mass_spectrum = print_numbers_and_masses(b);

    vector com_pos, com_vel;
    compute_com(b, com_pos, com_vel);
    cerr << "    center of mass position = " << com_pos << endl
	 << "                   velocity = " << com_vel << endl;

    bool heavy_stars = check_for_doubling(b);
    // PRI(4); PRL(heavy_stars);

    real kinetic_energy = 0, potential_energy = 0;

    real r_virial = -1;
    if (compute_energy && allow_n_sq_ops)	// Need virial radius for tR
	r_virial  = print_relaxation_time(b, kinetic_energy, potential_energy);

    if (r_virial == -1) {
        if (find_qmatch(b->get_dyn_story(), "energy_time")
	    && getrq(b->get_dyn_story(), "energy_time")
	                         == b->get_system_time()) {

	    // Take energies from the dyn story.

	    if (find_qmatch(b->get_dyn_story(), "kinetic_energy")
		&& find_qmatch(b->get_dyn_story(), "kinetic_energy")) {

		kinetic_energy = getrq(b->get_dyn_story(), "kinetic_energy");
		potential_energy = getrq(b->get_dyn_story(),
					 "potential_energy");
		if (potential_energy < 0)
		    r_virial = -0.5*b->get_mass()*b->get_mass()
					/ potential_energy;

	    }
	}
    }

    // Note: even if allow_n_sq_ops is false, we can still compute kT
    // from the kinetic energy.

    real kT = 0;

    if (compute_energy && allow_n_sq_ops)	// Recompute energies
        print_energies(b, kinetic_energy, potential_energy, kT);
    else if (b->get_oldest_daughter())
	kT = top_level_kinetic_energy(b) / (1.5*b->n_daughters());

    vector center = 0;
    real rcore = 0;

    bool has_densities = (getrq(b->get_dyn_story(), "density_time")
				 == b->get_system_time());

    if (has_densities || allow_n_sq_ops) {

	// Function compute_core_parameters will look in the dyn story
	// for densities and use them if they are current.  It will
	// only do n_sq operations if no current densities are found
	// and allow_n_sq_ops is true.

        if (verbose) {
	    cerr << "\n  Core parameters";
	    if (has_densities)
		cerr << " (densities from input snapshot)";
	    cerr << ":\n";
	}
        print_core_parameters(b, allow_n_sq_ops, center, rcore);

	if (rcore > 0 && r_virial > 0) {
            if (verbose) cerr << "\n  Dynamical King parameters:\n";
	    print_fitted_king_model(rcore/r_virial, rcore_rvirial);
           //print_king_parameters(rcore/r_virial); // Old version
	}
    }

    // If core parameters were computed, then center now is the mean
    // density center.  If not, then center is the geometric center.
    // However, this value of center is presently not used, as the
    // vector is set equal to lagr_pos below.

    // variable which:     which = 0 : single stars
    //                           = 1 : binaries
    //                           = 2 : light stars (heavy_stars = true)
    //                           = 3 : heavy stars (heavy_stars = true)

    real rhalf = print_lagrangian_radii(b, which_lagr, verbose, 0);

    // cerr << "first Lagrangian radii calc, all stars" << endl;

    // PRL(heavy_stars);

    if (verbose) {
        if (heavy_stars) {
	    cerr << "\n  Lagrangian radii (light stars):\n";
	    print_lagrangian_radii(b, which_lagr, verbose, 2);
	    cerr << "\n  Lagrangian radii (heavy stars):\n";
	    print_lagrangian_radii(b, which_lagr, verbose, 3);
	}
    }

    if (mass_spectrum) {

	if (verbose)
	    cerr << "\n  Mass distribution by Lagrangian zone (singles):\n";
	print_numbers_and_masses_by_radial_zone(b, 0);

	if (verbose)
	    cerr << "\n  Mass distribution by Lagrangian zone (multiples):\n";
	print_numbers_and_masses_by_radial_zone(b, 1);

    }

    if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (singles):\n";
    print_anisotropy_by_radial_zone(b, 0);

    if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (multiples):\n";
    print_anisotropy_by_radial_zone(b, 1);

    if (heavy_stars && verbose) {
	// if (verbose)
	cerr << "\n  Mass distribution by Lagrangian zone (light stars):\n";
	print_numbers_and_masses_by_radial_zone(b, 2);

	// if (verbose)
	cerr << "\n  Mass distribution by Lagrangian zone (heavy stars):\n";
	print_numbers_and_masses_by_radial_zone(b, 3);

	// if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (light stars):\n";
	print_anisotropy_by_radial_zone(b, 2);

	// if (verbose)
        cerr << "\n  Anisotropy by Lagrangian zone (heavy stars):\n";
	print_anisotropy_by_radial_zone(b, 3);
    }

    bool sstar = b->get_use_sstar();

    if (print_dstar_stats != NULL)
	sstar = !print_dstar_stats(b, mass_spectrum, center, verbose);

    if (sstar)
	  sstar_stats(b, mass_spectrum, center, verbose);

    if (binaries) {

	// Use the same center as used by the Lagrangian radii functions.

	if (find_qmatch(b->get_dyn_story(), "lagr_pos"))
	    center = getvq(b->get_dyn_story(), "lagr_pos");
	else
	    warning("sys_stats: lagr_pos not found");

	set_kepler_tolerance(1);

	if (verbose) cerr << "\n  Binaries/multiples:\n";
	print_binaries(b, kT, center, rcore, rhalf,
		       verbose, dstar_params);

	if (allow_n_sq_ops) {

	    if (verbose) {
		cerr << "\n  Other bound pairs with ";
		if (kT > 0)
		    cerr << "|E| > " << energy_cutoff << " kT:\n";
		else
		    cerr << "|E/mu| > " << energy_cutoff << ":\n";
	    }
	    search_for_binaries(b, energy_cutoff, kT, center, verbose);

	} else {

	    // Do an NN search for binaries (hdyn only).

	    int which = 0;
	    bool found = false;
	    for_all_daughters(dyn, b, bb)
		found |= bb->nn_stats(energy_cutoff, kT, // Dummy function for
				      center, verbose,   // dyn; get NN binary
				      which++);		 // info for hdyn
	    if (!found)
		cerr << "    (none)\n";
	}
    }
}

#else

main(int argc, char **argv)
{
    bool binaries = true, verbose = true, out = false,
         n_sq = true, calc_e = true;
    int which_lagr = 0;

    real T_start = 0;
    
    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "bneost";	// Note: "v" removed because only the
					// "verbose" option currently works.

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'b': binaries = !binaries;
		      break;
	    case 'e': calc_e = !calc_e;
		      break;
	    case 'n': n_sq = !n_sq;
		      break;
	    case 'o': out = !out;
		      break;
	    case 's': which_lagr = 2;
		      break;
	    case 't': which_lagr = 1;
		      break;
	    case 'v': verbose = !verbose;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	              get_help();
		      exit(1);
	}

    // Loop over input until no more data remain.

    dyn *b;
    int i = 0;
    while (b = get_dyn(cin)) {

        // NOTE:  get_xxx() reads NaN in as legal garbage...

      if(find_qmatch(b->get_oldest_daughter()
		->get_starbase()->get_star_story(), "Type")) {
	
	addstar(b,                            // Note that T_start and
		T_start,                       // Main_Sequence are
		Main_Sequence,                 // defaults. They are
		true);                         // ignored if a star
	b->set_use_sstar(true);                // is there.
	
      }
      
	if (i++ > 0) cerr << endl;
	sys_stats(b, 0.5, verbose, binaries,
		  which_lagr, true, calc_e, n_sq);

	if (out) put_node(cout, *b);
	rmtree(b);
    }
}

#endif
