
//// mksecondary_spz:  Create binary secondary components for randomly
////                   selected stars in the input snapshot, placing the
////                   results in a binary tree.  Only secondary masses
////                   are set here; orbital parameters are set by
////                   dyn::mkbinary.
////
//// Options:      -q    select choice of minimum mass ratio [false]
////                         if true, secondary mass ratio is chosen
////                             uniformly on [lower_limit, 1]
////                         if false, secondary mass ratio is chosen
////                             uniformly on [mmin, primary_mass], where
////                             mmin is the minimum of the lower limit
////                             specified on the command line and the
////                             minimum primary mass
////               -f    specify binary fraction [0.1]
////               -l    specify lower limit on mass ratio or secondary mass [0]
////               -s    specify random seed [random from system clock]

//		   Steve McMillan, July 1996
//                 Simon Portegies Zwart, Tokyo, December 1997

#include "node.h"

#define  SEED_STRING_LENGTH  256
char  tmp_string[SEED_STRING_LENGTH];

#ifdef TOOLBOX

// add_secondary: Add a secondary component to a primary to make
//		  a binary.  Only the mass, name and tree structure
//		  are modified at this stage.
//		  New primary and secondary components have names "na"
//		  and "nb", respectively, where "n" was the name of
//		  the original star.

local void add_secondary(node* original, real mass_ratio)
{
    node* primary = new node;
    node* secondary = new node;

    // Add new links.

    original->set_oldest_daughter(primary);

    primary->set_parent(original);
    secondary->set_parent(original);

    primary->set_younger_sister(secondary);
    secondary->set_elder_sister(primary);

    // Set new masses.

    primary->set_mass(original->get_mass());
    secondary->set_mass(mass_ratio*original->get_mass());
    original->inc_mass(secondary->get_mass());

    // Naming convention:

    if (original->get_name() == NULL)
	if (original->get_index() >= 0) {
	    char tmp[64];
	    sprintf(tmp, "%d", original->get_index());
	    original->set_name(tmp);
	}

    if (original->get_name() != NULL) {

	// Label components "a" and "b".

	primary->set_name(original->get_name());
	secondary->set_name(original->get_name());
	strcat(primary->get_name(), "a");
	strcat(secondary->get_name(), "b");

	// Make standard CM name.

	char tmp[256];
	sprintf(tmp, "(%s,%s)", primary->get_name(), secondary->get_name());
	original->set_name(tmp);
    }
}

local void warning(real mmin, real lower_limit) {

  cerr << "WARNING:\n";
  cerr << "       local void mksecondary(node*, real, real, bool)\n";
  cerr << "       minimum mass "<<mmin<<" larger than assumed lower limit "
       << lower_limit << endl;
  cerr << "       Continue execution with minimum mass.\n";

}

local real random_mass_ratio(const real qmin) {

  // For now, equal probability in q between qmin and 1.

  return randinter(qmin, 1.);
}

local void mksecondary(node* b, real binary_fraction,
		       real lower_limit, bool q_flag)
{
    // For now, use a flat distribution in secondary mass ratio.
    // Assume that the probability of a star being the primary of
    // a binary is independent of mass.

    real sum = 0;
    real m_prim = b->get_mass();
    b->set_mass(0);

    if(q_flag) {

	// Secondary mass ratio chosen uniformly on [lower_limit, 1].

	for_all_daughters(node, b, bi) {
	    sum += binary_fraction;
	    if (sum >= 1) {
		sum -= 1;

		add_secondary(bi, random_mass_ratio(lower_limit));
	    }
	    b->inc_mass(bi->get_mass());
	}
    } else {

	// Secondary mass ratio chosen uniformly on [mmin, primary_mass],
	// where mmin is the minimum of the lower limit specified on the
	// command line and the minimum primary mass.

	real mmin = 100, qmin = 0;
	for_all_daughters(node, b, bi)
	    mmin = min(mmin, bi->get_mass());

	if (mmin < lower_limit) {
	    warning(mmin, lower_limit);
	    sprintf(tmp_string,
		    "         -l = %5.4f adopted", mmin); 
	    b->log_comment(tmp_string);
	}
	else
	    mmin = lower_limit;

	for_all_daughters(node, b, bi) {
	    sum += binary_fraction;
	    if (sum >= 1) {
		sum -= 1;

		qmin = mmin/bi->get_mass();
		add_secondary(bi, random_mass_ratio(qmin));
	    }
	    b->inc_mass(bi->get_mass());
	}
    }
    real m_tot = b->get_mass();
    real m_sec = m_tot - m_prim;
    real old_mtot = b->get_starbase()->conv_m_dyn_to_star(1);
    if(old_mtot>0 && old_mtot<m_tot) {
      real old_r_vir= b->get_starbase()->conv_r_star_to_dyn(1);
      real old_t_vir= b->get_starbase()->conv_t_star_to_dyn(1);
      b->get_starbase()->set_stellar_evolution_scaling(m_tot,
						       old_r_vir,
						       old_t_vir);
    }

    sprintf(tmp_string,
	    "         total mass in secondaries = %8.2f Solar", m_sec); 
    b->log_comment(tmp_string);
}

void main(int argc, char ** argv)
{
    real binary_fraction = 0.1;
    real lower_limit = 0.0;
    real q_flag = false;	     // flag for mass-ratio selection
    int random_seed = 0;
    char seedlog[64];

    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "qf:l:s:";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

            case 'q': q_flag = true;
		      break;
	    case 'f': binary_fraction = atof(poptarg);
		      break;
	    case 'l': lower_limit = atof(poptarg);
		      break;
	    case 's': random_seed = atoi(poptarg);
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      exit(1);
	}

    if (binary_fraction < 0 || binary_fraction > 1)
	err_exit("mkbinary: Illegal binary fraction");
    if (q_flag && (lower_limit < 0 || lower_limit > 1))
	err_exit("mkbinary: Illegal minimum mass for secondary");

    node* b;
    b = get_node(cin);

    b->log_history(argc, argv);

    int actual_seed = srandinter(random_seed);

    sprintf(seedlog, "         random number generator seed = %d",actual_seed);
    b->log_comment(seedlog);

    mksecondary(b, binary_fraction, lower_limit, q_flag);

    put_node(cout, *b);
    rmtree(b);

}
#endif
