
// tree_io.C
//
//	The most important (pipeable) output routines have been
//	modified to work on HP/HPUX-8.x -- S. McMillan, 3/3/93
//
//	Modified to read/write compressed snapshots (cout only),
//	under control of environment variable STARLAB_ZIP.
//				-- S. McMillan, X. Zhuge, 6/22/98
//	Minor leak fix		-- PJT , 11/24/98

#include "starlab_vector.h"
#include "util_io.h"
#include "story.h"
#include "node.h"

void  node::log_comment(char * comment)
    {
    add_story_line(log_story, comment);
    }

void  node::log_history(int argc, char ** argv)	
    {
    char *hist = gethist(argc, argv);
    add_story_line(log_story, hist);
    delete hist;
    }

ostream& node::print_log_story(ostream& s)
{
    story * the_log_story = node::get_log_story();
    if (the_log_story)
        put_story(s, *the_log_story);
    return s;
}

ostream& node::print_hydro_story(ostream& s)
{
    hbase->print_hydro_story(s);
    return s;
}

ostream& node::print_star_story(ostream& s)
{
    sbase->print_star_story(s);
    return s;
}

istream & node::scan_log_story(istream& s, char * line)
{
    node::set_log_story(get_story(s, line));
    return s;
}

istream & node::scan_hydro_story(istream& s)
{
    hbase->scan_hydro_story(s);
    return s;
}

istream & node::scan_star_story(istream& s, int level)
{
    sbase->scan_star_story(s, level);
    return s;
}

#include <string.h>
#define BUF_SIZE 1024

static char format_string[BUF_SIZE];

// Note added by Steve 7/6/98.  Do NOT use format_label() to print out
// labels of more than one particle in the same print statement if the
// particles are identified by index rather than by label.  Because we
// use a single string here, only one format can be stored at any
// time.  Therefore, since all calls to format_label() will be executed
// before the string is sent to the output stream, a statement of the
// form
//
//	cerr << bi->format_label() << bj->format_label() << endl;
//
// will actually print out the label of bj twice!
//
// Better to use print_label() instead in those circumstances.

char* node::format_label()
{
    if (is_valid()) {
	if (name != NULL) {
	    strncpy(format_string, name, BUF_SIZE-1);
	    format_string[BUF_SIZE-1] = '\0';
	} else if(index >= 0) {
	    sprintf(format_string, "%d", index);	// SLWM removed leading
							// "#", July 1998
	} else {
	    sprintf(format_string, "?");
	}
    } else
	sprintf(format_string, "(invalid)");
	
    return format_string;
}

bool node::name_is(char* s)
{
    return streq(format_label(), s);
}

void node::print_label(ostream & s)
{
    s << format_label();
}

void node::pretty_print_node(ostream & s)
{
    print_label(s);
}

void node::pretty_print_tree(int depth_level, ostream & s)
    {
    int  k = depth_level;
    while (k--)
	s << "  ";
    pretty_print_node(s);
    if (mass != 1)
	s << "        m = " << mass;
    s << endl;
    if (is_parent())
	for_all_daughters(node, this, d)
	    d->pretty_print_tree(depth_level + 1, s);
    }

void node::pretty_print_tree(ostream & s)
    {
    pretty_print_tree(0, s);
    }

void pp(node * b, ostream & s)
{
    s << "(";
    b->pretty_print_node(s);
    for_all_daughters(node, b, daughter)
	pp(daughter, s);	
    s << ")";
}

void pp2(node * b, ostream & s,  int level)
{
    for (int i = 0; i<level*2; i++) {s << " ";}
    b->pretty_print_node(s);
    s << "\n";
    for_all_daughters(node, b, daughter)
	pp2(daughter, s, level + 1);	
}

local node * get_node_recursive(istream& s,
				npfp the_npfp,
				hbpfp the_hbpfp,
				sbpfp the_sbpfp,
				int level)
{
    node * b = (*the_npfp)(the_hbpfp, the_sbpfp);
    char line[MAX_INPUT_LINE_LENGTH];

    get_line(s, line);

    node * elder_sister = (node *)42;	// to make some compilers happy

    while (strcmp(")Particle",line)) {
	if (strcmp("(Dynamics", line) == 0) {
	    b->scan_dyn_story(s);
	} else if (strcmp("(Hydro", line) == 0) {
	    b->scan_hydro_story(s);
	} else if (strcmp("(Star", line) == 0) {
	    b->scan_star_story(s, level);
	} else if (strcmp("(Log", line) == 0) {
                // bug: every node gets a log story, but when you see
                // one from input, you set this to be the new one
                // and thus never dealloc the old one ???
	    b->scan_log_story(s, line);
	} else if (strcmp("(Particle", line) == 0) {
	    node * daughter =
		get_node_recursive(s, the_npfp, the_hbpfp, the_sbpfp, level+1);
	    if (b->get_oldest_daughter() == NULL) {
		b->set_oldest_daughter(daughter);
	    } else {
		daughter->set_elder_sister(elder_sister);
		elder_sister->set_younger_sister(daughter);
	    }
	    daughter->set_parent(b);
	    elder_sister = daughter;
	} else {
	    char keyword[MAX_INPUT_LINE_LENGTH];
	    char should_be_equal_sign[MAX_INPUT_LINE_LENGTH];
	    sscanf(line,"%s%s",keyword,should_be_equal_sign);
	    if (strcmp("=",should_be_equal_sign)) {
	        cerr << "Expected '=', but got '"
		     << should_be_equal_sign <<"'\n";
	    	exit(1);
	    }
    	    if (!strcmp("i",keyword)) {
		int index;
		sscanf(line,"%*s%*s%d",&index);
		b->set_label(index);
	    } else if (!strcmp("name",keyword)) {
 	        char cptr[MAX_INPUT_LINE_LENGTH];
		sscanf(line,"%*s%*s%s",cptr);
	        b->set_label(cptr);
	    } else if (!strcmp("N",keyword)) {   // N is not read in here;
		;                                // instead N will be computed
	    }                                    // anew at output time.
	    else{
		cerr << line <<" unexpected\n";
	        exit(1);
	    }
        }
        get_line(s, line);
    }
    return b;
}

local node * get_node_init(istream& s,
			   npfp the_npfp,
			   hbpfp the_hbpfp,
			   sbpfp the_sbpfp)
{

    if (!check_and_skip_input_line(s, "(Particle")) {
	return NULL;
    }
    return get_node_recursive(s, the_npfp, the_hbpfp, the_sbpfp, 0);
}

local void put_node_recursive(ostream & s, node & b)
{
    put_story_header(s, "Particle");

    if (b.get_index() >= 0) put_integer(s, "  i = ", b.get_index());
    if (b.get_name() != NULL) put_string(s, "  name = ", b.get_name());
    put_integer(s, "  N = ", b.n_leaves());

    b.print_log_story(s);	// HP OK -- story.C
    b.print_dyn_story(s);       // HP OK -- node_io.C
    b.print_hydro_story(s);     // HP OK -- *hydro_io.C
    b.print_star_story(s);      // HP OK -- *star_io.C

    for(node * daughter = b.get_oldest_daughter();
	daughter != NULL;
	daughter = daughter->get_younger_sister()){
	put_node_recursive(s, *daughter);	
    }

    put_story_footer(s, "Particle");
}

#ifdef HAS_GZIP
#include <pfstream.h>	// (May not exist on all systems...)
#endif

node * get_node(istream& s, npfp the_npfp, hbpfp the_hbpfp, sbpfp the_sbpfp)
{
    // If STARLAB_USE_GZIP is defined, the input will be decompressed
    // by gzip first.  This is accomplished by redefining the stream s.

    // Note that, for compression to be carried out, we must (1) compile
    // HAS_GZIP set, then (2) run with STARLAB_USE_GZIP defined.

#ifdef HAS_GZIP
    if (char * zip = getenv("STARLAB_USE_GZIP")) {
	ipfstream sz("|gzip -d -f");
	if (sz) {
	    return get_node_init(sz, the_npfp, the_hbpfp, the_sbpfp);
	}
    }
#endif

    return get_node_init(s, the_npfp, the_hbpfp, the_sbpfp);
}

void put_node(ostream & s, node & b)
{
    // If STARLAB_USE_GZIP is set, everything written to cout will be
    // compressed by gzip.

#ifdef HAS_GZIP
    if (&s == (ostream *) &cout) {
	if (char * zip = getenv("STARLAB_USE_GZIP")) {
	    opfstream sz("|gzip -c -f");
	    if (sz) {
		put_node_recursive(sz, b);
		return;
	    }
	}
    }
#endif

    put_node_recursive(s, b);
}

local void forget_node_recursive(istream& s)
    {
    char line[MAX_INPUT_LINE_LENGTH];

    get_line(s, line);

    while(strcmp(")Particle",line))
	{
	if(strcmp("(Particle", line) == 0)
	    forget_node_recursive(s);
	get_line(s, line);
	}
    }

// forget_node: reads in a complete node structure, just as get_node does, but 
//              without storing anything.  This is useful in the function
//              snapprune, which prunes a long list of snapshots.
//              Piet, 941125.

bool forget_node(istream& s)
{
    if(!check_and_skip_input_line(s,"(Particle")){
	return FALSE;
    }
    forget_node_recursive(s);
    return TRUE;
}

