
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//// addstar: Adds star class to excisting node structure.
////
//// Options:    -A    maximum initial semi-major axis [in solar units]
////             -a    minimum initial semi-major axis [in solar units]
////             -c    comment to put in the starbase log structure.
////             -E    maximum initial eccentricity
////             -e    minimum initial eccentricity
////             -M    Mass scaling [total cluster mass in solar units].
////             -Q    Virial ratio (used if time scaling is omitted)
////             -R    Dynamical size scaling for the star
////                   [in units of the virial radius].
////             -S    use internal scaling?
////             -s    Random seed.
////             -T    Dynamical time scaling
////                   [in units of the NBODY time].
////             -t    start time [in million years].
////
//++ Notes:
//++  Online arguments over rule parameters from input node.
//++
//++  If no time unit is explicitly specified it is derived
//++  from the other units.
//++
//++  The mass of the system can be set independedly from the total
//++  mass of the input N-body system (e.g. when modeling a large
//++  cluster by a smaller N-body system).
//++
//++ Example of usage:      
//++  adddouble
//++
//++ See also: addstar

//// Version 1.0: Feb 1995   Simon Portegies Zwart
//                           spz@grape.c.u-tokyo.ac.jp

#include "double_star.h"
#include "hdyn.h"
#include "util_io.h"
#include "dstar_to_dyn.h"

#ifndef TOOLBOX

#define MASS_UPDATE_LIMIT 0.0001
#define SMA_UPDATE_LIMIT 0.0001

#define REPORT_ADD_DOUBLE        false
#define REPORT_DEL_DOUBLE        false
#define REPORT_EVOLVE_DOUBLE     false

//This is a challenge, does it improve bugginess.

#define TRY_THIS true

local void evolve_an_unperturbed_binary(hdyn *bi, 
				  bool& check_binary_mass,
				  bool& check_binary_sma,
				  bool& check_binary_merger)
{
    check_binary_mass = check_binary_sma = check_binary_merger = false;
    
    real old_dyn_mass, old_dyn_sma;
    real new_dyn_mass_from_star, new_dyn_sma_from_star;
    
    //    real stellar_evolution_time =
    //      bi->get_starbase()->conv_t_dyn_to_star(((hdyn*)bi)->get_time());
    //    stellar_evolution_time = bi->get_starbase()->get_current_time();
    
    kepler * p_kep = bi->get_oldest_daughter()->get_kepler();
    old_dyn_mass = bi->get_mass();
    old_dyn_sma  = p_kep->get_semi_major_axis();
    real stellar_evolution_time = bi->get_starbase()
                                    ->conv_t_dyn_to_star(p_kep->get_time());

    real binary_age = max(bi->get_oldest_daughter()->get_starbase()
			  ->get_current_time(),
			  bi->get_oldest_daughter()->get_binary_sister()
			  ->get_starbase()->get_current_time());
    //cerr << "times: " << binary_age<<" "<<stellar_evolution_time<<endl;
    

    if (REPORT_EVOLVE_DOUBLE) {
      cerr << "Before evolution:"
	   << " sma= " << bi->get_starbase()->get_semi()
	   << " ecc= " << bi->get_starbase()->get_eccentricity() << "       ";
      put_state(make_state(dynamic_cast(double_star*, bi->get_starbase())));
      cerr << endl;
      ((double_star*) (bi->get_starbase()))->dump(cerr);
      cerr << endl;
    }

    bi->get_starbase()->evolve_element(stellar_evolution_time);

    new_dyn_mass_from_star = get_total_mass(bi);
    new_dyn_sma_from_star = bi->get_starbase()->
      conv_r_star_to_dyn(bi->get_starbase()->get_semi());

//		Relative position and velocity should be changed also
//		for the upper parent nodes!!!

    if (abs(new_dyn_mass_from_star-old_dyn_mass)/old_dyn_mass
	>=MASS_UPDATE_LIMIT){
	check_binary_mass=true;
    }
    if (abs(new_dyn_sma_from_star-old_dyn_sma)/old_dyn_sma
	>=SMA_UPDATE_LIMIT){
	check_binary_sma=true;
    }
    if (binary_is_merged(bi->get_oldest_daughter())) {
      cerr << "This is apparently a merged binary"<<endl;
      check_binary_merger=true;
    }
      
    star * primary = ((star*) (bi->get_starbase()))->get_primary();
    star * secondary = ((star*) (bi->get_starbase()))->get_secondary();
    real m1 = primary->get_total_mass();
    real m2 = secondary->get_total_mass();
    real semi = bi->get_starbase()->get_semi();
    real ecc = bi->get_starbase()->get_eccentricity();
    char * t1 = type_string(primary->get_element_type());
    char * t2 = type_string(secondary->get_element_type());
    if (REPORT_EVOLVE_DOUBLE) 
      if (check_binary_sma || check_binary_mass || check_binary_merger){
	cerr << " check: "
	     <<check_binary_mass<<" "
	     <<check_binary_sma<<" "
	     <<check_binary_merger<<endl;
	bi->pretty_print_node(cerr); PRC(bi->get_system_time());
	PRC(stellar_evolution_time);
	PRC(t1); PRC(m1); PRC(t2); PRC(m2); PRC(semi); PRL(ecc);
	  cerr << "After evolution:"
	       << " sma= " << semi
	       << " ecc= " << ecc << "       ";
	  put_state(make_state(dynamic_cast(double_star*, bi->get_starbase())));
	  cerr << endl;
	  ((double_star*) (bi->get_starbase()))->dump(cerr);
	  cerr << endl;
      }
}

bool create_or_delete_binary(hdyn *bi, 
			     bool * update_dynamics,
			     real eps,		      
			     real eta,
			     real dt_max) {

  bool delete_an_existing_binary = false;
  bool create_a_new_binary       = false;
  bool procedure_successful      = false;


  if (REPORT_ADD_DOUBLE || REPORT_DEL_DOUBLE )
      cerr << "create_or_delete_binary: params= " << endl;
  
  if (has_dstar(bi->get_oldest_daughter())) {
      if (REPORT_DEL_DOUBLE )
	  cerr << "Delete existing binary from "
	       << bi->format_label()<<endl;
    delete_an_existing_binary = true;
  }
  else {
      if (REPORT_ADD_DOUBLE)
	  cerr << "Create new binary to "
	       << bi->format_label()<<endl;
    create_a_new_binary=true;
  }

  if (create_a_new_binary) {
    adddouble(bi, bi->get_time());
    procedure_successful = true;
  }

  if (delete_an_existing_binary) {
    delete_double(bi, update_dynamics, eps, eta, dt_max);
    procedure_successful = true;
  }


  return procedure_successful;

}

local real initial_binary_period(double_star * ds)
{
 
    double_init *di = ds->get_initial_conditions();

    real pdays = 2*PI*sqrt(pow(di->semi*cnsts.parameters(solar_radius), 3.) /
			   (cnsts.physics(G)*di->mass_prim
			      *(1+di->q)*cnsts.parameters(solar_mass)))
			   /  cnsts.physics(days);
			   return pdays;

}

/*-----------------------------------------------------------------------------
 *  delete_double  -- "delete double()".
 *
 *   delete the double_star attached to the parent of the leaves.
 *   Does not check whether or not a double_star indeed exists.
 *-----------------------------------------------------------------------------
 */
void delete_double(hdyn *b,
		   bool * update_dynamics,
		   real eps,		      
		   real eta,
		   real max_timestep)
{
  if (b->get_oldest_daughter()->get_kepler()==NULL) {
    cerr << "Deleting double without a kepler pointer"<<endl;
    delete_double(b);
    return;
  }
  
  bool kepler_created=false;
  
  if (!b->get_oldest_daughter()->get_kepler()) {
      cerr << "No Kepler in delete_double; create one."<<endl;
      new_child_kepler(b, b->get_time(), MAX_CIRC_ECC);
      kepler_created=true;
  }
  else {
      b->get_oldest_daughter()->get_kepler()->
         set_circular_binary_limit(MAX_CIRC_ECC);	// Probably unnecessary
      dyn_to_child_kepler(b, b->get_time());
  }

  if(REPORT_DEL_DOUBLE) {
    cerr << "kepler created in delete_double ";
    cerr << "at address " << b->get_oldest_daughter()->get_kepler() <<endl;
  }
  
  //b->get_oldest_daughter()->update_kepler_from_hdyn();

  kepler* johannes = b->get_oldest_daughter()->get_kepler();
  
  if(REPORT_DEL_DOUBLE) {
    cerr << "delete double star "; b->pretty_print_node(cerr) ; cerr<<endl;
    ((double_star*)(b->get_starbase()))->dump(cerr);
  }
    
  double_star *the_binary = dynamic_cast(double_star*, b->get_starbase());

  if(REPORT_DEL_DOUBLE) 
      put_state(make_state(the_binary));

  // First make sure that the binary is evolved to exact time.

  real stellar_time = b->get_starbase()->
      conv_t_dyn_to_star(b->get_time());

  //    the_binary->evolve_the_binary(stellar_time);
  bool change_mass, change_sma, check_merger;
  evolve_an_unperturbed_binary(b,
			       change_mass,
			       change_sma,
			       check_merger);

    update_dynamics[0] = (change_mass || change_sma)?true:false;

  if(update_dynamics[0]) 
      cerr << "update_dynamics[0] is set to true"<<endl;


    if (REPORT_DEL_DOUBLE) {
      the_binary->dump(cerr);
      put_state(make_state(the_binary));
    }

    // Evolve the binary and update dynamics binary.

    if (check_merger) {
      cerr <<"MERGER! within: "
	   << b->format_label()
	   << " triggered by: ";
      cerr << b->format_label() <<endl;
      cerr << "MERGER should have ahppended but is commented out!!!!!!!"
	   <<endl;
#if 0
    if (check_merger) {
      cerr <<"MERGER! within: "
	   << b->format_label()
	   << " triggered by: " << b->format_label() <<endl;
      hdyn* bcoll = b->get_oldest_daughter()->get_binary_sister();
      b->get_oldest_daughter()->merge_nodes(bcoll,
					    eps,		      
					    eta,
					    max_timestep);
#endif

    }
    else {

      // Unperturbed motion is about to be terminated.  If the final orbit
      // is radically different from the initial one, we have to recompute
      // the time step...

      if (the_binary->get_period()
	    / initial_binary_period(the_binary) < 0.9) {
	  real dm_fast = sudden_mass_loss(b); 
	  update_kepler_from_binary_evolution(b, dm_fast);
	  update_dynamics[1] = true;
	  cerr << "update_dynamics[1] is set to true"<<endl;
      }

    }
      
    if (kepler_created) {

      if (b->get_oldest_daughter()->get_kepler()) {
	b->get_oldest_daughter()->set_kepler(NULL);
	b->get_oldest_daughter()->get_binary_sister()->set_kepler(NULL);
	delete johannes;
      }
    }
    
    // keep the story.
    // might add a story chapter about the binary evolution.
    story* old_story = the_binary->get_star_story();
    the_binary->set_star_story(NULL);

    // Now safely delete the story.
    delete old_story;
      
    // Now safely delete the binary.
    the_binary->set_node(NULL);
    delete the_binary;
      
    // create a new starbase to the dyn.
    b->set_starbase(new starbase(b));
}


void delete_double(dyn *b)
{
  if(REPORT_DEL_DOUBLE) {
    cerr << "Unsafe double_star deletion: "<<endl;
    b->pretty_print_node(cerr); cerr<<endl;
    ((star*)(b->get_starbase()))->dump(cerr);
  }
      double_star *the_binary = dynamic_cast(double_star*, b->get_starbase());

    // First make sure that the binary is evolved to exact time.

      // individual time of particle should be used here
      // NOT the system time!!!
    real stellar_time = b->get_starbase()->
	conv_t_dyn_to_star(b->get_system_time());
    the_binary->evolve_the_binary(stellar_time);
    the_binary->set_node(NULL);

    // keep the story.
    // might add a story chapter about the binary evolution.
    story* old_story = the_binary->get_star_story();
    the_binary->set_star_story(NULL);
      
    // Now safely delete the binary.
    delete the_binary;
      
    // create a new starbase to the dyn.
    b->set_starbase(new starbase(b));
}

/*-----------------------------------------------------------------------------
 *  adddouble  -- for all particles, add a star part using "new double()".
 *
 *  Adds a double_star to a parent node of two leaves, i.e.; the
 *  double_star is attached to the center-of-mass node of the
 *  two stars.
 *  Note that the kepler pointer is attached to the leave nodes.
 *  The function checks (too extensively) wheter or not a double_star
 *  is really needed. The function is fully self-supporting:
 *  call adddouble from any rood-node and it finds self what it needs,
 *  in contradiction to delete_double (see above). 
 *-----------------------------------------------------------------------------
 */

local void adddouble(hdyn *bi, real stellar_time,
		     binary_type type,
		     real sma, real ecc) {

  if (stellar_time<=0) {
    //cerr << "adding at Time=zero"<<endl;
    // Should be the individual bianries time!!
    stellar_time = bi->get_starbase()->conv_t_dyn_to_star(bi->get_system_time());
  }
  
  if (REPORT_ADD_DOUBLE) {
    cerr.precision(16);
    cerr<<"Before adddouble: at "<<stellar_time<< " to: ";
    bi->pretty_print_tree(cerr);
    cerr << " a=" << sma<< " e="<< ecc << endl;
    bi->get_oldest_daughter()->get_kepler()->print_all(cerr);
    cerr.precision(6);
  }
    
  //addstar(bi, stellar_time); Should not be needed.

  int id;
  if((bi->get_oldest_daughter()->is_low_level_node() &&
      bi->get_oldest_daughter()->get_younger_sister()->is_low_level_node()) &&
     (bi->get_oldest_daughter()->get_elder_sister() == NULL) && 
      bi->n_leaves()==2 ) {
    if (!has_dstar(bi->get_oldest_daughter())) {
       //bi->get_parent()->get_starbase()->get_element_type()!=Double) {

      //    if (bi->is_parent() && !bi->is_root())
      story * old_story = bi->get_starbase()->get_star_story();
      bi->get_starbase()->set_star_story(NULL);

      id = bi->get_index();

//      cerr << "Adding binary to "<< id<< "at time="<<stellar_time<<endl;
      double_star* new_double
	= new_double_star(bi, sma, ecc, stellar_time, id, type);
      // synchronize_binary_components(dynamic_cast(double_star*, bi->get_starbase()));

      if (REPORT_ADD_DOUBLE) {
	cerr<<"Adddouble: id, t, a, e: "
	    <<new_double->get_identity()<<" "
	    <<new_double->get_current_time() << " "
	    <<new_double->get_semi()<<" "
	    <<new_double->get_eccentricity()<<endl;
	put_state(make_state(new_double));
      }

      // give the new binary the old star_story.
      new_double->set_star_story(old_story);

      if (REPORT_ADD_DOUBLE) {
	cerr.precision(16);
	cerr<<"After adddouble: at "<<stellar_time<< " to: ";
	new_double->dump(cerr);
	bi->get_oldest_daughter()->get_kepler()->print_all(cerr);
	put_state(make_state(new_double));
	cerr << endl;
	cerr.precision(6);
      }
    }
    else {
      if (REPORT_ADD_DOUBLE) 
	cerr << "No double_star to node needed in adddouble"<<endl;
    }
  }
  else {
    if (REPORT_ADD_DOUBLE) 
      cerr << "Sorry, no binary node in adddouble"<<endl;
  }
}

void adddouble(hdyn *b, real dyn_time, binary_type type,
	       bool random_initialization,
	       real a_min, real a_max,
	       real e_min, real e_max)
{
  if (REPORT_ADD_DOUBLE) {
    cerr.precision(15);
    cerr<<"adddouble: "<<b<<" at T="<<dyn_time<<endl;
    cerr.precision(6);
  }
  
    real stellar_time = b->get_starbase()->conv_t_dyn_to_star(dyn_time);
    addstar(b, stellar_time);
    
    real ecc, sma;
    real binary_age=0;
//      binary_type local_type = Unknown_Binary_Type;
    real a_const = log(a_max) - log(a_min);
    int id;
//      kepler johannes; 
      
    for_all_nodes(hdyn, b, bi) { 
	if (bi->is_parent() && !bi->is_root()) {
            story * old_story = b->get_starbase()->get_star_story();
//            extract_story_chapter(local_type, sma, ecc, *old_story);
            b->get_starbase()->set_star_story(NULL);
//            delete old_story;

	    binary_age = max(bi->get_oldest_daughter()->get_starbase()
			       ->get_current_time(),
			     bi->get_oldest_daughter()->get_binary_sister()
			       ->get_starbase()->get_current_time());
	    //cerr << "times: " << binary_age<<" "<<stellar_time<<endl;
            id = bi->get_index();

            if (random_initialization) {

               sma = a_min*exp(randinter(0., a_const));
               do {
                  ecc = sqrt(randinter(0., 1.));
               }
               while (ecc<e_min && ecc>=e_max);

            } else { //if (local_type==Unknown_Binary_Type) 

	      // cerr << "Non-random creation"<<endl;
//               type = local_type;
//            else {

	        if (!bi->get_oldest_daughter()->get_kepler())
		    new_child_kepler(bi, dyn_time, MAX_CIRC_ECC);
		else {
		    bi->get_oldest_daughter()->get_kepler()->	  // Probably
			set_circular_binary_limit(MAX_CIRC_ECC);  // unnecessary
		    dyn_to_child_kepler(bi, dyn_time);
		}
		
		sma = b->get_starbase()->
		    		conv_r_dyn_to_star(bi->get_oldest_daughter()->
				get_kepler()->get_semi_major_axis());
		ecc = bi->get_oldest_daughter()->get_kepler()->
		    		get_eccentricity();
            }

	    adddouble(bi, binary_age, type, sma, ecc);
	    
            if (random_initialization) {
		randomize_created_binary(bi);
		//            update_dynamical_part_of_binary(bi);
		// The above function no longer exist
		// I hope the following does the expected
		// thing.... (JM)
		cerr << "try update dyn here\n";
		update_dyn_from_binary_evolution(bi);
		cerr << "return update dyn here\n";
	    }
	}
    }
}

#if 0
/* The old version of adddouble.
 * Not in use at the moment.
 */
void adddouble(hdyn *b, real dyn_time, binary_type type,
	       bool random_initialization,
	       real a_min, real a_max,
	       real e_min, real e_max)
{
    cerr.precision(15);
    cerr<<"adddouble: "<<b<<" "<<dyn_time<<endl;
    cerr.precision(6);
    
    real stellar_time = b->get_starbase()->conv_t_dyn_to_star(dyn_time);
    PRL(stellar_time);
    addstar(b, stellar_time);

    real ecc, sma;
//      binary_type local_type = Unknown_Binary_Type;
    real a_const = log(a_max) - log(a_min);
    int id;
//      kepler johannes; 
      
    for_all_nodes(dyn, b, bi) { 
	if (bi->is_parent() && !bi->is_root()) {
            story * old_story = b->get_starbase()->get_star_story();
//            extract_story_chapter(local_type, sma, ecc, *old_story);
            b->get_starbase()->set_star_story(NULL);
//            delete old_story;

            id = bi->get_index();

            if (random_initialization) {

               sma = a_min*exp(randinter(0., a_const));
               do {
                  ecc = sqrt(randinter(0., 1.));
               }
               while (ecc<e_min && ecc>=e_max);

            } else { //if (local_type==Unknown_Binary_Type) 

//               type = local_type;
//            else {

	        if (!bi->get_oldest_daughter()->get_kepler())
		    new_child_kepler(bi, dyn_time, MAX_CIRC_ECC);
		else {
		    bi->get_oldest_daughter()->get_kepler()->	  // Probably
			set_circular_binary_limit(MAX_CIRC_ECC);  // unnecessary
		    dyn_to_child_kepler(bi, dyn_time);
		}

		sma = b->get_starbase()->
		    		conv_r_dyn_to_star(bi->get_oldest_daughter()->
				get_kepler()->get_semi_major_axis());
		ecc = bi->get_oldest_daughter()->get_kepler()->
		    		get_eccentricity();
            }

            cerr<<"id, a, e: "<<id<<" "<<sma<<" "<<ecc<<endl;
	    bi->pretty_print_tree(cerr);
	    pretty_print_star(bi->get_starbase(), cerr);
	    
            double_star* new_double
		= new_double_star(bi, sma, ecc, stellar_time, id, type);

	    // give the new binary the old star_story.
            new_double->set_star_story(old_story);

            cerr<<"double: "<<new_double<<endl;
	    //	    new_double->dump(cerr);
	    put_state(make_state(new_double));
	    
            if (random_initialization) {
		randomize_created_binary(bi);
		//            update_dynamical_part_of_binary(bi);
		// The above function no longer exist
		// I hope the following does the expected
		// thing.... (JM)
		cerr << "try update dyn here\n";
		update_dyn_from_binary_evolution(bi);
		cerr << "return update dyn here\n";
	    }
	}
    }
}
#endif

void  adddouble(node * b, real dyn_time,
		binary_type type,		// Defaults set in
		bool random_initialization,	// double_star.h
		real a_min, real a_max,
		real e_min, real e_max)
{
  if (REPORT_ADD_DOUBLE) {
    cerr.precision(15);
    cerr<<"adddouble: "<<b<<" "<<dyn_time;
    cerr.precision(6);
    cerr <<" "<<a_min<<" "<<a_max<<" "<<e_min<<" "<<e_max << endl;
  }
    
    real stellar_time = b->get_starbase()->conv_t_dyn_to_star(dyn_time);
    addstar(b, stellar_time);

    real ecc, sma;
    // binary_type local_type = Unknown_Binary_Type;

    int id;
    real a_const = log(a_max) - log(a_min);
    for_all_nodes(node, b, bi) {
	if (bi->is_parent() && !bi->is_root()) {
            story * s = b->get_starbase()->get_star_story();
//            extract_story_chapter(local_type, sma, ecc, *s);
            b->get_starbase()->set_star_story(NULL);
            delete s;

            id = bi->get_index();

            if (random_initialization) {
               sma = a_min*exp(randinter(0., a_const));
               do {
                  ecc = sqrt(randinter(0., 1.));
               }
               while (ecc<e_min && ecc>=e_max);
            }
	    else { //if (local_type==Unknown_Binary_Type)
//               type = local_type;
//            else {
               id = bi->get_index();
               sma = a_min*exp(randinter(0., a_const));
               do {
                   ecc = sqrt(randinter(0., 1.));
               }
               while (ecc<e_min && ecc>=e_max);
            }
	    if (REPORT_ADD_DOUBLE) 
	      cerr<<"adddouble: iae: "<<id<<" "<<sma<<" "<<ecc<<endl;
	    
            double_star* new_double = new_double_star(bi, sma, ecc, stellar_time, id,
                         type);
	    if (REPORT_ADD_DOUBLE) {
	      cerr<<"double: "<<new_double<<endl;
	      new_double->dump(cerr);
	      put_state(make_state(new_double));
	    }
         }
      }
   }



#else

/*-----------------------------------------------------------------------------
 *  main  --
 *-----------------------------------------------------------------------------
 */
main(int argc, char ** argv)
    {
    int  c;
    bool  A_flag = FALSE;
    bool  a_flag = FALSE;
    bool  E_flag = FALSE;
    bool  e_flag = FALSE;
    bool  t_flag = FALSE;
    bool  M_flag = FALSE;
    bool  R_flag = FALSE;
    bool  Q_flag = FALSE;
    bool  T_flag = FALSE;
    bool  p_flag = FALSE;
    bool  S_flag = FALSE;
    bool  c_flag = FALSE;
    bool  random_initialization = false;
    real  m_tot = 1;
    real  r_vir = 1;
    real  t_hc = 1;
    real  Q_vir = 0.5; 
    real  a_min = 1;
    real  a_max = 1.e+6;
    real  e_min = 0;
    real  e_max = 1;
    real  t_start = 0;
    binary_type type = Detached;

    int random_seed = 0;
    char seedlog[64];
    char  *comment;
    extern char *poptarg;
    int  pgetopt(int, char **, char *);

    char * param_string = "A:a:E:e:M:R:Q:T:t:Ss:c:";
    check_help();
    
    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c)
	    {
            case 'A': A_flag = true;
		      a_max = atof(poptarg);
                      break;
            case 'a': a_flag = true;
                      a_min = atof(poptarg);
                      break;
            case 'E': E_flag = true;
                      e_max = atof(poptarg);
                      break;
            case 'e': e_flag = true;
                      e_min = atof(poptarg);
                      break;
            case 'M': M_flag = true;
		      m_tot = atof(poptarg);
                      break;
            case 'R': R_flag = true;
	              r_vir = atof(poptarg);
                      break;
            case 'Q': Q_flag = true;
	              Q_vir = atof(poptarg);
	              break; 
            case 'T': T_flag = true;
		      t_hc = atof(poptarg);
                      break;
	    case 't': t_start = atof(poptarg);
		      break;
            case 's': random_seed = atoi(poptarg);
                      break;
            case 'S': S_flag = true;
                      break;
	    case 'c': c_flag = true;
		      comment = poptarg;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	    	      get_help();
	    	      exit(1);
	    }            
    
    int actual_seed = srandinter(random_seed);
    sprintf(seedlog, "         random number generator seed = %d",actual_seed);

       hdyn* b = get_hdyn(cin);

       b->log_history(argc, argv);
       b->log_comment(seedlog);

       if (!T_flag) t_hc = 21.0 * sqrt(Q_vir/m_tot) * pow(r_vir, 1.5);

       if (S_flag)
          if (M_flag && R_flag && T_flag)
             b->get_starbase() ->set_stellar_evolution_scaling(m_tot, r_vir,
							       t_hc);
	   else {
	       cerr << "Proper scaling not set." << endl; cerr
	            << "Use the -M -R an -T options to set the scaling for "
		    << "the mass, \nthe size and the time" << endl; exit(1);
	   }

          if (A_flag || a_flag || E_flag || e_flag) 
             random_initialization=true;
       adddouble(b, t_start, type, random_initialization, 
	         a_min, a_max, e_min, e_max);
       put_hdyn(cout, *b);	
       delete b;
    }

#endif

/* endof: adddouble.c */
