//
// mkbinary.C:	Complete the binary formation process by adding binary
//		orbits to an existing binary tree.  It is assumed that
//		the binary tree structure and all masses have already
//		been set (e.g. by mksecondary), and that positions and
//		velocities of all top-level nodes are already known.
//
//		If possible, use the system parameters to scale the
//		binary parameters.
//
//		Steve McMillan, July 1996
//

#include "dyn.h"

#ifdef TOOLBOX

// The next two functions are stolen from sdyn/init/mkscat.C

local void make_standard_kepler(kepler &k, real t, real mass,
				real energy, real eccentricity,
				real q, real mean_anomaly) {

    // Note: energy is the true energy of the system
    // 	     q is the true periastron only in the case of a parabolic orbit.

    k.set_time(t);
    k.set_total_mass(mass);
    k.set_semi_major_axis((energy == 0 ? VERY_LARGE_NUMBER 
			               : .5*mass/abs(energy)));
    k.set_eccentricity(eccentricity);
    k.set_periastron(q);
    k.set_energy(energy);
    k.set_mean_anomaly(mean_anomaly);
    k.align_with_axes(1);
    k.initialize_from_shape_and_phase();	// Expects a, e [, q [, E]]
}

// set_random_orientation: randomly set the orientation of a kepler structure.

local void set_random_orientation(kepler &k) {

    real cos_theta, sin_theta;

    cos_theta = randinter(-1, 1);
    sin_theta = sqrt(1 - cos_theta * cos_theta);

    real phi = randinter(0, TWO_PI);
    real psi = randinter(0, TWO_PI);

    // Construct the normal vector:

    vector n = vector(sin_theta*cos(phi), sin_theta*sin(phi), cos_theta);

    // Construct unit vectors a and b perpendicular to n:

    vector temp = vector(1, 0, 0);
    if (abs(n[0]) > .5) temp = vector(0, 1, 0);	// temp is not parallel to n
    if (n[2] < 0) temp = -temp;

    vector b = n ^ temp;
    b /= abs(b);
    vector a = b ^ n;
    if (n[2] < 0) a = -a;	// Force (a, b) to be (x, y) for n = +/-z
    
    // Construct *random* unit vectors l and t perpendicular to each
    // other and to n (psi = 0 ==> periastron along a):

    vector l = cos(psi)*a + sin(psi)*b;
    vector t = n ^ l;

    k.set_orientation(l, t, n);
    k.initialize_from_shape_and_phase();
}

// add_secondary: Add a secondary component to a primary to make
//		  a binary with the specified parameters.
//		  All unspecified orbit elements are chosen randomly.
//		  New primary and secondary components have names "na"
//		  and "nb", respectively, where "n" was the name of
//		  the original star.

local void add_dynamics(dyn* cm, real ecc, real energy)
{
    dyn* primary = cm->get_oldest_daughter();
    dyn* secondary = primary->get_younger_sister();

    real m_total = cm->get_mass();
    real m1 = primary->get_mass();
    real m2 = secondary->get_mass();

    // Set internal orbital elements:

    kepler k;

    real peri = 1; // Default value (unimportant unless ecc = 1).
    if (ecc == 1) peri = 0;

    // For now, binary phase is random.

    real mean_anomaly = randinter(-PI, PI);

    make_standard_kepler(k, 0, m_total, energy, ecc,
			 peri, mean_anomaly);
    //    PRC(m_total);
    //    PRC(energy);
    //    PRL(ecc);

    set_random_orientation(k);

    // Set positions and velocities.

    primary->set_pos(-m2 * k.get_rel_pos() / m_total);
    primary->set_vel(-m2 * k.get_rel_vel() / m_total);

    secondary->set_pos(m1 * k.get_rel_pos() / m_total);
    secondary->set_vel(m1 * k.get_rel_vel() / m_total);

}

local void mkbinary(dyn* b, real e_lower, real e_upper, int option)
{
    // Binary energies are drawn from a 1/e distribution,
    // between the specified limits.
    // For now, use thermal eccentricities.

    real efac = e_upper/e_lower;

    for_all_daughters(dyn, b, bi)
	if (bi->get_oldest_daughter()) {

	    dyn* primary = bi->get_oldest_daughter();
	    dyn* secondary = primary->get_younger_sister();
	    real m_total = bi->get_mass();
	    real mu = primary->get_mass() * secondary->get_mass() / m_total;

	    // Choose binary parameters:
	    //
	    // Option 1 ==> limits refer to binary energy.
	    // Option 2 ==> limits refer to binary energy per unit reduced mass.
	    // Option 3 ==> limits refer to binary energy per unit binary mass.
	    //
	    // However, add_dynamics expects energy per unit reduced mass,
	    // so scale the energy in cases 1 and 3.

	    real ecc = sqrt(randinter(0,1));

	    real energy = e_lower * pow(efac, randinter(0,1));
	    if (option == 1)
		energy /= mu;
	    else if (option == 3)
		energy *= m_total / mu;		

//	    PR(m_total);
//	    PR(primary->get_mass());
//	    PRL(secondary->get_mass());
// 	    PRL(energy);
//	    real sma = 0.5 * primary->get_mass() * secondary->get_mass()
//	                   / energy;
//	    PRL(sma);

	    add_dynamics(bi, ecc, energy);
	}
}

local void scale_limits(real& e1, real& e2, int option,
			int N, real M, real E)
{
    real scale = 1.0;

    if (option == 3) {

	// Limits refer to kinetic energy per unit mass.  Scale to -E/M.

	scale = -E / M;

    } else if (option != 2) {

	// Limits refer to energy.  Scale to kT = -(2/3) E/N.

	scale = -E / (1.5*N);
    }

    e1 *= scale;
    e2 *= scale;
}

void main(int argc, char ** argv)
{
    real e_lower = 1.0, e_upper = 10.0;
    int energy_option = 1;
    int random_seed = 0;
    char seedlog[64];

    extern char *poptarg;
    int c;
    char* param_string = "l:s:o:u:";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'l': e_lower = atof(poptarg);
		      break;
	    case 'o': energy_option= atoi(poptarg);
		      break;
	    case 's': random_seed = atoi(poptarg);
		      break;
	    case 'u': e_upper = atof(poptarg);
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      exit(1);
	}

    if (e_lower <= 0 || e_upper <= 0 || e_lower > e_upper)
	err_exit("mkbinary: Illegal limits on semi-major axis");

    dyn* b;
    b = get_dyn(cin);

    b->log_history(argc, argv);

    int actual_seed = srandinter(random_seed);

    sprintf(seedlog, "         random number generator seed = %d",actual_seed);
    b->log_comment(seedlog);

    // Look in the dyn story to see if a total system energy is known.
    // If it is, scale e_lower and e_upper accordingly.

    // Thus, if the energy is known, e_lower and e_upper are specified
    // in "kT" units.  If the energy is not known, e_lower and e_upper
    // are absolute limits on the binary kinetic energy.

    char* energy_string = "total_energy";

    if (find_qmatch(b->get_dyn_story(), energy_string))
	scale_limits(e_lower, e_upper, energy_option,
		     b->n_daughters(), b->get_mass(),
		     getrq(b->get_dyn_story(), energy_string));
    else
	cerr << "mkbinary: Using unscaled energy limits.\n";

    mkbinary(b, e_lower, e_upper, energy_option);

    put_dyn(cout, *b);

}
#endif
