
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//// addstar:  Add star class to existing node structure.
////
//// Options:    -c    comment to put in the starbase log structure.
////             -M    Mass scaling [total cluster mass in solar units].
////             -Q    Virial ratio (used if time scaling is omitted)
////             -R    Dynamical size scaling for the star
////                   [in units of the virial radius].
////             -s    Initial types of stars [default is main_sequence].
////             -T    Dynamical time scaling
////                   [in units of the NBODY time].
////             -t    stellar age [in million years].
////
//++ Notes:
//++  Online arguments overrule parameters from input node.
//++
//++  If no time unit is explicitly specified it is derived
//++  from the other units.
//++
//++  The mass of the system can be set independently from the total
//++  mass of the input N-body system (e.g. when modeling a large
//++  cluster by a smaller N-body system).
//++
//++ Example of usage:      
//++  mknode -n 10 | mkmass -u 100 -l 10 | addstar -T 1 
//++
//++ See also: addstar
//++           mkmass
//++           mknode


//// Version 1.0: Apr 1993   Piet Hut, Steve McMillan, Jun Makino
//// Version 2.0: Mei 1993   Simon Portegies Zwart
//                           spz@grape.c.u-tokyo.ac.jp

#include "single_star.h"

#ifndef TOOLBOX

/*-----------------------------------------------------------------------------
 *  addstar  -- for all particles, add a star part using "new star()".
 *-----------------------------------------------------------------------------
 */
void  addstar(node * b, real t_current, stellar_type type, bool verbose) {

    if (b->get_oldest_daughter() != NULL) {

	real m_min = VERY_LARGE_NUMBER;
	real m_max = -VERY_LARGE_NUMBER;
	real m_av = 0;
	int  n_star = 0;

	for_all_daughters(node, b, bi) {
	    addstar(bi, t_current, type, verbose);

	    if (verbose) {
		real m = bi->get_starbase()
			   ->conv_m_dyn_to_star(bi->get_mass());
		m_min = min(m_min, m);
		m_max = max(m_max, m);
		m_av += m;
		n_star++;		
	    }
	}

	if (verbose) {
	    if (n_star > 0) m_av /= n_star;
	    //  PRC(n_star); PRC(m_min); PRC(m_max); PRL(m_av);
	}

    } else {

//	cerr << "addstar: ";
//	PRL(b->get_starbase()->get_element_type());

	// NAS is the default return value of get_element_type, in 
	// cases where only a starbase (but no star) exists.

	if (b->get_starbase()->get_element_type() != NAS) return;
	  
	int id = b->get_index();
	real t_cur=0, t_rel=0, m_rel=1, m_env=0, m_core=0.01;
	real T_eff, L_eff;
	real p_rot=0, b_fld=0; 
	real m_tot;

	// Create a (single) star part, using the unformation obtained from
	// the star story read in by get_node(), or created from scratch.

	stellar_type local_type = NAS;
	starbase * old_starbase = b->get_starbase();
	story * s = old_starbase->get_star_story();

	extract_story_chapter(local_type, t_cur, t_rel,  
			      m_rel, m_env, m_core, T_eff, L_eff,
			      p_rot, b_fld, *s);

	m_tot = m_env+m_core;
	old_starbase->set_star_story(NULL);
	delete s;

	if (local_type != NAS) {	// Is the star properly defined?

	    // Use the data extracted from the star story.  Note that the
	    // input values of the type and t_rel arguments are IGNORED.

	  //cerr << "addstar: " << type_string(local_type) 
	  //   << " t_rel = " << t_rel << " t_cur = " << t_cur << endl;
	  
	    type = local_type; 
	    single_star* new_star = new_single_star(type, id, t_cur,
						    t_rel, 
						    m_rel, m_tot, m_core,
						    p_rot, b_fld, b);

	    //new_star->set_current_time(t_current); // added (SPZ:2/1998)
	    //new_star->dump(cerr);
	    
	} else {		// No star story present, or at least no
	     			// proper definition for a single star.

	    // Create a default star of the specified type.

	    real t_cur=0, t_rel = 0, m_rel = 1, m_core = 0.01;
	    real p_rot=0, b_fld=0; 
	    real m_tot;
	    t_cur = t_current;
	    //cerr << "addstar2: " << type_string(type) << " t_rel = " << t_rel
	    // << " t_cur = " << t_cur << endl;

	    starbase * old_starbase = b->get_starbase();
	    id = b->get_index();
	    m_rel = m_tot = b->get_starbase()
			     ->conv_m_dyn_to_star(b->get_mass());

	    single_star* new_star = new_single_star(type, id, t_cur, t_rel, 
						    m_rel, m_tot, m_core,
						    p_rot, b_fld, b);

	}

	// Should not be needed sinfe new_single_star does the job.
	// (SPZ:25 Nov 1998)
	// delete old_starbase;
    }
}

#else

#  define  FALSE  0
#  define  TRUE   1

/*-----------------------------------------------------------------------------
 *  main  --
 *-----------------------------------------------------------------------------
 */
main(int argc, char ** argv)
    {
    int  c;
    bool  t_flag = FALSE;
    bool  M_flag = FALSE;
    bool  R_flag = FALSE;
    bool  T_flag = FALSE;
    bool  Q_flag = FALSE;
    bool  s_flag = FALSE;
    bool  c_flag = FALSE;
    real  m_tot = -1;
    real  r_vir = -1;
    real  t_vir = -1;
    real  t_rel = 0;
    real  q_vir = 0.5;
    real  T_start = 0;
    stellar_type type = Main_Sequence;

    char  *comment;
    extern char *poptarg;
    int  pgetopt(int, char **, char *);
    char * param_string = "M:R:Q:T:t:s:c:";
    
    check_help();
    
    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c)
	    {
	    case 'M': M_flag = true;
		      m_tot = atof(poptarg);
		      break;
            case 'R': R_flag = TRUE;
                      r_vir = atof(poptarg);
                      break;
	    case 'Q': Q_flag = TRUE;
	              q_vir = atof(poptarg);
	              break;
            case 'T': T_flag = TRUE;
                      t_vir = atof(poptarg);
                      break;
	    case 't': T_flag = TRUE;
		      t_rel = atof(poptarg);
		      break;
            case 's': s_flag = TRUE;
                      type = (stellar_type)atoi(poptarg);
                      break;
	    case 'c': c_flag = TRUE;
		      comment = poptarg;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	    	      get_help();
	    	      exit(1);
	    }            

    node *b;
    b = get_node(cin);
    if (c_flag == TRUE)
      b->log_comment(comment);
    b->log_history(argc, argv);

    real new_r_vir=-1, new_t_vir=-1, new_m_tot=-1;
    real old_r_vir = b->get_starbase()->conv_r_star_to_dyn(1);
    real old_t_vir = b->get_starbase()->conv_t_star_to_dyn(1);
    real old_m_tot = b->get_starbase()->conv_m_dyn_to_star(1);

    if(old_m_tot>0 && !M_flag) new_m_tot = old_m_tot;
    else if(M_flag)            new_m_tot =     m_tot;
    else err_exit("addstar: No mass scaling available.");

    if(old_r_vir>0 && !R_flag) new_r_vir = old_r_vir;
    else                       new_r_vir =     r_vir;

    if(old_t_vir>0 && !(T_flag || Q_flag)) new_t_vir = old_t_vir;
    else                                   new_t_vir =     t_vir;

    if (new_t_vir<=0 && new_m_tot>0 && new_r_vir>0) {
	
      // If no time unit is explicitly specified, derive it from
      // the other units.  However, allow the possibility that the
      // mass of the system being modeled and the total mass of the
      // N-body system are not the same (e.g. when modeling a large
      // cluster by a smaller N-body system).
	
      // Standard (N-body / Heggie & Mathieu) time unit, in Myr:
      
      new_t_vir = 21.0 * sqrt(q_vir/new_m_tot) * pow(new_r_vir, 1.5);
    }
    else if(new_t_vir<=0 && new_r_vir<0) 
      err_exit("addstar: Illegal limits time-scaling");

    if(old_m_tot>0 && !M_flag) {
      b->get_starbase()->set_stellar_evolution_scaling(1,1,1);
      addstar(b, T_start, Main_Sequence, true);   
      b->get_starbase()->set_stellar_evolution_scaling(new_m_tot,
						       new_r_vir,
						       new_t_vir);
    }
    else{
      b->get_starbase()->set_stellar_evolution_scaling(new_m_tot,
						       new_r_vir,
						       new_t_vir);
      addstar(b, T_start, Main_Sequence, false);  
    }
    
    put_node(cout, *b);	
    delete b;
}

#endif

/* endof: addstar.c */
