//  sstar_stats.C: Print out various diagnostic statistics on a system.

//#include "dstar_to_dyn.h"
#include "sstar_to_dyn.h"
#include "dyn.h"
#include "single_star.h"

#ifndef TOOLBOX

#define MINIMUM_LUMINOSITY 1.0e-4  // planets' luminosity

struct pop {
  int stypes[no_of_stellar_type];

  pop() {
    for (int i=0; i<no_of_stellar_type; i++)
      stypes[i]=0;
  }
};

struct nm_bin {
          int   no_of_stars;
          real  mass_limit;
          real  total_mass;
  nm_bin() {
    no_of_stars=0;
    mass_limit=total_mass=0;
  }
};

local int which_zone(dyn* bi, vector& dc_pos, int n_lagr, real* r_lagr)
{
    vector dr = bi->get_pos() - dc_pos;
    real dr2 = dr*dr;

    for (int j = 0; j < n_lagr; j++)
      if (r_lagr[j] > dr2) return j;

    return n_lagr;
}

local void print_stellar_content(dyn* b) {

  int n_lagr = 0;
  real *r_lagr = NULL;

  if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

    // Make a list of previously computed lagrangian radii.

    n_lagr = getiq(b->get_dyn_story(), "n_lagr");
    r_lagr = new real[n_lagr];
    getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

   }
  else 
    r_lagr = new real[n_lagr];

  
    pop *content = new pop[n_lagr+1];

    // Note that r_lagr omits the 0% and 100% radii.
    
    // Initialization:

    for (int i = 0; i <= n_lagr; i++)
      if (i < n_lagr) r_lagr[i] *= r_lagr[i];

    // As in lagrad, use the geometric center if 
    // density center is unknown.

    vector dc_pos = 0;
    if (find_qmatch(b->get_dyn_story(), "density_center_pos"))
      dc_pos = getvq(b->get_dyn_story(), "density_center_pos");
    
    int N_total = 0;
    stellar_type stype = NAS;
    for_all_daughters(dyn, b, bi) {

      // Count single stars and binaries separately.

      if (bi->get_oldest_daughter() != NULL)
	
	stype = Double;
    
      else {
	if (b->get_use_sstar()) 
	  stype = bi->get_starbase()->get_element_type();
	else if (find_qmatch(bi->get_star_story(), "Type")) 
	  stype = extract_stellar_type_string(
		  getsq(bi->get_star_story(), "Type")); 
	else {
	  cerr << "    No stellar information found for: ";
	  bi->pretty_print_node(cerr);
	  return;
	}
      }
      // Find which zone we are in.

      int i = which_zone(bi, dc_pos, n_lagr, r_lagr);

      // Update statistics.

      N_total++;

      content[i].stypes[(int)stype]++;
    }

    if (N_total > 0) {

      int j, k;
      char* ts;
      cerr << "           ("; 
      for (j=0; j<=dynamic_cast(int, Super_Giant); j++) 
	  cerr << " " << type_short_string(dynamic_cast(stellar_type, j));

      cerr << "     ";
      for (j=dynamic_cast(int, Carbon_Star); 
               j<dynamic_cast(int, no_of_stellar_type); j++) 
	  cerr << " " << type_short_string(dynamic_cast(stellar_type, j));

      cerr << ")" << endl;
      
      for (k = 0; k <= n_lagr; k += 5) {
	if (k > 0) cerr << endl;
	for (int i = k; i < k+5 && i <= n_lagr; i++) {
	  cerr << "    zone " << i << " ";
	  for (j=0; j<=Super_Giant; j++)
	    cerr << " " << content[i].stypes[j];
	  cerr << "     ";
	  for (j=Carbon_Star; j<no_of_stellar_type; j++)
	    cerr << " " << content[i].stypes[j];
	  cerr << endl;
	}
      }

    } else
      cerr << "            (none)\n";
  
    delete []content;
    delete []r_lagr;
//  }
//  else 
//    cerr << "     ---No Lagrangian radii specified--- " << endl;
  
}

local int print_special_stars(dyn* b) {

  if (!b->get_use_sstar())
    return -1;

  real M_to = turn_off_mass(b->get_starbase()
			     ->conv_t_dyn_to_star(b->get_system_time()));

  int N_special = 0;
  real f_Bs;
  for_all_leaves(dyn, b, bi) {

    // Count all single stars 

    stellar_type stype = NAS;
    star* ss = NULL;

    if (has_sstar(bi)) {
      star* ss = dynamic_cast(star*, bi->get_starbase());

      if (ss!=NULL) {
	star_state sts(ss);
      
	if (sts.special()) {

	  N_special++;

	  cerr << "    " << bi->format_label() 
	       << " is ";
	  put_short_state(sts, cerr);
	  cerr << " (" << type_dominant_state(sts)
	       << ")";
	  
	  if (sts.class_spec[Blue_Straggler]) {
	     f_Bs = sts.mass/M_to;
	    cerr << " [f_Bs = " << f_Bs << "]";
	  }
	    
	  if (ss->is_binary_component()) {
	    cerr << " binary companion is "
		 << bi->get_binary_sister()->format_label() << " ";
	    print_star(dynamic_cast(starbase*, ss->get_companion()));
	  }
	  cerr << endl;
	}
	else if (ss->get_element_type()==Disintegrated) {

	  N_special++;
	  
	  cerr << "    " << bi->format_label() 
	       << " is disintegrated star" << endl;
	}
      }
    }
  }
  
  if (N_special==0)
    cerr << "     ---No special stars---" <<endl;
}

local void print_compact_stars(dyn* b) {

  //  for_all_daugters(dyn, b, bi) {
  if (!b->get_use_sstar())
     return;

  int N_compact = 0;
  for_all_leaves(dyn, b, bi) {

    // Count all single stars 

    stellar_type stype = NAS;
    star* ss = NULL;
//    if (bi->get_oldest_daughter() != NULL)
    if (has_sstar(bi)) {
      
     ss = dynamic_cast(star*, bi->get_starbase());

    bool accreting = false;
    if (ss!=NULL) 
      if(ss->remnant()) {
	N_compact++;
	
	stype = ss->get_element_type();
	accreting = ((ss->get_spec_type(Accreting))?true:false);

	if (stype==Xray_Pulsar || stype==Radio_Pulsar) {
	  
	  cerr << "    " << bi->format_label()
	       << " is " << type_string(stype)
	       << " (Ps = " << ss->get_rotation_period()
	       << ", log B = " << ss->get_magnetic_field()
	       << ") ";
	  if (ss->is_binary_component()) {
	    cerr << " binary companion is "
	         << bi->get_binary_sister()->format_label() << " ";
	    print_star(dynamic_cast(starbase*, ss->get_companion()));
	  }
	  cerr << endl;
	}
	else if (stype==Helium_Giant) {

	  cerr << "    " << bi->format_label()
	       << " is " << type_string(stype)
	       << " (M =" << ss->get_relative_mass()
	       << ", Reff = " << ss->get_effective_radius()
	       << ", L = " << ss->get_luminosity()
	       << ") ";
	  if (ss->is_binary_component()) {
	    cerr << "binary companion is "
		 << bi->get_binary_sister()->format_label() << " ";
	    print_star(dynamic_cast(starbase*, ss->get_companion()));
	  }
	  cerr << endl;
	}
	else if (accreting) {

	  cerr << "    " << bi->format_label()
	       << " is accreting " << type_string(stype);
	  if (ss->is_binary_component()) {
	    cerr << " binary companion is "
	         << bi->get_binary_sister()->format_label() << " ";
	    print_star(dynamic_cast(starbase*, ss->get_companion()));
	  }
	  cerr << endl;
	}
      }
      else if (stype==Thorn_Zytkow) {

	cerr << "    " << bi->format_label()
	     << " is " << type_string(stype)
	     << " (M =" << ss->get_relative_mass()
	     << ", Reff = " << ss->get_effective_radius()
	     << ", L = " << ss->get_luminosity()
	     << ") ";
	if (ss->is_binary_component()) {
	  cerr << "binary companion is "
	       << bi->get_binary_sister()->format_label() << " ";
	  print_star(dynamic_cast(starbase*, ss->get_companion()));
	}
	cerr << endl;
      }
    }
  }
  
  if (N_compact==0)
    cerr << "     ---No compact stars---" <<endl;
  else
    cerr << "        Number of compact stars = " << N_compact << endl;
}

local void print_mass_over_light(dyn* b) {

  if (!b->get_use_sstar())
    return;

  if (find_qmatch(b->get_dyn_story(), "n_lagr")) {

    // Make a list of previously computed lagrangian radii.

    int n_lagr = getiq(b->get_dyn_story(), "n_lagr");
    real *r_lagr = new real[n_lagr];
    getra(b->get_dyn_story(), "r_lagr", r_lagr, n_lagr);

    real  *total_mass = new real[n_lagr+1];
    real  *total_lumi = new real[n_lagr+1];
    int i;
    for (i = 0; i <= n_lagr; i++) 
      total_mass[i]=total_lumi[i]=0;


    // Note that r_lagr omits the 0% and 100% radii.
    
    // Initialization:

    for (i = 0; i <= n_lagr; i++)
      if (i < n_lagr) r_lagr[i] *= r_lagr[i];

    // As in lagrad, use the geometric center if 
    // density center is unknown.

    vector dc_pos = 0;
    if (find_qmatch(b->get_dyn_story(), "density_center_pos"))
      dc_pos = getvq(b->get_dyn_story(), "density_center_pos");
    
    int N_total = 0;
    real L_star=0, M_star=0;

    for_all_leaves(dyn, b, bi) {

      // Count single stars and binaries separately.


      if (bi->get_oldest_daughter() == NULL) 
	if (has_sstar(bi)) {
	  
	  L_star = ((single_star*)bi->get_starbase())->get_luminosity();
	  M_star = bi->get_starbase()->get_total_mass();
	  if ( ((single_star*)bi->get_starbase())->remnant() )
	    L_star = 0;
	  
	}
	else if (bi->get_star_story()!=NULL) {
	  
	  L_star = getrq(bi->get_star_story(), "L_eff"); 
	  M_star = getrq(bi->get_star_story(), "M_env")
	         + getrq(bi->get_star_story(), "M_core"); 
	}


      // Find which zone we are in.

      i = which_zone(bi, dc_pos, n_lagr, r_lagr);

      // Update statistics.

      N_total++;

      total_mass[i] += M_star;
      total_lumi[i] += L_star;
    }

    if (N_total > 0) {

      for (i=0; i<=n_lagr; i++)
	if (total_lumi[i]>0)
	  cerr << "    zone " << i << "   "
	       << total_mass[i]/total_lumi[i]
	       << "    solar" <<  endl;
	else
      	  cerr << "    zone " << i << " "
	       << " no luminosity "
	       << endl;

    } else
      cerr << "            (none)\n";
  
    delete []r_lagr;
    delete []total_mass;
    delete []total_lumi;
  }
}

local void print_mass_spectrum(dyn* b, int nzones, bool verbose) {

    if (!b->get_use_sstar() || nzones < 2)
     return;

    nm_bin *mass_bins = new nm_bin[nzones+1];
    if (mass_bins == NULL) {
      cerr << "not enough memory left for print_mass_spectrum\n";
      return;
    }

    real m_min  =   b->get_starbase()->conv_m_star_to_dyn(0.1);   // Msun
    real m_max  =   b->get_starbase()->conv_m_star_to_dyn(100);   // Msun
    
    real log_m_min = log10(m_min);
    mass_bins[1].mass_limit = log_m_min;
    real mass_int = (log10(m_max)-mass_bins[1].mass_limit)/(nzones-2);

    int i;
    for (i=0; i<nzones; i++) {
      mass_bins[i].mass_limit = pow(10, log_m_min + (i-1)*mass_int);

      mass_bins[i].no_of_stars = 0;
      mass_bins[i].total_mass = 0;

    }
    mass_bins[0].mass_limit = mass_bins[nzones].mass_limit = 0;
    
    int N_total = 0;
    real mi, m_total=0;
    for_all_leaves(dyn, b, bi) {
      mi = bi->get_mass();
      m_total += mi;
      N_total++;
      if (mi > m_min && mi <= m_max) {

	for (i=2; i<nzones-1; i++)
	  
	  if (mi > mass_bins[i-1].mass_limit   &&
	      mi <= mass_bins[i].mass_limit) {

	    mass_bins[i-1].no_of_stars++;
	    mass_bins[i-1].total_mass  += mi;
	  }
      }
      else if (mi<=m_min) {
	
	mass_bins[0].no_of_stars++;
	mass_bins[0].total_mass  += mi;
      }
      else {
	
	mass_bins[nzones].no_of_stars++;
	mass_bins[nzones].total_mass  += mi;
      }
    }
    
    real scaling = 1;
    if (b->get_use_sstar()) {
      scaling = b->get_starbase()->conv_m_star_to_dyn(1);
      m_min = b->get_starbase()->conv_m_dyn_to_star(m_min);
      m_max = b->get_starbase()->conv_m_dyn_to_star(m_max);
      cerr.precision(3);
    }

    if (N_total>0) {

      cerr << endl;
      if (verbose)
	cerr << "  Mass function (dlog_m = " << mass_int
	     << "  min = " << m_min
	     << "  max = " << m_max << " [Msun]): "
	     << endl;

      cerr << "            " << mass_bins[0].no_of_stars
	   << "  <";
      for (i=1; i<=nzones-2; i++)
	cerr << "   " << mass_bins[i].no_of_stars;
      
      cerr << "   > " << mass_bins[nzones-1].no_of_stars
	   << endl;
	  
      cerr.precision(6);
    }
    else
      cerr << "            (none)\n";
    
    delete []mass_bins;
}


local void print_luminosity_function(dyn* b, int nzones, bool verbose) {

    if (!b->get_use_sstar() || nzones < 2)
     return;

    nm_bin *lstar_bins = new nm_bin[nzones+1];
    if (lstar_bins == NULL) {
      cerr << "not enough memory left for print_luminosity_function\n";
      return;
    }

    real l_min  = 0.01;
    real l_max  = 1.e+6;
    real log_l_min  = log10(l_min);
    real log_l_max  = log10(l_max);
    
    lstar_bins[1].mass_limit = log_l_min;
    real lstar_int = (log_l_max-log_l_min)/(nzones-2);

    int i;
    for (i=0; i<nzones; i++) {
      lstar_bins[i].mass_limit = pow(10, log_l_min + (i-1)*lstar_int);

      lstar_bins[i].no_of_stars = 0;
      lstar_bins[i].total_mass = 0;

    }
    lstar_bins[0].mass_limit = lstar_bins[nzones].mass_limit = 0;
    
    int N_total = 0;
    real li, l_total=0;
    for_all_leaves(dyn, b, bi) {

      li = ((star*)bi->get_starbase())->get_luminosity();
      l_total += li;
      N_total++;
      if (li > l_min && li <= l_max) {

	for (i=2; i<nzones-1; i++)
	  
	  if (li > lstar_bins[i-1].mass_limit   &&
	      li <= lstar_bins[i].mass_limit) {

	    lstar_bins[i-1].no_of_stars++;
	    lstar_bins[i-1].total_mass  += li;
	  }
      }
      else if (li<=l_min) {
	
	lstar_bins[0].no_of_stars++;
	lstar_bins[0].total_mass  += li;
      }
      else {
	
	lstar_bins[nzones].no_of_stars++;
	lstar_bins[nzones].total_mass  += li;
      }
    }
    
    if (N_total>0) {

      cerr << endl;
      if (verbose)
	cerr << "  Luminosity function (dlog_L = " << lstar_int
	     << "  min = 10^" << log_l_min
	     << "  max = 10^" << log_l_max << " [Lsun]): "
	     << endl;

      cerr << "            " << lstar_bins[0].no_of_stars
	   << "  <";
      for (i=1; i<=nzones-2; i++)
	cerr << "   " << lstar_bins[i].no_of_stars;
      
      cerr << "   > " << lstar_bins[nzones].no_of_stars
	   << endl;
	  
      cerr.precision(6);
    }
    else
      cerr << "            (none)\n";
    
    delete []lstar_bins;
}

local void print_massive_star(dyn *bi, vector dc_pos,
			      vector dc_vel, bool verbose) {

  real mass  = bi->get_starbase()->get_total_mass();
  real r_com = abs(bi->get_pos() - dc_pos);
  real v_com = abs(bi->get_vel() - dc_vel);

  cerr << "            " << bi->format_label()
       << "      " << mass << "  ("; put_state(make_star_state(bi), cerr);
  cerr << ") \t  " << r_com << "   " << v_com << endl;
	
}
    
local int print_massive_binary(dyn *bi,
				vector dc_pos, vector dc_vel,
				real mass_limit, real number_limit,
				bool verbose) {


  real m_tot  = bi->get_starbase()->conv_m_dyn_to_star(bi->get_mass());
  real mass1  = bi->get_oldest_daughter()
                  ->get_starbase()->get_total_mass();
  real mass2  = bi->get_younger_sister()
                  ->get_starbase()->get_total_mass();

  if ((mass1 >= mass_limit ||
       (bi->get_oldest_daughter()->get_index() > 0 &&
	bi->get_oldest_daughter()->get_index() <= number_limit)) ||
      (mass2 >= mass_limit ||
       (bi->get_younger_sister()->get_index() > 0 &&
	bi->get_younger_sister()->get_index() <= number_limit))) {
    
  real r_com  = abs(bi->get_pos() - dc_pos);
  real v_com  = abs(bi->get_vel() - dc_vel);

  cerr << "            " << bi->format_label()
       << "      " << m_tot 
       << "  (Multiple) \t  " << r_com << "   " << v_com << endl;

  cerr << "                     " << mass1 << "  (";
       put_state(make_star_state(bi->get_oldest_daughter()), cerr);
  cerr << ") " << endl;
  cerr << "                     " << mass2 << "  (";
       put_state(make_star_state(bi->get_younger_sister()), cerr);
  cerr << ") " << endl;

  }

  return 1;
}

local bool contains_massive_star(dyn *b,
				 real mass_limit,
				 real number_limit) {

    for_all_leaves(dyn, b, bi) 
      if ((bi->get_index()>0 && bi->get_index() <= number_limit) ||
	  bi->get_starbase()->get_total_mass() >= mass_limit) 
         return true;

  return false;

}

local int print_massive_binary_recursive(dyn *b,
				vector dc_pos, vector dc_vel,
				real mass_limit, real number_limit,
				bool verbose) {

  int nb = 0;
  if (b->get_oldest_daughter() &&
      contains_massive_star(b, mass_limit, number_limit)) {

    real r_com  = abs(b->get_pos() - dc_pos);
    real v_com  = abs(b->get_vel() - dc_vel);

    real m_tot = b->get_starbase()->conv_m_dyn_to_star(b->get_mass());
    cerr << "            " << b->format_label()
	 << "      " << m_tot 
	 << "  (Multiple) \t  " << r_com << "   " << v_com << endl;

    for_all_daughters(dyn, b, bb)
      if (bb->n_leaves() >= 2) 
	nb += print_massive_binary_recursive(bb, dc_pos, dc_vel,
					     mass_limit, number_limit,
					     verbose); 
      else
	print_massive_star(bb, dc_pos, dc_vel, verbose);
    
    //	nb += print_massive_binary(b, dc_pos, dc_vel,
    //				   mass_limit, number_limit,  verbose);
  }
  return nb;
}

local void print_massive_star_header(bool cod) {

  cerr << endl;
    cerr << "  Massive stars:" << endl;
    if (cod)
      cerr << "            Name    M(sun)   (type)   \t r_cod    v_cod"
	   << endl;
    else
      cerr << "            Name    M(sun)   (type)   \t r_com    v_com"
	   << endl;
}

local void print_most_massive_stars(dyn *b, real mass_limit,
				    real number_limit,
				    vector dc_pos,
				    bool verbose) {

  bool cod = false;
  
  vector dc_vel = 0;
  bool try_com = false;
  if(abs(dc_pos) == 0) {
    if (find_qmatch(b->get_dyn_story(), "density_center_pos")) {
      
      if (getrq(b->get_dyn_story(), "density_center_time")
	  != b->get_system_time()) {
	warning("lagrad: neglecting out-of-date density center");
	try_com = true;
      } else
	cod = true;
      
      dc_pos = getvq(b->get_dyn_story(), "density_center_pos");
      dc_vel = getvq(b->get_dyn_story(), "density_center_vel");
    }

    if (try_com && find_qmatch(b->get_dyn_story(), "com_pos")) {

      if (getrq(b->get_dyn_story(), "com_time")
	  != b->get_system_time()) {
	warning("lagrad: neglecting out-of-date center of mass");
      } else
	dc_pos = getvq(b->get_dyn_story(), "com_pos");
      dc_vel = getvq(b->get_dyn_story(), "com_vel");
    }
  }

  int ns=0, nb=0;
  for_all_daughters(dyn, b, bi) 
    if (bi->is_leaf()                    &&
	((bi->get_index()>0 && bi->get_index() <= number_limit) ||
	 bi->get_starbase()->get_total_mass() >= mass_limit)) {
	
      if (!ns)
	print_massive_star_header(cod);

      print_massive_star(bi, dc_pos, dc_vel, verbose);
      ns++;
    }

  for_all_daughters(dyn, b, bi) {
    if (!bi->is_leaf())                   
      nb += print_massive_binary_recursive(bi, dc_pos, dc_vel,
					   mass_limit, number_limit,
					   verbose); 
  }
  
  if (verbose && ns+nb==0) 
      cerr << "            (none)\n";
}

// print stellar timescale and turn-off mass.
void print_sstar_time_scales(dyn *b) {
  
      real time = b->get_system_time();
	cerr << "     = " << b->get_starbase()->conv_t_dyn_to_star(time)
	     << " Myr (Turn_off_mass = "
	     << turn_off_mass(b->get_starbase()->conv_t_dyn_to_star(time))
	     << ")" << endl;

	b->get_starbase()->print_stellar_evolution_scaling(cerr);
}

void sstar_stats(dyn* b, bool mass_spectrum, vector center,
		bool verbose) { 

  if (b->get_use_sstar()) {

    int axis = 0;   // x-axis
    if (verbose) cerr << "\n  Projected luminosity profile: ";
    real r1_r9 = compute_projected_luminosity_radii(b, axis, true, 10);

    if (verbose) cerr << "\n  Projected King profile:"; 
    print_fitted_king_model(r1_r9, proj_r10_r90_light);

    if (mass_spectrum) {
      
      print_luminosity_function(b, 12, verbose);

      print_mass_spectrum(b, 12, verbose);
	
      print_most_massive_stars(b, 10, 10, center, verbose);
    }
    
    if (verbose)
      cerr << "\n  Mass to light by Lagrangian zone (all stars):\n";
    print_mass_over_light(b);    

    if (verbose)
      cerr << "\n  Stellar content:\n";
    print_stellar_content(b);

    if (verbose)
      cerr << "\n  Special star content:\n";
    print_special_stars(b);

    if (verbose)
      cerr << "\n  Compact stars:\n";
    print_compact_stars(b);

  }
  else 
    cerr << "\n  No stellar evolution\n";
      
}

#else

main(int argc, char **argv)
{
    bool binaries = true, verbose = true, out = false,
         n_sq = true, calc_e = true;
    int which_lagr = 0;
    real T_start = 0;

    extern char *poptarg;
    int c;
    char* param_string = "bneost";	// Note: "v" removed because only the
					// "verbose" option currently works.

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'b': binaries = !binaries;
		      break;
	    case 'e': calc_e = !calc_e;
		      break;
	    case 'n': n_sq = !n_sq;
		      break;
	    case 'o': out = !out;
		      break;
	    case 's': which_lagr = 2;
		      break;
	    case 't': which_lagr = 1;
		      break;
	    case 'v': verbose = !verbose;
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
		      exit(1);
	}

    // Loop over input until no more data remain.

    dyn *b;
    int i = 0;
    vector zero = 0;
    bool mass_spectrum = true;
    while (b = get_dyn(cin)) {

        // NOTE:  get_xxx() reads NaN in as legal garbage...

      if(find_qmatch(b->get_oldest_daughter()
		 ->get_starbase()->get_star_story(), "Type")) {
	
	addstar(b,                            // Note that T_start and
		b->get_system_time(),          // Main_Sequence are
		Main_Sequence,                 // defaults. They are
		true);                         // ignored if a star
	b->set_use_sstar(true);                // is there.

      if (i++ > 0) cerr << endl;
      sstar_stats(b, mass_spectrum, zero,  verbose);

      }
      
      if (out) put_node(cout, *b);
      rmtree(b);
    }
}

#endif
